Package: depCensoring
Type: Package
Title: Statistical Methods for Survival Data with Dependent Censoring
Version: 0.1.10
Authors@R: c(person("Ilias", "Willems", email = "ilias.willems@kuleuven.be", role = "aut", comment = c(ORCID = "0009-0001-9463-9942")),
    person("Gilles", "Crommen", email = "gilles.crommen@kuleuven.be", role = "aut", comment = c(ORCID = "0000-0001-8380-1900")),
    person("Negera Wakgari", "Deresa", email = "negera.deresa@gmail.com", role = c("aut", "cre"), comment = c(ORCID = "0000-0002-1302-3725")),
    person("Jie", "Ding", email = "biostat_jieding@outlook.com", role = c("aut"), comment = c(ORCID = "0000-0002-6083-7529")),
    person("Claudia","Czado", email = "cczado@cit.tum.de", role = "aut", comment = c(ORCID = "0000-0002-6329-5438")),
    person("Myrthe", "D'Haen", email = "myrthe.dhaen@kuleuven.be", role = "aut", comment = c(ORCID = "0000-0003-0223-0267")),
    person("Ingrid", "Van Keilegom", email = "ingrid.vankeilegom@kuleuven.be", role = "aut", comment = c(ORCID = "0000-0001-8827-7642")))
Maintainer: Negera Wakgari Deresa <negera.deresa@gmail.com>
Description: Several statistical methods for analyzing survival data under various forms of dependent
    censoring are implemented in the package. In addition to accounting for dependent censoring, it 
    offers tools to adjust for unmeasured confounding factors. The implemented approaches allow 
    users to estimate the dependency between survival time and dependent censoring time, based 
    solely on observed survival data. For more details on the methods, refer to Deresa and Van 
    Keilegom (2021) <doi:10.1093/biomet/asaa095>, Czado and Van Keilegom (2023)
    <doi:10.1093/biomet/asac067>, Crommen et al. (2024) <doi:10.1007/s11749-023-00903-9>,
    Deresa and Van Keilegom (2024) <doi:10.1080/01621459.2022.2161387>, Willems et al. (2025) 
    <doi:10.48550/arXiv.2403.11860>, Ding and Van Keilegom (2025) and D'Haen et al. (2025)
    <doi:10.1007/s10985-025-09647-0>.
Imports: survival, foreach, parallel, doParallel, pbivnorm, stats,
        MASS, nleqslv, methods, Matrix, EnvStats, mvtnorm, rafalib,
        rvinecopulib, matrixcalc, nloptr, numDeriv, copula, R6,
        lubridate, splines2
License: GPL-3
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.3
Suggests: testthat (>= 3.0.0), rkriging, orthopolynom, OpenMx, stringr,
        LaplacesDemon
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-04-13 16:11:57 UTC; Negera Deresa
Author: Ilias Willems [aut] (ORCID: <https://orcid.org/0009-0001-9463-9942>),
  Gilles Crommen [aut] (ORCID: <https://orcid.org/0000-0001-8380-1900>),
  Negera Wakgari Deresa [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-1302-3725>),
  Jie Ding [aut] (ORCID: <https://orcid.org/0000-0002-6083-7529>),
  Claudia Czado [aut] (ORCID: <https://orcid.org/0000-0002-6329-5438>),
  Myrthe D'Haen [aut] (ORCID: <https://orcid.org/0000-0003-0223-0267>),
  Ingrid Van Keilegom [aut] (ORCID:
    <https://orcid.org/0000-0001-8827-7642>)
Depends: R (>= 3.5.0)
Repository: CRAN
Date/Publication: 2026-04-13 18:10:08 UTC
Built: R 4.5.2; ; 2026-04-13 19:18:22 UTC; unix

Package: depCensoring
Type: Package
Title: Statistical Methods for Survival Data with Dependent Censoring
Version: 0.1.10
Authors@R: c(person("Ilias", "Willems", email = "ilias.willems@kuleuven.be", role = "aut", comment = c(ORCID = "0009-0001-9463-9942")),
    person("Gilles", "Crommen", email = "gilles.crommen@kuleuven.be", role = "aut", comment = c(ORCID = "0000-0001-8380-1900")),
    person("Negera Wakgari", "Deresa", email = "negera.deresa@gmail.com", role = c("aut", "cre"), comment = c(ORCID = "0000-0002-1302-3725")),
    person("Jie", "Ding", email = "biostat_jieding@outlook.com", role = c("aut"), comment = c(ORCID = "0000-0002-6083-7529")),
    person("Claudia","Czado", email = "cczado@cit.tum.de", role = "aut", comment = c(ORCID = "0000-0002-6329-5438")),
    person("Myrthe", "D'Haen", email = "myrthe.dhaen@kuleuven.be", role = "aut", comment = c(ORCID = "0000-0003-0223-0267")),
    person("Ingrid", "Van Keilegom", email = "ingrid.vankeilegom@kuleuven.be", role = "aut", comment = c(ORCID = "0000-0001-8827-7642")))
Maintainer: Negera Wakgari Deresa <negera.deresa@gmail.com>
Description: Several statistical methods for analyzing survival data under various forms of dependent
    censoring are implemented in the package. In addition to accounting for dependent censoring, it 
    offers tools to adjust for unmeasured confounding factors. The implemented approaches allow 
    users to estimate the dependency between survival time and dependent censoring time, based 
    solely on observed survival data. For more details on the methods, refer to Deresa and Van 
    Keilegom (2021) <doi:10.1093/biomet/asaa095>, Czado and Van Keilegom (2023)
    <doi:10.1093/biomet/asac067>, Crommen et al. (2024) <doi:10.1007/s11749-023-00903-9>,
    Deresa and Van Keilegom (2024) <doi:10.1080/01621459.2022.2161387>, Willems et al. (2025) 
    <doi:10.48550/arXiv.2403.11860>, Ding and Van Keilegom (2025) and D'Haen et al. (2025)
    <doi:10.1007/s10985-025-09647-0>.
Imports: survival, foreach, parallel, doParallel, pbivnorm, stats,
        MASS, nleqslv, methods, Matrix, EnvStats, mvtnorm, rafalib,
        rvinecopulib, matrixcalc, nloptr, numDeriv, copula, R6,
        lubridate, splines2
License: GPL-3
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.3
Suggests: testthat (>= 3.0.0), rkriging, orthopolynom, OpenMx, stringr,
        LaplacesDemon
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-04-13 16:11:57 UTC; Negera Deresa
Author: Ilias Willems [aut] (ORCID: <https://orcid.org/0009-0001-9463-9942>),
  Gilles Crommen [aut] (ORCID: <https://orcid.org/0000-0001-8380-1900>),
  Negera Wakgari Deresa [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-1302-3725>),
  Jie Ding [aut] (ORCID: <https://orcid.org/0000-0002-6083-7529>),
  Claudia Czado [aut] (ORCID: <https://orcid.org/0000-0002-6329-5438>),
  Myrthe D'Haen [aut] (ORCID: <https://orcid.org/0000-0003-0223-0267>),
  Ingrid Van Keilegom [aut] (ORCID:
    <https://orcid.org/0000-0001-8827-7642>)
Depends: R (>= 3.5.0)
Repository: CRAN
Date/Publication: 2026-04-13 18:10:08 UTC
Built: R 4.6.0; ; 2026-04-26 06:54:59 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(summary,depFit)
S3method(summary,indepFit)
export(Chronometer)
export(NonParTrans)
export(ParamCop)
export(QRdepCens)
export(SolveL)
export(SolveLI)
export(SurvDC)
export(TCsim)
export(estimate.cmprsk)
export(fitDepCens)
export(fitIndepCens)
export(pi.surv)
import(MASS)
import(foreach)
import(matrixcalc)
import(mvtnorm)
import(nleqslv)
import(nloptr)
import(numDeriv)
import(parallel)
import(pbivnorm)
import(rafalib)
import(rvinecopulib)
import(stats)
importFrom(EnvStats,base)
importFrom(EnvStats,qlnormTrunc)
importFrom(MASS,mvrnorm)
importFrom(Matrix,bdiag)
importFrom(R6,R6Class)
importFrom(copula,archmCopula)
importFrom(copula,frankCopula)
importFrom(copula,gumbelCopula)
importFrom(copula,normalCopula)
importFrom(copula,pCopula)
importFrom(copula,rCopula)
importFrom(copula,tau)
importFrom(doParallel,registerDoParallel)
importFrom(foreach,foreach)
importFrom(grDevices,dev.off)
importFrom(graphics,abline)
importFrom(graphics,hist)
importFrom(graphics,legend)
importFrom(graphics,lines)
importFrom(graphics,points)
importFrom(lubridate,now)
importFrom(methods,is)
importFrom(nloptr,nloptr)
importFrom(parallel,detectCores)
importFrom(parallel,makeCluster)
importFrom(parallel,stopCluster)
importFrom(rvinecopulib,ktau_to_par)
importFrom(splines2,bSpline)
importFrom(stats,as.formula)
importFrom(stats,constrOptim)
importFrom(stats,dlnorm)
importFrom(stats,dlogis)
importFrom(stats,dnorm)
importFrom(stats,dweibull)
importFrom(stats,integrate)
importFrom(stats,lm)
importFrom(stats,na.omit)
importFrom(stats,nlminb)
importFrom(stats,plnorm)
importFrom(stats,plogis)
importFrom(stats,pnorm)
importFrom(stats,pweibull)
importFrom(stats,qnorm)
importFrom(stats,quantile)
importFrom(stats,sd)
importFrom(stats,uniroot)
importFrom(stats,var)
importFrom(survival,Surv)
importFrom(survival,coxph)
importFrom(survival,survreg)
importFrom(utils,install.packages)
importFrom(utils,tail)
importFrom(utils,write.csv)

#' @title Fit Dependent Censoring Models
#'
#' @description This function allows to estimate the dependency parameter along all other model parameters. First, estimates the cumulative hazard function, and
#' then at the second stage it estimates other model parameters assuming that the cumulative hazard function is known. The details for
#' implementing the dependent censoring methodology can be found in Deresa and Van Keilegom (2024).
#'
#' @references Deresa and Van Keilegom (2024). Copula based Cox proportional hazards models for dependent censoring, Journal of the American Statistical Association, 119:1044-1054.
#'
#' @param start Initial values for the finite dimensional parameters. If \code{start} is NULL, the initial values will be obtained
#' by fitting a Cox model for survival time T and a Weibull model for dependent censoring C.
#'
#' @param resData Data matrix with three columns;  Z = the observed survival time, d1 = the censoring indicator of T
#' and  d2 =  the censoring indicator of C.
#' @param X Data matrix with covariates related to T.
#' @param W Data matrix with covariates related to C. First column of W should be a vector of ones.
#' @param cop Which copula should be computed to account for dependency between T and C. This argument can take
#'   one of the values from \code{c("Gumbel", "Frank", "Normal")}.
#' @param dist The distribution to be used for the censoring time C. Only two distributions are allowed, i.e, Weibull
#' and lognormal distributions. With the value \code{"Weibull"} as the
#' default.
#' @param eps Convergence error. This is set by the user in such away that the desired convergence is met; the default is \code{eps = 1e-4}.
#' @param n.iter Number of iterations; the default is \code{n.iter = 50}. The larger the number of iterations, the longer the computational time.
#' @param bootstrap A boolean indicating whether to compute bootstrap standard errors for making inferences.
#' @param n.boot Number of bootstrap samples to use in the estimation of bootstrap standard errors if \code{bootstrap = TRUE}. The default is n.boot = 150. But, higher
#' values  of \code{n.boot} are recommended for obtaining good estimates of bootstrap standard errors.
#' @param ncore The number of cores to use for parallel computation in bootstrapping, with the default \code{ncore = 7}.
#' @importFrom copula pCopula frankCopula gumbelCopula tau
#' @importFrom stats nlminb pnorm  qnorm
#' @importFrom survival coxph survreg Surv
#'
#' @return This function returns a fit of dependent censoring model; parameter estimates, estimate of the cumulative hazard function, bootstrap standard
#' errors for finite-dimensional parameters, the nonparametric cumulative hazard function, etc.
#'
#'
#'
#' @examples
#' \donttest{
#' # Toy data example to illustrate implementation
#' n = 300
#' beta = c(0.5)
#' lambd = 0.35
#' eta = c(0.9,0.4)
#' X = cbind(rbinom(n,1,0.5))
#' W = cbind(rep(1,n),rbinom(n,1,0.5))
#' # generate dependency structure from Frank
#' frank.cop <- copula::frankCopula(param = 5,dim = 2)
#' U = copula::rCopula(n,frank.cop)
#' T1 = (-log(1-U[,1]))/(lambd*exp(X*beta))           # Survival time
#' T2 = (-log(1-U[,2]))^(1.1)*exp(W%*%eta)            # Censoring time
#' A = runif(n,0,15)                                  # administrative censoring time
#' Z = pmin(T1,T2,A)
#' d1 = as.numeric(Z==T1)
#' d2 = as.numeric(Z==T2)
#' resData = data.frame("Z" = Z,"d1" = d1, "d2" = d2)   # should be data frame
#' colnames(W) <- c("ones","cov1")
#' colnames(X) <- "cov.surv"
#'
#' # Fit dependent censoring model
#'
#'fit <- fitDepCens(resData = resData, X = X, W = W, bootstrap = FALSE)
#'
#' # parameter estimates
#'
#' fit$parameterEstimates
#'
#' # summary fit results
#' summary(fit)
#'
#' # plot cumulative hazard vs time
#'
#' plot(fit$observedTime, fit$cumhazardFunction, type = "l",xlab = "Time",
#' ylab = "Estimated cumulative hazard function")
#'}
#'
#' @export


fitDepCens = function(resData,X,W,
                      cop = c("Frank","Gumbel", "Normal"),
                      dist = c("Weibull", "lognormal"), start = NULL, n.iter = 50, bootstrap = TRUE, n.boot = 150, ncore = 7, eps = 1e-4){

  cop <- match.arg(cop)
  dist <- match.arg(dist)

  #  Verify column names of input dataframe resData
  if (!all(c("Z", "d1", "d2") %in% colnames(resData)))
    stop("Z, d1 and d2 arguments must be column names of resData")

  id <- match(c("Z", "d1", "d2"), colnames(resData))
  colnames(resData)[id] <- c("Z", "d1", "d2")

  Z = resData$Z
  d1 = resData$d1
  d2 = resData$d2

  if(!is.matrix(W)) W = as.matrix(W)
  if(!is.matrix(X)) X = as.matrix(X)
  k = dim(X)[2]
  l = dim(W)[2]

  if(l==1)
    stop("W should be a matrix of dimension larger than 1")

  if(!is.null(start)){
    if(start[(k+l+1)<=0]|start[(k+l+2)<=0])
      stop("Scale or dependency parameter cannot be negative")
  }
  if(is.null(start)){
    fit1 = coxph(Surv(Z,d1)~X)
    fit2 = survreg(Surv(Z,d2)~W-1)
  }

  if(is.null(start) & (cop == "Frank" | cop == "Gumbel")){                          # obtain initial values by fitting standard models
    start = c(fit1$coefficients,fit2$coefficients,fit2$scale,2)
  }else if (is.null(start) & (cop == "Normal")){
    start = c(fit1$coefficients,fit2$coefficients,fit2$scale,0.2)}

  # lb = lower boundary, ub = upper boundary
  if(cop == "Frank")                           # Frank copula
  {
    lb = c(rep(-Inf,k+l),0,1e-5)
    ub = c(rep(Inf,k+l+2))
  }else if(cop == "Gumbel")                    # Gumbel copula
  {lb = c(rep(-Inf,k+l),0,1.0001)
  ub = c(rep(Inf,k+l+2))
  } else if(cop == "Normal")                   # Normal
  {lb = c(rep(-Inf,k+l),0,-0.99)
  ub = c(rep(Inf,k+l+1),0.99)
  }else
    stop(paste0("The copula function",cop, "is not supported"))

  if(length(start)!= k+l+2)
    stop("The length of initial values is not equal to the number of parameters need to be estimated")


  res <- SolveL(start,resData,X,W,cop,dist)              # Estimate the cumulative hazard function
  lsml <- res$lambda
  Llrge <- res$cumhaz
  T1 <- res$times
  longfrm <- Longfun(Z,T1,lsml,Llrge)
  lhat <- longfrm[,1]
  cumL <- longfrm[,2]

  # initial step of estimation

  parhat = nlminb(start = start, PseudoL, resData = resData, X = X, W = W,lhat = lhat,cumL = cumL,cop = cop,dist = dist,lower = lb ,upper =  ub, control = list(eval.max = 300,iter.max = 200))$par
  a = start
  b = parhat
  flag = 0

  # Then do estimation iteratively until convergence

  while (Distance(b,a)>eps){                                             # doing this while loop until the desired convergence criteria are met
    a = b
    res <- SolveL(a,resData,X,W,cop,dist)
    lsml = res$lambda
    Llrge = res$cumhaz
    T1 = res$times
    longfrm = Longfun(Z,T1,lsml,Llrge)
    lhat = longfrm[,1]
    cumL = longfrm[,2]
    parhat = nlminb(start = a,PseudoL,resData = resData,X = X,W = W,lhat = lhat,cumL = cumL,cop = cop,dist = dist,lower = lb ,upper =  ub, control = list(eval.max=300,iter.max=200))$par
    b = parhat
    flag = flag+1;
    if (flag>n.iter)                                  # Stop after iteration 20; this usually gives sufficient convergence results
    {
      flag=0;
      #warning("The maximum number of iterations reached before convergence criteria is satisified. Better convergence may be obtained by increasing n.iter")
      warning(paste0("The current convergence error (eps) is ", Distance(b,a)))
      break;
    }
  }
  cumHat <- SolveL(b,resData,X,W,cop,dist)

  # nonparametric bootstrap for making inference

  paramsBootstrap = list("init" = parhat,"resData" = resData, "X" = X, "W" = W,"lhat" = lhat,"cumL" = cumL, "dist" = dist,"k" = k, "lb" = lb, "ub" = ub, "Obs.time" = T1, "cop" = cop, "n.boot" = n.boot, "n.iter" = n.iter, "ncore" = ncore, "eps" = eps)

  fitObj <- NULL

  if(bootstrap)                                       # Obtain bootstrap standard error
    fitObj <- do.call(boot.fun, paramsBootstrap)

  # Transform the copula parameter to tau scale
  nn <- length(parhat)
  if (cop == "Frank") {parhat[nn] <- tau(frankCopula(parhat[nn]))
  }else if (cop == "Gumbel"){ parhat[nn] <- tau(gumbelCopula(parhat[nn]))
  }else parhat[nn] <- 2*asin(parhat[nn])/pi
  names(parhat)[nn] <- "tau"
  names(parhat)[nn-1] <- "sigma"

  depObj <- c(list("parameterEstimates" = parhat,"copula" = cop, "censoringDistribution" = dist, "bootstrap" = bootstrap, "dimX" = k, "dimW" = l,"observedTime" = cumHat$times,"hazardFunction" = cumHat$lambda, "cumhazardFunction" = cumHat$cumhaz),fitObj)

  class(depObj) <- append(class(depObj), "depFit")                   # dependent censoring fit object

  return(depObj)
}





#' @title Fit Independent Censoring Models
#'
#' @description This function allows to estimate all model parameters under the assumption of independent censoring. First, estimates the cumulative hazard function, and
#' then at the second stage it estimates other model parameters assuming that the cumulative hazard is known.
#'
#'
#' @param start Initial values for the finite dimensional parameters. If \code{start} is NULL, the initial values will be obtained
#' by fitting a Cox model for survival time T and a Weibull model for censoring time C.
#'
#' @param resData Data matrix with three columns;  Z = the observed survival time, d1 = the censoring indicator of T
#' and  d2 =  the censoring indicator of C.
#' @param X Data matrix with covariates related to T.
#' @param W Data matrix with covariates related to C. First column of W should be ones.
#' @param dist The distribution to be used for the censoring time C. Only two distributions are allowed, i.e, Weibull
#' and lognormal distributions. With the value \code{"Weibull"} as the
#'   default.
#' @param eps Convergence error. This is set by the user in such away that the desired convergence is met; the default is \code{eps = 1e-4}.
#' @param n.iter Number of iterations; the default is \code{n.iter = 50}. The larger the number of iterations, the longer the computational time.
#' @param bootstrap A boolean indicating whether to compute bootstrap standard errors for making inferences.
#' @param n.boot Number of bootstrap samples to use in the estimation of bootstrap standard errors if \code{bootstrap = TRUE}. The default is n.boot = 150. But, higher
#' values  of \code{n.boot} are recommended for obtaining good estimates of bootstrap standard errors.
#' @param ncore The number of cores to use for parallel computation is configurable, with the default \code{ncore = 7}.
#' @importFrom stats nlminb pnorm  qnorm sd
#' @importFrom survival coxph survreg Surv
#'
#' @return This function returns a fit of independent censoring model; parameter estimates, estimate of the cumulative hazard function, bootstrap standard
#' errors for finite-dimensional parameters, the nonparametric cumulative hazard function, etc.
#'
#'
#'
#' @examples
#' \donttest{
#'
#' # Toy data example to illustrate implementation
#' n = 300
#' beta = c(0.5)
#' lambd = 0.35
#' eta = c(0.9,0.4)
#' X = cbind(rbinom(n,1,0.5))
#' W = cbind(rep(1,n),rbinom(n,1,0.5))
#' # generate dependency structure from Frank
#' frank.cop <- copula::frankCopula(param = 5,dim = 2)
#' U = copula::rCopula(n,frank.cop)
#' T1 = (-log(1-U[,1]))/(lambd*exp(X*beta))                  # Survival time'
#' T2 = (-log(1-U[,2]))^(1.1)*exp(W%*%eta)                   # Censoring time
#' A = runif(n,0,15)                                         # administrative censoring time
#' Z = pmin(T1,T2,A)
#' d1 = as.numeric(Z==T1)
#' d2 = as.numeric(Z==T2)
#' resData = data.frame("Z" = Z,"d1" = d1, "d2" = d2)     # should be data frame
#'
#' colnames(W) <- c("ones","cov1")
#' colnames(X) <- "cov.surv"
#'
#' # Fit independent censoring model
#'
#' fitI <- fitIndepCens(resData = resData, X = X, W = W, bootstrap = FALSE)
#'
#' # parameter estimates
#'
#' fitI$parameterEstimates
#'
#' # summary fit results
#' summary(fitI)
#'
#' # plot cumulative hazard vs time
#'
#'  plot(fitI$observedTime, fitI$cumhazardFunction, type = "l",xlab = "Time",
#'  ylab = "Estimated cumulative hazard function")
#'}
#'
#' @export
fitIndepCens = function(resData,X,W,
                        dist = c("Weibull", "lognormal"), start = NULL,
                        n.iter = 50, bootstrap = TRUE, n.boot = 150, ncore = 7, eps = 1e-4){

  dist <- match.arg(dist)

  if (!all(c("Z", "d1", "d2") %in% colnames(resData)))
    stop("Z, d1 and d2 arguments must be column names of resData")

  id <- match(c("Z", "d1", "d2"), colnames(resData))
  colnames(resData)[id] <- c("Z", "d1", "d2")

  Z = resData$Z
  d1 = resData$d1
  d2 = resData$d2

  if (is.vector(X)) k = 1
  else if(!is.vector(X)) k = dim(X)[2]
  l = dim(W)[2]

  if(!is.null(start)){
    if(start[(k+l+1)<=0])
      stop("Scale parameter cannot be negative")
  }

  if(is.null(start)){                                    # obtain initial values by fitting standard models
    fit1 = survival::coxph(Surv(Z,d1)~X)
    fit2 = survival::survreg(Surv(Z,d2)~W-1)
    start = c(fit1$coefficients,fit2$coefficients,fit2$scale)
  }
  if(length(start)!= k+l+1)
    stop("The length of initial values is not equal to the number of parameters need to be estimated")


  res <- SolveLI(start,resData,X)                 # estimate the cumulative hazard function
  lsml <- res$lambda
  Llrge <- res$cumhaz
  T1 <- res$times
  longfrm <- Longfun(Z,T1,lsml,Llrge)
  lhat <- longfrm[,1]
  cumL <- longfrm[,2]


  # lb = lower boundary, ub = upper boundary
  lb = c(rep(-Inf,k+l),0)
  ub = c(rep(Inf,k+l+1))

  # Initial estimate of theta

  parhat  = nlminb(start = start,LikCopInd, resData = resData,X = X,W = W,lhat = lhat,cumL = cumL, dist = dist, lower = lb ,upper =  ub, control = list(eval.max=300,iter.max=200))$par

  a = start
  b = parhat
  flag = 0

  while (Distance(b,a)>eps){                             # doing this while loop until the desired convergence criteria are met
    a = b
    res = SolveLI(a,resData,X)
    lsml = res$lambda
    Llrge = res$cumhaz
    T1 = res$times
    longfrm = Longfun(Z,T1,lsml,Llrge)
    lhat = longfrm[,1]
    cumL = longfrm[,2]
    parhat  = nlminb(start = a,LikCopInd,resData = resData,X = X,W = W,lhat = lhat,cumL = cumL, dist = dist, lower = lb ,upper =  ub, control = list(eval.max=300,iter.max=200))$par

    b = parhat
    flag = flag+1;
    if (flag>n.iter)                                    # stop after iteration 30; this usually gives sufficient convergence results
    {
      flag=0;
      #warning("The maximum number of iterations reached before convergence criteria is satisified. Better convergence may be obtained by increasing n.iter")
      warning(paste0("The current convergence error (eps) is", Distance(b,a)))
      break;
    }
  }
  cumHat = SolveLI(parhat,resData,X)

  # nonparametric bootstrap for making inference

  paramsBootstrap = list("init" = parhat,"resData" = resData, "X" = X, "W" = W,"lhat" = lhat,"cumL" = cumL, "dist" = dist,"k" = k, "lb" = lb, "ub" = ub, "Obs.time" = T1, "n.boot" = n.boot, "n.iter" = n.iter,"ncore" = ncore, "eps" = eps)

  fitObj <- NULL

  if(bootstrap)                                               # Obtain bootstrap standard error
    fitObj <- do.call(boot.funI, paramsBootstrap)

  nn <- length(parhat)
  names(parhat)[nn] <- "sigma"

  indObj <- c(list("parameterEstimates" = parhat, "censoringDistribution" = dist, "bootstrap" = bootstrap, "dimX" = k, "dimW" = l,"observedTime" = cumHat$times, "hazardFunction" = cumHat$lambda, "cumhazardFunction" = cumHat$cumhaz),fitObj)

  class(indObj) <- append(class(indObj), "indepFit")        # independent censoring fit object

  return(indObj)
}

library(depCensoring)
library(MASS)
library(survival)
library(pbivnorm)
library(nleqslv)

# data example

rho = 0.8
n = 300
beta = c(0.5, 1); eta = c(1,1.5)
sigma = matrix(c(1,rho,rho,1),ncol=2)
err = MASS::mvrnorm(n, mu = c(0,0) , Sigma=sigma)
err1 = err[,1]; err2 = err[,2]
x1 = rbinom(n,1,0.5); x2 = runif(n,-1,1)
X = matrix(c(x1,x2),ncol=2,nrow=n); W = X   # data matrix
T1 = X%*%beta+err1
C =  W%*%eta+err2
T1 = exp(T1); C = exp(C)
 A = runif(n,0,8); Y = pmin(T1,C,A)
d1 = as.numeric(Y==T1)
d2 = as.numeric(Y==C)
resData = data.frame("Z" = Y,"d1" = d1, "d2" = d2)   # should be data frame
colnames(X) = c("X1", "X2")
colnames(W) = c("W1","W2")

#  Bootstrap is False by default
output = NonParTrans(resData = resData, X = X, W = W, n.iter = 3)
l = length(output$parameterEstimates)
rhohat <- output$parameterEstimates[l]

library(depCensoring)
library(rvinecopulib)
library(rafalib)
tau = 0.75
Copula = "frank"
Dist.T = "weibull"
Dist.C = "lnorm"
par.T = c(2,1)
par.C = c(1,2)
n=1000
simdata<-TCsim(tau,Copula,Dist.T,Dist.C,par.T,par.C,n)
Y = simdata[[1]]
Delta = simdata[[2]]
fitPar <- ParamCop(Y,Delta,Copula,Dist.T,Dist.C)

#' @title Logarithmic transformation function.
#'
#' @description Computes the logarithm of a number.
#'
#' @param y Numerical value of which the logarithm is computed.
#'
#' @return This function returns the logarithm of the provided argument y
#' if it is greater than zero. If y is smaller than zero, it will return 0.
#'
#' @noRd
#'

log_transform = function(y) {
  transform_y = (y>0)*y+(y<=0)*1
  return(log(transform_y))
}

#' @title Power transformation function.
#'
#' @description Computes a given power of a number.
#'
#' @param y The number which one wants to raise to a certain power \code{pw}.
#' @param pw The power to which to raise \code{y}.
#'
#' @return This function returns the result of raising \code{y} to the power
#' \code{pw} when \code{y > 0}. Otherwise, it will return 1.
#'
#' @noRd
#'

power_transform = function(y,pw) {
  transform_y = (y>0)*y+(y<=0)*1
  return(transform_y^pw)
}

#' @title Yeo-Johnson transformation function
#'
#' @description Computes the Yeo-Johnson transformation of the provided argument.
#'
#' @param y The argument to be supplied to the Yeo-Johnson transformation.
#' @param theta The parameter of the Yeo-Johnson transformation. This should be
#' a number in the range [0,2].
#'
#' @return The transformed value of y.
#'
#' @noRd
#'

YJtrans = function(y,theta) {
  sg = y>=0
  if (theta==0) {temp = log_transform(y+1)*sg+(1-sg)*(0.5-0.5*(y-1)^2)}
  if (theta==2) {temp = sg*(-0.5+0.5*(y+1)^2)-log_transform(-y+1)*(1-sg)}
  if ((theta!=0) & (theta!=2)) {temp = sg*(power_transform(y+1,theta)-1)/theta+(1-sg)*(1-power_transform(-y+1,2-theta))/(2-theta)}
  return(temp)
}

#' @title Inverse Yeo-Johnson transformation function
#'
#' @description Computes the inverse Yeo-Johnson transformation of the provided
#' argument.
#'
#' @param y The argument to be supplied to the inverse Yeo-Johnson transformation.
#' @param theta The parameter of the inverted Yeo-Johnson transformation. This
#' should be a number in the range [0,2].
#'
#' @return The transformed value of y.
#'
#' @noRd
#'

IYJtrans = function(y,theta) {
  sg = y>=0
  if (theta==0) {temp =(exp(y)-1)*sg+(1-sg)*(1-power_transform(-2*y+1,0.5))}
  if (theta==2) {temp = sg*(-1+power_transform(2*y+1,0.5))+(1-exp(-y))*(1-sg)}
  if ((theta!=0) & (theta!=2)) {temp = sg*(power_transform(abs(theta)*y+1,1/theta)-1)+(1-sg)*(1-power_transform(1-(2-theta)*y,1/(2-theta)))}
  return(temp)
}

#' @title Derivative of the Yeo-Johnson transformation function
#'
#' @description Evaluates the derivative of the Yeo-Johnson transformation at
#' the provided argument.
#'
#' @param y The argument to be supplied to the derivative of the Yeo-Johnson
#' transformation.
#' @param theta The parameter of the Yeo-Johnson transformation. This should be
#' a number in the range [0,2].
#'
#' @return The transformed value of y.
#'
#' @noRd
#'

DYJtrans = function(y,theta) {
  sg = y>=0
  temp = power_transform(y+1,theta-1)*sg+power_transform(-y+1,1-theta)*(1-sg)
  return(temp)
}

#' @title Nonparametric bootstrap approach for the dependent censoring model
#'
#' @description This function estimates the bootstrap standard errors for the finite-dimensional model parameters and for the non-parametric cumulative
#' hazard function. Parallel computing using foreach has been used to speed up the estimation of standard errors.
#'
#'
#' @param init Initial values for the finite dimensional parameters obtained from the fit of \code{\link{fitDepCens}}
#' @param lhat Initial values for the hazard function obtained from the fit of \code{\link{fitDepCens}} based on the original data.
#' @param cumL Initial values for the cumulative hazard function obtained from the fit of \code{\link{fitDepCens}} based on the original data.
#' @param resData Data matrix with three columns;  Z = the observed survival time, d1 = the censoring indicator of T
#' and  d2 =  the censoring indicator of C.
#' @param X Data matrix with covariates related to T
#' @param W Data matrix with covariates related to C. First column of W should be a vector of ones
#' @param k Dimension of X
#' @param lb lower boundary for finite dimensional parameters
#' @param ub Upper boundary for finite dimensional parameters
#' @param cop Which copula should be computed to account for dependency between T and C. This argument can take
#'   one of the values from \code{c("Gumbel", "Frank", "Normal")}.
#' @param dist The distribution to be used for the  dependent censoring time C. Only two distributions are allowed, i.e, Weibull
#' and lognormal distributions. With the value \code{"Weibull"} as the
#'   default.
#' @param Obs.time Observed survival time, which is the minimum of T, C and A, where A is the administrative censoring time.
#' @param eps Convergence error. This is set by the user in such away that the desired convergence is met; the default is \code{eps = 1e-3}
#' @param n.iter Number of iterations; the default is \code{n.iter = 20}. The larger the number of iterations, the longer the computational time.
#' @param n.boot Number of bootstraps to use in the estimation of bootstrap standard errors.
#' @param ncore The number of cores to use for parallel computation is configurable, with the default \code{ncore = 7}.
#' @importFrom stats nlminb pnorm  qnorm sd
#' @importFrom copula pCopula frankCopula gumbelCopula tau
#' @import foreach
#' @import parallel
#' @return Bootstrap standard errors for parameter estimates and for estimated cumulative hazard function.
#' @noRd
boot.fun = function(init,resData,X,W,lhat, cumL,dist,k,lb, ub, Obs.time,cop,n.boot, n.iter, ncore, eps){
  B = n.boot                                     # number of bootstrap samples
  n.cores <- ncore

  my.cluster <- parallel::makeCluster(
    n.cores,
    type = "PSOCK"
  )
  doParallel::registerDoParallel(cl = my.cluster)

  boot = foreach(b = 1:B, .packages= c('survival', 'copula', 'pbivnorm'), .export = c("PseudoL", "SolveL","Distance", "CompC", "Longfun", "SearchIndicate")) %dopar% {
    samp1 = sample(length(resData$Z),replace = TRUE)
    resData_b = resData[samp1,]
    Zb = resData_b$Z
    if(k==1){
      Xb = X[samp1]
    }else{
      Xb = X[samp1,]
    }
    Wb = W[samp1,]

    # Initial step of estimation

    parhatb = nlminb(start = init, PseudoL, resData = resData_b,X = Xb,W = Wb,lhat = lhat,cumL = cumL,cop = cop,dist = dist,lower = lb ,upper =  ub, control = list(eval.max=300,iter.max=200))$par
    aboot = init
    bboot = parhatb
    flag = 0

    while (Distance(bboot,aboot)>eps){                            # doing this while loop until the desired convergence criteria are met
      aboot = bboot
      res = SolveL(aboot,resData_b,Xb,Wb,cop,dist)
      lsml = res$lambda
      Llrge = res$cumhaz
      T1 = res$times
      longfrm = Longfun(Zb,T1,lsml,Llrge)
      lhat = longfrm[,1]
      cumL = longfrm[,2]
      parhatb = nlminb(start = aboot, PseudoL, resData = resData_b ,X = Xb,W = Wb, lhat = lhat,cumL = cumL,cop = cop,dist = dist,lower = lb ,upper =  ub, control = list(eval.max=300,iter.max=200))$par

      bboot = parhatb
      flag = flag+1;
      if (flag>n.iter)                                                              # Stop after iteration 50; this usually gives sufficient convergence results
      {
        break;
      }
    }
    cumHat = SolveL(bboot,resData_b,Xb,Wb,cop,dist)
    parb = bboot
    nn = length(parb)
    lboot = cumHat$lambda
    Lboot = cumHat$cumhaz
    Tb1 = cumHat$times
    longB = Longfun(Obs.time,Tb1,lboot, Lboot)
    lhatb = longB[,1]
    cumH.boot = longB[,2]

    # Transform the copula parameter to tau scale
    if (cop == "Frank"){ parb[nn] = tau(frankCopula(parb[nn]))
    }else if (cop == "Gumbel"){ parb[nn] = tau(gumbelCopula(parb[nn]))
    }else parb[nn] = 2*asin(parb[nn])/pi
    names(parb)[nn] <- "tau"
    names(parb)[(nn-1)] <- "sigma"

    list("beta.star" = parb, cumH.star = cumH.boot)
  }
  parallel::stopCluster(cl = my.cluster)

  beta.star = t(sapply(boot, function(x) x$beta.star))
  cumH.star =  t(sapply(boot, function(x) x$cumH.star))
  Bootse = apply(beta.star,2,sd)
  cumHse = apply(cumH.star,2,sd)
  bootR = list("parEst.std.error" = Bootse, "cumHazard.std.error" = cumHse)
  return(bootR)
}



#' @title  Nonparametric bootstrap approach for the independent censoring model

#' @description This function estimates the bootstrap standard errors for the finite-dimensional model parameters and for the non-parametric cumulative
#' hazard function under the assumption of independent censoring. Parallel computing using foreach has been used to speed up the computation.
#'
#'
#' @param init Initial values for the finite dimensional parameters obtained from the fit of \code{\link{fitIndepCens}}
#' @param lhat Initial values for the hazard function obtained from the fit of \code{\link{fitIndepCens}} based on the original data
#' @param cumL Initial values for the cumulative hazard function obtained from the fit of \code{\link{fitIndepCens}} based on the original data
#' @param resData Data matrix with three columns;  Z = the observed survival time, d1 = the censoring indicator of T
#' and  d2 =  the censoring indicator of C.
#' @param X Data matrix with covariates related to T
#' @param k Dimension of X
#' @param lb lower boundary for finite dimensional parameters
#' @param ub Upper boundary for finite dimensional parameters
#' @param W Data matrix with covariates related to C. First column of W should be a vector of ones
#' @param dist The distribution to be used for the  dependent censoring time C. Only two distributions are allowed, i.e, Weibull
#' and lognormal distributions. With the value \code{"Weibull"} as the
#'   default
#' @param Obs.time Observed survival time, which is the minimum of T, C and A, where A is the administrative censoring time.
#' @param eps Convergence error. This is set by the user in such away that the desired convergence is met; the default is \code{eps = 1e-3}
#' @param n.iter Number of iterations; the default is \code{n.iter = 20}. The larger the number of iterations, the longer the computational time
#' @param n.boot Number of bootstraps to use in the estimation of bootstrap standard errors.
#' @param ncore The number of cores to use for parallel computation is configurable, with the default \code{ncore = 7}.
#' @importFrom stats nlminb pnorm  qnorm sd
#' @importFrom survival coxph survreg
#' @import foreach
#'
#' @return Bootstrap standard errors for parameter estimates and for estimated cumulative hazard function.
#' @noRd
boot.funI = function(init,resData,X,W,lhat, cumL,dist,k,lb,ub, Obs.time,n.boot, n.iter, ncore, eps){
  B = n.boot                                     # number of bootstrap samples
  n.cores <- ncore

  my.cluster <- parallel::makeCluster(
    n.cores,
    type = "PSOCK"
  )
  doParallel::registerDoParallel(cl = my.cluster)

  boot = foreach(b = 1:B, .packages= c('survival'), .export = c("LikCopInd", "SolveLI","Distance", "Longfun", "SearchIndicate")) %dopar% {
    samp1 = sample(length(resData$Z),replace = TRUE)
    resData_b = resData[samp1,]
    Zb = resData_b$Z
    if(k==1) {Xb = X[samp1]
    }else Xb = X[samp1,]
    Wb = W[samp1,]

    # Initial step of estimation
    parhatb  = nlminb(start = init,LikCopInd,resData = resData_b,X = Xb,W = Wb,lhat = lhat,cumL = cumL, dist = dist, lower = lb, upper =  ub, control = list(eval.max=300,iter.max=200))$par
    aboot = init
    bboot = parhatb
    flag = 0

    while (Distance(bboot,aboot)>eps){                                   # doing this while loop until the desired convergence criteria are met
      aboot = bboot
      res = SolveLI(aboot,resData_b,Xb)
      lsml = res$lambda
      Llrge = res$cumhaz
      T1 = res$times
      longfrm = Longfun(Zb,T1,lsml,Llrge)
      lhat = longfrm[,1]
      cumL = longfrm[,2]
      parhatb = nlminb(start = aboot, LikCopInd, resData = resData_b,X = Xb,W = Wb,lhat = lhat,cumL = cumL,dist = dist,lower = lb ,upper =  ub, control = list(eval.max=300,iter.max=200))$par

      bboot = parhatb
      flag = flag+1;
      if (flag>n.iter)
      {
        break;
      }
    }
    cumHat = SolveLI(bboot,resData_b,Xb)
    parb = bboot
    nn <- length(parb)
    names(parb)[nn] <- "sigma"

    lboot = cumHat$lambda
    Lboot = cumHat$cumhaz
    Tb1 = cumHat$times
    longB = Longfun(Obs.time,Tb1,lboot,Lboot)
    lhatb = longB[,1]
    cumH.boot = longB[,2]

    list("beta.star" = parb, cumH.star = cumH.boot)
  }
  parallel::stopCluster(cl = my.cluster)

  beta.star = t(sapply(boot, function(x) x$beta.star))
  cumH.star =  t(sapply(boot, function(x) x$cumH.star))
  Bootse = apply(beta.star,2,sd)
  cumHse = apply(cumH.star,2,sd)
  bootR = list("parEst.std.error" = Bootse, "cumHazard.std.error" = cumHse)
  return(bootR)
}





#' Summary of \code{depCensoringFit} object
#'
#' @param object Output of \code{\link{fitDepCens}} function
#' @param ... Further arguments
#' @return Summary of dependent censoring model fit in the form of table
#'
#' @export
#'
summary.depFit <- function(object, ...) {

  message(rep("-", 100))
  cat("Summary of dependent censoring model")

  # Sys.sleep(1)

  k = object$dimX
  l = object$dimW


  if(object$bootstrap){
    stError <- object$parEst.std.error
    stError_ch <- object$cumHazard.std.error
    Z <- object$parameterEstimates/stError
    pvalue <- 2*(1-pnorm(abs(Z)))
    out.res <- cbind(object$parameterEstimates,stError,pvalue)

    message(rep("-", 100))
    cat("\n")
    cat("Survival submodel: Cox proportional hazards model")
    cat("\n \n")

    surv.out <- out.res[1:k,]
    if(k>=2){
      if(is.null(rownames(surv.out))){
        colName <- cbind(c(rep("X",k)),c(seq(1:k)))
        rownames(surv.out) <- noquote(apply(colName,1,paste, collapse = "_"))
      }else {
        rownames(surv.out) <-  gsub("X", "", rownames(surv.out))
      }
      colnames(surv.out) <- c("Estimate", "Boot.SE", "Pvalue")
      surv.out <- round(surv.out, 3)
      print(surv.out)
    }else{
      names(surv.out) <- c("Estimate", "Boot.SE", "Pvalue")
      surv.out <- round(surv.out, 3)
      print(surv.out)
    }

    cat("\n \n")
    cat("Censoring submodel: ", object$censoringDistribution)
    cat("\n \n")
    cens.out <- out.res[(k+1):(l+k),]

    if(is.null(rownames(cens.out))){
      colName <- cbind(c(rep("W",l)),c(0,seq(1:(l-1))))
      colC <- noquote(apply(colName,1,paste, collapse = "_"))
      cens.out <- rbind(cens.out,out.res[(l+k+1),])
      rownames(cens.out) <- c("Intercept",colC[-1], "sigma")
    } else {
      colC <- rownames(cens.out)
      cens.out <- rbind(cens.out,out.res[(l+k+1),])
      rownames(cens.out) <- c("Intercept",colC[-1], "sigma")
      rownames(cens.out) <- gsub("W", "", rownames(cens.out))
    }
    colnames(cens.out) <- c("Estimate", "Boot.SE", "Pvalue")
    print(round(cens.out, 3))


    message(rep("-", 100))
    cat("Assumed copula model:", object$copula)
    cat("\n \n")

    cat("kendall's tau correlation :\n")
    cat("\n")
    tau = round(out.res[(l+k+2),],3)
    names(tau) <- c("tau","Boot.SE", "Pvalue")
    print(tau)
  }else{
    out.res <- object$parameterEstimates

    message(rep("-", 100))
    cat("Survival submodel: Cox proportional hazards model")
    cat("\n \n")
    cat("Parameter estimates:")
    cat("\n")
    surv.out <- out.res[1:k]
    if(k>=2){
      if(is.null(names(surv.out))){
        colName <- cbind(c(rep("X",k)),c(seq(1:k)))

        names(surv.out)<- noquote(apply(colName,1,paste, collapse = "_"))
      }else{
        names(surv.out) <- gsub("X", "", names(surv.out))
      }
    }else if(k==1){
      if(is.null(names(surv.out))){
        names(surv.out) <- "X1"
      }else  names(surv.out) <- names(surv.out)
    }
    print(round(surv.out, 3))

    cat("\n \n")
    cat("Censoring submodel: ", object$censoringDistribution)
    cat("\n \n")
    cens.out <- out.res[(k+1):(l+k+1)]
    if(is.null(names(cens.out))){
      colName <- cbind(c(rep("W",l)),c(seq(1:l)))

      colN <- noquote(apply(colName,1,paste, collapse = "_"))
      names(cens.out) <- c("Intercept",colN[-1],"sigma")
    }else{
      colC <- names(cens.out)
      names(cens.out) <- c("Intercept",colC[-1])
      names(cens.out) <- gsub("W","",names(cens.out))
    }
    print(round(cens.out, 3))

    message(rep("-", 100))
    cat("Assumed copula model:", object$copula)
    cat("\n \n")

    cat("kendall's tau correlation :")
    cat("\n")
    tau = round(out.res[(l+k+2)],3)
    names(tau) <- c("tau")
    print(tau)
  }
}


#' Summary of \code{indepCensoringFit} object
#'
#' @param object Output of \code{\link{fitIndepCens}} function
#' @param ... Further arguments
#'
#' @return Summary of independent censoring model fit in the form of table
#'
#'
#' @export
summary.indepFit <- function(object, ...) {

  message(rep("-", 100))
  message("Summary of independent censoring model")

  k = object$dimX
  l = object$dimW

  if(object$bootstrap){
    stError <- object$parEst.std.error
    stError_ch <- object$cumHazard.std.error
    Z <- object$parameterEstimates/stError
    pvalue <- 2*( 1-pnorm(abs(Z)))
    out.res <- cbind(object$parameterEstimates,stError,pvalue)

    message(rep("-", 100))
    cat("Survival submodel: Cox proportional hazards model")
    cat("\n \n")
    surv.out <- out.res[1:k,]
    if(dim(surv.out)[1]>=2){
      if(is.null(rownames(surv.out))){
        colName <- cbind(c(rep("X",k)),c(seq(1:k)))
        rownames(surv.out) <- noquote(apply(colName,1,paste, collapse = "_"))
      }else {
        rownames(surv.out) <- gsub("X", "", rownames(surv.out))
      }
      colnames(surv.out) <- c("Estimate", "Boot.SE", "Pvalue")
    }
    print(round(surv.out, 3))

    cat("\n \n")
    cat("Censoring submodel: ", object$censoringDistribution)
    cat("\n \n")
    cens.out <- out.res[(k+1):(l+k),]

    if(is.null(rownames(cens.out))){
      colName <- cbind(c(rep("W",l)),c(0,seq(1:(l-1))))
      colC <- noquote(apply(colName,1,paste, collapse = "_"))
      cens.out <- rbind(cens.out,out.res[(l+k+1),])
      rownames(cens.out) <- c("Intercept",colC[-1], "sigma")
    }else {
      colC <- rownames(cens.out)
      cens.out <- rbind(cens.out,out.res[(l+k+1),])
      rownames(cens.out) <- c("Intercept",colC[-1], "sigma")
      rownames(cens.out) <- gsub("W", "", rownames(cens.out))
    }
    colnames(cens.out) <- c("Estimate", "Boot.SE", "Pvalue")
    print(round(cens.out, 3))
  }else{
    out.res <- object$parameterEstimates

    message(rep("-", 100))
    cat("Survival submodel: Cox proportional hazards model")
    cat("\n\n")
    cat("Parameter estimates:")
    cat("\n")
    surv.out <- out.res[1:k]
    if(k>=2){
      if(is.null(names(surv.out))){
        colName <- cbind(c(rep("X",k)),c(seq(1:k)))

        names(surv.out)<- noquote(apply(colName,1,paste, collapse = "_"))
      }else{
        names(surv.out) <- gsub("X", "", names(surv.out))
      }
    }else if(k==1){
      if(is.null(names(surv.out))){
        names(surv.out) <- "X1"
      }else{
        names(surv.out) <- gsub("X", "", names(surv.out))
      }
    }
    print(round(surv.out, 3))
    cat("\n \n")
    # message(rep("-", 100))
    cat("Censoring submodel: ", object$censoringDistribution)
    cat("\n \n")
    cat("Parameter estimates:")
    cat("\n")
    cens.out <- out.res[(k+1):(l+k+1)]
    if(is.null(names(cens.out))){
      colName <- cbind(c(rep("W",l)),c(seq(1:l)))

      colN <- noquote(apply(colName,1,paste, collapse = "_"))
      names(cens.out) <- c("Intercept",colN[-1],"sigma")
    }else{
      colC <- names(cens.out)
      names(cens.out) <- c("Intercept",colC[-1])
      names(cens.out) <- gsub("W","",names(cens.out))
    }
    print(round(cens.out, 3))
  }
}

#' @title Data generation function for competing risks data
#'
#' @description This function generates competing risk data that can be used in
#' simulation studies.
#'
#' @param n The sample size of the generated data set.
#' @param par List of parameter vectors for each component of the transformation
#' model.
#' @param iseed Random seed.
#' @param s The number of competing risks. Note that the given parameter vector
#' could specify the parameters for the model with more than s competing risks,
#' but in this case only the first s sets of parameters will be considered.
#' @param conf Boolean value indicating whether the data set should contain
#' confounding.
#' @param Zbin Indicator whether the confounded variable is binary
#' \code{Zbin = 1} or not \code{Zbin = 0}. If \code{conf = FALSE}, this
#' variable is ignored.
#' @param Wbin Indicator whether the instrument is binary (\code{Zbin = 1}) or
#' not \code{Zbin = 0}.
#' @param type.cov Vector of characters "c" and "b", indicating which exogenous
#' covariates should be continuous \code{"c"} or binary \code{"b"}.
#' @param A.upper The upper bound on the support of the administrative censoring
#' distribution. This can be used to control for the amount of administrative
#' censoring in the data. Default is \code{A.upper = 15}. \code{A.upper = NULL}
#' will return a data set without administrative censoring.
#'
#' @import mvtnorm MASS stats
#'
#' @return A generated data set
#'
#' @noRd
#'

dat.sim.reg.comp.risks = function(n, par, iseed, s, conf, Zbin, Wbin, type.cov,
                                  A.upper = 15) {

  # Check for inconsistencies in the arguments
  n.cov <- length(type.cov) + as.numeric(conf)*2
  n.param <- n.cov + 1
  if (n.param > length(par[[1]])) {
    stop("Too few parameters provided in argument 'par'.")
  }

  # Set randomization seed
  set.seed(iseed)

  # Extract parameters
  sd.idx <- length(par) - 1
  gamma.idx <- length(par)
  beta.mat <- NULL
  for (i in 1:s) {
    beta.mat <- cbind(beta.mat, par[[i]])
  }
  sd <- par[[sd.idx]]
  gamma <- par[[gamma.idx]][1:(nrow(beta.mat) - 1)]

  # If there is no confounding, the parameters for Z and the control function
  # can be ignored.
  if (!conf) {
    beta.mat <- beta.mat[-c(nrow(beta.mat) - 1, nrow(beta.mat)), ]
  }

  # Set some useful parameters
  parl <- nrow(beta.mat)

  # Multivariate normal distribution for error terms
  mu <- rep(0, s)
  sigma <- diag(sd[1:s]^2, nrow = s)
  idx.start <- c(1)
  for (i in 1:(s - 2)) {
    idx.start <- c(idx.start, idx.start[length(idx.start)] + s - i)
  }
  for (i in 1:s) {
    for (j in 1:s) {
      if (i < j) {
        sigma[i, j] <- sd[i]*sd[j]*sd[s + idx.start[i] + j - i - 1]
        sigma[j, i] <- sigma[i, j]
      }
    }
  }
  err <- mvrnorm(n, mu = mu , Sigma = sigma)

  # Intercept
  X <- rep(1, n)

  # Covariates
  for (ct in type.cov) {
    if (ct == "b") {
      X <- cbind(X, sample(c(0, 1), n, replace = TRUE))
    } else {
      X <- cbind(X, rnorm(n = n))
    }
  }
  colnames(X) <- paste0("x", 0:(ncol(X) - 1))

  # If confounded data set, generate confounder and its instrument
  if (conf) {

    # Instrument
    if (Wbin) { # Bernoulli with p = 0.5
      W <- sample(c(0, 1), n, replace = TRUE)
    } else { # Uniform[0, 2]
      W <- runif(n, 0, 2)
    }

    XandW <- as.matrix(cbind(X, W))

    # Confounded variable
    if (Zbin) { # Z is binary
      V <- rlogis(n)
      Z <- as.matrix(as.numeric(XandW %*% gamma - V > 0))
      realV <- (1-Z)*((1+exp(XandW%*%gamma))*log(1+exp(XandW%*%gamma))-(XandW%*%gamma)*exp(XandW%*%gamma))-Z*((1+exp(-(XandW%*%gamma)))*log(1+exp(-(XandW%*%gamma)))+(XandW%*%gamma)*exp(-(XandW%*%gamma)))
    } else { # Z is continuous
      V <- rnorm(n, 0, 2)
      Z <- XandW %*% gamma + V
      realV <- Z - (XandW %*% gamma)
    }

  }

  # Matrix containing all covariates
  Mgen <- X
  if (conf) {
    Mgen <- cbind(Mgen, Z, realV)
  }

  # Latent times
  times <- matrix(nrow = n, ncol = s)
  for (j in 1:s) {
    times[,j] <- IYJtrans(Mgen %*% beta.mat[,j] + err[,j], sd[length(sd) - s + j])
  }
  if (!is.null(A.upper)) { # administrative censoring
    A <- runif(n, 0, A.upper)
    times <- cbind(times, A)
  }

  # Data matrix
  M <- X
  if (conf) {
    M <- cbind(M, Z, W)
  }

  # Observed time
  Y <- apply(X = times, MARGIN = 1, FUN = min)

  # Censoring indicators
  delta <- matrix(nrow = n, ncol = s)
  for (i in 1:s) {
    delta[,i] <- as.numeric(Y == times[,i])
  }
  if (!is.null(A.upper)) {
    delta <- cbind(delta, as.numeric(Y == times[, s + 1]))
  }

  # data consisting of observed time, censoring indicators, all data and the
  # control function.
  data <- cbind(Y, delta, M)
  if (conf) {
    data <- cbind(data, realV)
  }
  data <- as.data.frame(data)
  cens.names <- paste0("delta", 1:s)
  if (!is.null(A.upper)) {cens.names <- c(cens.names, "da")}
  cov.names <- paste0("X", 0:(ncol(X) - 1))
  if (conf) {conf.names <- c("Z", "W", "realV")} else {conf.names <- NULL}
  colnames(data) <- c("Y", cens.names, cov.names, conf.names)

  # Return the result
  return(data)
}

#' @title Competing risk likelihood function.
#'
#' @description This function implements the second step likelihood function of
#' the competing risk model defined in Willems et al. (2024+).
#'
#' @references Willems et al. (2024+). Flexible control function approach under competing risks (in preparation).
#'
#' @param n The sample size.
#' @param s The number of competing risks.
#' @param Y The observed times.
#' @param admin Boolean value indicating whether or not administrative censoring
#' should be taken into account.
#' @param cens.inds matrix of censoring indicators (each row corresponds to a
#' single observation).
#' @param M Design matrix, consisting of [intercept, exo.cov, Z, cf]. Note that
#' \code{cf} represents the multiple ways of 'handling' the endogenous covariate
#' Z, see also the documentation of 'estimate.cmprsk.R'. When there is no
#' confounding, \code{M} will be [intercept, exo.cov].
#' @param Sigma The covariance matrix.
#' @param beta.mat Matrix containing all of the covariate effects.
#' @param sigma.vct Vector of standard deviations. Should be equal to
#' \code{sqrt(diag(Sigma))}.
#' @param rho.mat The correlation matrix.
#' @param theta.vct Vector containing the parameters of the Yeo-Johnsontrans-
#' formations.
#'
#' @import mvtnorm pbivnorm
#' @importFrom stats dnorm
#'
#' @return Evaluation of the log-likelihood function
#'
#' @noRd
#'
cr.lik <- function(n, s, Y, admin, cens.inds, M, Sigma, beta.mat, sigma.vct,
                   rho.mat, theta.vct) {

  # Load dependency
  requireNamespace("OpenMx")

  # In the current implementation, Sigma should always be a positive definite
  # matrix. The if-clause is therefore superfluous.
  if (is.positive.definite(Sigma, tol = 1e-30)) {

    # Compute the (derivate of the) Yeo-Johnson transformations of Y
    transY.T <- matrix(nrow = n, ncol = s)
    DtransY.T <- matrix(nrow = n, ncol = s)
    for (i in 1:s) {
      transY.T[, i] <- YJtrans(Y, theta.vct[i])
      DtransY.T[, i] <- DYJtrans(Y, theta.vct[i])
    }

    # Precompute some values used in obtaining b_T and m_{k, j}.
    Mbeta <- M %*% beta.mat

    # Compute b_Tj/sigma_j, for each j.
    b_T <- (transY.T - Mbeta)/matrix(rep(sigma.vct, n), nrow = n, byrow = TRUE)

    # For each latent time T^j, compute the arguments to be supplied to the
    # multivariate normal distribution function.
    args.f_T <- array(Inf, dim = c(n, s, s + as.numeric(admin)))
    partial.correlations <- array(Inf, dim = c(s, s, s))
    for (j in 1:s) {

      # Compute the value of m_{k.j} for each observation
      mj <- matrix(Inf, nrow = n, ncol = s)
      for (k in 1:s) {
        if (k != j) {
          mj[, k] <- Mbeta[, k] + rho.mat[j, k] * sigma.vct[k] * b_T[, j]
        }
      }
      # Compute s_{k.j} for each observation
      sj <- rep(Inf, s)
      for (k in 1:s) {
        if (k != j) {
          sj[k] <- sigma.vct[k]*(1 - rho.mat[j, k]^2)^(1/2)
        }
      }

      # Compute rho_{kq.j}
      rhoj <- matrix(Inf, nrow = s, ncol = s)
      for (k in 1:s) {
        for (q in 1:s) {
          if ((k != j) & (q != j)) {
            rhoj[k, q] <- (rho.mat[k, q] - rho.mat[j, k]*rho.mat[j, q])/sqrt((1 - rho.mat[j, k]^2)*(1 - rho.mat[j, q]^2))
          }
        }
      }

      # Arguments to be supplied to the multivariate normal distribution
      args.f_Tj <- -(transY.T - mj)/matrix(rep(sj, n), nrow = n, byrow = TRUE)
      args.f_T[, , j] <- args.f_Tj

      # Partial correlations
      partial.correlations[, , j] <- rhoj
    }

    # Also compute the likelihood contributions when the times would correspond to
    # administrative censoring, if necessary
    if (admin) {
      args.f_Ta <- -b_T/matrix(rep(sigma.vct, n), nrow = n, byrow = TRUE)
      args.f_T[, , s + 1] <- args.f_Ta
    }

    # Matrix of arguments for which the multivariate normal distribution should
    # be evaluated.
    args.mvn <- matrix(nrow = n, ncol = s)
    for (i in 1:s) {
      args.mvn[, i] <- apply(args.f_T[, i ,]^cens.inds, 1, prod)
    }

    # Define function to be used to evaluate the multivariate normal distr.
    # function. The choice can greatly impact the computation time.
    if (s == 3) {
      mypmvnorm.cr <- function(arg.vct, cov.mat) {
        pbivnorm(arg.vct[1], arg.vct[2], rho = cov.mat[1, 2]/sqrt(cov.mat[1, 1]*cov.mat[2, 2]))
      }
    } else {
      mypmvnorm.cr <- function(arg.vct, cov.mat) {
        OpenMx::omxMnor(cov.mat, rep(0, length(arg.vct)), rep(-Inf, length(arg.vct)), arg.vct)
      }
    }
    mypmvnorm.ad <- function(arg.vct, cov.mat) {
      OpenMx::omxMnor(cov.mat, rep(0, length(arg.vct)), rep(-Inf, length(arg.vct)), arg.vct)
    }

    # Multivariate normal evaluations
    mvn.evals <- rep(Inf, n)
    cov.mat.array <- array(dim = c(s, s, s + as.numeric(admin)))
    cov.mat.array[, , 1:s] <- partial.correlations
    if (admin) {
      cov.mat.array[, , s + 1] <- rho.mat
    }

    for (row.idx in 1:nrow(args.mvn)) {
      event.type.idx <- which(cens.inds[row.idx, ] == 1)
      if (event.type.idx <= s) { # If it is a competing risks
        cov.mat <- cov.mat.array[-event.type.idx, -event.type.idx, event.type.idx]
        mvn.evals[row.idx] <- mypmvnorm.cr(args.mvn[row.idx, !is.na(args.mvn[row.idx, ])], cov.mat)
      } else {                  # If it is administrative censoring
        cov.mat <- cov.mat.array[, , event.type.idx]
        mvn.evals[row.idx] <- mypmvnorm.ad(args.mvn[row.idx, ], cov.mat)
      }
    }

    # Compute the likelihood contributions
    lik.contr <- rep(Inf, n)
    for (i in 1:n) {
      event.type.idx <- which(cens.inds[i, ] == 1)
      if (event.type.idx <= s) { # If it is a competing risks
        lik.contr[i] <- sigma.vct[event.type.idx]^(-1) * mvn.evals[i] * dnorm(b_T[i, event.type.idx]) * DtransY.T[i, event.type.idx]
      } else {                  # If it is administrative censoring
        lik.contr[i] <- mvn.evals[i]
      }
    }

    # Return the likelihood
    lik.contr <- pmax(lik.contr, 1e-100)
    Logn <- sum(log(lik.contr))
    return(-Logn)
  } else {
    return(2^{60}-1)                # if not positive definite --> return a very large value
  }
}



#' @title First step log-likelihood function for Z continuous
#'
#' @description This function defines the maximum likelihood used to estimate
#' the control function in the case of a continuous endogenous variable.
#'
#' @param gamma Vector of coefficients in the linear model used to estimate the
#' control function.
#' @param Z Endogenous covariate.
#' @param M Design matrix, containing an intercept, the exogenous covariates and
#' the instrumental variable.
#'
#' @return Returns the log-likelihood function corresponding to the data,
#' evaluated at the point \code{gamma}.
#'
#' @noRd
#'
LikGamma1 <- function(gamma, Z, M) {

  # Correctly format the variables
  W <- as.matrix(M)
  gamma <- as.matrix(gamma)

  # Vector of squared errors in the linear regression model
  tot <- (Z - W %*% gamma)^2

  # Prevent numerical issues
  p1 <- pmax(tot, 1e-100)

  # Sum of (log of) squared errors to be minimized.
  Logn <- sum(log(p1));

  # Return the results
  return(Logn)
}



#' @title First step log-likelihood function for Z binary.
#'
#' @description This function defines the maximum likelihood used to estimate
#' the control function in the case of a binary endogenous variable.
#'
#' @param gamma Vector of coefficients in the logistic model used to estimate
#' the control function.
#' @param Z Endogenous covariate.
#' @param M Design matrix, containing an intercept, the exogenous covariates and
#' the instrumental variable.
#'
#' @importFrom stats plogis
#'
#' @return Returns the log-likelihood function corresponding to the data,
#' evaluated at the point \code{gamma}.
#'
#' @noRd
#'
LikGamma2 <- function(gamma, Z, M) {

  # Correcly format the variables
  W <- as.matrix(M)
  gamma <- as.matrix(gamma)

  # Factors in likelihood function of logistic regression
  tot <- (plogis(W %*% gamma)^Z)*((1 - plogis(W %*% gamma))^(1-Z))

  # Prevent numerical issues
  p1 <- pmax(tot, 1e-100)

  # Log-likelihood function to be maximized when finding gamma
  Logn <- sum(log(p1))

  # In order to be consistent with 'LikGamma1.R' (which should be minimized to
  # find gamma), we return the negative of the log-likelihood, which should also
  # be minimized to find gamma.
  return(-Logn)
}



#' @title Second step log-likelihood function.
#'
#' @description This function defines the log-likelihood used to estimate
#' the second step in the competing risks extension of the model described in
#' Willems et al. (2024+).
#'
#'@references Willems et al. (2024+). Flexible control function approach under competing risks (in preparation).
#'
#' @param par Vector of all second step model parameters, consisting of the
#' regression parameters, variance-covariance matrix elements and transformation
#' parameters.
#' @param data Data frame resulting from the 'uniformize.data.R' function.
#' @param admin Boolean value indicating whether the data contains
#' administrative censoring.
#' @param conf Boolean value indicating whether the data contains confounding
#' and hence indicating the presence of Z and W.
#' @param cf "Control function" to be used. This can either be the (i) estimated
#' control function, (ii) the true control function, (iii) the instrumental
#' variable, or (iv) nothing (\code{cf = NULL}). Option (ii) is used when
#' comparing the two-step estimator to the oracle estimator, and option (iii) is
#' used to compare the two-step estimator with the naive estimator.
#'
#' @import mvtnorm pbivnorm
#'
#' @return Log-likelihood evaluation of the second step.
#'
#' @noRd
#'
LikF.cmprsk <- function(par, data, admin, conf, cf) {

  #### Unpack data ####

  # Observed times
  Y <- data[, "y"]

  # Censoring indicators
  cens.inds <- data[, grepl("delta[[:digit:]]+", colnames(data)), drop = FALSE]
  if (admin) {
    cens.inds <- cbind(cens.inds, data[, which(colnames(data) == "da")])
  }

  # Intercept
  intercept <- data[, "x0"]

  # Exogenous covariates
  exo.cov <- data[, grepl("x[1-9][[:digit:]]*", colnames(data)),
                  drop = FALSE]

  # Endogenous covariate
  if (conf) {
    Z <- data[, "z"]
  } else {
    Z <- NULL
    cf <- NULL
  }

  # Design matrix: [intercept, exogenous covariates, Z, cf]
  M <- as.matrix(cbind(intercept, as.matrix(exo.cov), Z, cf))

  #### Unpack parameters ####

  # Define some useful variables
  k <- ncol(M)                                             # Nbr. parameters
  s <- ifelse(admin, ncol(cens.inds) - 1, ncol(cens.inds)) # Nbr. comp. risks
  n <- nrow(M)                                             # Nbr. observations

  # Extract parameters for beta
  beta.vct <- par[1:(k*s)]
  beta.mat <- matrix(nrow = k, ncol = s)
  for (i in 1:s) {
    beta.mat[,i] <- beta.vct[((i - 1)*k + 1):(i*k)]
  }

  # Extract parameters for the variance-covariance matrix.
  # Structure of sd-vector:
  # [1:s]       -> standerd deviations
  # [(s+1):end] -> correlation; stored as
  #                [rho_{12}, rho_{13}, ..., rho_{1d}, rho_{23}, rho_{24}, ...]
  sd <- par[(k*s + 1):(length(par) - s)]

  # Construct correlation and covariance matrix
  rho.mat <- matrix(1, nrow = s, ncol = s)
  rho.mat[lower.tri(rho.mat, diag = FALSE)] <- sd[(s + 1):length(sd)]
  rho.mat[upper.tri(rho.mat, diag = FALSE)] <- t(rho.mat)[upper.tri(rho.mat, diag = FALSE)]
  Sigma <- rho.mat * outer(sd[1:s], sd[1:s], "*")
  sigma.vct <- sd[1:s]

  # Extract parameters for transformation function
  theta.vct <- par[(length(par) - s + 1):(length(par))]

  #### Compute log-likelihood evaluation ####

  cr.lik(n, s, Y, admin, cens.inds, M, Sigma, beta.mat, sigma.vct, rho.mat, theta.vct)
}



#' @title Wrapper implementing likelihood function using Cholesky factorization.
#'
#' @description This function parametrizes the covariance matrix using its
#' Cholesky decomposition, so that optimization of the likelihood can be done
#' based on this parametrization, and positive-definiteness of the covariance
#' matrix is guaranteed at each step of the optimization algorithm.
#'
#' @param par.chol Vector of all second step model parameters, consisting of the
#' regression parameters, Cholesky decomposition of the variance-covariance
#' matrix elements and transformation parameters.
#' @param data Data frame resulting from the 'uniformize.data.R' function.
#' @param admin Boolean value indicating whether the data contains
#' administrative censoring.
#' @param conf Boolean value indicating whether the data contains confounding
#' and hence indicating the presence of Z and W.
#' @param cf "Control function" to be used. This can either be the (i) estimated
#' control function, (ii) the true control function, (iii) the instrumental
#' variable, or (iv) nothing (\code{cf = NULL}). Option (ii) is used when
#' comparing the two-step estimator to the oracle estimator, and option (iii) is
#' used to compare the two-step estimator with the naive estimator.
#' @param eps Minimum value for the diagonal elements in the covariance matrix.
#' Default is \code{eps = 0.001}.
#'
#' @import mvtnorm pbivnorm
#'
#' @return Log-likelihood evaluation of the second step.
#'
#' @noRd
#'
likF.cmprsk.Cholesky <- function(par.chol, data, admin, conf, cf, eps = 0.001) {

  #### Unpack data ####

  # Censoring indicators
  cens.inds <- data[, grepl("delta[[:digit:]]+", colnames(data)), drop = FALSE]
  if (admin) {
    cens.inds <- cbind(cens.inds, data[, which(colnames(data) == "da")])
  }

  # Intercept
  intercept <- data[, "x0"]

  # Exogenous covariates
  exo.cov <- data[, grepl("x[1-9][[:digit:]]*", colnames(data)),
                  drop = FALSE]

  # Endogenous covariate and control function
  if (conf) {
    Z <- data[, "z"]
  } else {
    Z <- NULL
    cf <- NULL
  }

  # Design matrix: [intercept, exogenous covariates, Z, estimated cf]
  M <- as.matrix(cbind(intercept, as.matrix(exo.cov), Z, cf))

  #### Transform Cholesky factorization parameters ####

  # Define some useful variables
  k <- ncol(M)                                             # Nbr. parameters
  s <- ifelse(admin, ncol(cens.inds) - 1, ncol(cens.inds)) # Nbr. comp. risks

  # Extract parameters for the variance-covariance matrix. Note that in this
  # case, the parameters of the Cholesky factorization are provided.
  sd.chol <- par.chol[(k*s + 1):(length(par.chol) - s)]

  # Reconstruct upper diagonal matrix of Cholesky factorization
  chol.U <- matrix(0, nrow = s, ncol = s)
  chol.U[lower.tri(chol.U, diag = TRUE)] <- sd.chol

  # Construct variance-covariance matrix
  Sigma <- chol.U %*% t(chol.U)
  diag(Sigma) <- diag(Sigma) + eps

  # Correlation matrix
  rho.mat <- Sigma / outer(sqrt(diag(Sigma)), sqrt(diag(Sigma)), FUN = "*")

  # Get the variances and covariances
  sd <- rep(0, length(sd.chol))
  sd[1:s] <- sqrt(diag(Sigma))
  sd[(s+1):length(sd)] <- t(rho.mat)[lower.tri(rho.mat)]

  # Contruct vector to be supplied to likF.cmprsk.R
  par <- par.chol
  par[(k*s + 1):(length(par) - s)] <- sd

  #### Run regular likelihood function ####

  LikF.cmprsk(par, data, admin, conf, cf)
}



#' @title Second likelihood function needed to fit the independence model in the
#' second step of the estimation procedure.
#'
#' @description This function defines the log-likelihood used in estimating
#' the second step in the competing risks extension of the model described in
#' Willems et al. (2024+). The results of this function will serve as
#' starting values for subsequent optimizations (LikI.comprsk.R and
#' LikF.cmprsk.R)
#'
#' @references Willems et al. (2024+). Flexible control function approach under competing risks (in preparation).
#'
#' @param par Vector of all second step model parameters, consisting of the
#' regression parameters, variance-covariance matrix elements and transformation
#' parameters.
#' @param data Data frame resulting from the 'uniformize.data.R' function.
#' @param admin Boolean value indicating whether the data contains
#' administrative censoring.
#' @param conf Boolean value indicating whether the data contains confounding
#' and hence indicating the presence of Z and W
#' @param cf "Control function" to be used. This can either be the (i) estimated
#' control function, (ii) the true control function, (iii) the instrumental
#' variable, or (iv) nothing (\code{cf = NULL}). Option (ii) is used when
#' comparing the two-step estimator to the oracle estimator, and option (iii) is
#' used to compare the two-step estimator with the naive estimator.
#'
#' @importFrom stats dnorm
#'
#' @return Starting values for subsequent optimization function used in the
#' second step of the estimation procedure.
#'
#' @noRd
#'
LikI.bis <- function(par, data, admin, conf, cf) {

  #### Unpack data ####

  # Observed times
  Y <- data[, "y"]

  # Censoring indicators
  cens.inds <- data[, grepl("delta[[:digit:]]+", colnames(data)), drop = FALSE]
  cens.inds <- cbind(cens.inds, data[, which(colnames(data) == "da")])

  # Intercept
  intercept <- data[, "x0"]

  # Exogenous covariates
  exo.cov <- data[, grepl("x[1-9][[:digit:]]*", colnames(data)),
                  drop = FALSE]

  # Endogenous covariate
  if (conf) {
    Z <- data[, "z"]
  } else {
    Z <- NULL
    cf <- NULL
  }

  # Design matrix: [intercept, exogenous covariates, Z, cf]
  M <- as.matrix(cbind(intercept, as.matrix(exo.cov), Z, cf))

  #### Unpack parameters ####

  # Define some useful variables
  k <- ncol(M)
  s <- ifelse(admin, ncol(cens.inds) - 1, ncol(cens.inds))
  n <- nrow(M)

  # Extract parameters for beta
  beta.vct <- par[1:(k*s)]
  beta.mat <- matrix(nrow = k, ncol = s)
  for (i in 1:s) {
    beta.mat[,i] <- beta.vct[((i - 1)*k + 1):(i*k)]
  }

  # Extract parameters for the variance-covariance matrix
  sigma.vct <- par[(k*s + 1):(k*s + s)]

  # Construct variance-covariance matrix (in this function, all correlations are
  # assumed to equal zero)
  Sigma <- diag(sigma.vct[1:s]^2, nrow = s)

  # Extract parameters for transformation function
  theta.vct <- par[(k*s + s + 1):(length(par))]

  #### Compute the likelihood function evaluation ####

  # For computational efficiency, we de not use the function 'cr.lik.contr.R' to
  # obtain the likelihood contributions. Computational speed-up can be achieved
  # by exploiting the independence assumption, which allows to evaluate the
  # multivariate normal distribution in a more efficient way.

  # Compute the (derivate of the) Yeo-Johnson transformations of Y
  transY.T <- matrix(nrow = n, ncol = s)
  DtransY.T <- matrix(nrow = n, ncol = s)
  for (i in 1:s) {
    transY.T[, i] <- YJtrans(Y, theta.vct[i])
    DtransY.T[, i] <- DYJtrans(Y, theta.vct[i])
  }

  # Compute b_Tj, for each j.
  b_T <- matrix(nrow = n, ncol = s)
  for (i in 1:s) {
    b_T[, i] <- (transY.T[,i] - (M %*% beta.mat[,i]))/sigma.vct[i]
  }

  # Precompute the values m_{k, j}. Note that these values are equal for all j
  # because of the independence (second equation on page 22).
  m <- M %*% beta.mat

  # For each latent time T^j, compute the likelihood contribution
  f_T <- matrix(nrow = n, ncol = s)
  for (j in 1:s) {
    mj <- m
    mj[, j] <- Inf
    args <- (transY.T - mj)/matrix(rep(sigma.vct, n), nrow = n, byrow = TRUE)
    f_T[,j] <- 1/sigma.vct[j] * dnorm(b_T[, j]) * DtransY.T[, j] *
      apply(pnorm(-args), 1, prod)
  }

  # Also compute the likelihood contributions when the times would correspond to
  # administrative censoring
  if (admin) {
    args <- b_T/matrix(rep(sigma.vct, n), nrow = n, byrow = TRUE)
    f_Ta <- apply(pnorm(-args), 1, prod)
    f_T <- cbind(f_T, f_Ta)
  }

  tot <- apply(f_T^cens.inds, 1, prod)
  p1 <- pmax(tot, 1e-100)
  Logn <- sum(log(p1))
  return(-Logn)
}



#' @title Second step log-likelihood function under independence assumption.
#'
#' @description This function defines the log-likelihood used to estimate
#' the second step in the competing risks extension assuming independence of
#' some of the competing risks in the model described in Willems et al. (2024+).
#'
#' @references Willems et al. (2024+). Flexible control function approach under competing risks (in preparation).
#'
#' @param par Vector of all second step model parameters, consisting of the
#' regression parameters, variance-covariance matrix elements and transformation
#' parameters.
#' @param data Data frame resulting from the 'uniformize.data.R' function.
#' @param eoi.indicator.names Vector of names of the censoring indicator columns
#' pertaining to events of interest. Events of interest will be modeled allowing
#' dependence between them, whereas all censoring events (corresponding to
#' indicator columns not listed in \code{eoi.indicator.names}) will be treated
#' as independent of every other event.
#' @param admin Boolean value indicating whether the data contains
#' administrative censoring.
#' @param conf Boolean value indicating whether the data contains confounding
#' and hence indicating the presence of Z and W
#' @param cf "Control function" to be used. This can either be the (i) estimated
#' control function, (ii) the true control function, (iii) the instrumental
#' variable, or (iv) nothing (\code{cf = NULL}). Option (ii) is used when
#' comparing the two-step estimator to the oracle estimator, and option (iii) is
#' used to compare the two-step estimator with the naive estimator.
#'
#' @import mvtnorm pbivnorm
#'
#' @return Log-likelihood evaluation for the second step in the esimation
#' procedure.
#'
#' @noRd
#'
LikI.cmprsk <- function(par, data, eoi.indicator.names, admin, conf, cf) {

  #### Unpack data ####

  # Observed times
  Y <- data[, "y"]

  # Censoring indicators
  cens.inds <- data[, grepl("delta[[:digit:]]+", colnames(data)), drop = FALSE]
  if (admin) {
    cens.inds <- cbind(cens.inds, data[, which(colnames(data) == "da")])
  }

  # Intercept
  intercept <- data[, "x0"]

  # Exogenous covariates
  exo.cov <- data[, grepl("x[1-9][[:digit:]]*", colnames(data)),
                  drop = FALSE]

  # Endogenous covariate and instrumental variable
  if (conf) {
    Z <- data[, "z"]
  } else {
    Z <- NULL
    cf <- NULL
  }

  # Design matrix: [intercept, exogenous covariates, Z, estimated cf]
  M <- as.matrix(cbind(intercept, as.matrix(exo.cov), Z, cf))

  #### Unpack parameters ####

  # Define some useful variables
  k <- ncol(M)                                             # Nbr. parameters
  s <- ifelse(admin, ncol(cens.inds) - 1, ncol(cens.inds)) # Nbr. comp. risks
  n <- nrow(M)                                             # Nbr. observations

  # Extract parameters for beta
  beta.vct <- par[1:(k*s)]
  beta.mat <- matrix(nrow = k, ncol = s)
  for (i in 1:s) {
    beta.mat[,i] <- beta.vct[((i - 1)*k + 1):(i*k)]
  }

  # Extract parameters for the variance-covariance matrix. Note that in this
  # case, the correlations between the competing risks and the censoring is zero.
  # Structure of sd-vector:
  # [1:s]       -> standerd deviations
  # [(s+1):end] -> correlation; stored as
  #                [rho_{ij}, rho_{ik}, ..., rho_{iq} rho_{jk}, ...],
  #                where eoi.inds = c(i, j, k, ..., q)
  sd <- par[(k*s + 1):(length(par) - s)]

  # Quick compatibility check
  eoi.inds <- as.numeric(gsub("delta", "", eoi.indicator.names))
  if (length(sd) != s + choose(length(eoi.inds), 2)) {
    stop("Supplied parameter vector is not consistent with necessary model parameters.")
  }

  # Construct correlation and covariance matrix
  rho.mat <- matrix(0, nrow = s, ncol = s)
  diag(rho.mat) <- 1
  sd.idx <- s + 1
  for (i in eoi.inds) {
    for (j in eoi.inds) {
      if (i < j) {
        rho.mat[i, j] <- sd[sd.idx]
        rho.mat[j, i] <- sd[sd.idx]
        sd.idx <- sd.idx + 1
      }
    }
  }
  Sigma <- rho.mat * outer(sd[1:s], sd[1:s], FUN = "*")
  sigma.vct <- sd[1:s]

  # Extract parameters for transformation function
  theta.vct <- par[(length(par) - s + 1):(length(par))]

  #### Compute log-likelihood evaluation ####

  cr.lik(n, s, Y, admin, cens.inds, M, Sigma, beta.mat, sigma.vct, rho.mat, theta.vct)

}



#' @title Wrapper implementing likelihood function assuming independence between
#' competing risks and censoring using Cholesky factorization.
#'
#' @description This function does the same as LikI.cmprsk (in fact, it even
#' calls said function), but it parametrizes the covariance matrix using its
#' Cholesky decomposition in order to guarantee positive definiteness. This
#' function is never used, might not work and could be deleted.
#'
#' @param par.chol Vector of all second step model parameters, consisting of the
#' regression parameters, Cholesky decomposition of the variance-covariance
#' matrix elements and transformation parameters.
#' @param data Data frame resulting from the 'uniformize.data.R' function.
#' @param eoi.indicator.names Vector of names of the censoring indicator columns
#' pertaining to events of interest. Events of interest will be modeled allowing
#' dependence between them, whereas all censoring events (corresponding to
#' indicator columns not listed in \code{eoi.indicator.names}) will be treated
#' as independent of every other event.
#' @param admin Boolean value indicating whether the data contains
#' administrative censoring.
#' @param conf Boolean value indicating whether the data contains confounding
#' and hence indicating the presence of Z and W.
#' @param cf "Control function" to be used. This can either be the (i) estimated
#' control function, (ii) the true control function, (iii) the instrumental
#' variable, or (iv) nothing (\code{cf = NULL}). Option (ii) is used when
#' comparing the two-step estimator to the oracle estimator, and option (iii) is
#' used to compare the two-step estimator with the naive estimator.
#' @param eps Minimum value for the diagonal elements in the covariance matrix.
#' Default is \code{eps = 0.001}.
#'
#' @import mvtnorm pbivnorm
#'
#' @return Log-likelihood evaluation for the second step in the estimation
#' procedure.
#'
#' @noRd
#'
LikI.cmprsk.Cholesky <- function(par.chol, data, eoi.indicator.names, admin,
                                 conf, cf, eps = 0.001) {

  ## Unpack data ##

  # Censoring indicators
  cens.inds <- data[, grepl("delta[[:digit:]]+", colnames(data)), drop = FALSE]
  if (admin) {
    cens.inds <- cbind(cens.inds, data[, which(colnames(data) == "da")])
  }

  # Intercept
  intercept <- data[, "x0"]

  # Exogenous covariates
  exo.cov <- data[, grepl("x[1-9][[:digit:]]*", colnames(data)),
                  drop = FALSE]

  # Endogenous covariate and control function
  if (conf) {
    Z <- data[, "z"]
  } else {
    Z <- NULL
    cf <- NULL
  }

  # Design matrix: [intercept, exogenous covariates, Z, estimated cf]
  M <- as.matrix(cbind(intercept, as.matrix(exo.cov), Z, cf))

  #### Transform Cholesky factorization parameters ####

  # Define some useful variables
  k <- ncol(M)                                             # Nbr. parameters
  s <- ifelse(admin, ncol(cens.inds) - 1, ncol(cens.inds)) # Nbr. comp. risks

  # Extract parameters for the variance-covariance matrix. Note the particular
  # form of the covariance matrix, which is (assuming for simplicity of
  # representation that the first s1 events are the events of interest):
  #
  #      s1        s2
  #  ---------  ---------
  # [s x ... x  0 0 ... 0]
  #           ...
  # [x s ... x  0 0 ... 0]
  #
  # [0 0 ... 0  s 0 ... 0]
  # [0 0 ... 0  0 s ... 0]
  #           ...
  # [0 0 ... 0  0 0 ... s]
  #
  # where 'x' represents a possibly non-zero element and 's' represents a
  # positive element

  # Let s1 equal the number of events of interest. The vector sd.chol has the
  # following structure:
  #
  # [1:(choolse(s1, 2) + s1]: Parameters of the Cholesky decomposition of the
  #                           sub matrix of the covariance matrix pertaining to
  #                           the events of interest.
  # [choose(s1, 2) + s1:end]: Standard deviations pertaining to the events that
  #                           are to be modeled independently from every other
  #                           event.
  sd.chol <- par.chol[(k*s + 1):(length(par.chol) - s)]

  # Extract parameters for the events of interest
  s1 <- length(eoi.indicator.names)
  sd1.chol <- sd.chol[1:(choose(s1, 2) + s1)]
  chol1.U <- matrix(0, nrow = s1, ncol = s1)
  chol1.U[lower.tri(chol1.U, diag = TRUE)] <- sd1.chol
  Sigma1 <- chol1.U %*% t(chol1.U) + diag(eps, nrow = s1)
  sd1 <- sqrt(diag(Sigma1))
  rho.mat1 <- Sigma1 / outer(sd1, sd1)
  cor1 <- t(rho.mat1)[lower.tri(rho.mat1)]

  # Extract parameters for the independent events
  sd2 <- NULL
  if (s1 < s) {
    sd2 <- sd.chol[(choose(s1, 2) + s1 + 1):length(sd.chol)]
  }

  sd <- c(sd1, sd2, cor1)

  # Construct vector to be supplied to likI.cmprsk.R
  par <- par.chol
  par[(k*s + 1):(length(par.chol) - s)] <- sd

  #### Run regular likelihood function ####

  LikI.cmprsk(par, data, eoi.indicator.names, admin, conf, cf)
}



#' @title Full likelihood (including estimation of control function).
#'
#' @description This function defines the 'full' likelihood of the model.
#' Specifically, it includes the estimation of the control function in the
#' computation of the likelihood. This function is used in the estimation of the
#' variance of the estimates (variance.cmprsk.R).
#'
#' @param parhatG The full parameter vector.
#' @param data Data frame.
#' @param eoi.indicator.names Vector of names of the censoring indicator columns
#' pertaining to events of interest. Events of interest will be modeled allowing
#' dependence between them, whereas all censoring events (corresponding to
#' indicator columns not listed in \code{eoi.indicator.names}) will be treated
#' as independent of every other event. If \code{eoi.indicator.names == NULL},
#' all events will be modelled dependently.
#' @param admin Boolean value indicating whether the data contains
#' administrative censoring.
#' @param conf Boolean value indicating whether the data contains confounding
#' and hence indicating the presence of Z and W.
#' @param Zbin Boolean value indicating whether the confounding variable is
#' binary.
#' @param inst Type of instrumental function to be used.
#'
#' @import mvtnorm pbivnorm
#'
#' @return Full model log-likelihood evaluation.
#'
#' @noRd
#'
likIFG.cmprsk.Cholesky <- function(parhatG, data, eoi.indicator.names, admin,
                                   conf, Zbin, inst) {

  #### Precondition checks ####

  if (!conf) {
    stop("Only implemented when there is confounding in the data set.")
  }
  if (inst != "cf") {
    stop("Only implemented when the control function should be estimated.")
  }

  #### Extract parameters ####

  # Censoring indicators
  cens.inds <- data[, grepl("delta[[:digit:]]+", colnames(data)), drop = FALSE]
  if (admin) {
    cens.inds <- cbind(cens.inds, data[, which(colnames(data) == "da")])
  }

  # Design matrix of exogenous covariates
  intercept <- data[, "x0"]
  exo.cov <- data[, grepl("x[1-9][[:digit:]]*", colnames(data)), drop = FALSE]
  W <- data[, "w"]
  XandW <- as.matrix(cbind(intercept, exo.cov, W))

  # Endogenous covariate
  Z <- data[, "z"]

  # Set useful variables and extract parameter vector
  k <- ncol(XandW) + 1                                     # Nbr. parameters
  s <- ifelse(admin, ncol(cens.inds) - 1, ncol(cens.inds)) # Nbr. comp. risks
  par.chol <- parhatG[1:(length(parhatG) - k + 1)]
  gammaest <- parhatG[(length(parhatG) - k + 2):(length(parhatG))]

  #### Estimate control function ####

  cf <- estimate.cf(XandW, Z, Zbin, gammaest)$cf

  #### Return full likelihood ####

  if (length(eoi.indicator.names) == s) {
    likF.cmprsk.Cholesky(par.chol, data, admin, conf, cf)
  } else {
    LikI.cmprsk.Cholesky(par.chol, data, eoi.indicator.names, admin, conf, cf)
  }

}



#' @title Estimate the control function
#'
#' @description This function estimates the control function for the endogenous
#' variable based on the provided covariates. This function is called inside
#' estimate.cmprsk.R.
#'
#' @param XandW Design matrix of exogenous covariates.
#' @param Z Endogenous covariate.
#' @param Zbin Boolean value indicating whether endogenous covariate is binary.
#' @param gammaest Vector of pre-estimated parameter vector. If \code{NULL},
#' this function will first estimate \code{gammaest}. Default value is
#' \code{gammaest = NULL}.
#' @import nloptr
#' @importFrom stats lm
#'
#' @return List containing the vector of values for the control function and
#' the regression parameters of the first step.
#'
#' @noRd
#'
estimate.cf <- function(XandW, Z, Zbin, gammaest = NULL) {

  # Define useful parameter
  parlgamma <- ncol(XandW)

  # Fit appropriate model for Z and obtain control function
  if (Zbin) {
    if (is.null(gammaest)) {
      gammaest <- nloptr(x0 = rep(0, parlgamma),
                         eval_f = LikGamma2,
                         Z = Z,
                         M = XandW,
                         lb = c(rep(-Inf, parlgamma)),
                         ub = c(rep(Inf, parlgamma)),
                         eval_g_ineq = NULL,
                         opts = list(algorithm = "NLOPT_LN_BOBYQA",
                                     "ftol_abs" = 1.0e-30,
                                     "maxeval" = 100000,
                                     "xtol_abs" = rep(1.0e-30))
      )$solution
    }
    cf <- (1-Z)*((1+exp(XandW%*%gammaest))*log(1+exp(XandW%*%gammaest))-(XandW%*%gammaest)*exp(XandW%*%gammaest))-Z*((1+exp(-(XandW%*%gammaest)))*log(1+exp(-(XandW%*%gammaest)))+(XandW%*%gammaest)*exp(-(XandW%*%gammaest)))
  } else {
    if (is.null(gammaest)) {
      gammaest <- lm(Z~XandW - 1)$coefficients
    }
    cf <- Z-(XandW%*%gammaest)
  }

  # Return the results
  list(cf = cf, gammaest = gammaest)
}



#' @title Estimate the competing risks model of Willems et al. (2025).
#'
#' @description This function estimates the parameters in the competing risks
#' model described in Willems et al. (2025). Note that this model
#' extends the model of Crommen, Beyhum and Van Keilegom (2024) and as such, this
#' function also implements their methodology.
#'
#' @references Willems et al. (2025). Flexible control function approach under competing risks (submitted).
#' @references Crommen, G., Beyhum, J., and Van Keilegom, I. (2024). An instrumental variable approach under dependent censoring. Test, 33(2), 473-495.
#'
#' @param data A data frame, adhering to the following formatting rules:
#' \itemize{
#'    \item The first column, named \code{"y"}, contains the observed times.
#'    \item The next columns, named \code{"delta1"}, \code{delta2}, etc. contain
#'    the indicators for each of the competing risks.
#'    \item The next column, named \code{da}, contains the censoring indicator
#'    (independent censoring).
#'    \item The next column should be a column of all ones (representing the
#'    intercept), names \code{x0}.
#'    \item The subsequent columns should contain the values of the covariates,
#'    named \code{x1}, \code{x2}, etc.
#'    \item When applicable, the next column should contain the values of the
#'    endogenous variable. This column should be named \code{z}.
#'    \item When \code{z} is provided and an instrument for \code{z} is
#'    available, the next column, named \code{w}, should contain the values for
#'    the instrument.
#' }
#' @param admin Boolean value indicating whether the data contains
#' administrative censoring.
#' @param conf Boolean value indicating whether the data contains confounding
#' and hence indicating the presence of \code{z} and, possibly, \code{w}.
#' @param eoi.indicator.names Vector of names of the censoring indicator columns
#' pertaining to events of interest. Events of interest will be modeled allowing
#' dependence between them, whereas all censoring events (corresponding to
#' indicator columns not listed in \code{eoi.indicator.names}) will be treated
#' as independent of every other event. If \code{eoi.indicator.names == NULL},
#' all events will be modeled dependently.
#' @param Zbin Indicator value indicating whether (\code{Zbin = TRUE}) or not
#' \code{Zbin = FALSE} the endogenous covariate is binary. Default is
#' \code{Zbin = NULL}, corresponding to the case when \code{conf == FALSE}.
#' @param inst Variable encoding which approach should be used for dealing with
#' the confounding. \code{inst = "cf"} indicates that the control function
#' approach should be used. \code{inst = "W"} indicates that the instrumental
#' variable should be used 'as is'. \code{inst = "None"} indicates that Z will
#' be treated as an exogenous covariate. Finally, when \code{inst = "oracle"},
#' this function will access the argument \code{realV} and use it as the values
#' for the control function. Default is \code{inst = "cf"}.
#' @param realV Vector of numerics with length equal to the number of rows in
#' \code{data}. Used to provide the true values of the instrumental function
#' to the estimation procedure.
#' @param compute.var Boolean value indicating whether the variance of the
#' parameter estimates should be computed as well (this can be very
#' computationally intensive, so may want to be disabled). Default is
#' \code{estimate.var = TRUE}.
#' @param eps Value that will be added to the diagonal of the covariance matrix
#' during estimation in order to ensure strictly positive variances.
#'
#' @import matrixcalc nloptr numDeriv pbivnorm mvtnorm stats
#' @importFrom MASS mvrnorm
#'
#' @return A list of parameter estimates in the second stage of the estimation
#' algorithm (hence omitting the estimates for the control function), as well
#' as an estimate of their variance and confidence intervals.
#'
#' @examples
#' \donttest{
#'
#' n <- 200
#'
#' # Set parameters
#' gamma <- c(1, 2, 1.5, -1)
#' theta <- c(0.5, 1.5)
#' eta1 <- c(1, -1, 2, -1.5, 0.5)
#' eta2 <- c(0.5, 1, 1, 3, 0)
#'
#' # Generate exogenous covariates
#' x0 <- rep(1, n)
#' x1 <- rnorm(n)
#' x2 <- rbinom(n, 1, 0.5)
#'
#' # Generate confounder and instrument
#' w <- rnorm(n)
#' V <- rnorm(n, 0, 2)
#' z <- cbind(x0, x1, x2, w) %*% gamma + V
#' realV <- z - (cbind(x0, x1, x2, w) %*% gamma)
#'
#' # Generate event times
#' err <- MASS::mvrnorm(n, mu = c(0, 0), Sigma =
#' matrix(c(3, 1, 1, 2), nrow = 2, byrow = TRUE))
#' bn <- cbind(x0, x1, x2, z, realV) %*% cbind(eta1, eta2) + err
#' Lambda_T1 <- bn[,1]; Lambda_T2 <- bn[,2]
#' x.ind = (Lambda_T1>0)
#' y.ind <- (Lambda_T2>0)
#' T1 <- rep(0,length(Lambda_T1))
#' T2 <- rep(0,length(Lambda_T2))
#' T1[x.ind] = ((theta[1]*Lambda_T1[x.ind]+1)^(1/theta[1])-1)
#' T1[!x.ind] = 1-(1-(2-theta[1])*Lambda_T1[!x.ind])^(1/(2-theta[1]))
#' T2[y.ind] = ((theta[2]*Lambda_T2[y.ind]+1)^(1/theta[2])-1)
#' T2[!y.ind] = 1-(1-(2-theta[2])*Lambda_T2[!y.ind])^(1/(2-theta[2]))
#' # Generate adminstrative censoring time
#' C <- runif(n, 0, 40)
#'
#' # Create observed data set
#' y <- pmin(T1, T2, C)
#' delta1 <- as.numeric(T1 == y)
#' delta2 <- as.numeric(T2 == y)
#' da <- as.numeric(C == y)
#' data <- data.frame(cbind(y, delta1, delta2, da, x0, x1, x2, z, w))
#' colnames(data) <- c("y", "delta1", "delta2", "da", "x0", "x1", "x2", "z", "w")
#'
#' # Estimate the model
#' admin <- TRUE                # There is administrative censoring in the data.
#' conf <- TRUE                 # There is confounding in the data (z)
#' eoi.indicator.names <- NULL  # We will not impose that T1 and T2 are independent
#' Zbin <- FALSE                # The confounding variable z is not binary
#' inst <- "cf"                 # Use the control function approach
#' compute.var <- TRUE          # Variance of estimates should be computed.
#' # Since we don't use the oracle estimator, this argument is ignored anyway
#' realV <- NULL
#' estimate.cmprsk(data, admin, conf, eoi.indicator.names, Zbin, inst, realV,
#'                 compute.var)
#'
#' }
#'
#' @export
#'
estimate.cmprsk <- function(data, admin, conf,
                            eoi.indicator.names = NULL, Zbin = NULL,
                            inst = "cf", realV = NULL, compute.var = TRUE,
                            eps = 0.001) {

  #### Conformity checks ####

  if (conf) {
    if (!is.logical(Zbin)) {
      stop("When conf = TRUE, a valid value for Zbin should be supplied.")
    }
    if (!("z" %in% colnames(data))) {
      stop("If there is confounding, the confounding variable should be names 'z'.")
    }
    if (!((inst %in% c("cf", "W", "None")) | is.vector(inst))) {
      stop("When 'conf' is TRUE, inst should be 'cf', 'W', 'None' or a vector (see documentation).")
    }
  } else {
    if (!is.null(Zbin)) {
      stop("Argument mismatch: Zbin is supplied (not NULL), yet conf = FALSE.")
    }
  }
  if ((inst == "oracle") & (is.null(realV))) {
    stop("When inst == oracle, a vector of values to be used as the control function should be provided through realV.")
  }

  # Indicator whether to use Cholesky decomposition for estimating the covariance
  # matrix. (Function will currently throw an error if use.chol == FALSE).
  use.chol <- TRUE

  #### Unpack data ####

  # Observed times
  Y <- data[, "y"]

  # Censoring indicators
  cens.inds <- data[, grepl("delta[[:digit:]]+", colnames(data)), drop = FALSE]
  cens.inds <- cbind(cens.inds, data[, which(colnames(data) == "da")])

  # Intercept
  intercept <- data[, "x0"]

  # Exogenous covariates
  exo.cov <- data[, grepl("x[1-9][[:digit:]]*", colnames(data)),
                  drop = FALSE]

  # Endogenous covariate and Instrumental variable
  if (conf) {
    Z <- data[, "z"]
    W <- data[, "w"]
  }

  # Indices of the events of interest
  if (is.null(eoi.indicator.names)) {
    eoi.indicator.names <- colnames(data)[grepl("delta[[:digit:]]", colnames(data))]
  }
  eoi.inds <- as.numeric(gsub("delta", "", eoi.indicator.names))

  # Design matrix without endogenous covariate
  XandW <- as.matrix(cbind(intercept, exo.cov))
  if (conf) {
    XandW <- cbind(XandW, W)
  }

  #### Step 1: estimate control function if necessary ####

  if (conf & (inst == "cf")) {
    est.cf.out <- estimate.cf(XandW, Z, Zbin)
    cf <- est.cf.out$cf
    gammaest <- est.cf.out$gammaest
  } else if (conf & (inst == "w")) {
    cf <- W
  } else if (conf & (inst == "none")) {
    cf <- NULL
  } else if (conf & (inst == "oracle")) {
    cf <- realV
  } else {
    cf <- NULL
  }

  #### Step 2: estimate model parameters ####

  # Define useful parameters
  n.trans <- sum(grepl("delta[1-9][[:digit:]]*", colnames(data)))
  if (conf & !is.null(cf)) {
    # When there is confounding, we have [intercept, exo.cov, Z, cf]
    totparl <- (1 + ncol(exo.cov) + 2) * n.trans
  } else if (conf & is.null(cf)) {
    # When there is confounding but no cf, we have [intercept, exo.cov, Z]
    totparl <- (1 + ncol(exo.cov) + 1) * n.trans
  } else {
    # When there is no confounding, we have [intercept exo.cov]
    totparl <- (1 + ncol(exo.cov)) * n.trans
  }

  sd.lb <- 1e-05
  sd.ub <- Inf
  cor.lb <- -0.99
  cor.ub <- 0.99
  chol.lb <- -Inf
  chol.ub <- Inf
  transfo.param.lb <- 0
  transfo.param.ub <- 2

  # Initial estimate for the parameters assuming independence between competing
  # risks.
  init.covariate.effects <- rep(0, totparl)
  init.sds <- rep(1, n.trans)
  init.transfo.pararms <- rep(1, n.trans)
  init.par <- c(init.covariate.effects, init.sds, init.transfo.pararms)

  lb.vec <- c(rep(-Inf, totparl), rep(sd.lb, n.trans), rep(transfo.param.lb, n.trans))
  ub.vec <- c(rep(Inf, totparl), rep(sd.ub, n.trans), rep(transfo.param.ub, n.trans))

  parhat.init <- nloptr(x0 = init.par,
                        eval_f = LikI.bis,
                        data = data,
                        admin = admin,
                        conf = conf,
                        cf = cf,
                        lb = lb.vec,
                        ub = ub.vec,
                        eval_g_ineq = NULL,
                        opts = list(algorithm = "NLOPT_LN_BOBYQA",
                                    "ftol_abs" = 1.0e-30,
                                    "maxeval" = 100000,
                                    "xtol_abs" = rep(1.0e-30))
                        )$solution

  ## Estimate the model parameters, possibly under the assumption of solely
  ## independent censoring if indicated. Use the pre-estimates of the model that
  ## assumes independent competing risks.

  # Set pre-estimated values as initial values
  pre.covariate.effects <- parhat.init[1:totparl]
  pre.sds <- parhat.init[(totparl + 1):(totparl + n.trans)]
  pre.transfo.params <- parhat.init[(totparl + n.trans + 1):length(parhat.init)]

  # Estimate the model
  if (!use.chol) {

    # Prevent this option from being used, as it is deprecated
    stop("Cholesky factorization should be used in estimating the model.")

    # Add initial value for the correlations between the competing risks
    pre.correlations <- rep(0, choose(length(eoi.indicator.names), 2))
    init <- c(pre.covariate.effects, pre.sds, pre.correlations, pre.transfo.params)

    lb.vec <- c(rep(-Inf, totparl), rep(sd.lb, n.trans), rep(cor.lb, choose(n.trans, 2)),
                rep(transfo.param.lb, n.trans))
    ub.vec <- c(rep(Inf, totparl), rep(sd.ub, n.trans), rep(cor.ub, choose(n.trans, 2)),
                rep(transfo.param.ub, n.trans))

    # Obtain the model parameters
    parhat1 <- nloptr(x0 = init,
                      eval_f = LikI.cmprsk,
                      data = data,
                      eoi.indicator.names = eoi.indicator.names,
                      admin = admin,
                      conf = conf,
                      cf = cf,
                      lb = lb.vec,
                      ub = ub.vec,
                      eval_g_ineq = NULL,
                      opts = list(algorithm = "NLOPT_LN_BOBYQA",
                                  "ftol_abs" = 1.0e-30,
                                  "maxeval" = 100000,
                                  "xtol_abs" = 1.0e-30)
                      )$solution

    } else {

    ## Add initial value for the correlations between the competing risks
    pre.cov.mat <- diag(pre.sds[eoi.inds]^2, nrow = length(eoi.inds),
                        ncol = length(eoi.inds))
    pre.chol1 <- chol(pre.cov.mat)[lower.tri(pre.cov.mat, diag = TRUE)]
    pre.sd2 <- pre.sds[setdiff(1:n.trans, eoi.inds)]
    init <- c(pre.covariate.effects, pre.chol1, pre.sd2, pre.transfo.params)

    s1 <- length(eoi.inds)
    s2 <- n.trans - s1
    lb.vec <- c(rep(-Inf, totparl), rep(chol.lb, choose(s1, 2) + s1),
                rep(sd.lb, s2), rep(transfo.param.lb, n.trans))
    ub.vec <- c(rep(Inf, totparl), rep(chol.ub, choose(s1, 2) + s1),
                rep(sd.ub, s2), rep(transfo.param.ub, n.trans))

    ## Obtain the model parameters
    parhatc <- nloptr(x0 = init,
                      eval_f = LikI.cmprsk.Cholesky,
                      data = data,
                      eoi.indicator.names = eoi.indicator.names,
                      admin = admin,
                      conf = conf,
                      cf = cf,
                      eps = eps,
                      lb = lb.vec,
                      ub = ub.vec,
                      eval_g_ineq = NULL,
                      opts = list(algorithm = "NLOPT_LN_BOBYQA",
                                  "ftol_abs" = 1.0e-30,
                                  "maxeval" = 100000,
                                  "xtol_abs" = 1.0e-30)
    )$solution

    ## Obtain variance of estimates, if necessary.

    if (compute.var) {

      # Compute the variance
      var.out <- variance.cmprsk(parhatc, gammaest, data, admin, conf, inst, cf,
                                 eoi.indicator.names, Zbin, use.chol, n.trans,
                                 totparl)

      # Store the results
      res <- var.out
    }

    ## Backtransform the cholesky parameters
    #
    # (if not already done in variance computations)

    if (!compute.var) {

      # Extract parameters
      par.chol <- parhatc[(totparl + 1):(length(parhatc) - n.trans)]
      par.chol1 <- par.chol[1:(choose(s1, 2) + s1)]
      par.sd2 <- par.chol[(choose(s1, 2) + s1 + 1):length(par.chol)]

      # Reconstruct covariance matrix for events of interest
      mat.chol1 <- matrix(0, nrow = s1, ncol = s1)
      mat.chol1[lower.tri(mat.chol1, diag = TRUE)] <- par.chol1
      var1 <- mat.chol1 %*% t(mat.chol1)

      # Reconstruct full covariance matrix and correlation matrix
      var <- matrix(0, nrow = n.trans, ncol = n.trans)
      var[eoi.inds, eoi.inds] <- var1
      diag(var)[setdiff(1:n.trans, eoi.inds)] <- par.sd2^2
      rho.mat <- var / sqrt(outer(diag(var), diag(var), FUN = "*"))

      # Construct full parameter vector.
      parhat1 <- c(parhatc[1:totparl], diag(var),
                   rho.mat[lower.tri(rho.mat, diag = FALSE)],
                   parhatc[(length(parhatc) - n.trans + 1):length(parhatc)])

      # Store the results
      res <- parhat1

      }
    }

  #### Return the results ####

  res
}



#' @title Transform Cholesky decomposition to covariance matrix parameter element.
#'
#' @description This function transforms the parameters of the Cholesky de-
#' composition to a covariance matrix element. This function is used in
#' chol2par.R.
#'
#' @param a The row index of the covariance matrix element to be computed.
#' @param b The column index of the covariance matrix element to be computed.
#' @param par.chol1 The vector of Cholesky parameters.
#'
#' @return Specified element of the covariance matrix resulting from the
#' provided Cholesky decomposition.
#'
#' @noRd
#'
chol2par.elem <- function(a, b, par.chol1) {

  # Create matrix with Cholesky parameters
  p <- (-1 + sqrt(1 + 8*length(par.chol1)))/2
  chol.mat <- matrix(0, nrow = p, ncol = p)
  chol.mat[lower.tri(chol.mat, diag = TRUE)] <- par.chol1
  chol.mat[upper.tri(chol.mat, diag = TRUE)] <- t(chol.mat)[upper.tri(chol.mat, diag = TRUE)]

  # Return covariance element
  cov.elem <- 0
  for (i in 1:min(a, b)) {
    cov.elem <- cov.elem + chol.mat[a, i]*chol.mat[b, i]
  }
  cov.elem
}



#' @title Transform Cholesky decomposition to covariance matrix
#'
#' @description This function transforms the parameters of the Cholesky de-
#' composition to the covariance matrix, represented as a the row-wise con-
#' catenation of the upper-triangular elements.
#'
#' @param par.chol1 The vector of Cholesky parameters.
#'
#' @return Covariance matrix corresponding to the provided Cholesky decomposition.
#'
#' @noRd
#'
chol2par <- function(par.chol1) {

  # Create matrix with Cholesky parameters
  p <- (-1 + sqrt(1 + 8*length(par.chol1)))/2
  chol.mat <- matrix(0, nrow = p, ncol = p)
  chol.mat[lower.tri(chol.mat, diag = TRUE)] <- par.chol1
  chol.mat[upper.tri(chol.mat, diag = TRUE)] <- t(chol.mat)[upper.tri(chol.mat, diag = TRUE)]

  # Define matrix of double indices corresponding to the single index rep-
  # resentation of the elements in par.chol.
  double.idx.mat <- matrix(1, nrow = 1, ncol = 2)
  for (i in 2:length(par.chol1)) {
    second <- ifelse(double.idx.mat[nrow(double.idx.mat), 2] < p,
                     double.idx.mat[nrow(double.idx.mat), 2] + 1,
                     double.idx.mat[nrow(double.idx.mat), 1] + 1)
    first <- ifelse(double.idx.mat[nrow(double.idx.mat), 2] < p,
                    double.idx.mat[nrow(double.idx.mat), 1],
                    double.idx.mat[nrow(double.idx.mat), 1] + 1)
    double.idx.mat <- rbind(double.idx.mat, c(first, second))
  }

  # Switch columns in double.idx.mat
  double.idx.mat <- cbind(double.idx.mat[, 2], double.idx.mat[, 1])

  # Compute transformation
  par.vec <- rep(NA, nrow(double.idx.mat))
  for (i in 1:nrow(double.idx.mat)) {
    par.vec[i] <- chol2par.elem(double.idx.mat[i, 1], double.idx.mat[i, 2],
                                par.chol1)
  }

  # Give appropriate row and column names
  names(par.vec) <- apply(double.idx.mat,
                          1,
                          function(row) {sprintf("(%s)", paste(row, collapse= ", "))})

  # Return the result
  par.vec
}



#' @title Derivative of transform Cholesky decomposition to covariance matrix
#' element.
#'
#' @description This function defines the derivative of the transformation
#' function that maps Cholesky parameters to a covariance matrix element. This
#' function is used in dchol2par.R.
#'
#' @param k The row index of the parameter with respect to which to take the
#' derivative.
#' @param q the column index of the parameter with respect to which to take the
#' derivative.
#' @param a The row index of the covariance matrix element to be computed.
#' @param b The column index of the covariance matrix element to be computed.
#' @param par.chol1 The vector of Cholesky parameters.
#'
#' @return Derivative of the function that transforms the cholesky parameters
#' to the specified element of the covariance matrix, evaluated at the specified
#' arguments.
#'
#' @noRd
#'
dchol2par.elem <- function(k, q, a, b, par.chol1) {

  # Create matrix with Cholesky parameters
  p <- (-1 + sqrt(1 + 8*length(par.chol1)))/2
  chol.mat <- matrix(0, nrow = p, ncol = p)
  chol.mat[lower.tri(chol.mat, diag = TRUE)] <- par.chol1
  chol.mat[upper.tri(chol.mat, diag = TRUE)] <- t(chol.mat)[upper.tri(chol.mat, diag = TRUE)]

  # Return derivative of the transformation function, evaluated as par.chol1
  if (k != a & q != b) {
    der <- 0
  } else if (k == a & q == b & a == b) {
    der <- 2*chol.mat[k, q]
  } else if (k == a & k == q & a > b) {
    der <- 0
  } else if (k == a & q == b & k > q) {
    der <- chol.mat[q, q]
  } else if (k == q & a > b & b == q) {
    der <- chol.mat[a, b]
  } else if (k > q & q == b & a == b) {
    der <- 0
  } else if (k > q & q < b & a == b & k == a) {
    der <- 2*chol.mat[k, q]
  } else if (k > q & q < b & a == b & k != a) {
    der <- 0
  } else if (k > q & q == b & b < a & k != a) {
    der <- 0
  } else if (k > q & k == a & a > b & q < b) {
    der <- chol.mat[b, q]
  } else if (k > q & k == a & a > b & q > b) {
    der <- 0
  } else if (k > q & k == a & a > b & q == b) {
    der <- 0
  }
  der
}



#' @title Derivative of transform Cholesky decomposition to covariance matrix.
#'
#' @description This function defines the derivative of the transformation
#' function that maps Cholesky parameters to the full covariance matrix.
#'
#' @param par.chol1 The vector of Cholesky parameters.
#'
#' @return Derivative of the function that transforms the cholesky parameters
#' to the full covariance matrix.
#'
#' @noRd
#'
dchol2par <- function(par.chol1) {

  # Create matrix with Cholesky parameters
  p <- (-1 + sqrt(1 + 8*length(par.chol1)))/2
  chol.mat <- matrix(0, nrow = p, ncol = p)
  chol.mat[lower.tri(chol.mat, diag = TRUE)] <- par.chol1
  chol.mat[upper.tri(chol.mat, diag = TRUE)] <- t(chol.mat)[upper.tri(chol.mat, diag = TRUE)]

  # Define matrix of double indices corresponding to the single index rep-
  # resentation of the elements in par.chol.
  double.idx.mat <- matrix(1, nrow = 1, ncol = 2)
  for (i in 2:length(par.chol1)) {
    second <- ifelse(double.idx.mat[nrow(double.idx.mat), 2] < p,
                     double.idx.mat[nrow(double.idx.mat), 2] + 1,
                     double.idx.mat[nrow(double.idx.mat), 1] + 1)
    first <- ifelse(double.idx.mat[nrow(double.idx.mat), 2] < p,
                    double.idx.mat[nrow(double.idx.mat), 1],
                    double.idx.mat[nrow(double.idx.mat), 1] + 1)
    double.idx.mat <- rbind(double.idx.mat, c(first, second))
  }

  # Switch columns in double.idx.mat
  double.idx.mat <- cbind(double.idx.mat[, 2], double.idx.mat[, 1])

  # Compute derivative of transformation
  d.mat <- matrix(NA, nrow = nrow(double.idx.mat), ncol = nrow(double.idx.mat))
  for (i in 1:nrow(double.idx.mat)) {
    for (j in 1:nrow(double.idx.mat)) {
      d.mat[i, j] <- dchol2par.elem(double.idx.mat[j, 1], double.idx.mat[j, 2],
                                    double.idx.mat[i, 1], double.idx.mat[i, 2],
                                    par.chol1)
    }
  }

  # Give appropriate row and column names
  rownames(d.mat) <- apply(double.idx.mat,
                           1,
                           function(row) {sprintf("(%s)", paste(row, collapse= ", "))})
  colnames(d.mat) <- apply(double.idx.mat,
                           1,
                           function(row) {sprintf("(%s)", paste(row, collapse= ", "))})

  # Return the matrix of derivatives
  d.mat
}



#' @title Compute the variance of the estimates.
#'
#' @description This function computes the variance of the estimates computed
#' by the 'estimate.cmprsk.R' function.
#'
#' @param parhatc Vector of estimated parameters, computed in the first part of
#' \code{estimate.cmprsk.R}.
#' @param gammaest Vector of estimated parameters in the regression model for
#' the control function.
#' @param data A data frame.
#' @param admin Boolean value indicating whether the data contains
#' administrative censoring.
#' @param conf Boolean value indicating whether the data contains confounding
#' and hence indicating the presence of \code{z} and, possibly, \code{w}.
#' @param inst Variable encoding which approach should be used for dealing with
#' the confounding. \code{inst = "cf"} indicates that the control function
#' approach should be used. \code{inst = "W"} indicates that the instrumental
#' variable should be used 'as is'. \code{inst = "None"} indicates that Z will
#' be treated as an exogenous covariate. Finally, when \code{inst = "oracle"},
#' this function will access the argument \code{realV} and use it as the values
#' for the control function. Default is \code{inst = "cf"}.
#' @param cf The control function used to estimate the second step.
#' @param eoi.indicator.names Vector of names of the censoring indicator columns
#' pertaining to events of interest. Events of interest will be modeled allowing
#' dependence between them, whereas all censoring events (corresponding to
#' indicator columns not listed in \code{eoi.indicator.names}) will be treated
#' as independent of every other event. If \code{eoi.indicator.names == NULL},
#' all events will be modeled dependently.
#' @param Zbin Indicator value indicating whether (\code{Zbin = TRUE}) or not
#' \code{Zbin = FALSE} the endogenous covariate is binary. Default is
#' \code{Zbin = NULL}, corresponding to the case when \code{conf == FALSE}.
#' @param use.chol Boolean value indicating whether the cholesky decomposition
#' was used in estimating the covariance matrix.
#' @param n.trans Number of competing risks in the model (and hence, number of
#' transformation models).
#' @param totparl Total number of covariate effects (including intercepts) in
#' all of the transformation models combined.
#'
#' @import numDeriv
#' @importFrom stats dlogis plogis var
#'
#' @return Variance estimates of the provided vector of estimated parameters.
#'
#' @noRd
#'
variance.cmprsk <- function(parhatc, gammaest, data, admin, conf, inst, cf,
                            eoi.indicator.names, Zbin, use.chol, n.trans, totparl) {

  ## Precondition checks

  requireNamespace("stringr")
  if (!use.chol) {
    stop("Only implemented for use.chol == TRUE.")
  }
  if (!conf) {
    stop("Only implemented when in cases where there is confounding.")
  }

  ## Define useful variables

  # Number of observations
  n <- nrow(data)

  # Indicator variable whether eoi.indicator.names is specified. If necessary,
  # determine the indices of the events of interest.
  indep <- !is.null(eoi.indicator.names)
  if (indep) {
    cens.ind.columns <- colnames(data)[grepl("delta[1-9][[:digit:]]*", colnames(data))]
    eoi.idxs <- which(cens.ind.columns %in% eoi.indicator.names)
  }

  # Intercept
  intercept <- data[, "x0"]

  # Exogenous covariates
  exo.cov <- data[, grepl("x[1-9][[:digit:]]*", colnames(data)),
                  drop = FALSE]

  # Endogenous covariate and instrumental variable
  Z <- data[, "z"]
  W <- data[, "w"]

  # Design matrix without endogenous covariate
  XandIntercept <- as.matrix(cbind(intercept, exo.cov))
  XandW <- cbind(XandIntercept, W)

  #### Compute the Hessian matrix of the full likelihood at parhatG ####

  # Construct full vector of estimates
  parhatG <- c(parhatc, gammaest)

  # Numerical approximation of the Hessian at parhatG
  Hgamma <- hessian(likIFG.cmprsk.Cholesky,
                    parhatG,
                    data = data,
                    eoi.indicator.names = eoi.indicator.names,
                    admin = admin,
                    conf = conf,
                    Zbin = Zbin,
                    inst = inst,
                    method = "Richardson",
                    method.args = list(eps = 1e-4, d = 0.0001,
                                       zer.tol = sqrt(.Machine$double.eps/7e-7),
                                       r = 6, v = 2, show.details = FALSE)
  )


  # Omit the part of the variance pertaining to the parameters of the control
  # function and compute the inverse.
  H <- Hgamma[1:length(parhatc),1:length(parhatc)]
  HI <- ginv(H)

  #### Compute the variance of estimates ####

  # H_gamma
  Vargamma <- Hgamma[1:length(parhatc),
                     (length(parhatc) + 1):(length(parhatc) + length(gammaest))]

  # Compute all products [1*1, 1*X1, 1*X2, ..., 1*W, X1*X1, X1*X2, ..., X1*W,
  #                       X2*X2, ..., W*W]
  prodvec <- XandW[, 1]
  for (i in 1:length(gammaest)) {
    for (j in 2:length(gammaest)) {
      if (i <= j){
        prodvec <- cbind(prodvec, diag(XandW[, i] %*% t(XandW[, j])))
      }
    }
  }

  # M-matrix: second derivative of m(W, Z, gamma).
  if (Zbin) {

    secder <- t(-dlogis(XandW %*% gammaest)) %*% prodvec

    WM <- matrix(0, nrow = length(gammaest), ncol = length(gammaest))
    WM[lower.tri(WM, diag = TRUE)] <- secder
    WM[upper.tri(WM)] <- t(WM)[upper.tri(WM)]

  } else {

    # Negative of the column sums of prodvec.
    sumsecder <- c(rep(0, ncol(prodvec)))
    for (i in 1:length(sumsecder)) {
      sumsecder[i] <- -sum(prodvec[, i])
    }

    # In this case m(W, Z, gamma) = (Z - W^\top \gamma)². Note that a factor
    # '-2' is omitted.
    WM <- matrix(0, nrow = length(gammaest), ncol = length(gammaest))
    WM[lower.tri(WM, diag = TRUE)] <- sumsecder
    WM[upper.tri(WM)] <- t(WM)[upper.tri(WM)]
  }

  # Inverse of M-matrix
  WMI <- ginv(WM)

  # h_m: first derivative of m(W,Z,gamma). Note that a factor '-2' is omitted.
  mi <- c()
  if (Zbin) {
    diffvec <- Z - plogis(XandW %*% gammaest)
    for(i in 1:n) {
      newrow <- diffvec[i] %*% XandW[i, ]
      mi <- rbind(mi, newrow)
    }
  } else {
    for(i in 1:n) {
      newrow <- cf[i] %*% XandW[i, ]
      mi <- rbind(mi, newrow)
    }
  }
  mi <- t(mi)

  # psi_i-matrix (omitted factors cancel out)
  psii <- -WMI %*% mi

  # h_l(S_i, gamma, delta)
  gi <- c()
  if (indep) {
    for (i in 1:n) {
      J1 <- jacobian(LikI.cmprsk.Cholesky,
                     parhatc,
                     data = data[i, ],
                     eoi.indicator.names = eoi.indicator.names,
                     admin = admin,
                     conf = conf,
                     cf = cf[i],
                     method = "Richardson",
                     method.args = list(eps = 1e-4, d = 0.0001,
                                        zer.tol = sqrt(.Machine$double.eps/7e-7),
                                        r = 6, v = 2, show.details = FALSE))
      gi <- rbind(gi,c(J1))
    }
  } else {
    for (i in 1:n) {
      J1 <- jacobian(likF.cmprsk.Cholesky,
                     parhatc,
                     data = data[i, ],
                     admin = admin,
                     conf = conf,
                     cf = cf[i],
                     method = "Richardson",
                     method.args = list(eps = 1e-4, d = 0.0001,
                                        zer.tol = sqrt(.Machine$double.eps/7e-7),
                                        r = 6, v = 2, show.details = FALSE))
      gi <- rbind(gi,c(J1))
    }
  }

  gi <- t(gi)

  # h_l(S, gamma, delta) + H_gamma %*% Psi_i
  partvar <- gi + Vargamma %*% psii
  Epartvar2 <- (partvar %*% t(partvar))
  totvarex <- HI %*% Epartvar2 %*% t(HI)
  se <- sqrt(abs(diag(totvarex)))

  #### Construct confidence intervals ####

  ## Covariate effect parameters

  # Extract (the variances of) the coefficients. Give appropriate names.
  coeffs <- parhatc[1:totparl]
  coeff.vars <- diag(totvarex[1:totparl, 1:totparl])

  coeff.names <- c()
  for (i in 1:n.trans) {
    for (j in 0:ncol(exo.cov)) {
      coeff.names <- c(coeff.names, sprintf("beta_{%d, %d}", i, j))
    }
    if (conf) {
      coeff.names <- c(coeff.names, sprintf("alpha_{%d}", i))
    }
    if (conf & (inst != "none")) {
      coeff.names <- c(coeff.names, sprintf("lambda_{%d}", i))
    }
  }
  names(coeffs) <- coeff.names
  names(coeff.vars) <- coeff.names

  # Construct the confidence intervals
  coeffs.lb <- coeffs - 1.96 * coeff.vars
  coeffs.ub <- coeffs + 1.96 * coeff.vars

  ## Covariance matrix parameters

  if (indep) {
    s1 <- length(eoi.indicator.names)
  } else {
    s1 <- n.trans
  }

  # Extract Cholesky decomposition parameters
  par.chol <- parhatc[(totparl + 1):(length(parhatc) - n.trans)]
  par.chol1 <- par.chol[1:(choose(s1, 2) + s1)]
  par.chol2 <- NULL
  if ((choose(s1, 2) + s1) < length(par.chol)) {
    par.chol2 <- par.chol[(choose(s1, 2) + s1 + 1):length(par.chol)]
  }

  # Delta method for transforming Cholesky parameter variances to covariance
  # element variances.
  chol.covar.mat <- totvarex[(totparl + 1):(length(parhatc) - n.trans),
                             (totparl + 1):(length(parhatc) - n.trans)]

  chol1.covar.mat <- chol.covar.mat[1:(choose(s1, 2) + s1), 1:(choose(s1, 2) + s1)]
  cov1.elems <- chol2par(par.chol1)
  cov1.vars <- t(dchol2par(par.chol1)) %*% chol1.covar.mat %*% dchol2par(par.chol1)

  if ((choose(s1, 2) + s1) < length(par.chol)) {
    cov2.elems <- par.chol2^2
    chol2.covar.mat <- chol.covar.mat[(choose(s1, 2) + s1 + 1):length(par.chol),
                                      (choose(s1, 2) + s1 + 1):length(par.chol)]
    cov2.vars <- diag(as.matrix(chol2.covar.mat)) * (2 * par.chol2)^2
  }

  # Get all column (row) indices corresponding to variance elements
  var.idxs <- unlist(lapply(names(cov1.elems),
                            function(elem) {
                              var(as.numeric(stringr::str_split(gsub("\\(|\\)", "", elem), ", ")[[1]])) == 0
                            }))

  # Extract (the variances of) the variance elements. Give appropriate names.
  var1.elems <- unname(cov1.elems[var.idxs])
  var1.vars <- unname(diag(cov1.vars)[var.idxs])
  var2.elems <- var2.vars <- NULL
  if ((choose(s1, 2) + s1) < length(par.chol)) {
    var2.elems <- cov2.elems
    var2.vars <- cov2.vars
  }

  var.elems <- rep(0, n.trans)
  var.vars <- rep(0, n.trans)
  var.elems[eoi.idxs] <- var1.elems
  var.elems[setdiff(1:n.trans, eoi.idxs)] <- var2.elems
  var.vars[eoi.idxs] <- var1.vars
  var.vars[setdiff(1:n.trans, eoi.idxs)] <- var2.vars

  var.names <- unlist(lapply(1:n.trans, function(i) {sprintf("sigma^2_{%d}", i)}))
  names(var.elems) <- var.names
  names(var.vars) <- var.names

  # Construct the confidence interval for variances in the log-space. Then
  # backtransform.
  log.var.vars <- var.vars / var.elems
  log.var.lb <- log(var.elems) - 1.96 * sqrt(log.var.vars)
  log.var.ub <- log(var.elems) + 1.96 * sqrt(log.var.vars)

  var.lb <- exp(log.var.lb)
  var.ub <- exp(log.var.ub)

  # Extract (the variances of) the covariance elements
  cov1.offdiag.elems <- cov1.elems[!var.idxs]
  cov1.offdiag.vars <- diag(cov1.vars)[!var.idxs]

  # Construct the full covariance matrix (offdiagonal elements)
  cov.mat <- matrix(0, nrow = n.trans, ncol = n.trans)
  cov.mat[eoi.idxs, eoi.idxs][lower.tri(cov.mat[eoi.idxs, eoi.idxs])] <- cov1.offdiag.elems
  cov.mat[upper.tri(cov.mat)] <- t(cov.mat)[upper.tri(cov.mat)]

  # Construct the full variance of covariance matrix (offdiagonal elements)
  cov.var.mat <- matrix(0, nrow = n.trans, ncol = n.trans)
  cov.var.mat[eoi.idxs, eoi.idxs][lower.tri(cov.mat[eoi.idxs, eoi.idxs])] <- cov1.offdiag.vars
  cov.var.mat[upper.tri(cov.var.mat)] <- t(cov.var.mat)[upper.tri(cov.var.mat)]

  # Vectorize the off-diagonal elements. Name the elements
  cov.offdiag.elems <- cov.mat[lower.tri(cov.mat)]
  cov.offdiag.vars <- cov.var.mat[lower.tri(cov.var.mat)]

  cov.offdiag.names <- c()
  for (i in 1:n.trans) {
    for (j in 1:n.trans) {
      if (i < j) {
        cov.offdiag.names <- c(cov.offdiag.names, sprintf("sigma_{%d, %d}", i, j))
      }
    }
  }
  names(cov.offdiag.vars) <- cov.offdiag.names
  names(cov.offdiag.elems) <- cov.offdiag.names

  # Construct the confidence interval for the covariances.
  cov.lb <- cov.offdiag.elems - 1.96 * sqrt(cov.offdiag.vars)
  cov.ub <- cov.offdiag.elems + 1.96 * sqrt(cov.offdiag.vars)

  ## Transformation parameters

  # Extract (the variances of) the transformation parameters. Give appropriate
  # names.
  theta.elems <- parhatc[(length(parhatc) - n.trans + 1):length(parhatc)]
  theta.sds <- se[(length(parhatc) - n.trans + 1):length(parhatc)]

  theta.names <- unlist(lapply(1:n.trans, function(i) {sprintf("theta_{%d}", i)}))
  names(theta.elems) <- theta.names
  names(theta.sds) <- theta.names

  # Confidence interval for transformation parameters
  theta.lb <- theta.elems - 1.96 * theta.sds
  theta.ub <- theta.elems + 1.96 * theta.sds

  # Matrix with all confidence intervals
  CI.mat <- cbind(c(coeffs.lb, var.lb, cov.lb, theta.lb),
                  c(coeffs.ub, var.ub, cov.ub, theta.ub))

  ## Return the results

  # Vector of (transformed) parameters
  params <- c(coeffs, var.elems, cov.offdiag.elems, theta.elems)

  # Vector of variances of (transformed) parameters
  param.vars <- c(coeff.vars, var.vars, cov.offdiag.vars, theta.sds^2)

  # Return results
  list(params = params, param.vars = param.vars, CI.mat = CI.mat)
}



#' @title Print the results from a ptcr model.
#'
#' @description This function takes the output of the function estimate.cmprsk.R
#' as an argument and prints a summary of the results.
#' Note: this function is not yet finished! It should be made a proper S3 method
#' in the future.
#'
#' @param output.estimate.cmprsk The output of the estimate.cmprsk.R function.
#'
#' @noRd
#'
mysummary.ptcr <- function(output.estimate.cmprsk) {

  # Load dependency
  requireNamespace("stringr")

  # For now, work with dummy values for the variables
  var.names <- c("(intercept)", "var1", "var2", "var3")
  estimates <- c(1, 20, 300, 4)
  std.error <- c(0.5, 0.141579390248, 200, 4)
  sign <- c("*", "**", ".", "")

  cr.names <- c("cr1", "cr2", "cr3", "cr4")
  cr.vars <- c(1, 0.1234567, 300.1, 4)
  cr.corr <- c(0.1, 0.22, 0.333, 0.4444, 0.55555, 0.666666)
  cr.vars.std <- c(1, 1, 1, 1)
  cr.corr.std <- 1:6

  cr.transfo.param <- c(0, 0.6666666666666, 1.3, 2)

  # Compute the width of each column
  header.names <- c("estimate", "std.error", "sign.")
  col.var.names.width <- max(sapply(var.names, nchar))
  col.estimates.width <- max(sapply(as.character(estimates), nchar), nchar(header.names[1]))
  col.std.error.width <- max(sapply(as.character(std.error), nchar), nchar(header.names[2]))
  col.sign.width <- nchar(header.names[3])

  # Construct header for parameters pertaining to covariates
  header <- paste0(stringr::str_pad("", col.var.names.width), "  ",
                   stringr::str_pad(header.names[1], col.estimates.width, "right"), "  ",
                   stringr::str_pad(header.names[2], col.std.error.width, "right"), "  ",
                   stringr::str_pad(header.names[3], col.sign.width, "right"))

  # Construct body for parameters pertaining to covariates
  body <- ""
  for (i in 1:length(estimates)) {
    line <- paste0(stringr::str_pad(var.names[i], col.var.names.width, "right"), "  ",
                   stringr::str_pad(estimates[i], col.estimates.width, "right"), "  ",
                   stringr::str_pad(std.error[i], col.std.error.width, "right"), "  ",
                   stringr::str_pad(sign[i], col.sign.width, "right"))

    body <- paste0(body, line)
    if (i < length(estimates)) {
      body <- paste0(body, "\n")
    }
  }

  # Construct header for parameters pertaining to competing risks models

  # Construct body for parameters pertaining to competing risks models

  # Make and print summary
  summary <- paste0(header, "\n", body)
  message(summary)
}

#' @title Likelihood function under dependent censoring
#'
#' @description
#' The \code{PseudoL} function  is maximized in order to
#' estimate the finite dimensional model parameters, including the dependency parameter.
#' This function assumes that the cumulative hazard function is known.
#'
#' @inheritParams SolveL
#' @param lhat The estimated hazard function obtained from the output of \code{\link{SolveL}}.
#' @param cumL The estimated cumulative hazard function from the output of \code{\link{SolveL}}.
#' @importFrom copula pCopula frankCopula gumbelCopula tau
#' @import pbivnorm
#' @return maximized log-likelihood value
#' @noRd
PseudoL = function(theta,resData,X,W,lhat,cumL,cop,dist){
  Z = resData$Z
  d1 = resData$d1
  d2 = resData$d2

  if (is.vector(X)){
    k = 1
    beta = theta[k]
    mu1 = X*beta
  }else if(!is.vector(X)){
    k = dim(X)[2]
    beta = theta[1:k]
    mu1 = X%*%beta}
  l = dim(W)[2]
  eta = theta[(k+1):(l+k)]
  nu = theta[(l+k+1)]
  gm = theta[(l+k+2)]
  y = log(Z)

  # distribution of T -- Cox PH

  g1 = lhat*exp(mu1)*exp(-cumL*exp(mu1))
  G1 = 1-exp(-cumL*exp(mu1))

  # dist of C

  if (dist == "Weibull"){
    g2 = 1/nu*exp((y-W%*%eta)/nu)*exp(-exp((y-W%*%eta)/nu))           # log-time scale
    G2 = 1-exp(-exp((y-W%*%eta)/nu))
  }else if (dist == "lognormal"){   # Lognormal
    m = (y-W%*%eta)/nu
    g2 = 1/(sqrt(2*pi)*nu*Z)*exp(-(m^2/2))
    G2 = pnorm(m)
  }else
    stop("The distribution of C is not supported")

  # Joint distribution
  # avoid NA

  g1[is.na(g1)] = 1e-10
  g2[is.na(g2)] = 1e-10
  g1[is.nan(g1)] = 1e-10
  g2[is.nan(g2)] = 1e-10
  G1[is.na(G1)] = 1e-10
  G2[is.na(G2)] = 1e-10
  G1[is.nan(G1)] = 1e-10
  G2[is.nan(G2)] = 1e-10
  g1[!is.finite(g1)] = 0
  g2[!is.finite(g2)] = 0
  G1 = pmax(G1,1e-10)
  G1 = pmin(G1,0.99999999)
  G2 = pmax(G2,1e-10)
  G2 = pmin(G2,0.99999999)

  PD = cbind(G1,G2)

  if (cop == "Frank")                                         # Frank copula
  {
    frank.cop = copula::frankCopula(gm, dim = 2)
    cp1 = (exp(- gm*G1)*(exp(- gm*G2)-1))/(exp(-gm)-1 + (exp(-gm*G1)-1)*(exp(-gm*G2)-1))
    cp2 = (exp(- gm*G2)*(exp(- gm*G1)-1))/(exp(-gm)-1 + (exp(-gm*G1)-1)*(exp(-gm*G2)-1))
    z6 =  1-G1-G2+copula::pCopula(PD,frank.cop)
    z6 =   pmax(z6,1e-10)
  }else if (cop == "Gumbel")                                       # Gumbel copula
  {
    gumb.cop = copula::gumbelCopula(gm, dim = 2)
    cp1 = copula::pCopula(PD,gumb.cop)*((-log(G1))^gm+(-log(G2))^gm)^(-1+1/gm)*(-log(G1))^(gm-1)/G1
    cp2 = copula::pCopula(PD,gumb.cop)*((-log(G1))^gm+(-log(G2))^gm)^(-1+1/gm)*(-log(G2))^(gm-1)/G2
    z6 =  1-G1-G2+copula::pCopula(PD,gumb.cop)
    z6 =   pmax(z6,1e-10)
  }else if (cop == "Normal")     # normal
  {
    cp1 = pnorm((qnorm(G2) - gm * qnorm(G1))/sqrt(1 - gm^2))
    cp2 = pnorm((qnorm(G1) - gm * qnorm(G2))/sqrt(1 - gm^2))
    p1 = qnorm(G1)
    p2 = qnorm(G2)
    c2 = cbind(p1,p2)
    z6 =  1-G1-G2+pbivnorm(c2,rho = gm)          # contribution to the likelihood from A
    z6 =   pmax(z6,1e-10)
  }
  cp1[is.na(cp1)] = 1e-20
  cp2[is.na(cp2)] = 1e-20
  cp1[is.nan(cp1)] = 1e-20
  cp2[is.nan(cp2)] = 1e-20
  cp1[!is.finite(cp1)] = 0
  cp2[!is.finite(cp2)] = 0
  d3 =  (1-d1-d2)                # Censoring
  term1 <- (1 - cp1)*g1          #  Contribution from T
  term2 <- (1 - cp2)*g2          #  Contribution from C

  Logn <- -sum(d1 * log(pmax(term1, 1e-10)) + d2 * log(pmax(term2, 1e-10)) + d3*log(z6))
  return(Logn)
}




#' @title Loglikehood function under independent censoring
#'
#' @inheritParams SolveLI
#' @param W Data matrix with covariates related to C. First column of W should be ones
#' @param lhat The estimated hazard function obtained from the output of \code{\link{SolveLI}}.
#' @param cumL The estimated cumulative hazard function from the output of \code{\link{SolveLI}}.
#' @param dist The distribution to  be used for the dependent censoring C. Only two distributions are allowed, i.e, Weibull
#' and lognormal distributions. With the value \code{"Weibull"} as the
#'   default.
#'  @importFrom stats nlminb pnorm  qnorm sd
#'
#' @return Maximized log-likelihood value
#' @noRd
LikCopInd <- function(theta,resData,X,W,lhat,cumL,dist){ # gamma = 0
  Z = resData$Z
  d1 = resData$d1
  d2 = resData$d2

  if (is.vector(X)){
    k = 1
    beta = theta[k]
    mu1 = X*beta
  }else if(!is.vector(X)){
    k = dim(X)[2]
    beta = theta[1:k]
    mu1 = X%*%beta
  }

  l = dim(W)[2]
  eta = theta[(k+1):(l+k)]
  mu2 = W%*%eta
  nu = theta[(l+k+1)]
  y = log(Z)


  # distribution of T

  g1 = lhat*exp(mu1)*exp(-cumL*exp(mu1))
  G1 = 1-exp(-cumL*exp(mu1))

  # dist of C

  if (dist == "Weibull"){                                           # Weibull
    g2 = 1/nu*exp((y-mu2)/nu)*exp(-exp((y-mu2)/nu))*(1/Z)           # density of C
    G2 = 1-exp(-exp((y-mu2)/nu))                                    # dist of C
  } else if (dist == "lognormal"){                                  # Log-normal
    m = (y-mu2)/nu
    g2 = 1/(sqrt(2*pi)*nu*Z)*exp(-(m^2/2))                          # density of C
    G2 = pnorm(m)                                                   # dist of C
  } else
    stop("The distribution of C is not supported")

  # Joint dist,

  G1 = pmax(1e-10,G1)
  G2 = pmax(1e-10,G2)

  Logn <- -sum(log(pmax(g1[d1==1], 1e-10)))-sum((log(1-G1[(1-d1)==1])))-sum(log(pmax(g2[d2==1], 1e-10)))-sum(log(1-G2[(1-d2)==1]))
  return(Logn)
}

#' @title Cumulative hazard function of survival time under dependent censoring

#' @description  This function estimates the cumulative hazard function of survival time (T) under dependent censoring (C). The estimation
#' makes use of the estimating equations derived based on martingale ideas.
#'
#' @param theta Estimated parameter values/initial values for finite dimensional parameters
#' @param resData Data matrix with three columns;  Z = the observed survival time, d1 = the censoring indicator of T
#' and  d2 =  the censoring indicator of C.
#' @param X Data matrix with covariates related to T
#' @param W Data matrix with covariates related to C. First column of W should be ones
#' @param cop Which copula should be computed to account for dependency between T and C. This argument can take
#'   one of the values from \code{c("Gumbel", "Frank", "Normal")}. The default copula model is "Frank".
#' @param dist The distribution to  be used for the dependent censoring C. Only two distributions are allowed, i.e, Weibull
#' and lognormal distributions. With the value \code{"Weibull"} as the
#'   default.
#' @importFrom copula pCopula frankCopula gumbelCopula
#'
#' @return This function returns an estimated hazard function, cumulative hazard function and distinct observed survival times;
#'
#' @examples
#' \donttest{
#' n = 200
#' beta = c(0.5)
#' lambd = 0.35
#' eta = c(0.9,0.4)
#' X = cbind(rbinom(n,1,0.5))
#' W = cbind(rep(1,n),rbinom(n,1,0.5))
#' frank.cop <- copula::frankCopula(param = 5,dim = 2)
#' U = copula::rCopula(n,frank.cop)
#' T1 = (-log(1-U[,1]))/(lambd*exp(X*beta))         # Survival time'
#' T2 = (-log(1-U[,2]))^(1.1)*exp(W%*%eta)          # Censoring time
#' A = runif(n,0,15)                                # administrative censoring time
#' Z = pmin(T1,T2,A)
#' d1 = as.numeric(Z==T1)
#' d2 = as.numeric(Z==T2)
#' resData = data.frame("Z" = Z,"d1" = d1, "d2" = d2)
#' theta = c(0.3,1,0.3,1,2)
#'
#' # Estimate cumulative hazard function
#'cumFit <- SolveL(theta, resData,X,W)
#'cumhaz = cumFit$cumhaz
#'time = cumFit$times
#'
#' # plot hazard vs time
#'
#'plot(time, cumhaz, type = "l",xlab = "Time",
#'ylab = "Estimated cumulative hazard function")
#'
#'}
#'
#' @export
SolveL = function(theta,resData,X,W,
                  cop = c("Frank", "Gumbel",  "Normal"),
                  dist = c("Weibull", "lognormal")){

  cop <- match.arg(cop)
  dist <- match.arg(dist)


  #  Verify column names of input dataframe resData
  if (!all(c("Z", "d1", "d2") %in% colnames(resData)))
    stop("Z, d1 and d2 arguments must be column names of resData")

  id <- match(c("Z", "d1", "d2"), colnames(resData))
  colnames(resData)[id] <- c("Z", "d1", "d2")

  if (is.vector(X)){
    k = 1
    mu1 = theta[1]*X
  }else {
    k = dim(X)[2]
    mu1 = X%*%theta[1:k]
  }

  Z = resData$Z
  d1 = resData$d1
  d2 = resData$d2

  T1 <- c(0,sort(unique(Z[d1 == 1])))
  m = length(T1)
  L = rep(0,m)
  L[1] = 0
  L[2] = sum((Z==T1[2]))/sum((Z>=T1[2])*exp(mu1))
  for (i in 3:m){
    csum = sum(L[1:(i-1)])
    psi = CompC(theta,T1[i],X,W,csum,cop,dist)
    L[i] <- sum(Z == T1[i])/sum((Z>=T1[i])*exp(psi))
  }
  res <- list(lambda = L,cumhaz = cumsum(L), times = T1)
}





#' @title Cumulative hazard function of survival time under independent censoring
#'
#' @description
#' This function estimates the cumulative hazard function of survival time (T) under the assumption of independent censoring.
#' The estimating equation is derived based on martingale ideas.
#'
#' @param theta Estimated parameter values/initial values for finite dimensional parameters
#' @param resData Data matrix with three columns;  Z = the observed survival time, d1 = the censoring indicator of T
#' and  d2 =  the censoring indicator of C.
#' @param X Data matrix with covariates related to T
#'
#'
#' @return This function returns an estimated hazard function,  cumulative hazard function and distinct observed survival times;
#'
#' @examples
#' \donttest{
#' n = 200
#' beta = c(0.5)
#' lambd = 0.35
#' eta = c(0.9,0.4)
#' X = cbind(rbinom(n,1,0.5))
#' W = cbind(rep(1,n),rbinom(n,1,0.5))
#' frank.cop <- copula::frankCopula(param = 5,dim = 2)
#' U = copula::rCopula(n,frank.cop)
#' T1 = (-log(1-U[,1]))/(lambd*exp(X*beta))           # Survival time'
#' T2 = (-log(1-U[,2]))^(1.1)*exp(W%*%eta)            # Censoring time
#' A = runif(n,0,15)                                  # administrative censoring time
#' Z = pmin(T1,T2,A)
#' d1 = as.numeric(Z==T1)
#' d2 = as.numeric(Z==T2)
#' resData = data.frame("Z" = Z,"d1" = d1, "d2" = d2)
#' theta = c(0.3,1,0.3,1)
#'
#' # Estimate cumulative hazard function
#'
#'cumFit_ind <- SolveLI(theta, resData,X)
#'
#' cumhaz = cumFit_ind$cumhaz
#' time = cumFit_ind$times
#'
#'# plot hazard vs time
#'
#'plot(time, cumhaz, type = "l",xlab = "Time",
#' ylab = "Estimated cumulative hazard function")
#'
#'}
#'
#'
#' @export
SolveLI = function(theta,resData,X){

  #  Verify column names of input dataframe resData
  if (!all(c("Z", "d1", "d2") %in% colnames(resData)))
    stop("Z, d1 and d2 arguments must be column names of resData")

  id <- match(c("Z", "d1", "d2"), colnames(resData))
  colnames(resData)[id] <- c("Z", "d1", "d2")

  if (is.vector(X)){
    k = 1
    mu1 = theta[1]*X
  }else{
    k = dim(X)[2]
    mu1 = X%*%theta[1:k]
  }
  Z = resData$Z
  d1 = resData$d1

  T1 <- c(sort(unique(Z[d1 == 1])))
  m = length(T1)
  L = rep(0,m)
  for (i in 1:m){
    L[i] <- sum(Z == T1[i])/sum((Z>=T1[i])*exp(mu1))
  }
  res <- list(lambda = L,cumhaz = cumsum(L), times = T1)
}



#' @title Compute phi function
#'
#' @description  This function estimates phi function at fixed time point t
#'
#' @inheritParams SolveL
#' @param t A fixed time point
#' @param ld Output of \code{\link{SolveL}} function at a fixed time t
#' @importFrom copula pCopula frankCopula gumbelCopula tau
#' @import pbivnorm
#' @noRd
CompC = function(theta,t,X,W,ld,cop,dist){

  if (is.vector(X)){
    k = 1
    beta = theta[1]
    # Distribution of T
    G1 = 1-exp(-ld*exp(X*beta))            # Cox model for T
    B0 = X*beta-ld*exp(X*beta)
  }else{
    k = dim(X)[2]
    beta = theta[1:k]
    # Distribution of T
    G1 = 1-exp(-ld*exp(X%*%beta))          # Cox model for T
    B0 = X%*%beta-ld*exp(X%*%beta)
  }
  l = dim(W)[2]
  eta = theta[(k+1):(l+k)]
  nu = theta[(l+k+1)]
  gm = theta[(l+k+2)]
  lfun = (W%*%eta)

  # Distribution of C

  if (dist == "Weibull"){                        # Weibull margin
    G2 = 1-exp(-exp((log(t)-W%*%eta)/nu))
  }else if (dist == "lognormal"){                # Log-normal margin
    G2 = pnorm((log(t)-W%*%eta)/nu)
  }else
    stop("Only Weibull and lognormal distributions are supported for C")


  # Avoid numerical issues
  G1[is.nan(G1)] = 1e-10
  G2[is.nan(G2)] = 1e-10

  G1[is.na(G1)] <- 1e-10
  G2[is.na(G2)] <- 1e-10
  G1 = pmax(1e-6,G1)
  G2 = pmax(1e-6,G2)
  G1 = pmin(G1,0.999999)
  G2 = pmin(G2,0.999999)

  PD = cbind(G1,G2)                         # Join distributions

  if (cop == "Frank")                       # Frank copula
  {
    frank.cop = frankCopula(gm, dim = 2)
    cp1 = (exp(- gm*G1)*(exp(- gm*G2)-1))/(exp(-gm)-1 + (exp(-gm*G1)-1)*(exp(-gm*G2)-1))
    cp1 = pmax(1e-10,cp1)
    z6 =  1-G1-G2+pCopula(PD,frank.cop)
    z6 =   pmax(z6,1e-10)
  }else if (cop == "Gumbel")                    # Gumbel copula
  {
    gumb.cop = gumbelCopula(gm, dim = 2)
    cp1 = pCopula(PD,gumb.cop)*((-log(G1))^gm+(-log(G2))^gm)^(-1+1/gm)*(-log(G1))^(gm-1)/G1
    cp1[is.na(cp1)] <- 1e-5
    cp1 = pmax(1e-5,cp1)
    z6 =  1-G1-G2+pCopula(PD,gumb.cop)
    z6 =   pmax(z6,1e-10)
  }else if (cop == "Normal")              # normal
  {
    cp1 = pnorm((qnorm(G2)-gm * qnorm(G1))/sqrt(1 - gm^2))
    p1 = qnorm(G1)
    p2 = qnorm(G2)
    c2 = cbind(p1,p2)

    z6 =  1-G1-G2+pbivnorm(c2,rho = gm)       # contribution to the likelihood from A
    z6 =   pmax(z6,1e-20)
  }else
    stop("Entered copula function is not supported")

  B1 = log(pmax(1e-10,1-cp1))-log(z6)
  tot = B0+B1
  return(tot)
}


#' @title Long format
#'
#' @description
#'  Change hazard and cumulative hazard to long format
#'
#' @param Z Observed survival time, which is the minimum of T, C and A, where A is the administrative censoring time.
#' @param T1 Distinct observed survival time
#' @param lhat Hazard function estimate
#' @param Lhat Cumulative hazard function estimate
#'
#' @noRd
Longfun = function(Z,T1,lhat,Lhat){
  n = length(Z)
  llong = rep(0,n)
  Llong = rep(0,n)

  for(i in 1:n){
    j = SearchIndicate(Z[i],T1)
    llong[i] = lhat[j]
    Llong[i] = Lhat[j]
  }
  long = cbind(llong,Llong)
}


#' @title Distance between vectors
#' @description This function computes distance between two vectors based on L2-norm
#' @param a First vector
#' @param b Second vector
#' @noRd
Distance = function(a,b){
  x = b-a
  n = length(x)
  l2norm = sqrt(sum(x^2)/n)
  return(l2norm)
}

#' @title Estimate a nonparametric transformation function
#'
#' @description This function estimates the nonparametric transformation  function H when the survival time and censoring time
#' are dependent given covariates.  The estimating equation of H was derived based on the martingale ideas. More details about
#' the derivation of a nonparmaetric estimator of H and its estimation algorithm can be found in Deresa and Van Keilegom (2021).
#'
#' @references  Deresa, N. and Van Keilegom, I. (2021). On semiparametric modelling, estimation and inference for survival data subject to dependent censoring, Biometrika, 108, 965–979.
#'
#' @param theta Vector of parameters in the semiparametric transformation model.
#' @param resData Data matrix with three columns;  Z = the observed survival time, d1 = the censoring indicator of T
#'   and  d2 =  the censoring indicator of C.
#' @param X Data matrix with covariates related to T.
#' @param W Data matrix with covariates related to C.
#' @importFrom stats pnorm  dnorm qnorm sd
#' @import pbivnorm
#' @import nleqslv
#'
#' @return Returns the estimated transformation function H for a fixed value of parameters theta.
#'
#' @noRd
SolveH <- function(theta,resData, X, W){                                 # Z,nu,X,W,beta,eta,rho
  k = ncol(X)
  l = ncol(W)
  beta = theta[1:k]
  eta = theta[(k+1):(l+k)]
  rho = theta[(l+k+1)]
  Z = resData$Z
  d1 = resData$d1
  d2 = resData$d2
  n = length(Z)
  nu = d1+d2
  T1 = c(-Inf,Z[nu==1])
  T1 = unique(T1)                                                     # distinct observed values of Y
  m = length(T1);
  H = rep(0,m);
  H[1] = -Inf;                                                        # for the first observed time, H goes to negative infinity by assumption
  # H[2] = uniroot(SolveHt1,interval = c(-7,7),Z=Z,nu=nu,t=T1[2],X=X,W=W,beta=beta,eta=eta,rho=rho)$root
  H[2] = nleqslv(0,SolveHt1,jac = NULL,Z=Z,nu=nu,t=T1[2],X=X,W=W,theta)$x;

  for (k in 3:m)
  {
    Sum = 0
    j = k-1
    for (i in 1:n)
    {
      upx = H[j]-X[i,]%*%beta
      upy = H[j]-W[i,]%*%eta
      bvprob = Bvprob(upx,upy,rho);
      der = (-dnorm(upy)*(1-pnorm((upx-rho*upy)/(1-rho^2)^0.5))-dnorm(upx)*(1-pnorm((upy-rho*upx)/(1-rho^2)^0.5)))
      if (bvprob!=0)
        Sum = Sum +(Z[i]>=T1[k])/bvprob*der
      else
        Sum = Sum
    }
    if (Sum!=0)
    {
      H[k] = H[j]-sum((Z==T1[k])*(nu==1))/Sum
    }else {
      H[k] = H[j]
    }
  }
  return(cbind(T1,H))
}

#' @title Estimating equation for Ht1
#'
#' @description This function obtains an estimating equation of H at the first observed survival time t1.
#'
#' @param Ht1 The solver solves for an optimal value of Ht1 by equating the estimating equation to zero.
#' @param Z The observed survival time, which is the minimum of T, C and A.
#' @param nu The censoring indicator for T or C
#' @param t A fixed time point
#' @param theta Vector of parameters
#' @param X Data matrix with covariates related to T.
#' @param W Data matrix with covariates related to C.
#'
#' @noRd
SolveHt1 <- function(Ht1,Z,nu,t,X,W,theta){
  k = ncol(X)
  l = ncol(W)
  beta = theta[1:k]
  eta = theta[(k+1):(l+k)]
  rho = theta[(l+k+1)]
  upx = Ht1-X%*%beta;
  upy = Ht1-W%*%eta;
  dN1 = sum((Z==t)*nu)
  c = cbind(upx,upy)
  temp = 1- pnorm(upx)-pnorm(upy)+ pbivnorm(c,rho=rho)
  return(dN1+sum((Z>=t)*log(temp)))
}


#' @title Search function
#'
#' @description Function to indicate position of t in observed survival time
#'
#' @param t fixed time t
#' @param T1 distinct observed survival time
#'
#' @noRd
SearchIndicate = function(t,T1){
  i = 1;
  m = length(T1);

  while(T1[i+1]<=t){
    i = i+1;
    if (i == m) break;
  }
  return(i)
}

#' @title Distance between vectors
#' @description This function computes distance between two vectors based on L2-norm
#' @param a First vector
#' @param b Second vector
#'
#' @noRd
Distance = function(b,a){   # L2-norm of difference
  x = b-a
  n = length(x)
  l2norm = sqrt(sum(x^2)/n)
  return(l2norm)
}

#' @title Compute bivariate survival probability
#' @description This function calculates a bivariate survival probability based on multivariate normal distribution.
#' @param lx The first lower bound of integration
#' @param ly The second lower bound
#' @param rho  Association parameter
#' @import MASS
#' @import mvtnorm
#'
#' @noRd
Bvprob = function(lx,ly,rho) { # compute bivariate prob.
  cor = diag(2)
  cor[1,2] = rho
  cor[2,1] = rho
  bvprob = pmvnorm(lower=c(lx,ly), upper=c(Inf,Inf), mean=c(0,0), corr = cor,algorithm = "GenzBretz");
  return(bvprob)
}

#' @title Change H to long format
#'
#' @description
#'  Change a nonparametric transformation function to long format
#'
#' @param Z Observed survival time, which is the minimum of T, C and A, where A is the administrative censoring time.
#' @param T1 Distinct observed survival time
#' @param H Nonparametric transformation function estimate
#'
#' @noRd
LongNPT = function(Z,T1,H){
  n = length(Z)
  Hlong = rep(0,n)

  for(i in 1:n){
    j = SearchIndicate(Z[i],T1)
    Hlong[i] = H[j]
  }
return(Hlong)
}


#' @title Score equations of finite parameters
#' @description This function computes the score vectors  and the Jacobean matrix for finite model parameters.
#'
#' @param theta Vector of parameters in the semiparametric transformation model.
#' @param resData Data matrix with three columns;  Z = the observed survival time, d1 = the censoring indicator of T
#'   and  d2 =  the censoring indicator of C.
#' @param X Data matrix with covariates related to T.
#' @param W Data matrix with covariates related to C.
#' @param H The estimated non-parametric transformation function for a given value of theta
#' @importFrom stats pnorm  dnorm qnorm sd
#'
#' @noRd
ScoreEqn = function(theta,resData,X,W,H){
  k = ncol(X)
  l = ncol(W)
  u = k+l
  beta = theta[1:k]
  eta = theta[(k+1):u]
  rho = theta[(u+1)]
  Z = resData$Z
  d1 = resData$d1
  d2 = resData$d2
  nu = d1+d2
  T1 = c(-Inf,Z[nu==1])
  T1 = unique(T1)               # distinct observed survival time
  n = length(Z);
  nu = d1+d2;

  VBeta = rep(0,k); VEta = rep(0,l);   VRho = 0;
  V = diag(1,(u+1));

  for(i in 1:n){

    up1 = H[SearchIndicate(Z[i],T1)]-X[i,]%*%beta;
    up2 = H[SearchIndicate(Z[i],T1)]-W[i,]%*%eta;

    if (up1>4) up1 = Inf
    if (up2>4) up2 = Inf
    if (up1<=-4.1) up1 = -Inf
    if (up2<=-4.1) up2 = -Inf
    up21 = (up2-up1*rho)/(1-rho^2)^0.5
    up12 = (up1-up2*rho)/(1-rho^2)^0.5
    if (up1==-Inf & up2==-Inf){
      up21 = 0
      up12 = 0
    }
    if (up1==Inf & up2==Inf){
      up21 = 0
      up12 = 0
    }
    if (up1>-Inf & up2>-Inf&up1<Inf&up2<Inf)
    {
      if(d1[i]==1 & d2[i] == 0)
      {
        SB = c(up1)*c(X[i,])-c(dnorm(up21)/(1-pnorm(up21)))*rho/(1-rho^2)^0.5*c(X[i,]);
        SE = c(1/(1-pnorm(up21))*dnorm(up21))*c(W[i,])/(1-rho^2)^0.5;
        SR = 1/(1-pnorm(up21))*(-dnorm(up21))*(rho/(1-rho^2)^1.5*up2-up1/(1-rho^2)^0.5-rho^2/(1-rho^2)^1.5*up1);
        VBeta = VBeta+SB;
        VEta = VEta+SE;
        VRho = VRho+SR;
        S = c(SB,SE,SR);
        V = V + S%*%t(S);
      }
      if(d1[i]==0 & d2[i] == 1)
      {
        SB = c(1/(1-pnorm(up12))*dnorm(up12))*c(X[i,])/(1-rho^2)^0.5
        SE = c(up2)*W[i,]-c(dnorm(up12)/(1-pnorm(up12)))*rho/(1-rho^2)^0.5*W[i,]
        SR = 1/(1-pnorm(up12))*(-dnorm(up12))*(rho/(1-rho^2)^1.5*up1-up2/(1-rho^2)^0.5-rho^2/(1-rho^2)^1.5*up2);
        VBeta = VBeta+SB;
        VEta = VEta+SE;
        VRho = VRho+SR;
        S = c(SB,SE,SR);
        V = V + S%*%t(S);
      }
      if(d1[i]==0 & d2[i] == 0)
      {
        temp = Bvprob(up1,up2,rho)[1];
        SB = c(dnorm(up1))*X[i,]*c(1-pnorm(up21))/temp;
        SE = c(dnorm(up2))*W[i,]*c(1-pnorm(up12))/temp;
        SR = 1/(1-rho^2)^0.5*dnorm(up1)*dnorm(up21)/temp;
        VBeta = VBeta+SB;
        VEta = VEta+SE;
        VRho = VRho+SR;
        S = c(SB,SE,SR);
        V = V + S%*%t(S);
      }
    }
    if (up1>-Inf & up2 == -Inf)
    {
      if(d1[i]==1 & d2[i] == 0)
      {
        SB = c(up1)*X[i,];
        SE = rep(0,k);
        SR = 0;
        VBeta = VBeta+SB;
        VEta = VEta+SE;
        VRho = VRho+SR;
        S = c(SB,SE,SR);
        V = V + S%*%t(S);
      }
      if(d1[i]== 0 & d2[i] == 1)
      {
        SB = rep(0,k);
        SE = rep(0,k);
        SR = 0;
        VBeta = VBeta+SB;
        VEta = VEta+SE;
        VRho = VRho+SR;
        S = c(SB,SE,SR);
        V = V + S%*%t(S);
      }
      if(d1[i]== 0 & d2[i] == 0)
      {
        SB = c(dnorm(up1))*X[i,]/c(1-pnorm(up1))
        SE = rep(0,k);
        SR = 0;
        VBeta = VBeta+SB;
        VEta = VEta+SE;
        VRho = VRho+SR;
        S = c(SB,SE,SR);
        V = V + S%*%t(S);
      }
    }
    if (up1 == -Inf & up2>-Inf)
    {
      if(d1[i]== 1 & d2[i] == 0)
      {
        SB = rep(0,k);
        SE = rep(0,k);
        SR = 0;
        VBeta = VBeta+SB;
        VEta = VEta+SE;
        VRho = VRho+SR;
        S = c(SB,SE,SR);
        V = V + S%*%t(S);
      }
      if(d1[i]== 0 & d2[i] == 1)
      {
        SB = rep(0,k);
        SE = c(up2)*W[i,];
        SR = 0;
        VBeta = VBeta+SB;
        VEta = VEta+SE;
        VRho = VRho+SR;
        S = c(SB,SE,SR);
        V = V + S%*%t(S);
      }
      if (d1[i] == 0 & d2[i] == 0)
      {
        SB = rep(0,k);
        SE = c(dnorm(up2))*W[i,]/c(1-pnorm(up2))
        SR = 0;
        VBeta = VBeta+SB;
        VEta = VEta+SE;
        VRho = VRho+SR;
        S = c(SB,SE,SR);
        V = V + S%*%t(S);
      }
    }
    if(up1 == -Inf & up2 == -Inf)
    {
      SB = rep(0,k);
      SE = rep(0,k);
      SR = 0;
      SR = 0;
      VBeta = VBeta+SB;
      VEta = VEta+SE;
      VRho = VRho+SR;
      S = c(SB,SE,SR);
      V = V + S%*%t(S);
    }
    if (up1 == Inf & up2>-4& up2<4)
    {
      if(d1[i] == 1 & d2[i] == 0)
      {
        SB = rep(0,k);
        SE = rep(0,k);
        SR = 0;
        VBeta = VBeta+SB;
        VEta = VEta+SE;
        VRho = VRho+SR;
        S = c(SB,SE,SR);
        V = V + S%*%t(S);
      }
      if(d1[i]== 0 & d2[i] == 1)
      {
        SB = rep(0,k);
        SE = c(up2)*W[i,]
        SR = 0;
        VBeta = VBeta+SB;
        VEta = VEta+SE;
        VRho = VRho+SR;
        S = c(SB,SE,SR);
        V = V + S%*%t(S);
      }
      if (d1[i] ==0 & d2[i] == 0)
      {
        SB = rep(0,k);
        SE = rep(0,k)
        SR = 0;
        VBeta = VBeta+SB;
        VEta = VEta+SE;
        VRho = VRho+SR;
        S = c(SB,SE,SR);
        V = V + S%*%t(S);
      }
    }
    if (up1 == Inf & up2 == Inf)
    {
      if(d1[i] == 1 & d2[i] == 0)
      {
        SB = rep(0,k);
        SE = rep(0,k);
        SR = 0;
        VBeta = VBeta+SB;
        VEta = VEta+SE;
        VRho = VRho+SR;
        S = c(SB,SE,SR);
        V = V + S%*%t(S);
      }
      if(d1[i]== 0 & d2[i] == 1)
      {
        SB = rep(0,k);
        SE = rep(0,k);
        SR = 0;
        VBeta = VBeta+SB;
        VEta = VEta+SE;
        VRho = VRho+SR;
        S = c(SB,SE,SR);
        V = V + S%*%t(S);
      }
      if (d1[i] ==0 & d2[i] == 0)
      {
        SB = rep(0,k);
        SE = rep(0,k)
        SR = 0;
        VBeta = VBeta+SB;
        VEta = VEta+SE;
        VRho = VRho+SR;
        S = c(SB,SE,SR);
        V = V + S%*%t(S);
      }
    }
  }
  out = list(Vbeta = VBeta,Veta = VEta, Vrho = VRho,V1 = V)
  return(out)
}


#' @title  Estimate finite parameters based on score equations
#' @description This function estimates the model parameters
#'
#' @inheritParams SolveH
#' @param H The estimated non-parametric transformation function for a given value of theta.
#' @param eps Convergence error.
#'
#' @noRd
SolveScore = function(theta,resData,X,W,H, eps = 1e-3){   # Estimate model parameters

  PEst = ScoreEqn(theta,resData,X,W,H)

  VBeta = PEst$Vbeta
  VEta = PEst$Veta
  VRho = PEst$Vrho
  S = c(VBeta,VEta,VRho)
  V = PEst$V1
  b = theta+solve(V)%*%S
  m = length(b)
  if(b[m]>0.98) b[m] = 0.98

  step = 0
  a = theta
  while (Distance(b,a)>eps){
    a = b
    step = step+1
    PEst = ScoreEqn(a,resData,X,W,H)
    VBeta = PEst$Vbeta
    VEta = PEst$Veta
    VRho = PEst$Vrho
    S = c(VBeta,VEta,VRho)
    V = PEst$V1
    b = a+solve(V)%*%S
    if(b[m]>0.98) b[m] = 0.98    # Perfect correlation can lead to error
    if (step>=10)                # Max ten steps
      break;
  }
  return(b)
}

#' @title Fit a semiparametric transformation model for dependent censoring
#'
#' @description This function allows to estimate the dependency parameter along all other model parameters. First, estimates a non-parametric transformation function, and
#' then at the second stage it estimates other model parameters assuming that the non-parametric function is known. The details for
#' implementing the dependent censoring methodology can be found in Deresa and Van Keilegom (2021).
#'
#' @references Deresa, N. and Van Keilegom, I. (2021). On semiparametric modelling, estimation and inference for survival data subject to dependent censoring, Biometrika, 108, 965–979.
#'
#' @param start Initial values for the finite dimensional parameters. If \code{start} is NULL, the initial values will be obtained
#' by fitting an Accelerated failure time models.
#' @param resData Data matrix with three columns;  Z = the observed survival time, d1 = the censoring indicator of T
#' and  d2 =  the censoring indicator of C.
#' @param X Data matrix with covariates related to T
#' @param W Data matrix with covariates related to C
#' @param eps Convergence error. This is set by the user in such away that the desired convergence is met; the default is \code{eps = 1e-3}.
#' @param n.iter Number of iterations; the default is \code{n.iter = 20}. The larger the number of iterations, the longer the computational time.
#' @param bootstrap A boolean indicating whether to compute bootstrap standard errors for making inferences.
#' @param n.boot Number of bootstrap samples to use in the estimation of bootstrap standard errors if \code{bootstrap = TRUE}. The default is n.boot = 50. But, higher
#' values  of \code{n.boot} are recommended for obtaining good estimates of bootstrap standard errors.
#' @importFrom stats pnorm  qnorm sd
#' @importFrom survival survreg Surv
#' @importFrom MASS mvrnorm
#' @import pbivnorm nleqslv
#'
#' @return This function returns a fit of a semiparametric transformation model; parameter estimates, estimate of the non-parametric transformation function, bootstrap standard
#' errors for finite-dimensional parameters, the nonparametric cumulative hazard function, etc.
#'
#' @examples
#' \donttest{
#' # Toy data example to illustrate implementation
#' n = 300
#' beta = c(0.5, 1); eta = c(1,1.5); rho = 0.70
#' sigma = matrix(c(1,rho,rho,1),ncol=2)
#' err = MASS::mvrnorm(n, mu = c(0,0) , Sigma=sigma)
#' err1 = err[,1]; err2 = err[,2]
#' x1 = rbinom(n,1,0.5); x2 = runif(n,-1,1)
#' X = matrix(c(x1,x2),ncol=2,nrow=n); W = X   # data matrix
#' T1 = X%*%beta+err1
#' C =  W%*%eta+err2
#' T1 = exp(T1); C = exp(C)
#' A = runif(n,0,8); Y = pmin(T1,C,A)
#' d1 = as.numeric(Y==T1)
#' d2 = as.numeric(Y==C)
#' resData = data.frame("Z" = Y,"d1" = d1, "d2" = d2)   # should be data frame
#' colnames(X) = c("X1", "X2")
#' colnames(W) = c("W1","W2")
#'
#' #  Bootstrap is false by default
#' output = NonParTrans(resData = resData, X = X, W = W, n.iter = 2)
#' output$parameterEstimates
#'
#' }
#
#' @export
NonParTrans = function(resData, X, W, start = NULL, n.iter = 15,  bootstrap = FALSE, n.boot = 50, eps = 1e-3){
  X = as.matrix(X)
  W = as.matrix(W)

  #  Verify column names of input dataframe resData
  if (!all(c("Z", "d1", "d2") %in% colnames(resData)))
    stop("Z, d1 and d2 arguments must be column names of resData")

  id <- match(c("Z", "d1", "d2"), colnames(resData))
  colnames(resData)[id] <- c("Z", "d1", "d2")

  # Make  sure that the data is ordered
  resData= resData[order(resData$Z),]
  Z = resData$Z
  d1 = resData$d1
  d2 = resData$d2
  X = X[order(Z),]
  W = W[order(Z),]

  if(is.null(start)){
    fit1 = survreg(Surv(Z,d1)~X, dist ="lognormal")
    fit2 = survreg(Surv(Z,d2)~W, dist = "lognormal")
    start = c(fit1$coefficients[-1],fit2$coefficients[-1],0.3)
  }
  l = length(start)
  if(start[l] < -1 | start[l]>1) start[l] = 0.3

  TransHat = SolveH(start, resData, X, W)
  T1 = TransHat[,1]
  H = TransHat[,2]
  parhat = SolveScore(start, resData,X,W, H, eps = eps)

  flag = 0;
  b = parhat
  a = start
  m = length(b)
  if(b[m]>0.99){
    b[m] = 0.98
  }else if (b[m] <= -0.99){
    b[m] = -0.98
    }else b = b
  while (Distance(b,a)>eps){
    a = b
    TransHat = SolveH(a, resData, X, W)
    H = TransHat[,2]
    parhat = SolveScore(a, resData, X, W, H, eps = eps)
    b = parhat
    if(b[m]>0.99){
      b[m] = 0.98
    }else if (b[m]<= -0.99){
      b[m] = -0.98
    }else b[m] = b[m]

    flag = flag+1
    if (flag>n.iter)
    {
      flag=0;
      break;
    }
  }
  TransHat = SolveH(parhat, resData, X, W)

  # Nonparametric bootstrap for making inference

  paramsBootstrap = list("init" = parhat,"resData" = resData, "X" = X, "W" = W,"n.boot" = n.boot, "n.iter" = n.iter, "eps" = eps)

  fitNonTrans <- NULL
  if(bootstrap) {                                                          # Obtain bootstrap standard error
    fitNonTrans <- do.call(boot.nonparTrans, paramsBootstrap)
    stError <- fitNonTrans$parEst.std.error
    Z <- parhat/stError
    pvalue <- 2*(1-pnorm(abs(Z)))
    out.res <- cbind(parhat,stError,pvalue)
    rownames(out.res) <- c(colnames(X), colnames(W),"rho")
    colnames(out.res) <- c("coef.", "BootSE", "Pvalue")
    output <- list("parameterEstimates" = out.res, "bootstrap" = bootstrap, "dimX" = ncol(X), "dimW" = ncol(W),"observedTime" = T1,"H" = H)
  }else{
    rownames(parhat) <- c(colnames(X), colnames(W),"rho")
    output <- c(list("parameterEstimates" = parhat, "bootstrap" = bootstrap, "dimX" = ncol(X), "dimW" = ncol(W),"observedTime" = T1,"H" = H),fitNonTrans)
  }
  class(output) <- append(class(output), "fitNonPar")                              # dependent censoring fit object

  return(output)

}


#' @title Nonparametric bootstrap approach for a Semiparametric transformation model under dependent censpring

#' @description This function estimates the bootstrap standard errors for the finite-dimensional model parameters and for the non-parametric transformation
#' function. Parallel computing using foreach has been used to speed up the estimation of standard errors.
#'
#'
#' @param init Initial values for the finite dimensional parameters obtained from the fit of \code{\link{NonParTrans}}
#' @param resData Data matrix with three columns;  Z = the observed survival time, d1 = the censoring indicator of T
#' and  d2 =  the censoring indicator of C.
#' @param X Data matrix with covariates related to T
#' @param W Data matrix with covariates related to C.
#' @param eps Convergence error. This is set by the user in such away that the desired convergence is met; the default is \code{eps = 1e-3}
#' @param n.iter Number of iterations; the default is \code{n.iter = 15}. The larger the number of iterations, the longer the computational time.
#' @param n.boot Number of bootstraps to use in the estimation of bootstrap standard errors.
#' @importFrom stats pnorm  qnorm sd
#' @importFrom survival survreg Surv
#' @import pbivnorm MASS mvtnorm nleqslv foreach parallel
#'
#' @return Bootstrap standard errors for parameter estimates and for estimated cumulative hazard function.
#'
#' @noRd
boot.nonparTrans = function(init,resData,X,W,n.boot, n.iter, eps){
  B = n.boot                                               # number of bootstrap samples
  n.cores <- parallel::detectCores() - 1

  my.cluster <- parallel::makeCluster(
    n.cores,
    type = "PSOCK"
  )
  doParallel::registerDoParallel(cl = my.cluster)

  boot = foreach(b = 1:B, .packages= c('survival', 'pbivnorm','nleqslv','MASS','mvtnorm'), .export = c("SolveH","SolveScore","Bvprob","Distance","ScoreEqn","LongNPT" ,"SearchIndicate","SolveHt1")) %dopar% {
    samp1 = sample(length(resData$Z),replace = TRUE)
    resData_b = resData[samp1,]
    resData_b = resData_b[order(resData_b$Z),]
    Zb = resData_b$Z
    Xb = X[samp1,]
    Wb = W[samp1,]
    Xb = Xb[order(Zb),]
    Wb = Wb[order(Zb),]
    nu = resData_b$d1+resData_b$d2
    Tb1 = c(-Inf,Zb[nu==1])
    Tb1 = unique(Tb1)       # distinct observed survival time

    TransHat = SolveH(init, resData_b, Xb, Wb)
    T1 = TransHat[,1]
    H = TransHat[,2]
    parhat = SolveScore(init, resData_b,Xb,Wb, H, eps)

    flag = 0;
    b = parhat
    a = init
    m = length(b)
    if(b[m]>0.99){
      b[m] = 0.98
    }else if (b[m] <= -0.99){
      b[m] = -0.98
    }else b = b

    while (Distance(b,a)>eps){
      a = b
      TransHat = SolveH(a, resData_b, Xb, Wb)
      H = TransHat[,2]
      parhat = SolveScore(a, resData_b, Xb, Wb, H, eps)
      b = parhat
      if(b[m]>0.99){
        b[m] = 0.98
      }else if (b[m]<= -0.99){
        b[m] = -0.98
      }else b[m] = b[m]

      flag = flag+1
      if (flag>n.iter)
      {
        flag=0;
        break;
      }
    }
    longB = LongNPT(Zb,Tb1,H)
    longB[longB == -Inf] = -5
    list("beta.star" = b, "H.star" = longB)
  }
  parallel::stopCluster(cl = my.cluster)

  beta.star = t(sapply(boot, function(x) x$beta.star))
  H.star =  t(sapply(boot, function(x) x$H.star))
  Bootse = apply(beta.star,2,sd)
  Hse = apply(H.star,2,sd)
  bootR = list("parEst.std.error" = Bootse, "H.std.error" = Hse)
  return(bootR)
}

#' @title Semiparametric Estimation of the Survival Function under Dependent Censoring
#' @description Provide semiparametric approaches that can be used to model right-censored survival data under dependent censoring (without covariates).
#'   The copula-based approach is adopted and there is no need to explicitly specify the association parameter.
#'   One of the margins can be modeled nonparametrically. As a byproduct, both marginal distributions of survival and
#'   censoring times can be considered as fully parametric. The existence of a cured fraction concerning survival time
#'   can also be taken into consideration.
#'
#'  @references Czado and Van Keilegom (2023). Dependent censoring based on parametric copulas. Biometrika, 110(3), 721-738.
#'  @references Delhelle and Van Keilegom (2024). Copula based dependent censoring in cure models. TEST (to appear).
#'  @references Ding and Van Keilegom (2024). Semiparametric estimation of the survival function under dependent censoring (in preparation).
#'
#' @param yobs a numeric vector that indicated the observed survival times.
#' @param delta a numeric vector that stores the right-censoring indicators.
#' @param tm a numeric vector that contains interested non-negative time points at which the survival probabilities will be evluated.
#' Note that if we omit the definition of this argument (the default value becomes \code{NULL}), our function will automatically output survival probabilities at all oberserved time points, that is, \code{yobs}.
#' @param copfam a character string that specifies the copula family.
#'   Currently, it supports Archimedean copula families, including \code{"frank"} (the default value), \code{"clayton"}, \code{"gumbel"}, and \code{"joe"}.
#'   The degenerated independent censoring case can be considered as well by setting \code{"indep"}.
#'   (other options will be added in the near future!)
#' @param margins a list used to define the distribution structures of both the survival and censoring margins.
#'   Specifically, it contains the following elements:
#'   \describe{
#'     \item{\code{survfam}}{ a character string that defines the assumed distribution for the survival time random variable,
#'       including \code{"lnorm"} for log-normal distribution, \code{"weibull"} for weibull distribution (other options will be added in the near future).}
#'     \item{\code{censfam}}{ a character string that defines the assumed distribution for the censoring time random variable, and the details are the same as those shown in \code{survfam}.}
#'     \item{\code{survtrunc}}{ a positive numeric value thats denotes the value of truncation for the assumed distribution, that is, \code{survfam}.}
#'     \item{\code{censtrunc}}{ a positive numeric value thats denotes the value of truncation for the assumed distribution, that is, \code{censfam}.}
#'   }
#'   Note if one of the marginal distributions should be modeled nonparametrically, one can let the corresponding argument to be \code{NULL} directly.
#'   For example if a semiparametric framework that defines the survival margin to be nonparametric and the censoring margin to be parametric, say log-normal, is desired,
#'   we can let \code{survfam = NULL} and \code{censfam = "lnorm"}, which is indeed the default value.
#'   Furthermore, if no truncation is imposed in \code{survfam} (or \code{censfam}), one can directly omit the specification of \code{survtrunc} (or \code{censtrunc}), which is the default specification.
#'   We also remark here that when a cured fraction is included (\code{cure = TRUE}), if \code{survfam} is not \code{NULL} and \code{survtrunc = NULL}, we will automatically let \code{survtrunc} to be \code{max(yobs)}.
#'   If we wants to model the data with a non-truncated survival distribution when there is a cured fraction, we can set \code{survtrunc = Inf}.
#' @param cure a logical value that indicates whether the existence of a cured fraction should be considered.
#' @param Var a list that controls the execution of the bootstrap for variance estimation,
#'     and it contains two elements:
#'     \code{do} is a logical value with default \code{FALSE} to tell the function whether the boostrap-based variances should be calculated;
#'     \code{nboot} is a numeric integer that specifies the number of bootstrap samples.
#' @param control indicates more detailed control of the underlying model fitting procedures.
#'   It is a list of the following three arguments:
#'   \describe{
#'     \item{\code{maxit}}{ a positive integer that denotes the maximum iteration number in optimization. The default value is \code{300}.}
#'     \item{\code{eps}}{ a positive small numeric value that denotes the tolerance for convergence. The default value is \code{1e-6}.}
#'     \item{\code{trace}}{ a logical value that judges whereh the tracing information on the progress of the model fitting should be produced. The default value if \code{TRUE}.}
#'     \item{\code{ktau.inits}}{ a numeric vector that contains initial values of the Kendall's tau.
#'       The default value is \code{NULL}, meaning that a grids of initial values will be automatically generated within our function.}
#'   }
#'
#' @details
#'
#'   This unified function provide approaches that can be used to model right-censored survival data under dependent censoring (without covariates).
#'   Various specifications of marginal distributions can be considered by choosing different combinations of the provided arguments.
#'   Generally speaking, the following two scenarios are what we mainly focused on:
#'   \describe{
#'     \item{\code{nonparametric survival margin and parametric censoring margin (without cure)}}{
#'       \code{survfam = NULL}, \code{censfam} is not \code{NULL} and \code{cure = FALSE}.}
#'     \item{\code{nonparametric survival margin and parametric censoring margin (with cure)}}{
#'       \code{survfam = NULL}, \code{censfam} is not \code{NULL} and \code{cure = TRUE}.}
#'   }
#'   As byproducts, several other scenarios (the distribution of the underlying survival time is not nonparametric
#'   but fully parametric) can also be considered by this R function:
#'   \describe{
#'     \item{\code{parametric survival and censoring margins (without cure)}}{
#'       both \code{survfam} and \code{censfam} are not \code{NULL} and \code{cure = FALSE}.}
#'     \item{\code{parametric survival and censoring margins (with cure)}}{
#'       both \code{survfam} and \code{censfam} are not \code{NULL} and \code{cure = TRUE}.}
#'     \item{\code{parametric survival margin and nonparametric censoring margin (without cure)}}{
#'       \code{survfam} is not \code{NULL}, \code{censfam = NULL} and \code{cure = FALSE}.}
#'   }
#'   Furthermore, one might expect that a scenario with "parametric survival margin and nonparametric censoring margin
#'   (with cure)" can also be included. Indeed, it can be done based on: \code{survfam} is not \code{NULL}, \code{censfam = NULL}
#'   and \code{cure = TRUE}. However, from a theoretical perspective of view, whether this type of modeling is reasonable or not
#'   still needs further investigations.
#'
#'   We emphasize that the first scenario (in byproducts) has also be considered in another function of this package.
#'   Specifically, the scenario of "parametric survival margin and nonparametric censoring margin (without cure)" can be
#'   fitted based on \code{ParamCop()}. However, the default joint modeling of survival and censoring times are based on
#'   their joint survival function in line with the semiparametric case (instead of modeling joint distribution function
#'   directly as in  Czado and Van Keilegom (2023) <doi:10.1093/biomet/asac067>), but the idea of estimation methodology
#'   are exactly the same.
#'
#' @return
#'   A list of fitted results is returned.
#'   Within this outputted list, the following elements can be found:
#'   \describe{
#'     \item{\code{probs}}{survival probabilities of the survial margin at \code{tm}.}
#'     \item{\code{ktau}}{Kendall's tau.}
#'     \item{\code{parapar}}{estimation of all parameters (except Kendall's tau) contained in the parametric part.}
#'     \item{\code{GoF}}{goodness-of-test results.}
#'     \item{\code{curerate}}{cure rate. If \code{cure = FALSE}, it is \code{NULL}.}
#'   }
#'
#' @importFrom methods is
#' @importFrom stats dlnorm dweibull integrate plnorm pnorm pweibull sd uniroot na.omit quantile constrOptim
#' @importFrom Matrix bdiag
#' @importFrom EnvStats qlnormTrunc
#' @importFrom copula rCopula archmCopula
#'
#' @examples
#'
#' \donttest{
#' #----------------------------------------------------------#
#' # Basic preparations before running subsequent examples ####
#' #----------------------------------------------------------#
#'
#' # library necessary packages
#'
#' #------------------------------------------------------------------------#
#' # simulated data from Frank copula log-Normal margins (without cure)
#' #------------------------------------------------------------------------#
#'
#' # generate the simulated data
#'
#' # - the sample size of the generated data
#' n <- 1000
#'
#' # information on the used copula
#' copfam.true <- "frank"
#' ktau.true <- 0.5
#' coppar.true <- 5.74
#'
#' # parameters of the underlying log-normal marginal distributions
#' survpar.true <- c(2.20,1.00)
#' censpar.true <- c(2.20,0.25)
#'
#' # - true underlying survival and censoring times
#' set.seed(1)
#' u.TC <- copula::rCopula(
#'   n        = n,
#'   copula   = copula::archmCopula(
#'     family = copfam.true,
#'     param  = coppar.true,
#'     dim    = 2
#'   )
#' )
#' yobs.T <- qlnorm(1-u.TC[,1],survpar.true[1],survpar.true[2])
#' yobs.C <- qlnorm(1-u.TC[,2],censpar.true[1],censpar.true[2])
#'
#' # observations
#' yobs  <- pmin(yobs.T,yobs.C)
#' delta <- as.numeric(yobs.T<=yobs.C)
#' cat("censoring rate is", mean(1-delta))
#'
#' # model the data under different scenarios
#'
#' # scenario 1: nonparametric survival margin and parametric censoring margin
#' set.seed(1)
#' sol.scenario1 <- SurvDC(
#'   yobs    = yobs,
#'   delta   = delta,
#'   tm      = quantile(yobs, c(0.25,0.50,0.75)),
#'   copfam  = copfam.true,
#'   margins = list(survfam = NULL, censfam = "lnorm"),
#'   Var     = list(do = FALSE, nboot = 50)
#' )
#' sol.scenario1$probs
#' sol.scenario1$ktau
#' sol.scenario1$parapar
#'
#' # scenario 2: parametric survival and censoring margins
#' set.seed(1)
#' sol.scenario2 <- SurvDC(
#'   yobs    = yobs,
#'   delta   = delta,
#'   tm      = quantile(yobs, c(0.25,0.50,0.75)),
#'   copfam  = copfam.true,
#'   margins = list(survfam = "lnorm", censfam = "lnorm"),
#'   Var     = list(do = FALSE, nboot = 50)
#' )
#' sol.scenario2$probs
#' sol.scenario2$ktau
#' sol.scenario2$parapar
#'
#' # scenario 3: parametric survival margin and nonparametric censoring margin
#' set.seed(1)
#' sol.scenario3 <- SurvDC(
#'   yobs    = yobs,
#'   delta   = delta,
#'   tm      = quantile(yobs, c(0.25,0.50,0.75)),
#'   copfam  = copfam.true,
#'   margins = list(survfam = "lnorm", censfam = NULL),
#'   Var     = list(do = FALSE, nboot = 50)
#' )
#' sol.scenario3$probs
#' sol.scenario3$ktau
#' sol.scenario3$parapar
#'
#' #------------------------------------------------------------------------
#' # simulated data from Frank copula log-Normal margins (with cure)
#' #------------------------------------------------------------------------
#'
#' # generate the simulated data
#'
#' #  true underlying cure rate
#' curerate.true <- 0.2
#'
#' # true underlying survival and censoring times
#' set.seed(1)
#' u.TC <- copula::rCopula(
#'   n        = n,
#'   copula   = copula::archmCopula(
#'     family = copfam.true,
#'     param  = coppar.true,
#'     dim    = 2
#'   )
#' )
#' yobs.T <- sapply(u.TC[,1],function(uT){
#'   if(uT<=curerate.true){ val <- Inf }else{
#'     val <- EnvStats::qlnormTrunc((1-uT)/(1-curerate.true),survpar.true[1],survpar.true[2],0,15)
#'   }
#'   return(val)
#' })
#' yobs.C <- qlnorm(1-u.TC[,2],censpar.true[1],censpar.true[2])
#' cat("cure rate is",mean(yobs.T==Inf))
#'
#' #  observations
#' yobs  <- pmin(yobs.T,yobs.C)
#' delta <- as.numeric(yobs.T<=yobs.C)
#' cat("censoring rate is",mean(1-delta))
#'
#' # model the data under different scenarios (with cure)
#'
#' # scenario 4: parametric survival and censoring margins
#' set.seed(1)
#' sol.scenario4 <- SurvDC(
#'   yobs    = yobs,
#'   delta   = delta,
#'   tm      = quantile(yobs, c(0.25,0.50,0.75)),
#'   copfam  = copfam.true,
#'   margins = list(survfam = "lnorm", censfam = "lnorm"),
#'   Var     = list(do = FALSE, nboot = 50),
#'   cure    = TRUE
#' )
#' sol.scenario4$probs
#' sol.scenario4$ktau
#' sol.scenario4$parapar
#' sol.scenario4$curerate
#'
#' # scenario 5: nonparametric survival margin and parametric censoring margin
#' set.seed(1)
#' sol.scenario5 <- SurvDC(
#'   yobs    = yobs,
#'   delta   = delta,
#'   tm      = quantile(yobs, c(0.25,0.50,0.75)),
#'   copfam  = copfam.true,
#'   margins = list(survfam = NULL, censfam = "lnorm"),
#'   Var     = list(do = FALSE, nboot = 50),
#'   cure    = TRUE
#' )
#' sol.scenario5$probs
#' sol.scenario5$ktau
#' sol.scenario5$parapar
#' sol.scenario5$curerate
#' }
#'
#' @export SurvDC
#'
#'
SurvDC <- function(
    yobs,
    delta,
    tm      = NULL,
    copfam  = "frank",
    margins = list(survfam = NULL, censfam = "lnorm"),
    cure    = FALSE,
    Var     = list(do = TRUE, nboot = 200, level=0.05),
    control = list(
      maxit      = 300,
      eps        = 1e-6,
      trace      = TRUE,
      ktau.inits = NULL
    )
){

  ## preparations
  n <- length(yobs)
  margins$type <- ifelse(any(c(is.null(margins$survfam),is.null(margins$censfam))),
                         "semiparametric","parametric")
  control.pre <- control.arguments()
  control <- lapply(names(control.pre),function(argi){
    if(is.null(control[[argi]])){
      val <- control.pre[[argi]]
    }else{ val <- control[[argi]] }
    return(val)
  })
  names(control) <- names(control.pre)

  ## set the value of truncation point (if NULL and cure)
  if(is.null(margins$survfam)){
    margins$survtrunc <- NULL
  }else{
    if(is.null(margins$survtrunc) & cure==TRUE){
      margins$survtrunc <- max(yobs)
    }
  }
  if(is.null(margins$censfam)){margins$censtrunc <- NULL}

  ## prepare the constraint matrix and vector
  constraints <- Parameters.Constraints(copfam=copfam,margins=margins,cure=cure)

  ## prepare initial values of the Kendall's tau
  if(is.null(control$ktau.inits)==TRUE | copfam=="indep"){
    if(copfam == "indep"){ control$ktau.inits <- 0 }else{
      ktau.left  <- ifelse(copfam=="frank",-1+1e-5,1e-5)
      control$ktau.inits <- seq(ktau.left,1-1e-5,ifelse(copfam=="frank",0.4,0.2))[-1]
    }
  }

  ## fit the semi-parametric model across many initial values
  if(control$trace==TRUE){cat("- Fitting the Model ...\n")}
  fit.model  <- list(value = -Inf)
  for(iktau in 1:length(control$ktau.inits)){

    # fit the semi-parametric model within try() to avoid error
    cytry <- try({
      sol.c <- Likelihood.Profile.Solve(
        yobs=yobs,delta=delta,copfam=copfam,margins=margins,
        ktau.init=control$ktau.inits[iktau],parapar.init=NULL,
        cure=cure,curerate.init=NULL,constraints=constraints,
        maxit=control$maxit,eps=control$eps
      )
      param       <- sol.c$param
    }, silent = TRUE) # end try

    # - judge the current fit
    if(is(cytry, "try-error") == FALSE){
      # __ calculate the value of likelihood
      if(margins$type=="parametric"){
        value <- Likelihood.Parametric(param=param,yobs=yobs,delta=delta,copfam=copfam,margins=margins,cure=cure)
      }else{
        value <- Likelihood.Profile.Kernel(param=param,yobs=yobs,delta=delta,copfam=copfam,margins=margins,cure=cure)
      }
      # __ judge the current fit formally
      if(value > fit.model$value){
        fit.model$param <- param
        fit.model$value <- value
      }
    }

  }

  ## extract estimates and prepare other needed ones

  # parametric part
  if(is.null(fit.model$param)==TRUE){
    stop("Fitting Failed for Current Setting!")
  }else{
    (ktau   <- fit.model$param[1])
    if(cure==TRUE & is.null(margins$survfam)==FALSE){
      parapar <- fit.model$param[-c(1,2)]
      curerate <- fit.model$param[2]
    }else{
      parapar <- fit.model$param[-1]
      curerate <- NULL
    }

  }

  # survival probabilities of survival time at interested points (and cure rate)
  if(is.null(tm)==TRUE){tm <- yobs}
  if(is.null(margins$survfam)==TRUE){
    probs.all <- SurvFunc.CG(tm=tm,yobs=yobs,delta=delta,copfam=copfam,ktau=ktau)$Shat
    if(cure == TRUE){
      curerate <- SurvFunc.CG(tm=max(yobs[delta==1]),yobs=yobs,delta=delta,copfam=copfam,ktau=ktau)$Shat
      probs    <- (probs.all-curerate)/(1-curerate)
    }else{ curerate <- NULL; probs <- probs.all }
  }else{
    if(margins$survfam %in% c("lnorm","weibull")){ parapar.surv <- parapar[1:2] }
    probs <- 1-parafam.p(x=tm,parameter=parapar.surv,distribution=margins$survfam,truncation=margins$survtrunc)
  }

  # calculate goodness-of-fit statistic
  GoF <- SurvDC.GoF(yobs=yobs,delta=delta,copfam=copfam,margins=margins,ktau=ktau,parapar=parapar,cure=cure,curerate=curerate)

  # obtain the AIC
  degree.free <- 0
  for(fam in c("survfam","censfam")){
    if(is.null(margins[[fam]])==TRUE){ next }
    if(margins[[fam]] %in% c("lnorm","weibull")){
      degree.free <- degree.free + 2
    }
  }
  if(copfam!="indep"){degree.free <- degree.free + 1}
  if(cure == TRUE){degree.free <- degree.free + 1}
  AIC <- -2*fit.model$value + 2*degree.free

  ## bootstrap procedure for variance estimation
  if(Var$do==TRUE){

    if(control$trace==TRUE){cat("- Doing Bootstrap to Quantify Variation ...\n")}

    # preparations

    # start the bootstrap
    results.boots <- rep(list(list()),Var$nboot)
    iboot <- 1; nwrong <- 0
    while(iboot <= Var$nboot){

      if(nwrong > ceiling(Var$nboot*0.1)){stop("Too Many Failed Fittings in Bootstrap!")}
      if(control$trace==TRUE){if(iboot%%10==0){cat("  + Bootstrap",iboot,"/",Var$nboot,"\n")}}

      # - generate the current bootstrap sample
      idx.iboot   <- sort(sample(1:n,n,replace=TRUE))
      yobs.iboot  <- yobs[idx.iboot]
      delta.iboot <- delta[idx.iboot]

      # - fit the current model and store the results
      try.iboot <- try({

        # __ fit the model and extract estimates
        sol.iboot <- Likelihood.Profile.Solve(
          yobs=yobs.iboot,delta=delta.iboot,
          copfam=copfam,margins=margins,
          ktau.init=ktau,parapar.init=parapar,
          cure=cure,curerate.init=curerate,
          constraints=constraints,
          maxit=control$maxit,eps=control$eps
        )
        ktau.iboot        <- sol.iboot$param[1]
        if(cure==TRUE & is.null(margins$survfam)==FALSE){
          parapar.iboot  <- sol.iboot$param[-c(1,2)]
          curerate.iboot <- sol.iboot$param[2]
        }else{
          parapar.iboot  <- sol.iboot$param[-1]
          curerate.iboot <- NULL
        }

        # __ survival probabilities at interested points (and cure rate)
        if(is.null(margins$survfam)==TRUE){
          probs.all.iboot <- SurvFunc.CG(tm=tm,yobs=yobs.iboot,delta=delta.iboot,copfam=copfam,ktau=ktau.iboot)$Shat
          if(cure == TRUE){
            curerate.iboot <- SurvFunc.CG(tm=max(yobs.iboot[delta.iboot==1]),
                                          yobs=yobs.iboot,delta=delta.iboot,copfam=copfam,ktau=ktau.iboot)$Shat
            probs.iboot    <- (probs.all.iboot-curerate.iboot)/(1-curerate.iboot)
          }else{ curerate.iboot <- NULL; probs.iboot <- probs.all.iboot }
        }else{
          if(margins$survfam %in% c("lnorm","weibull")){ parapar.surv.iboot <- parapar.iboot[1:2] }
          probs.iboot <- 1-parafam.p(x=tm,parameter=parapar.surv.iboot,distribution=margins$survfam,truncation=margins$survtrunc)
        }

        # __ calculate goodness-of-fit statistic
        GoF.iboot <- SurvDC.GoF(
          yobs=yobs.iboot,delta=delta.iboot,copfam=copfam,
          margins=margins,ktau=ktau.iboot,parapar=parapar.iboot,
          cure=cure,curerate=curerate.iboot)

      }, silent = T)

      # - judge the current fit
      if(is(try.iboot,"try-error") == FALSE){
        results.boots[[iboot]]$ktau    <- ktau.iboot
        results.boots[[iboot]]$parapar <- parapar.iboot
        results.boots[[iboot]]$probs   <- probs.iboot
        results.boots[[iboot]]$curerate <- curerate.iboot
        results.boots[[iboot]]$GoF <- GoF.iboot
      }else{ nwrong <- nwrong + 1; next }

      # - prepare for next iteration
      iboot <- iboot + 1

    }

    # get estimtes of standard errors

    # - all bootstrap estimates
    ktau.boots    <- do.call(c,lapply(1:Var$nboot,function(iboot){results.boots[[iboot]]$ktau}))
    parapar.boots <- do.call(cbind,lapply(1:Var$nboot,function(iboot){results.boots[[iboot]]$parapar}))
    probs.boots   <- do.call(cbind,lapply(1:Var$nboot,function(iboot){results.boots[[iboot]]$probs}))

    # - standard errors
    ktau.se       <- sd(ktau.boots)
    parapar.se    <- apply(parapar.boots,1,sd)
    probs.se      <- apply(probs.boots,1,sd)
    if(cure==TRUE){
      curerate.boots <- do.call(c,lapply(1:Var$nboot,function(iboot){results.boots[[iboot]]$curerate}))
      curerate.se    <- sd(curerate.boots)
    }else{
      curerate.boots <- curerate.se <- NULL
    }

    # - confidence intervals
    ktau.upper       <- quantile(ktau.boots,1-Var$level/2,names=FALSE)
    ktau.lower       <- quantile(ktau.boots,Var$level/2,names=FALSE)
    parapar.upper    <- apply(parapar.boots,1,function(x){quantile(x,1-Var$level/2,names=FALSE)})
    parapar.lower    <- apply(parapar.boots,1,function(x){quantile(x,Var$level/2,names=FALSE)})
    probs.upper    <- apply(probs.boots,1,function(x){quantile(x,1-Var$level/2,names=FALSE)})
    probs.lower    <- apply(probs.boots,1,function(x){quantile(x,Var$level/2,names=FALSE)})
    if(cure==TRUE){
      curerate.upper <- quantile(curerate.boots,1-Var$level/2,names=FALSE)
      curerate.lower <- quantile(curerate.boots,Var$level/2,names=FALSE)
    }else{
      curerate.upper <- curerate.lower <- NULL
    }

    # calculate p-value for goodness-of-fit statistic
    GoF.boots  <- do.call(c,lapply(1:Var$nboot,function(iboot){results.boots[[iboot]]$GoF}))
    GoF.pvalue <- mean(GoF.boots>=GoF)

  }else{
    ktau.se <- parapar.se <- probs.se <- curerate.se <- GoF.pvalue <- NULL
    ktau.boots <- parapar.boots <- probs.boots <- curerate.boots <- GoF.boots <- NULL
    ktau.upper <- parapar.upper <- probs.upper <- curerate.upper <- NULL
    ktau.lower <- parapar.lower <- probs.lower <- curerate.lower <- NULL
  }

  ## summary all fitted results

  # - Kendall's tau
  ktau <- c(
    Est=ktau,
    SE=ktau.se,
    zvalue=ktau/ktau.se,
    pvalue=2*(1-pnorm(abs(ktau/ktau.se))),
    LowerCI = ktau.lower,
    UpperCI = ktau.upper
  )

  # - parametric parameters
  parapar <- as.data.frame(cbind(
    Est    = parapar,
    SE     = parapar.se,
    zvalue = parapar/parapar.se,
    pvalue = 2*(1-pnorm(abs(parapar/parapar.se))),
    LowerCI = parapar.lower,
    UpperCI = parapar.upper
  ))

  # - survival probabilities
  probs <- as.data.frame(cbind(
    time   = tm,
    Est    = probs,
    SE     = probs.se,
    zvalue = probs/probs.se,
    pvalue = 2*(1-pnorm(abs(probs/probs.se))),
    LowerCI = probs.lower,
    UpperCI = probs.upper
  ))[order(tm),,drop=FALSE]

  # - cure rate
  if(cure==TRUE){
    curerate <- c(
      Est=curerate,
      SE=curerate.se,
      zvalue=curerate/curerate.se,
      pvalue=2*(1-pnorm(abs(curerate/curerate.se))),
      LowerCI = curerate.lower,
      UpperCI = curerate.upper
    )
  }else{
    curerate <- NULL
  }

  # - goodness-of-fit
  GoF <- c(statistic=GoF,pvalue=GoF.pvalue)

  ## output
  out <- list(
    probs     = probs,
    ktau      = ktau,
    parapar   = parapar,
    GoF       = GoF,
    AIC       = AIC,
    curerate  = curerate,
    boots     = list(ktau     = ktau.boots,
                     parapar  = parapar.boots,
                     probs    = probs.boots,
                     curerate = curerate.boots,
                     GoF      = GoF.boots),
    inputs    = list(
      yobs    = yobs,
      delta   = delta,
      copfam  = copfam,
      margins = margins
    )
  )
  return(out)

}



#' @title Solve the profiled likelihood function
#'
#' @param yobs a numeric vector that indicated the observed survival times.
#' @param delta a numeric vector that stores the right-censoring indicators.
#' @param copfam a character string that specifies the copula family.
#' @param margins a list used to define the distribution structures of both the survival and censoring margins.
#' @param ktau.init initial value of Kendall's tau.
#' @param parapar.init initial value of parametric parameters.
#' @param cure a logical value that indicates whether the existence of a cured fraction should be considered.
#' @param curerate.init initial value of cure rate.
#' @param constraints constraints of parameters.
#' @param maxit a positive integer that denotes the maximum iteration number in optimization.
#' @param eps a positive small numeric value that denotes the tolerance for convergence.
#'
#' @noRd
#'
Likelihood.Profile.Solve <- function(
    yobs,delta,copfam,margins,ktau.init,parapar.init,cure,curerate.init,constraints,maxit,eps
){

  # prepare initial values for parametric parameters (if it is not nonparametric)
  if(is.null(parapar.init)==TRUE | copfam=="indep"){

    # calculate initial estimates for survival or censoring part
    parapar.init <- NULL
    for(fam in c("survfam","censfam")){

      # - if this part is NULL, omit the following calculations
      if(is.null(margins[[fam]])==TRUE){next}

      # - preparations
      if(fam=="survfam"){
        truncation <- margins$survtrunc
      }else{
        truncation <- margins$censtrunc
      }
      delta.parametric <- ifelse(rep(fam=="survfam",length(yobs)),delta,1-delta)

      # - current constraints!!!
      # ui      = constraints$A
      # ci      = constraints$b

      # obtain initial values of parametric parameters by assuming independence
      sol.fam.init <- SurvMLE(
        yobs=yobs,delta=delta.parametric,distribution=margins[[fam]],
        truncation=truncation,cure=(cure & fam=="survfam"),maxit=maxit)


      # - further refine the initial value (if not independent copula)
      if(copfam=="indep"){
        if(cure==TRUE & fam=="survfam"){
          curerate.init <- sol.fam.init$par[1]
          parapar.init <- c(parapar.init,sol.fam.init$par[-1])
        }else{
          parapar.init <- c(parapar.init,sol.fam.init$par)
        }
      }else{
        probs.parametric <- SurvFunc.CG(tm=NULL,yobs=yobs,delta=delta.parametric,copfam=copfam,ktau=ktau.init)$Shat
        if(cure==TRUE & fam=="survfam"){
          sol.parametric.refine <- stats::constrOptim(
            theta = c(sol.fam.init$curerate,sol.fam.init$parapar),
            f     = function(parapar,probs.parametric,yobs){
              mean((1-probs.parametric-parapar[1]*parafam.p(x=yobs,parameter=parapar[-1],distribution=margins[[fam]],truncation=truncation))^2)},
            grad  = NULL, ui = sol.fam.init$constraints$A, ci = sol.fam.init$constraints$b,
            control = list(trace=0,maxit=maxit), yobs = yobs, probs.parametric = probs.parametric
          )
          curerate.init <- sol.parametric.refine$par[1]
          parapar.init <- c(parapar.init,sol.parametric.refine$par[-1])
        }else{
          sol.parametric.refine <- stats::constrOptim(
            theta = sol.fam.init$parapar,
            f     = function(parapar,probs.parametric,yobs){
              mean((1-probs.parametric-parafam.p(x=yobs,parameter=parapar,distribution=margins[[fam]],truncation=truncation))^2)},
            grad  = NULL, ui = sol.fam.init$constraints$A, ci = sol.fam.init$constraints$b,
            control = list(trace=0,maxit=maxit), yobs = yobs, probs.parametric = probs.parametric
          )
          parapar.init <- c(parapar.init,sol.parametric.refine$par)
        }
      }

    }

  }

  # fit the model formally
  if(copfam=="indep"){
    if(cure==TRUE & is.null(margins$survfam)==FALSE){
      param <- c(0,curerate.init,parapar.init)
    }else{
      param <- c(0,parapar.init)
    }
    convergence <- TRUE
  }else{

    # - initial values
    if(cure==TRUE & is.null(margins$survfam)==FALSE){
      param.init <- c(ktau.init,curerate.init,parapar.init)
    }else{
      param.init <- c(ktau.init,parapar.init)
    }

    # - fit the model formally
    if(margins$type=="parametric"){

      sol.model <- stats::constrOptim(
        theta   = param.init,
        f       = Likelihood.Parametric,
        grad    = NULL,
        ui      = constraints$A,
        ci      = constraints$b,
        control = list(trace=0,fnscale=-1,maxit=maxit),
        yobs    = yobs, delta = delta, copfam = copfam, margins = margins, cure = cure
      )
      param <- sol.model$par
      convergence <- (sol.model$convergence == 0)

    }else if(margins$type=="semiparametric"){

      param.old <- param.init
      delta.nonparametric <- ifelse(rep(is.null(margins$survfam),length(yobs)),delta,1-delta)
      Syobs.old <- SurvFunc.CG(tm=NULL,yobs=yobs,delta=delta.nonparametric,
                               copfam=copfam,ktau=param.old[1],coppar=NULL)$Shat
      numit     <- 1
      repeat{
        sol.model <- stats::constrOptim(
          theta   = param.old,
          f       = Likelihood.Semiparametric,
          grad    = NULL,
          ui      = constraints$A,
          ci      = constraints$b,
          control = list(trace=0,fnscale=-1,maxit=maxit),
          Syobs   = Syobs.old, yobs = yobs, delta = delta, copfam = copfam,
          margins = margins, cure = cure
        )
        param  <- sol.model$par
        Syobs  <- SurvFunc.CG(tm=NULL,yobs=yobs,delta=delta.nonparametric,copfam=copfam,ktau=param[1],coppar=NULL)$Shat
        epsmax <- max(c(abs(param-param.old),abs(Syobs-Syobs.old)))
        if( epsmax>eps & numit<maxit ){
          param.old <- param
          Syobs.old <- Syobs
          numit     <- numit + 1
        }else{
          break
        }
      }
      convergence <- (numit<maxit)

    }

  }

  # output
  out <- list(
    param       = param,
    convergence = convergence
  )
  return(out)

}

#' @title Calculate the goodness-of-fit test statistic
#'
#' @param yobs a numeric vector that indicated the observed survival times.
#' @param delta a numeric vector that stores the right-censoring indicators.
#' @param copfam a character string that specifies the copula family.
#' @param margins a list used to define the distribution structures of both the survival and censoring margins.
#' @param ktau Kendall's tau.
#' @param parapar parametric parameters.
#' @param cure a logical value that indicates whether the existence of a cured fraction should be considered.
#' @param curerate value of cure rate.
#'
#' @noRd
#'
SurvDC.GoF <- function(yobs,delta,copfam,margins,ktau,parapar,cure=FALSE,curerate=NULL){

  coppar  <- ktau.to.coppar(ktau=ktau,copfam=copfam)

  # distribution-related functions
  if(margins$type=="parametric"){

    # - length of parameters for two parts
    if(margins$survfam %in% c("lnorm","weibull")){ parapar.surv.num <- 2 }

    # - probabilities for both survival and censoring parts
    STC <- NULL
    for(fam in c("survfam","censfam")){
      if(fam=="survfam"){
        parapar.temp <- parapar[1:parapar.surv.num]
        truncation.temp <- margins$survtrunc
      }else{
        parapar.temp <- parapar[-c(1:parapar.surv.num)]
        truncation.temp <- margins$censtrunc
      }
      S.temp <- 1-parafam.p(x=yobs,parameter=parapar.temp,distribution=margins[[fam]],truncation=truncation.temp)
      STC <- cbind(STC,S.temp)
    }

    # - define probabilities formally
    if(cure==TRUE){
      ST <- curerate + (1-curerate)*STC[,1]
    }else{
      ST <- STC[,1]
    }
    SC <- STC[,2]

  }else if(margins$type=="semiparametric"){

    # - probabilities for parametric part time (judge later)
    S.temp <- 1-parafam.p(
      x=yobs,parameter=parapar,
      distribution=c(margins$survfam,margins$censfam),
      truncation=c(margins$survtrunc,margins$censtrunc)
    )

    # - probabilities for both survival and censoring parts
    if(is.null(margins$survfam) == TRUE){
      ST <- SurvFunc.CG(tm=NULL,yobs=yobs,delta=delta,copfam=copfam,ktau=NULL,coppar=coppar)$Shat
      SC <- S.temp
    }else{
      if(cure==TRUE){
        ST <- curerate + (1-curerate)*S.temp
      }else{
        ST <- S.temp
      }
      SC <- SurvFunc.CG(tm=NULL,yobs=yobs,delta=1-delta,copfam=copfam,ktau=NULL,coppar=coppar)$Shat
    }

  }

  SY.Model <- copdist.Archimedean(x=cbind(ST,SC),copfam=copfam,ktau=NULL,coppar=coppar)
  SY.Empir <- sapply(yobs,function(yobsi){mean(yobs>yobsi)})
  GoF      <- sum((SY.Model-SY.Empir)^2)

  # output
  return(GoF)

}


#' @title Likelihood for a given parametric distribution
#'
#' @param param a vector contains all parametric parameters.
#' @param yobs a numeric vector that indicated the observed survival times.
#' @param delta a numeric vector that stores the right-censoring indicators.
#' @param distribution the specified distribution function.
#' @param truncation a positive numeric value thats denotes the value of truncation for the assumed distribution.
#' @param cure a logical value that indicates whether the existence of a cured fraction should be considered.
#'
#' @noRd
SurvMLE.Likelihood <- function(param,yobs,delta,distribution,truncation=NULL,cure=FALSE){

  # extract parameters
  if(cure==TRUE){
    curerate <- param[1]
    parameter <- param[-1]
  }else{
    parameter <- param
  }

  # basic distribution function
  FT.0 <- parafam.p(x=yobs[delta==0],parameter=parameter,distribution=distribution,truncation=truncation)
  fT.1 <- parafam.d(x=yobs[delta==1],parameter=parameter,distribution=distribution,truncation=truncation)

  # revision based on the cure status
  if(cure==TRUE){
    FT.0 <- FT.0*(1-curerate)
    fT.1 <- fT.1*(1-curerate)
  }

  # calculate likelihood
  Lik1  <- sum(log(pmax(1e-20,fT.1)),na.rm=TRUE)
  Lik0  <- sum(log(pmax(1e-20,1-FT.0)),na.rm=TRUE)
  Lik   <- Lik1 + Lik0

  # return
  return(Lik)

}

#' @title Maximum likelihood estimator for a given parametric distribution
#'
#' @param yobs a numeric vector that indicated the observed survival times.
#' @param delta a numeric vector that stores the right-censoring indicators.
#' @param distribution the specified distribution function.
#' @param truncation a positive numeric value thats denotes the value of truncation for the assumed distribution.
#' @param cure a logical value that indicates whether the existence of a cured fraction should be considered.
#' @param maxit a positive integer that denotes the maximum iteration number in optimization.
#'
#' @noRd
SurvMLE <- function(yobs,delta,distribution,truncation=NULL,cure=FALSE,maxit=300){

  # setup initial values and constraints
  if(distribution %in% c("lnorm")){
    A <- rbind(c(0,1))
    b <- 1e-5
    param.init <- c(0,1)
  }else if(distribution %in% c("weibull")){
    A <- rbind(c(1,0),c(0,1))
    b <- c(1e-5,1e-5)
    param.init <- c(1,1)
  }

  # add information about cure
  if(cure==TRUE){
    A = as.matrix(Matrix::bdiag(rbind(1,-1),A))
    b = c(1e-5,-1+1e-5,b)
    param.init <- c(0.5,param.init)
  }

  # maximizing the likelihood function
  sol.model <- stats::constrOptim(
    theta   = param.init,
    f       = SurvMLE.Likelihood,
    grad    = NULL,
    ui      = A,
    ci      = b,
    control = list(trace=0,fnscale=-1,maxit=maxit),
    yobs    = yobs, delta = delta, distribution = distribution,
    truncation = truncation, cure = cure
  )
  if(cure==TRUE){
    curerate <- sol.model$par[1]
    parapar  <- sol.model$par[-1]
  }else{
    curerate <- NULL
    parapar  <- sol.model$par
  }
  convergence <- (sol.model$convergence == 0)

  # output
  out <- list(
    parapar     = parapar,
    curerate    = curerate,
    convergence = convergence,
    constraints = list(A=A,b=b)
  )

}

#' @title Obtain the adjustment value of truncation
#'
#' @param truncation a positive numeric value thats denotes the value of truncation for the assumed distribution.
#' @param parameter the parameter of the specified distribution
#' @param distribution the specified distribution function.
#'
#' @noRd
parafam.trunc <- function(truncation,parameter,distribution){

  if(distribution=="lnorm"){
    val.adjust <- plnorm(truncation,parameter[1],parameter[2])
  }else if(distribution=="weibull"){
    val.adjust <- pweibull(truncation,parameter[1],parameter[2])
  }
  return(val.adjust)

}

#' @title Obtain the value of the distribution function
#'
#' @param x the value in which the distribution function will be calculated at.
#' @inheritParams parafam.trunc
#'
#' @noRd
parafam.p <- function(x,parameter,distribution,truncation=NULL){

  # values with no truncation
  if(distribution=="lnorm"){
    val <- plnorm(x,parameter[1],parameter[2])
  }else if(distribution=="weibull"){
    val <- pweibull(x,parameter[1],parameter[2])
  }

  # truncation
  if(is.null(truncation)==FALSE){
    val.adjust <- parafam.trunc(truncation=truncation,parameter=parameter,distribution=distribution)
    val <- pmin(val/val.adjust,1)
  }

  # output
  return(val)

}

#' @title Obtain the value of the density function
#'
#' @param x the value in which the density function will be calculated at.
#' @inheritParams parafam.trunc
#'
#' @noRd
parafam.d <- function(x,parameter,distribution,truncation=NULL){

  # values with no truncation
  if(distribution=="lnorm"){
    val <- dlnorm(x,parameter[1],parameter[2])
  }else if(distribution=="weibull"){
    val <- dweibull(x,parameter[1],parameter[2])
  }

  # truncation
  if(is.null(truncation)==FALSE){
    val.adjust <- parafam.trunc(truncation=truncation,parameter=parameter,distribution=distribution)
    val <- val/val.adjust
    val[x>truncation] <- 0
  }

  # output
  return(val)

}

#' @title Calculate the semiparametric version of profiled likelihood function
#'
#' @param param a vector contains all parametric parameters.
#' @param Syobs values of survival function at observed time points.
#' @param yobs a numeric vector that indicated the observed survival times.
#' @param delta a numeric vector that stores the right-censoring indicators.
#' @param copfam a character string that specifies the copula family.
#' @param margins a list used to define the distribution structures of both the survival and censoring margins.
#' @param cure a logical value that indicates whether the existence of a cured fraction should be considered.
#'
#' @noRd
Likelihood.Semiparametric <- function(param,Syobs,yobs,delta,copfam,margins,cure=FALSE){

  # extract current values
  if(cure==TRUE & is.null(margins$survfam)==FALSE){
    curerate <- param[2]
    parapar  <- param[-c(1,2)]
  }else{
    parapar <- param[-1]
  }
  ktau    <- param[1]
  coppar  <- ktau.to.coppar(ktau=ktau,copfam=copfam)

  # - probabilities for parametric part time (judge later)
  delta.parametric <- ifelse(rep(is.null(margins$survfam),length(yobs)),1-delta,delta)
  S.temp <- 1-parafam.p(x=yobs,parameter=parapar,distribution=c(margins$survfam,margins$censfam),truncation=c(margins$survtrunc,margins$censtrunc))
  f.temp <- parafam.d(x=yobs[delta.parametric==1],parameter=parapar,distribution=c(margins$survfam,margins$censfam),truncation=c(margins$survtrunc,margins$censtrunc))

  # distribution-related functions
  if(is.null(margins$survfam)==TRUE){

    # basic quantities
    ST   <- Syobs
    SC   <- S.temp
    fC.0 <- f.temp

    # calculate likelihood
    hCT.1 <- cophfunc(x=cbind(ST,SC)[delta==1,,drop=FALSE],coppar=coppar,copfam=copfam,condvar=1)
    hTC.0 <- cophfunc(x=cbind(ST,SC)[delta==0,,drop=FALSE],coppar=coppar,copfam=copfam,condvar=2)
    Lik1  <- sum(log(pmax(1e-20,hCT.1)),na.rm=TRUE)
    Lik0  <- sum(log(pmax(1e-20,fC.0*hTC.0)),na.rm=TRUE)
    Lik   <- Lik1 + Lik0

  }else{

    # basic quantities
    SC <- Syobs
    if(cure==TRUE){
      ST <- curerate + (1-curerate)*S.temp
      fT.1 <- (1-curerate)*f.temp
    }else{
      ST <- S.temp
      fT.1 <- f.temp
    }

    # calculate likelihood
    hCT.1 <- cophfunc(x=cbind(ST,SC)[delta==1,,drop=FALSE],coppar=coppar,copfam=copfam,condvar=1)
    hTC.0 <- cophfunc(x=cbind(ST,SC)[delta==0,,drop=FALSE],coppar=coppar,copfam=copfam,condvar=2)
    Lik1  <- sum(log(pmax(1e-20,fT.1*hCT.1)),na.rm=TRUE)
    Lik0  <- sum(log(pmax(1e-20,hTC.0)),na.rm=TRUE)
    Lik   <- Lik1 + Lik0

  }

  # return
  return(Lik)

}


#' @title Calculate the likelihood function for the fully parametric joint distribution
#'
#' @inheritParams Likelihood.Semiparametric
#'
#' @noRd
Likelihood.Parametric <- function(param,yobs,delta,copfam,margins,cure=FALSE){

  # extract current values
  ktau <- param[1]
  if(cure==TRUE){
    curerate <- param[2]
    parapar  <- param[-c(1,2)]
  }else{
    parapar <- param[-1]
  }
  coppar <- as.vector(ktau.to.coppar(ktau=ktau,copfam=copfam))

  # - survival and censoring time
  if(margins$survfam %in% c("lnorm","weibull")){
    parapar.surv <- parapar[1:2]
    parapar.cens <- parapar[-c(1:2)]
  }

  # distribution-related functions

  # - for survival time

  # __ basic distribution function
  FT   <- parafam.p(x=yobs,parameter=parapar.surv,distribution=margins$survfam,truncation=margins$survtrunc)
  fT.1 <- parafam.d(x=yobs[delta==1],parameter=parapar.surv,distribution=margins$survfam,truncation=margins$survtrunc)

  # __ revision based on the cure status
  if(cure==TRUE){
    FT   <- FT*(1-curerate)
    fT.1 <- fT.1*(1-curerate)
  }

  # - for censoring time
  FC   <- parafam.p(x=yobs,parameter=parapar.cens,distribution=margins$censfam,truncation=margins$censtrunc)
  fC.0 <- parafam.d(x=yobs[delta==0],parameter=parapar.cens,distribution=margins$censfam,truncation=margins$censtrunc)

  # calculate likelihood
  hCT.1 <- cophfunc(x=cbind(1-FT,1-FC)[delta==1,,drop=FALSE],coppar=coppar,copfam=copfam,condvar=1)
  hTC.0 <- cophfunc(x=cbind(1-FT,1-FC)[delta==0,,drop=FALSE],coppar=coppar,copfam=copfam,condvar=2)
  Lik1  <- sum(log(pmax(1e-20,fT.1*hCT.1)),na.rm=TRUE)
  Lik0  <- sum(log(pmax(1e-20,fC.0*hTC.0)),na.rm=TRUE)
  Lik   <- Lik1 + Lik0

  # return
  return(Lik)

}


#' @title Generate constraints of parameters
#'
#' @inheritParams Likelihood.Semiparametric
#'
#' @noRd
Parameters.Constraints <- function(copfam,margins,cure){

  eps.fluate <- 1e-3

  # produce constraints
  if(copfam == "indep"){
    constraints <- NULL
  }else{

    # - bounds for the Kendall's tau
    ktau.left  <- ifelse(copfam=="frank",-1+eps.fluate,eps.fluate)
    ktau.right <- 1-eps.fluate
    constraints <- list(
      A = rbind(1,-1),
      b = c(ktau.left,-ktau.right)
    )

    # - bounds for the cure rate
    if(cure == TRUE & is.null(margins$survfam) == FALSE){
      constraints <- list(
        A = as.matrix(Matrix::bdiag(constraints$A,rbind(1,-1))),
        b = c(constraints$b,eps.fluate,-1+eps.fluate)
      )
    }

    # - define the constraints for other parametric parts

    for(fam in c("survfam","censfam")){

      # - if this part is NULL, omit the following calculations
      if(is.null(margins[[fam]])==TRUE){next}

      # - add constraints based on different parametric families
      if(margins[[fam]] %in% c("lnorm")){
        constraints <- list(
          A = as.matrix(Matrix::bdiag(constraints$A,rbind(c(0,1)))),
          b = c(constraints$b,eps.fluate)
        )
      }else if(margins[[fam]] %in% c("weibull")){
        constraints <- list(
          A = as.matrix(Matrix::bdiag(constraints$A,rbind(c(1,0),c(0,1)))),
          b = c(constraints$b,eps.fluate,eps.fluate)
        )
      }

    }

  }

  # output
  return(constraints)

}


#' @title Convert the Kendall's tau into the copula parameter
#'
#' @param ktau a numeric value that denotes the Kendall's tau.
#' @param copfam a character string that denotes the copula family.
#' @noRd

ktau.to.coppar <- function(ktau,copfam){

  if(copfam=="frank"){

    coppar <- uniroot(
      f=function(coppar,ktau){
        if(abs(coppar)>=1e-5){
          int  <- integrate(f=function(t){t/(exp(t)-1)},0,coppar)$value
          ktauc <- 1-4*(1-int/coppar)/coppar
        }else{ ktauc <- 0 }
        int  <- integrate(f=function(t){t/(exp(t)-1)},0,coppar)$value
        ktauc - ktau
      },
      interval=c(-1e5,1e5),ktau=ktau
    )$root

  }else if(copfam=="gumbel"){

    coppar <- 1/(1-ktau)

  }else if(copfam=="joe"){

    coppar <- uniroot(
      f=function(coppar,ktau){
        int <- integrate(f=function(t,coppar){exp(-2*t)*t*(1-exp(-t))^(2/coppar-2)},0,Inf,coppar)$value
        1 - 4*int/(coppar^2) - ktau
      },
      interval=c(1,1e5),ktau=max(ktau,1e-5)
    )$root

  }else if(copfam=="clayton"){

    coppar <- 2*ktau/(1-ktau)

  }else if(copfam=="indep"){

    coppar <- NULL

  }

  return(coppar)

}

#' @title Convert the copula parameter the Kendall's tau
#'
#' @param coppar a numeric value that denotes the copula parameter.
#' @param copfam a character string that denotes the copula family.
#'
#' @noRd
coppar.to.ktau <- function(coppar,copfam){

  # coppar's allowed range:
  # frank   : [-infty,infty]
  # gumbel  : [1,infty)
  # clayton : (0,infty)
  # joe     : [1,infty)

  if(copfam=="frank"){

    if(abs(coppar)>=1e-5){
      int  <- integrate(f=function(t){t/(exp(t)-1)},0,coppar)$value
      ktau <- 1-4*(1-int/coppar)/coppar
    }else{
      ktau <- 0
    }

  }else if(copfam=="gumbel"){

    ktau <- (coppar-1)/coppar

  }else if(copfam=="joe"){

    int  <- integrate(f=function(t,coppar){exp(-2*t)*t*(1-exp(-t))^(2/coppar-2)},0,Inf,coppar)$value
    ktau <- 1-4*int/(coppar^2)

  }else if(copfam=="clayton"){

    ktau <- coppar/(coppar+2)

  }else if(copfam=="indep"){

    ktau <- 0

  }

  return(ktau)

}

#' @title The generator function of the Archimedean copula
#'
#' @param x the value at which the generator function will be calculated at.
#' @inheritParams coppar.to.ktau
#' @param inverse a logical value that specifies whether the inverse function will be used.
#'
#' @noRd
generator.Archimedean <- function(x,coppar,copfam,inverse=FALSE){

  if(copfam=="frank"){
    if(inverse==FALSE){
      val <- -log((exp(-coppar*x)-1)/(exp(-coppar)-1))
    }else{
      val <- -log(1+exp(-x)*(exp(-coppar)-1))/coppar
    }
  }else if(copfam=="gumbel"){
    if(inverse==FALSE){
      val <- (-log(x))^coppar
    }else{
      val <- exp(-x^(1/coppar))
    }
  }else if(copfam=="joe"){
    if(inverse==FALSE){
      val <- -log(1-(1-x)^coppar)
    }else{
      val <- 1-(1-exp(-x))^(1/coppar)
    }
  }else if(copfam=="clayton"){
    if(inverse==FALSE){
      val <- (x^(-coppar)-1)/coppar
    }else{
      val <- (1+x*coppar)^(-1/coppar)
    }
  }else if(copfam=="indep"){
    if(inverse==FALSE){
      val <- -log(x)
    }else{
      val <- exp(-x)
    }
  }

  return(val)

}


#' @title The h-function of the copula
#'
#' @param x the value at which the h-function will be calculated at.
#' @inheritParams coppar.to.ktau
#' @param condvar a numeric value that specifies the type of the h-function.
#'
#' @noRd
cophfunc <- function(x,coppar,copfam,condvar=1){

  # preparations
  x <- pmin(pmax(x,1e-20),1-1e-20)
  numx <- nrow(x)
  if(condvar==1){
    u <- x[,1]; v <- x[,2]
  }else{
    u <- x[,2]; v <- x[,1]
  }

  # get the value of h function under different copulas
  if(copfam=="frank"){
    val <- exp(-coppar*u)*(exp(-coppar*v)-1)/(exp(-coppar)-1+(exp(-coppar*u)-1)*(exp(-coppar*v)-1))
  }else if(copfam=="gumbel"){
    uvpar <- (-log(u))^coppar + (-log(v))^coppar
    val <- exp(-uvpar^(1/coppar))*(uvpar)^(1/coppar-1)*(-log(u))^(coppar-1)/u
  }else if(copfam=="joe"){
    val <- ((1-u)^coppar+(1-v)^coppar-(1-u)^coppar*(1-v)^coppar)^(1/coppar-1)*(1-u)^(coppar-1)*(1-(1-v)^coppar)
  }else if(copfam=="clayton"){
    val <- u^(-coppar-1)*(u^(-coppar)+v^(-coppar)-1)^(-1/coppar-1)
  }else if(copfam=="indep"){
    val <- v
  }

  # output
  return(val)

}

#' @title The distribution function of the Archimedean copula
#'
#' @param x the value at which the distribution function will be calculated at.
#' @param ktau a numeric value that denotes the Kendall's tau.
#' @inheritParams coppar.to.ktau
#'
#' @noRd
copdist.Archimedean <- function(x,copfam,ktau,coppar = NULL){

  # preparations
  u <- x[,1]; v <- x[,2]
  if(is.null(coppar)==TRUE){ coppar <- ktau.to.coppar(ktau=ktau,copfam=copfam) }

  # distribution function
  val <- generator.Archimedean(
    x=generator.Archimedean(x=u,coppar,copfam)+generator.Archimedean(x=v,coppar,copfam),
    coppar,copfam,inverse=TRUE)

  # output
  return(val)

}


#' @title Estimated survival function based on copula-graphic estimator (Archimedean copula only)
#'
#' @param tm a vector contains all time points that the survival function will be calculated at.
#' @inheritParams copdist.Archimedean
#' @inheritParams Likelihood.Semiparametric
#'
#' @noRd
SurvFunc.CG <- function(tm=NULL,yobs,delta,copfam,ktau,coppar=NULL){

  # preparation
  n <- length(yobs)
  yobs.order <- order(yobs)
  if(is.null(coppar)==TRUE){ coppar <- ktau.to.coppar(ktau=ktau,copfam=copfam) }

  # calculate values of survival function
  pai.yobs      <- (n-rank(yobs,ties.method="first")+1)/n
  pai.yobs.left <- pmax((n-rank(yobs,ties.method="first"))/n,0)
  diff.yobs     <- ifelse(delta==0,0,generator.Archimedean(x=pai.yobs,coppar=coppar,copfam=copfam)-generator.Archimedean(x=pai.yobs.left,coppar=coppar,copfam=copfam))
  if(is.null(tm)==FALSE){
    sum.tm <- sapply(tm,function(tmi){sum(diff.yobs[yobs<=tmi])})
  }else{
    sum.tm <- cumsum(diff.yobs[yobs.order])
    sum.tm[yobs.order] <- sum.tm
  }
  Shat <- generator.Archimedean(x=-sum.tm,coppar=coppar,copfam=copfam,inverse=TRUE)

  # return
  out <- list(Shat = Shat)
  return(out)

}


#' @title Estimated survival function based on Kaplan-Meier estimator
#'
#' @inheritParams SurvFunc.CG
#' @param type a character string that specifies the type of the step function. If \code{type="right"}, it will be a right-continuous function.
#'
#' @noRd
SurvFunc.KM <- function(tm=NULL,yobs,delta,type="right"){

  # preparation
  if(is.null(tm)==TRUE){tm <- yobs}
  N <- length(yobs)
  yobs.order <- order(yobs)
  yobs.sort <- yobs[yobs.order]
  delta.sort <- delta[yobs.order]
  yobs.1max <- max(yobs[delta==1])

  # calculate the values of KM at pre-specified time points tm
  prods <- 1-delta.sort/(N-(1:N)+1)
  if(type=="right"){
    KMt <- sapply(tm,function(tmi){prod(prods[yobs.sort<=tmi])}) # right-continuous
  }else{
    KMt <- sapply(tm,function(tmi){prod(prods[yobs.sort<tmi])}) # left-continuous
  }

  # output
  return(KMt)

}

#' @title Calculate the profiled likelihood function with kernel smoothing
#'
#' @inheritParams Likelihood.Semiparametric
#'
#' @noRd
Likelihood.Profile.Kernel <- function(param,yobs,delta,copfam,margins,cure=FALSE){

  # extract current values
  if(cure==TRUE & is.null(margins$survfam)==FALSE){
    curerate <- param[2]
    parapar <- param[-c(1,2)]
  }else{
    parapar <- param[-1]
  }
  ktau    <- param[1]
  coppar  <- ktau.to.coppar(ktau=ktau,copfam=copfam)

  # probabilities for parametric part time (judge later)
  delta.parametric <- ifelse(rep(is.null(margins$survfam),length(yobs)),1-delta,delta)
  S.temp <- 1-parafam.p(x=yobs,parameter=parapar,distribution=c(margins$survfam,margins$censfam),truncation=c(margins$survtrunc,margins$censtrunc))
  f.temp <- parafam.d(x=yobs[delta.parametric==1],parameter=parapar,distribution=c(margins$survfam,margins$censfam),truncation=c(margins$survtrunc,margins$censtrunc))

  # distribution-related functions
  if(is.null(margins$survfam)){

    # - for survival time
    ST           <- SurvFunc.CG(tm=NULL,yobs=yobs,delta=delta,copfam=copfam,ktau=NULL,coppar=coppar)$Shat
    yobs.1.order <- order(yobs[delta==1])
    yobs.1.sort  <- yobs[delta==1][yobs.1.order]
    ST.1.sort    <- ST[delta==1][yobs.1.order]
    jumpT.1      <- c(1-ST.1.sort[1],-diff(ST.1.sort))
    h            <- (8*sqrt(2)/3)^(1/5)*sd(yobs.1.sort)*length(yobs)^(-1/5)
    fT.1         <- sapply(yobs[delta==1],function(x){sum(Kernel((x-yobs.1.sort)/h)*jumpT.1)/h})

    # - for censoring time
    SC   <- S.temp
    fC.0 <- f.temp

  }else{

    # - for survival time
    if(cure==TRUE){
      ST <- curerate + (1-curerate)*S.temp
      fT.1 <- (1-curerate)*f.temp
    }else{
      ST   <- S.temp
      fT.1 <- f.temp
    }

    # - for censoring time
    SC           <- SurvFunc.CG(tm=NULL,yobs=yobs,delta=1-delta,copfam=copfam,ktau=NULL,coppar=coppar)$Shat
    yobs.0.order <- order(yobs[delta==0])
    yobs.0.sort  <- yobs[delta==0][yobs.0.order]
    SC.0.sort    <- SC[delta==0][yobs.0.order]
    jumpC.0      <- c(1-SC.0.sort[1],-diff(SC.0.sort))
    h            <- (8*sqrt(2)/3)^(1/5)*sd(yobs.0.sort)*length(yobs)^(-1/5)
    fC.0         <- sapply(yobs[delta==0],function(x){sum(Kernel((x-yobs.0.sort)/h)*jumpC.0)/h})

  }

  # calculate likelihood
  hCT.1 <- cophfunc(x=cbind(ST,SC)[delta==1,,drop=FALSE],coppar=coppar,copfam=copfam,condvar=1)
  hTC.0 <- cophfunc(x=cbind(ST,SC)[delta==0,,drop=FALSE],coppar=coppar,copfam=copfam,condvar=2)
  Lik1  <- sum(log(pmax(1e-20,fT.1*hCT.1)),na.rm=TRUE)
  Lik0  <- sum(log(pmax(1e-20,fC.0*hTC.0)),na.rm=TRUE)
  Lik   <- Lik1 + Lik0

  # return
  return(Lik)

}

#' @title Calculate the kernel function
#'
#' @param u the value in which the kernel function will be calculated at.
#' @param name a character used to specify the type of kernel function.
#'
#' @noRd
Kernel <- function(u, name="Gaussian"){
  # the kernel function (not re-scaled) with only one continuous variable

  if(name=="Uniform"){
    ret <- 0.5*(abs(u)<=1)
  }else if(name=="Gaussian"){
    ret <- exp(-u^2/2)/sqrt(2*pi)
  }else if(name=="Epanechnikov"){
    ret <- 0.75*(1-u^2)*(abs(u)<=1)
  }else if(name=="Quartic"){
    ret <- (15/16)*(1-u^2)^2*(abs(u)<=1)
  }
  return(ret)
}

#' @title Prepare initial values within the control arguments
#'
#' @param maxit a positive integer that denotes the maximum iteration number in optimization.
#' @param eps a positive small numeric value that denotes the tolerance for convergence.
#' @param trace a logical value that judges whereh the tracing information on the progress of the model fitting should be produced. The default value if \code{TRUE}.
#' @param ktau.inits a numeric vector that contains initial values of the Kendall's tau.
#'
#' @noRd
control.arguments <- function(
    maxit      = 300,
    eps        = 1e-6,
    trace      = TRUE,
    ktau.inits = NULL
){
  return(list(
    maxit      = maxit,
    eps        = eps,
    trace      = trace,
    ktau.inits = ktau.inits
  ))
}

# Documentation for data sets.

#' @title Liver cirrhosis data set.
#'
#' @description End stage liver disease data set, as for example analyzed by
#' D'Haen et al. (2025).
#'
#' @format A data frame with 281 rows and 7 variables:
#' \describe{
#'  \item{patient}{patient ID.}
#'  \item{time}{Survival time.}
#'  \item{status}{Censoring indicator.}
#'  \item{age}{Age of patient.}
#'  \item{gender}{Gender of patient (1 - male, 0 - female).}
#'  \item{bmi}{Body mass index.}
#'  \item{ukeld}{UK End-stage Liver Disease score. Higher scores indicate more
#'  severe disease state.}
#' }
"liver"

#' @title Obtain vector of dependencies of the QRdepCens code base.
#'
#' @description
#' This function returns a character vector containing all required dependencies
#' for the correct operation of the QRdepCens code base.
#'
#' @noRd
#'
QRdepCens.getDependencies <- function() {
  dep.vec <- c("orthopolynom", "rvinecopulib", "nloptr", "MASS", "numDeriv")
}


#' @title Variance function.
#'
#' @description
#' This function describes the heteroscedasticity in the quantile regression
#' model of D\'Haen et al. (2025).
#'
#' @param X Vector of covariates. Default is \code{X = 1}, representing just
#' an intercept entry (i.e. presenting the homoscedastic case).
#' @param gam Vector of regression coefficients in the model for the variance.
#'
#' @noRd
#'
sigma_fun <- function(X = 1, gam) {
  if (length(X) == 1) {
    return(as.numeric(exp(X %*% gam)))
  } else {
    return(exp(X %*% gam))
  }
}

#' @title Squared L2-norm.
#'
#' @description
#' This function implements the sqaured L2-norm of a vector.
#'
#' @param vect Vector of numerics.
#'
#' @noRd
#'
sqnorm <- function(vect){
  return(sum(vect^2))
}

#' @title Laguerre coefficients
#'
#' @description
#' This function computes the coefficients of the Laguerre polynomial of a
#' Laguerre basis polynomial of a certain degree.
#'
#' @param n Degree of the Laguerre polynomial.
#' @param j Coefficient of \eqn{x^j} in the Laguerre basis polynomial.
#'
#' @noRd
#'
Lag_coeff <- function(n, j){
  choose(n,j)*(-1)^j/factorial(j)
}

#' @title Perturb covariates
#'
#' @description
#' This function perturbs a given vector of covariates using multiplicative
#' uniform noise sampled from \eqn{U[0.5, 2]}. Used to perturb initial values
#' for optimization routine.
#'
#' @param x Vector of covariates.
#'
#' @noRd
#'
perturb <- function(x) {
  return(x*runif(length(x), 0.5, 2))
}

#' @title Slightly perturb covariates
#'
#' @description
#' This function perturbs a given vector of covariates using multiplicative
#' uniform noise sampled from \eqn{U[2/3, 4/3]}. Used to perturb initial values
#' for optimization routine.
#'
#' @param x Vector of numerics.
#'
#' @seealso perturb
#'
#' @noRd
#'
slight_perturb <- function(x) {
  return(x*runif(length(x), 2/3, 4/3))
}

#' @title Partial derivative of Frank copula (wrt second argument).
#'
#' @description
#' This function computes the partial derivative wrt the second argument of a
#' Frank copula, evaluated at a certain point.
#'
#' @param u First argument to partial derivative of copula.
#' @param v Second argument to partial derivative of copula.
#' @param theta Copula parameter.
#'
#' @seealso [h_CT_frank()]
#'
#' @noRd
#'
h_TC_frank <- function(u, v, theta){
  res <- (exp(-theta*v)*(exp(-theta*u) - 1))/
    (exp(-theta) - 1 + (exp(-theta*u) - 1)*(exp(-theta*v) - 1))
  return(pmax(pmin(res, 0.99999),0.00001))
}

#' @title Partial derivative of Frank copula (wrt first argument).
#'
#' @description
#' This function computes the partial derivative wrt the first argument of a
#' Frank copula, evaluated at a certain point.
#'
#' @param u First argument to partial derivative of copula.
#' @param v Second argument to partial derivative of copula.
#' @param theta Copula parameter.
#'
#' @seealso [h_TC_frank()]
#'
#' @noRd
#'
h_CT_frank <- function(u,v,theta){
  res <- (exp(-theta*u)*(exp(-theta*v) - 1))/
    (exp(-theta) - 1 + (exp(-theta*v) - 1)*(exp(-theta*u) - 1))
  return(pmax(pmin(res, 0.99999),0.00001))
}

#' @title Partial derivative of Gumbel copula (wrt second argument).
#'
#' @description
#' This function computes the partial derivative wrt the second argument of a
#' Frank copula, evaluated at a certain point.
#'
#' @param u First argument to partial derivative of copula.
#' @param v Second argument to partial derivative of copula.
#' @param theta Copula parameter.
#'
#' @seealso [h_CT_gumbel()]
#'
#' @noRd
#'
h_TC_gumbel <- function(u,v,theta){
  res <- ((log(u)/log(v))^theta+1)^(-1 + 1/theta)*
    exp(-log(v) -((-log(u))^theta + (-log(v))^theta)^(1/theta))
  return(pmax(pmin(res, 0.99999),0.00001))
}

#' @title Partial derivative of Gumbel copula (wrt first argument).
#'
#' @description
#' This function computes the partial derivative wrt the first argument of a
#' Frank copula, evaluated at a certain point.
#'
#' @param u First argument to partial derivative of copula.
#' @param v Second argument to partial derivative of copula.
#' @param theta Copula parameter.
#'
#' @seealso [h_TC_gumbel()]
#'
#' @noRd
#'
h_CT_gumbel <- function(u,v,theta){
  res <- ((log(v)/log(u))^theta+1)^(-1 + 1/theta)*
    exp(-log(u) -((-log(v))^theta + (-log(u))^theta)^(1/theta))
  return(pmax(pmin(res, 0.99999),0.00001))
}

#' @title Partial derivative of Clayton copula (wrt second argument).
#'
#' @description
#' This function computes the partial derivative wrt the second argument of a
#' Frank copula, evaluated at a certain point.
#'
#' @param u First argument to partial derivative of copula.
#' @param v Second argument to partial derivative of copula.
#' @param theta Copula parameter.
#'
#' @seealso [h_CT_clayton()]
#'
#' @noRd
#'
h_TC_clayton <- function(u,v,theta){
  res <- (1 + (v/u)^theta - v^theta)^(-(theta+1)/theta)
  return(pmax(pmin(res, 0.99999),0.00001))
}

#' @title Partial derivative of Clayton copula (wrt first argument).
#'
#' @description
#' This function computes the partial derivative wrt the first argument of a
#' Frank copula, evaluated at a certain point.
#'
#' @param u First argument to partial derivative of copula.
#' @param v Second argument to partial derivative of copula.
#' @param theta Copula parameter.
#'
#' @seealso [h_TC_clayton()]
#'
#' @noRd
#'
h_CT_clayton <- function(u,v,theta){
  res <- (1 + (u/v)^theta - u^theta)^(-(theta+1)/theta)
  return(pmax(pmin(res, 0.99999),0.00001))
}

#' @title Sample covariates from specified distribution.
#'
#' @description
#' This function samples covariates from a specified distribution. Note that it
#' returns the vector with a prepended intercept term.
#'
#' @param size Number of covariate vectors to sample.
#' @param X_params List of parameters pertaining to the covariate distribution.
#' @param seed Initial seed. Default is \code{seed = NULL}.
#'
#' @noRd
#'
covariate_sampling <- function(size, params_X, seed = NULL) {

  # Set random seed if necessary.
  if (!is.null(seed)) {
    set.seed(seed)
  }

  # Extract parameter settings for the covariate distribution.
  X_distribution <- params_X[["X_distribution"]]
  X_dimension <- params_X[["X_dimension"]]
  X_mean <- params_X[["X_mean"]]
  X_var <- params_X[["X_var"]]
  X_lower <- params_X[["X_lower"]]
  X_upper <- params_X[["X_upper"]]

  # Generate the covariates.
  if (X_distribution == "norm") {
    Sigma <- matrix(X_var, nrow = X_dimension)
    return(cbind(1, mvrnorm(n = size, mu = X_mean, Sigma = Sigma)))
  }
  if (X_distribution == "unif") {
    X_dat <- cbind(1, runif(n = size, X_lower, X_upper))
    if (X_dimension == 2) {
      X_dat <- cbind(X_dat, runif(n = size, X_lower, X_upper))
    }
    return(X_dat)
  }
}

#' @title Extract covariates from data frame.
#'
#' @description
#' This function extracts the covariates from the given data frame,
#' corresponding to the column names X1, X2, etc.
#'
#' @param data Data frame.
#'
#' @return Data frame containing the covariates, with column names X1, X2, etc.
#'
#' @noRd
#'
extract.covariates <- function(data) {
  data[, grepl("X[1-9][[:digit:]]*$", colnames(data)), drop = FALSE]
}

#' @title Test if left-hand side is not contained in right-hand side.
#'
#' @description
#' Test whether the elements of the left-hand side vector are contained in the
#' right-hand side vector. This function can be seen as an analogue to the
#' built-in function \code{\%in\%}.
#'
#' @param lhs Left-hand side.
#' @param rhs Right-hand side.
#'
#' @noRd
#'
`%notin%` <- function(lhs, rhs) {
  unlist(lapply(lhs, function(num) {!(num %in% rhs)}))
}

#' @title Check if numeric is integer.
#' @description
#' This function checks if a given numeric value is an integer.
#'
#' @param nbr Numeric.
#'
#' @noRd
#'
is_integer <- function(nbr) {
  nbr == round(nbr)
}

#' @title Compute confidence intervals.
#'
#' @description
#' This function computes 95% confidence intervals based on the provided point
#' estimates and standard deviation estimates.
#'
#' @param est Point estimates.
#' @param sd Standard deviations.
#' @param as.string Boolean flag indicating whether the result should be
#' returned as a sting. Default is \code{as.string = TRUE}.
#'
#' @noRd
#'
make_ci <- function(est, sd, as.string = TRUE) {

  # Construct matrix of confidence intervals.
  ci.mat <- cbind(est - qnorm(0.975) * sd, est + qnorm(0.975) * sd)

  # Construct confidence intervals as strings.
  ci.mat.str <- apply(ci.mat, 1, function(row) {sprintf("[%.2f, %.2f]", round(row[1], 3), round(row[2], 3))})

  # Return the results
  if (as.string) {
    return(ci.mat.str)
  } else {
    return(ci.mat)
  }
}


#' @title Compute initial values for mll optimization for the heteroscedastic
#' model.
#'
#' @description
#' This function determines appropriate starting values for the mll optimization
#' function in the case heteroscedasticity is assumed.
#'
#' @param T1 Subvector of observed times corresponding to \eqn{Delta = 1}, i.e.
#' vector of observed event times.
#' @param X1 Subvector of covariates corresponding to \eqn{Delta = 1}.
#' @param C0 Subvector of observed times corresponding to \eqn{Delta = 0}, i.e.
#' vector of observed censoring times.
#' @param X0 Subvector of covariates corresponding to \eqn{Delta = 1}.
#' @param Delta Vector of censoring indicators.
#' @param cop_name Name of the copula to be used.
#' @param hp List of hyperparameters.
#'
#' @seealso perform_mll_optimisation, determine_init_basis
#'
#' @noRd
#'
determine_init_basis_het <- function(T1, X1, C0, X0, Delta, cop_name, hp) {

  # Extract hyperparameters
  init_value_number_basis <- hp[["init_value_number_basis"]]

  ## data-driven part: beta, gamma and paras_C ##

  # initial parameters for beta and paras_C
  # X already has intercept column: ~ X - 1
  alpha_init <- as.vector(lm(C0 ~ X0 - 1)$coefficients)
  beta_init <- as.vector(lm(T1 ~ X1 - 1)$coefficients)

  # Standard deviation of residuals in linear model for C.
  C_res <- C0 - as.matrix(X0) %*% alpha_init
  sigma_C_init <- sd(C_res)

  # to get an approximate value for gamma
  Z <- (T1 - X1 %*% beta_init)^2

  # avoid near-zero values
  Z[Z < 10^(-100)] <- rep(10^(-100), length((Z[Z < 10^(-100)])))
  gamma_init <- as.vector(lm(log(Z) ~ X1 -1)$coefficients/2)
  gamma_init_update <- c(gamma_init[1] - log(8)/2, gamma_init[-1])

  # used further on in loop for adjusting beta_0
  T_res_no_intercept <- T1 - as.matrix(X1[,-1]) %*% beta_init[-1]

  ## other parameters: random values in admissible/some plausible range
  init_values <- vector(mode = "list", length = init_value_number_basis)

  for (j in 1:init_value_number_basis){
    lambda_init <- runif(n = 1, min = 0.05, max = 0.95)
    transf_lambda_init <- lambda_transform(lambda_init)

    # adjust beta_0 based on the current initial value for lambda
    beta_intercept <- as.numeric(quantile(T_res_no_intercept, lambda_init))
    beta_init_update <- c(beta_intercept, beta_init[-1])

    if (cop_name == "frank"){ # ktau between -1 and 1
      eta_init <- ktau_to_eta(ktau = runif(1, -1, 1), cop_name = cop_name)
    }
    if (cop_name == "frankPos") { #ktau between 0 and 1
      eta_init <- ktau_to_eta(ktau = runif(1, 0, 1), cop_name = cop_name)
    }
    if (cop_name == "gumbel"){ # ktau between 0 and 1
      eta_init <- ktau_to_eta(ktau = runif(1,0,1), cop_name = cop_name)
    }
    if (cop_name == "clayton"){ # ktau between 0 and 1
      eta_init <- ktau_to_eta(ktau = runif(1,0,1), cop_name = cop_name)
    }
    if (cop_name == "indep"){
      eta_init <- 0 # not used during optimisation anyway
    }

    init_values[[j]] <- c(eta_init, perturb(beta_init_update),
                          perturb(gamma_init_update),
                          transf_lambda_init,
                          perturb(alpha_init),
                          log(perturb(sigma_C_init)))
  }
  return(init_values)
}

#' @title Compute initial values for mll optimization (homoscedastic case).
#'
#' @description
#' This function determines appropriate starting values for the mll optimization
#' function in the case that homoscedasticity can be assumed.
#'
#' @param T1 Subvector of observed times corresponding to \eqn{Delta = 1}, i.e.
#' vector of observed event times.
#' @param X1 Subvector of covariates corresponding to \eqn{Delta = 1}.
#' @param C0 Subvector of observed times corresponding to \eqn{Delta = 0}, i.e.
#' vector of observed censoring times.
#' @param X0 Subvector of covariates corresponding to \eqn{Delta = 1}.
#' @param Delta Vector of censoring indicators.
#' @param cop_name Name of the copula to be used.
#' @param hp List of hyperparameters.
#'
#' @seealso perform_mll_optimisation, determine_init_basis
#'
#' @noRd
#'
determine_init_basis_hom <- function(T1, X1, C0, X0, Delta, cop_name, hp) {

  # Extract hyperparameters
  init_value_number_basis <- hp[["init_value_number_basis"]]
  lambda <- hp[["lambda"]]

  ## data-driven part: beta, gamma and paras_C ##

  # initial parameters for beta and paras_C
  # X already has intercept column: ~ X - 1
  alpha_init <- as.vector(lm(C0 ~ X0 - 1)$coefficients)
  beta_init <- as.vector(lm(T1 ~ X1 - 1)$coefficients)

  # adjust beta_0 based on current initial values & (fixed) lambda
  T_res_no_intercept <- T1 - as.matrix(X1[,-1]) %*% beta_init[-1]
  beta_intercept <- as.numeric(quantile(T_res_no_intercept, lambda))
  beta_init_update <- c(beta_intercept, beta_init[-1])

  C_res <- C0 - as.matrix(X0) %*% alpha_init
  sigma_C_init <- sd(C_res)


  # initial parameter for gamma_T
  sigma_T_init <- sqrt(var(T_res_no_intercept) *
                         (lambda^2*(1 - lambda)^2)/(lambda^2 + (1-lambda)^2))
  gamma_T_init <- log(sigma_T_init)

  ## remaining parameter ktau/eta: random values in admissible range
  init_values <- vector(mode = "list", length = init_value_number_basis)

  for (j in 1:init_value_number_basis){
    if (cop_name == "frank"){ # ktau between -1 and 1
      eta_init <- ktau_to_eta(ktau = runif(1, -1, 1), cop_name = cop_name)
    }
    if (cop_name == "frankPos") { #ktau between 0 and 1
      eta_init <- ktau_to_eta(ktau = runif(1, 0, 1), cop_name = cop_name)
    }
    if (cop_name == "gumbel"){ # ktau between 0 and 1
      eta_init <- ktau_to_eta(ktau = runif(1,0,1), cop_name = cop_name)
    }
    if (cop_name == "clayton"){ # ktau between 0 and 1
      eta_init <- ktau_to_eta(ktau = runif(1,0,1), cop_name = cop_name)
    }
    if (cop_name == "indep"){
      eta_init <- 0 # not used during optimisation anyway
    }

    init_values[[j]] <- c(eta_init, perturb(beta_init_update),
                          perturb(gamma_T_init),
                          perturb(alpha_init),
                          log(perturb(sigma_C_init)))
  }
  return(init_values)
}

#' @title Compute initial values for mll optimization.
#'
#' @description
#' This function determines appropriate starting values for the mll optimization
#' function. This function is simply a wrapper for the homoscedastic and
#' heteroscedastic case.
#'
#' @param T1 Subvector of observed times corresponding to \eqn{Delta = 1}, i.e.
#' vector of observed event times.
#' @param X1 Subvector of covariates corresponding to \eqn{Delta = 1}.
#' @param C0 Subvector of observed times corresponding to \eqn{Delta = 0}, i.e.
#' vector of observed censoring times.
#' @param X0 Subvector of covariates corresponding to \eqn{Delta = 1}.
#' @param Delta Vector of censoring indicators.
#' @param cop_name Name of the copula to be used.
#' @param hp List of hyperparameters.
#'
#' @seealso perform_mll_optimisation, determine_init_basis_hom,
#' determine_init_basis_het
#'
#' @noRd
#'
determine_init_basis <- function(T1, X1, C0, X0, Delta, cop_name, hp) {
  if (hp$homoscedastic) {
    determine_init_basis_hom(T1, X1, C0, X0, Delta, cop_name, hp)
  } else {
    determine_init_basis_het(T1, X1, C0, X0, Delta, cop_name, hp)
  }
}

#' @title Determine list of initial values.
#'
#' @description
#' This function determines a list of initial values based on slightly
#' perturbing the given vector of parameters.
#'
#' @param current_paras Parameter vector. In the implementation, this is the
#' intermediate result after degree selection of the Laguerre polynomials.
#' @param hp List of hyperparameters.
#' @param indep_assumption Boolean flag indicating whether independence can be
#' assumed. Default is \code{indep_assumption = FALSE}.
#'
#' @noRd
#'
determine_init_final <- function(current_paras, hp, indep_assumption = FALSE) {
  init_value_number_final <- hp[["init_value_number_final"]]
  new_init_values <- vector(mode = "list", length = init_value_number_final)
  if (!indep_assumption) {
    for (j in seq_len(init_value_number_final)) {
      new_init_values[[j]] <- slight_perturb(x = current_paras)
    }
  } else {
    for (j in seq_len(init_value_number_final)) {
      new_init_values[[j]] <- c(0, slight_perturb(x = current_paras))
    }
  }
  return(new_init_values)
}


#' @title Density function of error term for C (homoscedastic + Gaussian case).
#'
#' @description
#' This function defines the density function of the error term in the linear
#' model for C, in the homoscedastic case and when a normal distribution is
#' assumed on C.
#'
#' @param y Outcome variables; i.e. censoring times.
#' @param x Predictors; i.e. covariates.
#' @param alpha_C Regression coefficients.
#' @param sigma_C Standard deviation.
#'
#' @noRd
#'
dCens_norm <- function(y, x, alpha_C, sigma_C){
  z <- 1/sigma_C*(y -  x %*% alpha_C)
  return(1/sigma_C*dnorm(z))
}

#' @title Distribution function of error term for C (homoscedastic + Gaussian
#' case).
#'
#' @description
#' This function defines the distribution function of the error term in the linear
#' model for C, in the homoscedastic case and when a normal distribution is
#' assumed on C.
#'
#' @inheritParams dCens_norm
#'
#' @noRd
#'
pCens_norm <- function(y, x, alpha_C, sigma_C){
  z <- 1/sigma_C*(y -  x %*% alpha_C)
  return(pnorm(z))
}

#' @title Quantile function of error term for C (homoscedastic + Gaussian case).
#'
#' @description
#' This function defines the quantile function of the error term in the linear
#' model for C, in the homoscedastic case and when a normal distribution is
#' assumed on C.
#'
#' @inheritParams dCens_norm
#'
#' @noRd
#'
qCens_norm <- function(w, x, alpha_C, sigma_C) {
  return(qnorm(w)*sigma_C + x %*% alpha_C)
}

#' @title Density function of error term for C (homoscedastic case).
#'
#' @description
#' This function defines the density function of the error term in the linear
#' model for C, in the homoscedastic case.
#'
#' @param y Outcome variables; i.e. censoring times.
#' @param x Predictors; i.e. covariates.
#' @param distr_C Name of distribution for C.
#' @param params_C List of parameters for the model for C. Default is
#' \code{params_C = NULL}. However, a \code{NULL} value is only valid in the
#' specific case where the function is called with only three arguments.
#'
#' @noRd
#'
dCens <- function(y, x, distr_C, params_C = NULL) {

  # -- For compatibility with estimation methodology --
  # Throughout the implementation of the estimation procedure, this function
  # is used with arguments dCens(y, x, transf_params_C) and the distribution of
  # C is assumed to be normal.
  if (is.null(params_C) & !is.character(distr_C)) {
    transf_paras_C <- distr_C
    alpha_C <- transf_paras_C[-length(transf_paras_C)]
    sigma_C <- exp(transf_paras_C[length(transf_paras_C)])
    distr_C <- "norm"
    params_C <- list(alpha_C = alpha_C, sigma_C = sigma_C)
  }

  # If C follows a normal distribution...
  if (distr_C == "norm") {
    alpha_C <- params_C[["alpha_C"]]
    sigma_C <- params_C[["sigma_C"]]
    return(dCens_norm(y, x, alpha_C, sigma_C))
  }

  # If non of the above distribution applied, throw an error.
  stop("Distribution for C not implemented yet.")
}

#' @title Distribution function of error term for C (homoscedastic case).
#'
#' @description
#' This function defines the distribution function of the error term in the linear
#' model for C, in the homoscedastic case.
#'
#' @inheritParams dCens
#'
#' @noRd
#'
pCens <- function(y, x, distr_C, params_C = NULL) {

  # -- For compatibility with estimation methodology --
  # Throughout the implementation of the estimation procedure, the this function
  # is used with arguments dCens(y, x, transf_params_C) and the distribution of
  # C is assumed to be normal.
  if (is.null(params_C) & !is.character(distr_C)) {
    transf_paras_C <- distr_C
    alpha_C <- transf_paras_C[-length(transf_paras_C)]
    sigma_C <- exp(transf_paras_C[length(transf_paras_C)])
    distr_C <- "norm"
    params_C <- list(alpha_C = alpha_C, sigma_C = sigma_C)
  }

  # If C follows a normal distribution...
  if (distr_C == "norm") {
    alpha_C <- params_C[["alpha_C"]]
    sigma_C <- params_C[["sigma_C"]]
    return(pCens_norm(y, x, alpha_C, sigma_C))
  }

  # If non of the above distribution applied, throw an error.
  stop("Distribution for C not implemented yet.")
}

#' @title Quantile function of error term for C (homoscedastic case).
#'
#' @description
#' This function defines the quantile function of the error term in the linear
#' model for C, in the homoscedastic case.
#'
#' @inheritParams dCens
#'
#' @noRd
#'
qCens <- function(y, x, distr_C, params_C = NULL) {

  # -- For compatibility with estimation methodology --
  # Throughout the implementation of the estimation procedure, the this function
  # is used with arguments dCens(y, x, transf_params_C) and the distribution of
  # C is assumed to be normal.
  if (is.null(params_C) & !is.character(distr_C)) {
    transf_paras_C <- distr_C
    alpha_C <- transf_paras_C[-length(transf_paras_C)]
    sigma_C <- exp(transf_paras_C[length(transf_paras_C)])
    distr_C <- "norm"
    params_C <- list(alpha_C = alpha_C, sigma_C = sigma_C)
  }

  # If C follows a normal distribution...
  if (distr_C == "norm") {
    alpha_C <- params_C[["alpha_C"]]
    sigma_C <- params_C[["sigma_C"]]
    return(qCens_norm(y, x, alpha_C, sigma_C))
  }

  # If non of the above distribution applied, throw an error.
  stop("Distribution for C not implemented yet.")
}

#### 4)  Transformation functions ####

#' @title Compute Kendall's tau based on \eqn{eta}.
#'
#' @param eta Transformed Kendall's tau parameter, which can take values on the
#' entire real line.
#' @param cop_name Name of copula under consideration. Could be either
#' \code{"frank"}, \code{"frankPos"}, \code{"gumbel"}, \code{"clayton"} or
#' \code{"indep"}.
#'
#' @returns Kendall's tau.
#'
#' @noRd
#'
eta_to_ktau <- function(eta, cop_name){
  if (cop_name == "frank"){ # ktau between -1 and 1
    ktau <- tanh(eta)
  }
  if (cop_name %in% c("frankPos", "gumbel", "clayton")) { #ktau between 0 and 1
    ktau <- 1/(1 + exp(-eta))
  }
  if (cop_name == "indep"){ # no ktau present
    ktau <- 0
  }
  return(ktau)
}

#' @title Compute \eqn{eta} based on Kendall's tau.
#'
#' @param ktau Value of Kendall's tau.
#' @param cop_name Name of copula under consideration. Could be either
#' \code{"frank"}, \code{"frankPos"}, \code{"gumbel"}, \code{"clayton"} or
#' \code{"indep"}.
#'
#' @returns eta.
#'
#' @noRd
#'
ktau_to_eta <- function(ktau, cop_name){
  if (cop_name == "frank"){ # ktau between -1 and 1
    eta <- 0.5* log((1 + ktau)/(1 - ktau))
  }
  if (cop_name %in% c("frankPos", "gumbel", "clayton")) { # ktau between 0 and 1
    eta <- log(ktau/(1 - ktau))
  }
  if (cop_name == "indep"){ # no ktau present
    eta <- 0
  }
  return(eta)
}

#' @title Compute copula parameter based on Kendall's tau.
#'
#' @param ktau Value of Kendall's tau.
#' @param cop_name Name of copula under consideration. Could be either
#' \code{"frank"}, \code{"frankPos"}, \code{"gumbel"}, \code{"clayton"} or
#' \code{"indep"}.
#'
#' @importFrom rvinecopulib ktau_to_par
#'
#' @returns Copula parameter.
#'
#' @noRd
#'
ktau_to_theta <- function(ktau, cop_name) {
  cop_name_rvcl <- cop_name
  if (cop_name == "frankPos") {
    cop_name_rvcl <- "frank"
  }
  theta <- as.numeric(rvinecopulib::ktau_to_par(family = cop_name_rvcl, tau = ktau))
  return(theta)
}

#' @title Transform quantile value to whole real line.
#'
#' @description
#' This function transforms the quantile value, which is probability and hence
#' a value between [0, 1] to the whole real line.
#'
#' @param lambda Quantile value.
#'
#' @returns Transformed quantile value.
#'
#' @seealso lambda_inverse_transform.
#'
#' @noRd
#'
lambda_transform <- function(lambda){
  return(log(lambda/(1 - lambda)))
}

#' @title Back-transform quantile value
#'
#' @description
#' This function transforms the quantile value represented as a number on the
#' whole real line to one that takes values in [0, 1].
#'
#' @param transf_lambda Transformed quantile value.
#'
#' @returns Back-transformed quantile value.
#'
#' @seealso lambda_transform.
#'
#' @noRd
#'
lambda_inverse_transform <- function(transf_lambda){
  return(1/(1 + exp(-transf_lambda)))
}

#' @title Back-transform the parameter vector.
#'
#' @description
#' This function back-transforms certain elements of the parameter vector. In
#' particular, it targets the dependence parameter, variance parameter and - if
#' applicable - lambda. I.e., all parameters that are restricted to some proper
#' subset of the real line.
#'
#' @param transf_para_vec Transformed parameter vector.
#' @param cop_name Name of considered copula.
#' @param hp List of hyperparameters.
#'
#' @returns Transformed parameter vector.
#'
#' @noRd
#'
para_inv_transform <- function(transf_para_vec, cop_name, hp) {

  # Extract hyperparameter
  lambda_index <- hp[["lambda_index"]]
  homoscedastic <- hp[["homoscedastic"]]

  para_vec <- transf_para_vec

  if (cop_name == "indep"){
    ktau_length <- 0
  } else {
    ktau_length <- 1
  }

  # replace eta by ktau, if present
  if (ktau_length == 1){
    para_vec[1] <- eta_to_ktau(eta = para_vec[1], cop_name = cop_name)
  }

  # replace transf_lambda by lambda
  if (!homoscedastic) {
    para_vec[lambda_index + ktau_length] <- lambda_inverse_transform(
      para_vec[lambda_index + ktau_length])
  }

  # replace log_sigma_C by sigma_C
  para_vec[length(para_vec)] <- exp(para_vec[length(para_vec)])

  return(para_vec)
}


#' @title Density function of EAL distribution.
#'
#' @description
#' This function implements the density function of an Enriched Asymetric
#' Laplace (EAL) distribution.
#'
#' @param y (vector of) Observed time(s).
#' @param x Covariate vector/matrix.
#' @param beta Coefficients in linear quantile regression model.
#' @param gam_par Parameters for variance model (heteroscedastic case.)
#' @param lambda Quantile of interest.
#' @param phi_til Laguerre expansion coefficients for y <= 0.
#' @param phi Laguerre expansion coefficients for y > 0.
#' @param hp List of hyperparameters.
#'
#' @returns EAL density function evaluation.
#'
#' @noRd
#'
dEAL <- function(y, x, beta, gam_par, lambda, phi_til, phi, hp) {

  # Load dependency if necessary
  requireNamespace("orthopolynom", quitely = TRUE)

  # Standard deviation in quantile regression model for T.
  x.simga_fun <- if (hp$homoscedastic) {1} else {x}
  s <- sigma_fun(x.simga_fun, gam_par)

  # Point at which to evaluate the EAL distribution; residuals in linear
  # quantile regression model.
  z <- 1/s * (y - x %*% beta)

  # Laguerre polynomial coefficients.
  phi_til_ext <- c(1, phi_til)
  phi_ext <- c(1, phi)

  # Lag of degree 0 is constant 1, so first term in sum is always fixed constant
  left_sum <- 1
  right_sum <- 1
  poly_list <- orthopolynom::laguerre.polynomials(max(length(phi),length(phi_til)),
                                                  normalized = TRUE)

  if (length(y) == 1){ # avoid indexing one vector x as x[pos_indices,]
    if (z <= 0){
      for (k in seq_along(phi_til)){
        poly <- as.function(poly_list[[k+1]])
        left_sum <- left_sum + phi_til[k]*poly((lambda-1)*z)
      }
      result <- (1-lambda)*lambda/s*1/sqnorm(phi_til_ext)*exp(-z *(lambda-1)) * left_sum^2

    } else {
      for (k in seq_along(phi)){
        poly <- as.function(poly_list[[k+1]])
        right_sum <- right_sum + phi[k]*poly(lambda*z)
      }
      result <- lambda*(1-lambda)/s*1/sqnorm(phi_ext)*exp(-z*lambda) * right_sum^2
    }
  } else {
    pos_indices <- which(z > 0)

    z_pos <- z[pos_indices]
    z_neg <- z[-pos_indices]

    if (hp$homoscedastic) {
      s_pos <- s
      s_neg <- s
    } else {
      s_pos <- s[pos_indices]
      s_neg <- s[-pos_indices]
    }

    for (k in seq_along(phi_til)){
      poly <- as.function(poly_list[[k+1]])
      left_sum <- left_sum + phi_til[k]*poly((lambda-1)*z_neg)
    }

    for (k in seq_along(phi)){
      poly <- as.function(poly_list[[k+1]])
      right_sum <- right_sum + phi[k]*poly(lambda*z_pos)
    }

    result <- rep(0, length(z))
    result[pos_indices] <- lambda*(1-lambda)/s_pos*1/sqnorm(phi_ext)*
      exp(-z_pos*lambda) * right_sum^2
    result[-pos_indices] <- (1-lambda)*lambda/s_neg*1/sqnorm(phi_til_ext)*
      exp(-z_neg *(lambda-1)) * left_sum^2
  }
  return(as.matrix(result))
}

#' @title Left side of EAL distribution function.
#'
#' @description
#' This function implements the left side of the distribution function of an
#' Enriched Asymetric Laplace (EAL) distribution.
#'
#' @param y (vector of) Observed time(s).
#' @param x Covariate vector/matrix.
#' @param beta Coefficients in linear quantile regression model.
#' @param gam_par Parameters for variance model (heteroscedastic case.)
#' @param lambda Quantile of interest.
#' @param phi_til Laguerre expansion coefficients for y <= 0.
#' @param phi Laguerre expansion coefficients for y > 0.
#'
#' @returns EAL distribution function evaluation.
#'
#' @seealso pEAL
#'
#' @noRd
#'
pEAL_neg <- function(y, x, beta, gam_par, lambda, phi_til) {

  # Determine if homoscedasticity is assumed
  homoscedastic <- length(gam_par) == 1

  # Standard deviation in quantile regression model for T.
  x.simga_fun <- if (homoscedastic) {1} else {x}
  s <- sigma_fun(x.simga_fun, gam_par)

  z <- 1/s * (y - x %*% beta)

  phi_til_ext <- c(1, phi_til)
  n <- length(phi_til)

  sum <- 0
  for (a in 0:n){
    for (b in 0:n){
      for (c in 0:a){
        for (d in 0:b){
          inner_sum <- 0
          for (j in 0:(c+d)){
            inner_sum <- inner_sum + factorial(c+d)/factorial(j)*
              (-(1-lambda)*z)^j
          }
          sum <- sum + phi_til_ext[a+1]*phi_til_ext[b+1] *
            Lag_coeff(a,c)*Lag_coeff(b,d)* inner_sum
        }
      }
    }
  }
  res <- lambda*exp((1-lambda)*z) * sum/sqnorm(phi_til_ext)

  res_corrected <- pmin(pmax(0, res), 1)
  return(res_corrected) # could still be NaN !
}

#' @title Right side of EAL distribution function.
#'
#' @description
#' This function implements the right side of the distribution function of an
#' Enriched Asymetric Laplace (EAL) distribution.
#'
#' @param y (vector of) Observed time(s).
#' @param x Covariate vector/matrix.
#' @param beta Coefficients in linear quantile regression model.
#' @param gam_par Parameters for variance model (heteroscedastic case.)
#' @param lambda Quantile of interest.
#' @param phi_til Laguerre expansion coefficients for y <= 0.
#' @param phi Laguerre expansion coefficients for y > 0.
#'
#' @returns EAL distribution function evaluation.
#'
#' @seealso pEAL
#'
#' @noRd
#'
pEAL_pos <- function(y, x, beta, gam_par, lambda, phi) {

  # Determine if homoscedasticity is assumed
  homoscedastic <- length(gam_par) == 1

  # Standard deviation in quantile regression model for T.
  x.simga_fun <- if (homoscedastic) {1} else {x}
  s <- sigma_fun(x.simga_fun, gam_par)

  z <- 1/s * (y - x %*% beta)

  phi_ext <- c(1, phi)
  n <- length(phi)

  sum <- 0
  for (a in 0:n){
    for (b in 0:n){
      for (c in 0:a){
        for (d in 0:b){
          inner_sum <- 0
          for (j in 0:(c+d)){
            inner_sum <- inner_sum + factorial(c+d)/factorial(j)*(lambda*z)^j
          }
          sum <- sum + phi_ext[a+1]*phi_ext[b+1]*Lag_coeff(a,c)*Lag_coeff(b,d)*
            (factorial(c+d)- exp(-lambda*z)*inner_sum)
        }
      }
    }
  }
  res <- lambda + (1-lambda) * sum/sqnorm(phi_ext)

  res_corrected <- pmin(pmax(0, res), 1)
  return(res_corrected) # could still be NaN !
}

#' @title EAL distribution function.
#'
#' @description
#' This function implements the distribution function of an Enriched Asymetric
#' Laplace (EAL) distribution. It is mainly a wrapper around the left and right
#' side distribution function.
#'
#' @param y (vector of) Observed time(s).
#' @param x Covariate vector/matrix.
#' @param beta Coefficients in linear quantile regression model.
#' @param gam_par Parameters for variance model (heteroscedastic case.)
#' @param lambda Quantile of interest.
#' @param phi_til Laguerre expansion coefficients for y <= 0.
#' @param phi Laguerre expansion coefficients for y > 0.
#'
#' @returns EAL distribution function evaluation.
#'
#' @seealso [pEAL_neg(), pEAL_pos()]
#'
#' @noRd
#'
pEAL <- function(y, x, beta, gam_par, lambda, phi_til, phi) {

  # Determine if homoscedasticity is assumed
  homoscedastic <- length(gam_par) == 1

  # Standard deviation in quantile regression model for T.
  x.simga_fun <- if (homoscedastic) {1} else {x}
  s <- sigma_fun(x.simga_fun, gam_par)

  z <- 1/s * (y - x %*% beta)


  if (length(y) == 1){ # avoid indexing one vector x as x[pos_indices,]
    if (z < 0){
      res_vec <- pEAL_neg(y, x, beta, gam_par, lambda, phi_til)
    } else {
      res_vec <- pEAL_pos(y, x, beta, gam_par, lambda, phi)
    }
  } else { # general case, correct subsetting dimensions
    pos_indices <- which(z > 0)
    res_vec <- as.matrix(rep(0, length(z)))

    if (length(pos_indices) == 0){
      res_vec <- pEAL_neg(y, x, beta, gam_par, lambda, phi_til)
    } else {
      res_vec[pos_indices,] <- pEAL_pos(y[pos_indices], x[pos_indices,],
                                        beta, gam_par, lambda, phi)


      if (length(pos_indices) < length(z)){
        res_vec[-pos_indices,] <- pEAL_neg(y[-pos_indices], x[-pos_indices,],
                                           beta, gam_par, lambda, phi_til)
      }
    }
  }

  # correct for numerical errors that lead to values outside [0,1]
  res_vec <- pmin(pmax(0, res_vec), 1)

  return(as.matrix(res_vec)) # could still be NaN !
}

#' @title EAL quantile function.
#'
#' @description
#' This function implements the quantile function of an Enriched Asymetric
#' Laplace (EAL) distribution.
#'
#' @param w (vector of) Observed probability/ies.
#' @param x Covariate vector/matrix.
#' @param beta Coefficients in linear quantile regression model.
#' @param gam_par Parameters for variance model (heteroscedastic case.)
#' @param lambda Quantile of interest.
#' @param phi_til Laguerre expansion coefficients for y <= 0.
#' @param phi Laguerre expansion coefficients for y > 0.
#'
#' @returns EAL quantile function evaluation.
#'
#' @noRd
#'
qEAL <- function(w, x, beta, gam_par, lambda, phi_til, phi) {

  # Set some useful parameters
  X_dimension <- length(beta) - 1

  # Determine if homoscedasticity is assumed
  homoscedastic <- length(gam_par) == 1

  # determine basis quantile from EAL, only afterwards shift
  x_null = matrix(rep(1, length(w)*(X_dimension + 1)), nrow = length(w))
  beta_null = rep(0, X_dimension + 1)
  gam_par_null = if (homoscedastic) {0} else {rep(0, X_dimension + 1)}

  f_null <- function(y){
    return(pEAL(y, x_null, beta_null, gam_par_null, lambda, phi_til, phi) - w)
  }

  basis_quantile <- uniroot(f_null, lower = -10, upper = 10, extendInt = "yes")$root

  x.sigma_fun <- if (homoscedastic) {1} else {x}
  s <- sigma_fun(x.sigma_fun, gam_par)

  return(x %*% beta + s*basis_quantile)
}



#' @title Log-likelihood function of D'Haen et al. (2025).
#'
#' @description This function implements the log-likelihood function used in
#' the method of D'Haen et al. (2025). In particular, it implements equation
#' (4.1) of their paper.
#'
#' @param all_paras Vector containing all parameters.
#' @param lag_degs Degrees of Laguerre polynomials.
#' @param cop_name Name of considered copula.
#' @param Y Vector of observed times.
#' @param X Design matrix.
#' @param Delta Vector of censoring indicators.
#' @param hp List of hyperparameters.
#'
#' @returns Log-likelihood evaluation.
#'
#' @noRd
#'
#' @references D'Haen, M., Van Keilegom, I. and Verhasselt, A. (2025). Quantile
#' regression under dependent censoring with unknown association. Lifetime Data
#' Analysis 31(2):253-299.
#'
log_lik_fun <- function(all_paras, lag_degs, cop_name, Y, X, Delta, hp) {

  # 0) Extract hyperparameters

  beta_full_index <- hp[["beta_full_index"]]
  gamma_index <- hp[["gamma_index"]]
  lambda_index <- hp[["lambda_index"]]
  phi_til_index <- hp[["phi_til_index"]]
  phi_index <- hp[["phi_index"]]
  lambda <- hp[["lambda"]]
  homoscedastic <- hp[["homoscedastic"]]
  heteroscedastic <- !homoscedastic

  # 1) Retrieving current parameters & corresponding copula object

  paras_T_indices <- c(beta_full_index, gamma_index, lambda_index,
                       phi_til_index(lag_degs),
                       phi_index(lag_degs)) + 1

  eta <- all_paras[1]
  transf_paras_T <- all_paras[paras_T_indices]
  transf_paras_C <- all_paras[-c(1, paras_T_indices)]

  beta_full <- transf_paras_T[beta_full_index]
  if (heteroscedastic) {
    transf_lam <- transf_paras_T[lambda_index]
    lambda <- lambda_inverse_transform(transf_lam)
  }
  gam_par <- transf_paras_T[gamma_index]
  phi_til <- transf_paras_T[phi_til_index(lag_degs)]
  phi <- transf_paras_T[phi_index(lag_degs)]

  # 2) Construct argument matrices for censored & uncensored cases
  Y_un <- Y[Delta == 1]
  Y_cens <- Y[Delta == 0]

  X_un <- X[Delta == 1, ]
  X_cens <- X[Delta == 0, ]

  # 3) Density, distribution & copula partial derivative evaluation

  ## density
  dens_T_un <- dEAL(Y_un, X_un, beta_full, gam_par, lambda, phi_til, phi, hp)
  dens_C_cens <- dCens(Y_cens, X_cens, transf_paras_C)

  ## distribution
  distr_T_un <- pEAL(Y_un, X_un, beta_full, gam_par, lambda, phi_til, phi)
  distr_T_cens <- pEAL(Y_cens, X_cens, beta_full, gam_par, lambda, phi_til, phi)

  distr_C_un <- pCens(Y_un, X_un, transf_paras_C)
  distr_C_cens <- pCens(Y_cens, X_cens, transf_paras_C)


  ## copula partial derivatives
  theta <- ktau_to_theta(eta_to_ktau(eta, cop_name), cop_name)

  if (cop_name %in% c("frank", "frankPos")) {
    h_C_given_T_uncens <- h_CT_frank(u = distr_T_un, v = distr_C_un, theta)
    h_T_given_C_cens <- h_TC_frank(u = distr_T_cens, v = distr_C_cens, theta)
  }
  if (cop_name == "gumbel"){
    h_C_given_T_uncens <- h_CT_gumbel(u = distr_T_un, v = distr_C_un, theta)
    h_T_given_C_cens <- h_TC_gumbel(u = distr_T_cens, v = distr_C_cens, theta)
  }
  if (cop_name == "clayton"){
    h_C_given_T_uncens <- h_CT_clayton(u = distr_T_un, v = distr_C_un, theta)
    h_T_given_C_cens <- h_TC_clayton(u = distr_T_cens, v = distr_C_cens, theta)
  }

  # Return the results

  lh_uncens <- dens_T_un * (1 - h_C_given_T_uncens)
  lh_cens <- dens_C_cens * (1 - h_T_given_C_cens)

  llh <- sum(log(lh_uncens)) + sum(log(lh_cens))

  res <- min(max(llh, -100000), 100000) # avoid returning +- Inf
  if (is.nan(res)){ # avoid NaN (due to Inf - Inf by num. errors) as well
    res <- -100000
  }
  return(res)
}

#' @title Negative log-likelihood function of D'Haen et al. (2025).
#'
#' @description This function implements the log-likelihood function used in
#' the method of D'Haen et al. (2025). In particular, it implements equation
#' (4.1) of their paper. It will return the negative of the
#' log-likelihood. This function is simply a wrapper around
#' \code{log_lik_fun}.
#'
#' @inheritParams log_lik_fun
#' @returns Negative of the log-likelihood evaluation
#' @seealso log_lik_fun
#'
#' @noRd
#'
neg_log_lik_fun <- function(all_paras, lag_degs, cop_name, Y, X, Delta, hp) {
  -log_lik_fun(all_paras, lag_degs, cop_name, Y, X, Delta, hp)
}

#' @title Continuity constraint of EAL distribution.
#'
#' @description
#' This function implements the equality constraint (6.1) of D'Haen et al.
#' (2025) that ensures continuity of the EAL distribution. It is implemented
#' as a combination of two inequality constraints.
#'
#' @param all_paras Vector containing all parameters.
#' @param lag_degs Laguerre polynomial maximal degrees.
#' @param cop_name Name of considered copula.
#' @param Y Vector of observed times.
#' @param X Design matrix.
#' @param Delta Vector of censoring indicators.
#' @param hp List of hyperparameters.
#'
#' @returns Vector of continuity constraint violations.
#'
#' @noRd
#'
ctuity_constr_fun <- function(all_paras, lag_degs,
                              cop_name, Y, X, Delta, hp) {

  # Extract hyperparameters
  phi_til_index <- hp[["phi_til_index"]]
  phi_index <- hp[["phi_index"]]

  # args. on second line not needed, but should be the same as for obj. fun.

  phi_til <- all_paras[phi_til_index(lag_degs) + 1] # omit eta
  phi <- all_paras[phi_index(lag_degs) + 1] # omit eta

  phi_til_ext <- c(1, phi_til)
  phi_ext <- c(1, phi)

  # combination of Lag. polynomials evaluated in 0 on neg./pos. axis:
  left <- sum(phi_til_ext)^2 / sqnorm(phi_til_ext)
  right <- sum(phi_ext)^2 / sqnorm(phi_ext)

  jump <- left - right

  return(c(jump, - jump))
}

#' @title Optimize the log-likelihood function.
#'
#' @description
#' This function optimizes the likelihood function using the specified settings.
#' In particular, it implements the the (Nelder-Mead + COBYLA)-hybrid algorithm
#' as described in D'Haen et al. (2025).
#'
#' @param lag_degs Laguerre polynomial maximal degrees.
#' @param cop_name Name of considered copula.
#' @param Y Vector of observed times.
#' @param X Design matrix.
#' @param Delta Vector of censoring indicators.
#' @param init_val_list List of initial values.
#' @param opt_algo Algorithm to be used for optimization. Could be one of
#' \code{"NM"} or \code{"NMCob"}, corresponding to Nelder-Mead and the
#' (Nelder-Mead + COBYLA)-hybrid, respectively.
#' @param hp List of hyperparameters.
#' @param indep_assumption Boolean indicator whether independence can be
#' assumed. Default is \code{indep_assumption = FALSE}.
#'
#' @returns List of program value and argument minimizers (i.e. ML estimates).
#'
#' @noRd
#'
perform_mll_optimisation <- function(lag_degs, cop_name, Y, X, Delta,
                                     init_val_list,
                                     opt_algo, hp,
                                     indep_assumption = FALSE) {

  # Set some parameters
  it_num_unconstr <- hp[["it_NMCob_unconstr"]]
  it_num_constr <- hp[["it_NMCob_constr"]]

  # Set initial values for max/argmax log-likelihood.
  current_mllh <- -Inf
  current_paras <- init_val_list[[1]]

  # Starting from each initial value, optimize the log-likelihood using the
  # Nelder-Mead algorithm.
  if (opt_algo == "NM") {

    # If independence is assumed between T and C...
    if (indep_assumption) {

      # In the case where independence between T and C can be assumed, there is
      # no dependence parameter (first entry of argmax vector) to be estimated.
      current_paras <- current_paras[-1]

      # Optimize the likelihood for each starting value.
      for (init_vals in init_val_list) {
        res <- optim(init_vals[-1], log_lik_fun_indep, method = "Nelder-Mead",
                     control = list(fnscale = -1, maxit = it_num_unconstr),
                     lag_degs = lag_degs, cop_name = cop_name,
                     Y = Y, X = X, Delta = Delta, hp = hp)
        new_paras <- res[[1]]
        new_mllh <- res[[2]]

        # If necessary, update max and argmax.
        if (new_mllh > current_mllh){
          current_mllh <- new_mllh
          current_paras <- new_paras
        }
      }

      # If independence is not assumed between T and C...
    } else {

      # Optimize the likelihood for each starting value.
      for (init_vals in init_val_list) {
        res <- optim(init_vals, log_lik_fun, method = "Nelder-Mead",
                     control = list(fnscale = -1, maxit = it_num_unconstr),
                     lag_degs = lag_degs, cop_name = cop_name,
                     Y = Y, X = X, Delta = Delta, hp = hp)
        new_paras <- res[[1]]
        new_mllh <- res[[2]]

        # If necessary, update max and argmax.
        if (new_mllh > current_mllh){
          current_mllh <- new_mllh
          current_paras <- new_paras
        }
      }
    }
  }

  # Starting from each initial value, optimize the log-likelihood using the
  # Nelder-Mead -- COBYLA hybrid algorithm.
  if (opt_algo == "NMCob"){

    # Set optimization hyperparameters.
    opts <- list("algorithm"= "NLOPT_LN_COBYLA", "maxeval"= it_num_constr)

    if (indep_assumption) {

      # In the case where independence between T and C can be assumed, there is
      # no dependence parameter (first entry of argmax vector) to be estimated.
      current_paras <- current_paras[-1]

      # Optimize the likelihood for each starting value.
      for (init_vals in init_val_list) {

        # Nelder-Mead first stage
        intermed_para <- optim(init_vals[-1], log_lik_fun_indep,
                               method = "Nelder-Mead",
                               control = list(fnscale = -1,
                                              maxit = it_num_unconstr),
                               lag_degs = lag_degs, cop_name = cop_name,
                               Y = Y, X = X, Delta = Delta, hp = hp)$par

        # COBYLA second stage
        res <- nloptr(x0 = intermed_para,
                      eval_f = neg_log_lik_fun_indep, # max. rather than min.!
                      eval_g_ineq = ctuity_constr_fun_indep,
                      opts = opts,
                      lag_degs = lag_degs, cop_name = cop_name,
                      Y = Y, X = X, Delta = Delta, hp = hp)
        new_paras <- res$solution
        new_mllh <- - res$objective # Compare with llh rather than neg_llh

        # Update max/argmax if necessary
        if (new_mllh > current_mllh){
          current_mllh <- new_mllh
          current_paras <- new_paras
        }
      }
    } else {
      for (init_vals in init_val_list){
        intermed_para <- optim(init_vals, log_lik_fun,
                               method = "Nelder-Mead",
                               control = list(fnscale = -1,
                                              maxit = it_num_unconstr),
                               lag_degs = lag_degs, cop_name = cop_name,
                               Y = Y, X = X, Delta = Delta, hp = hp)$par

        res <- nloptr(x0 = intermed_para,
                      eval_f = neg_log_lik_fun, # max. rather than min.!
                      eval_g_ineq = ctuity_constr_fun,
                      opts = opts,
                      lag_degs = lag_degs, cop_name = cop_name,
                      Y = Y, X = X, Delta = Delta, hp = hp)
        new_paras <- res$solution
        new_mllh <- - res$objective # Compare with llh rather than neg_llh

        if (new_mllh > current_mllh){
          current_mllh <- new_mllh
          current_paras <- new_paras
        }
      }
    }
  }

  return(list(current_mllh, current_paras))
}


#' @title Log-likelihood function with subvector of fixed parameters.
#'
#' @description
#' This function is a wrapper around \code{log_lik_fun} implementing the case
#' where a part of the parameter vector is fixed and hence excluded from
#' optimization.
#'
#' @param para_variable Parameters to be estimated.
#' @param para_fixed Fixed parameters.
#' @param fixed_indices Indices of fixed parameters.
#' @param lag_degs Maximal degrees of Laguerre polynomials.
#' @param cop_name Name of considered copula.
#' @param Y Vector of observed times.
#' @param X Design matrix.
#' @param Delta Vector of censoring indicators.
#' @param hp List of hyperparameters.
#'
#' @returns Log-likelihood function evaluation.
#'
#' @seealso log_lik_fun
#'
#' @noRd
#'
log_lik_partial <- function(para_variable, para_fixed, fixed_indices,
                            lag_degs, cop_name, Y, X, Delta, hp) {
  all_paras <- rep(0, length(para_variable) + length(para_fixed))
  all_paras[fixed_indices] <- para_fixed
  all_paras[-fixed_indices] <- para_variable
  return(log_lik_fun(all_paras, lag_degs, cop_name, Y, X, Delta, hp))
}

#' @title Negative of partial log-likelihood function with subvector of fixed
#' parameters.
#'
#' @description
#' This function is a wrapper around \code{log_lik_partial} implementing the case
#' where a part of the parameter vector is fixed and hence excluded from
#' optimization.
#'
#' @inheritParams log_lik_partial
#' @returns Negative log-likelihood function evaluation.
#' @seealso log_lik_partial
#'
#' @noRd
#'
neg_log_lik_partial <- function(para_variable, para_fixed, fixed_indices,
                                lag_degs, cop_name, Y, X, Delta, hp) {
  -log_lik_partial(para_variable, para_fixed, fixed_indices, lag_degs, cop_name,
                   Y, X, Delta, hp)
}

#' @title Continuity constraint of EAL distribution (some fixed parameters).
#'
#' @description
#' This function implements the equality constraint (6.1) of D'Haen et al.
#' (2025) in the case where some parameters are fixed rather than estimated.
#'
#' @param para_variable Parameters to be estimated.
#' @param para_fixed Fixed parameters.
#' @param fixed_indices Indices of fixed parameters.
#' @param lag_degs Maximal degrees of Laguerre polynomials.
#' @param cop_name Name of considered copula.
#' @param Y Vector of observed times.
#' @param X Design matrix.
#' @param Delta Vector of censoring indicators.
#' @param hp List of hyperparameters.
#'
#' @returns Vector of continuity constraint violations.
#'
#' @seealso ctuity_constr_fun
#'
#' @noRd
#'
ctuity_constr_partial <- function(para_variable, para_fixed, fixed_indices,
                                  lag_degs, cop_name, Y, X, Delta, hp) {

  # Extract hyperparameters
  phi_til_index <- hp[["phi_til_index"]]
  phi_index <- hp[["phi_index"]]

  if (cop_name == "indep"){
    ktau_length <- 0
  } else {
    ktau_length <- 1
  }

  all_paras <- rep(0, length(para_variable) + length(para_fixed))
  all_paras[fixed_indices] <- para_fixed
  all_paras[-fixed_indices] <- para_variable

  phis <- all_paras[c(phi_til_index(lag_degs), phi_index(lag_degs))
                    + ktau_length]

  if (lag_degs[1] == 0){
    phi_til_ext <- c(1)
    phi_ext <- c(1, phis)
  } else {
    if (lag_degs[2] == 0){
      phi_til_ext <- c(1, phis)
      phi_ext <- c(1)
    } else {
      phi_til_ext <- c(1, phis[1:lag_degs[1]])
      phi_ext <- c(1, phis[(lag_degs[1]+1):(lag_degs[1] + lag_degs[2])])
    }
  }

  # combination of Lag. polynomials evaluated in 0 on neg./pos. axis:
  left <- sum(phi_til_ext)^2 / sqnorm(phi_til_ext)
  right <- sum(phi_ext)^2 / sqnorm(phi_ext)

  jump <- left - right

  return(c(jump, - jump))
}

#' @title Optimize the log-likelihood function.
#'
#' @description
#' This function optimizes the likelihood function using the specified settings.
#' In particular, it implements the the (Nelder-Mead + COBYLA)-hybrid algorithm
#' as described in D'Haen et al. (2025).
#'
#' @param lag_degs Laguerre polynomial maximal degrees.
#' @param cop_name Name of considered copula.
#' @param Y Vector of observed times.
#' @param X Design matrix.
#' @param Delta Vector of censoring indicators.
#' @param para_fixed Fixed parameters.
#' @param fixed_indices Indices of fixed parameters.
#' @param init_values_variable Initial values for variable parameters.
#' @param opt_algo Algorithm to be used for optimization. Could be one of
#' \code{"NM"} or \code{"NMCob"}, corresponding to Nelder-Mead and the
#' (Nelder-Mead + COBYLA)-hybrid, respectively.
#' @param hp List of hyperparameters.
#' @param indep_assumption Boolean indicator whether independence can be
#' assumed. Default is \code{indep_assumption = FALSE}.
#'
#' @returns List of program value and argument minimizers (i.e. ML estimates).
#'
#' @noRd
#'
perform_mll_partial_optimisation <- function(lag_degs, cop_name, Y, X, Delta,
                                             para_fixed, fixed_indices,
                                             init_values_variable,
                                             opt_algo, hp,
                                             indep_assumption = FALSE) {

  it_num_constr <- hp[["it_NMCob_constr"]]
  it_num_unconstr <- hp[["it_NMCob_unconstr"]]

  current_mllh <- -100000
  current_para_variable <- init_values_variable[[1]]

  if (opt_algo == "NM") {
    if (indep_assumption) {
      current_para_variable <- current_para_variable[-1]
      for (init_vals in init_values_variable){
        res <- optim(init_vals[-1], log_lik_indep_partial, method = "Nelder-Mead",
                     control = list(fnscale = -1, maxit = it_num_unconstr),
                     para_fixed = para_fixed, fixed_indices = fixed_indices,
                     lag_degs = lag_degs, cop_name = cop_name,
                     Y = Y, X = X, Delta = Delta, hp = hp)
        new_para_variable <- res[[1]]
        new_mllh <- res[[2]]

        if (new_mllh > current_mllh){
          current_mllh <- new_mllh
          current_para_variable <- new_para_variable
        }
      }
    } else {
      for (init_vals in init_values_variable){
        res <- optim(init_vals, log_lik_partial, method = "Nelder-Mead",
                     control = list(fnscale = -1, maxit = it_num_unconstr),
                     para_fixed = para_fixed, fixed_indices = fixed_indices,
                     lag_degs = lag_degs, cop_name = cop_name,
                     Y = Y, X = X, Delta = Delta, hp)
        new_para_variable <- res[[1]]
        new_mllh <- res[[2]]

        if (new_mllh > current_mllh){
          current_mllh <- new_mllh
          current_para_variable <- new_para_variable
        }
      }
    }
  }

  if (opt_algo == "NMCob"){

    opts <- list("algorithm"= "NLOPT_LN_COBYLA", "maxeval"= it_num_constr)

    if (indep_assumption) {
      current_para_variable <- current_para_variable[-1]
      for (init_vals in init_values_variable){
        intermed_para_variable <- optim(init_vals[-1], log_lik_indep_partial,
                                        method = "Nelder-Mead",
                                        control = list(fnscale = -1,
                                                       maxit = it_num_unconstr),
                                        para_fixed = para_fixed,
                                        fixed_indices = fixed_indices,
                                        lag_degs = lag_degs,
                                        cop_name = cop_name,
                                        Y = Y, X = X, Delta = Delta, hp = hp)$par

        res <- nloptr(x0 = intermed_para_variable,
                      eval_f = neg_log_lik_indep_partial, # max. rather than min.!
                      eval_g_ineq = ctuity_constr_partial,
                      opts = opts,
                      para_fixed = para_fixed, fixed_indices = fixed_indices,
                      lag_degs = lag_degs, cop_name = cop_name,
                      Y = Y, X = X, Delta = Delta, hp = hp)
        new_para_variable <- res$solution
        new_mllh <- - res$objective # Compare with llh rather than neg_llh

        if (new_mllh > current_mllh){
          current_mllh <- new_mllh
          current_para_variable <- new_para_variable
        }
      }
    } else {
      for (init_vals in init_values_variable){
        intermed_para_variable <- optim(init_vals, log_lik_partial,
                                        method = "Nelder-Mead",
                                        control = list(fnscale = -1,
                                                       maxit = it_num_unconstr),
                                        para_fixed = para_fixed,
                                        fixed_indices = fixed_indices,
                                        lag_degs = lag_degs,
                                        cop_name = cop_name,
                                        Y = Y, X = X, Delta = Delta, hp = hp)$par

        res <- nloptr(x0 = intermed_para_variable,
                      eval_f = neg_log_lik_partial, # max. rather than min.!
                      eval_g_ineq = ctuity_constr_partial,
                      opts = opts,
                      para_fixed = para_fixed, fixed_indices = fixed_indices,
                      lag_degs = lag_degs, cop_name = cop_name,
                      Y = Y, X = X, Delta = Delta, hp = hp)
        new_para_variable <- res$solution
        new_mllh <- - res$objective # Compare with llh rather than neg_llh

        if (new_mllh > current_mllh){
          current_mllh <- new_mllh
          current_para_variable <- new_para_variable
        }
      }
    }
  }

  # Final step: combine results in one vector
  para_all <- rep(0, length(current_para_variable) + length(para_fixed))
  para_all[fixed_indices] <- para_fixed
  para_all[-fixed_indices] <- current_para_variable

  return(list(current_mllh, para_all))
}



#' @title Log-likelihood under independence and with a fixed subvector.
#'
#' @description
#' This function defines the likelihood under the independence assumption and
#' with some elements of the covariate vector being fixed. Specifically, note
#' that \code{para_variable} nor \code{para_fixed} has an entry for \eqn{\eta},
#' the dependence coefficient.
#'
#' @param para_variable Parameters to be estimated.
#' @param para_fixed Fixed parameters.
#' @param fixed_indices Indices of fixed parameters.
#' @param lag_degs Maximal degrees of Laguerre polynomials.
#' @param cop_name Name of considered copula.
#' @param Y Vector of observed times.
#' @param X Design matrix.
#' @param Delta Vector of censoring indicators.
#' @param hp List of hyperparameters.
#'
#' @noRd
#'
log_lik_indep_partial <- function(para_variable, para_fixed, fixed_indices,
                                  lag_degs, cop_name, Y, X, Delta, hp) {
  all_paras <- rep(0, length(para_variable) + length(para_fixed))
  all_paras[fixed_indices] <- para_fixed
  all_paras[-fixed_indices] <- para_variable
  return(log_lik_fun_indep(all_paras, lag_degs, cop_name, Y, X, Delta, hp))
}

#' @title Negative log-likelihood under independence and with a fixed subvector.
#'
#' @description
#' This function defines the negative of the log-likelihood under the
#' independence assumption and with some elements of the covariate vector being
#' fixed. This function is simply a wrapper around \code{log_lik_indep_partial}.
#'
#' @inheritParams log_lik_indep_partial
#' @seealso [log_lik_indep_partial()]
#'
#' @noRd
#'
neg_log_lik_indep_partial <- function(para_variable, para_fixed, fixed_indices,
                                      lag_degs, cop_name, Y, X, Delta, hp) {
  -log_lik_indep_partial(para_variable, para_fixed, fixed_indices, lag_degs,
                         cop_name, Y, X, Delta, hp)
}


#' @title Log-likelihood function under independence.
#'
#' @description
#' This function defines the log-likelihood for all parameters under the
#' independence assumption. Specifically, note that \code{para_variable} nor
#' \code{para_fixed} has an entry for \eqn{\eta}, the dependence coefficient.
#'
#' @param all_paras_indep Vector of parameters (without an entry for \eqn{\eta}
#' in the first position).
#' @param lag_degs Maximal degrees of Laguerre polynomials.
#' @param cop_name Name of considered copula.
#' @param Y Vector of observed times.
#' @param X Design matrix.
#' @param Delta Vector of censoring indicators.
#' @param hp List of hyperparameters.
#'
#' @noRd
#'
log_lik_fun_indep <- function(all_paras_indep, lag_degs, cop_name, Y, X, Delta,
                              hp) {

  #### 0) Extract hyperparameters ####

  beta_full_index <- hp[["beta_full_index"]]
  gamma_index <- hp[["gamma_index"]]
  lambda_index <- hp[["lambda_index"]]
  phi_til_index <- hp[["phi_til_index"]]
  phi_index <- hp[["phi_index"]]
  lambda <- hp[["lambda"]]
  homoscedastic <- hp[["homoscedastic"]]
  heteroscedastic <- !homoscedastic

  #### 1) Retrieving current parameters ####

  paras_T_indices <- c(beta_full_index, gamma_index, lambda_index,
                       phi_til_index(lag_degs), phi_index(lag_degs))

  # no eta present, so no shift + 1
  transf_paras_T <- all_paras_indep[paras_T_indices]
  transf_paras_C <- all_paras_indep[-paras_T_indices]

  beta_full <- transf_paras_T[beta_full_index]
  if (heteroscedastic) {
    transf_lam <- transf_paras_T[lambda_index]
    lambda <- lambda_inverse_transform(transf_lam)
  }
  gam_par <- transf_paras_T[gamma_index]
  phi_til <- transf_paras_T[phi_til_index(lag_degs)]
  phi <- transf_paras_T[phi_index(lag_degs)]

  #### 2) Construct argument matrices for censored & uncensored cases ####

  Y_un <- Y[Delta == 1]
  Y_cens <- Y[Delta == 0]

  X_un <- X[Delta == 1, , drop = FALSE]
  X_cens <- X[Delta == 0, , drop = FALSE]

  #### 3) Density & distribution evaluation ####

  # density
  dens_T_un <- dEAL(Y_un, X_un, beta_full, gam_par, lambda, phi_til, phi, hp)
  dens_C_cens <- dCens(Y_cens, X_cens, transf_paras_C)

  # distribution
  distr_T_cens <- pEAL(Y_cens, X_cens, beta_full, gam_par, lambda, phi_til, phi)
  distr_C_un <- pCens(Y_un, X_un, transf_paras_C)

  #### 4) Compute and return log-likelihood ####

  lh_uncens <- dens_T_un * (1 - distr_C_un)
  lh_cens <- dens_C_cens * (1 - distr_T_cens)

  llh <- sum(log(lh_uncens)) + sum(log(lh_cens))

  res <- min(max(llh, -100000), 100000) # avoid returning +- Inf
  if (is.nan(res)){ # avoid NaN (due to Inf - Inf by num. errors) as well
    res <- -100000
  }
  return(res)
}

#' @title Negative log-likelihood function under independence.
#'
#' @description
#' This function defines the negative log-likelihood for all parameters under
#' the independence assumption. Specifically, note that \code{para_variable} nor
#' \code{para_fixed} has an entry for \eqn{\eta}, the dependence coefficient.
#' This function is simply a wrapper around \code{log_lik_fun_indep}.
#'
#' @inheritParams log_lik_fun_indep
#' @seealso [neg_log_lik_fun_indep()]
#'
#' @noRd
#'
neg_log_lik_fun_indep <- function(all_paras_indep, lag_degs, cop_name, Y, X,
                                  Delta, hp) {
  -log_lik_fun_indep(all_paras_indep, lag_degs, cop_name, Y, X, Delta, hp)
}

#' @title Continuity constraint of EAL distribution under independence.
#'
#' @description
#' Functionality is similar to \code{ctuity_constr_fun}.
#'
#' @param all_paras_indep Parameters to be estimated.
#' @param lag_degs Maximal degrees of Laguerre polynomials.
#' @param cop_name Name of considered copula.
#' @param Y Vector of observed times.
#' @param X Design matrix.
#' @param Delta Vector of censoring indicators.
#' @param hp List of hyperparameters.
#'
#' @returns Vector of continuity constraint violations.
#'
#' @seealso ctuity_constr_fun
#'
#' @noRd
#'
ctuity_constr_fun_indep <- function(all_paras_indep, lag_degs, cop_name, Y, X,
                                    Delta, hp) {

  # Extract hyperparameters
  phi_til_index <- hp[["phi_til_index"]]
  phi_index <- hp[["phi_index"]]

  phi_til <- all_paras_indep[phi_til_index(lag_degs)]
  phi <- all_paras_indep[phi_index(lag_degs)]

  phi_til_ext <- c(1, phi_til)
  phi_ext <- c(1, phi)

  # combination of Lag. polynomials evaluated in 0 on neg./pos. axis:
  left <- sum(phi_til_ext)^2 / sqnorm(phi_til_ext)
  right <- sum(phi_ext)^2 / sqnorm(phi_ext)

  jump <- left - right

  return(c(jump, - jump))
}


#' @title Degree selection based on AIC.
#'
#' @description
#' This function uses AIC to select the degree of the Laguerre polynomials in
#' the enriched Laplace distribution.
#'
#' @param Y Vector of observed times.
#' @param X matrix of observed covariate values.
#' @param T1 Subvector of observed times corresponding to \eqn{Delta = 1}, i.e.
#' vector of observed event times.
#' @param X1 Subvector of covariates corresponding to \eqn{Delta = 1}.
#' @param C0 Subvector of observed times corresponding to \eqn{Delta = 0}, i.e.
#' vector of observed censoring times.
#' @param X0 Subvector of covariates corresponding to \eqn{Delta = 1}.
#' @param Delta Vector of censoring indicators.
#' @param cop_name Name of the copula to use.
#' @param hp List of hyperparameters.
#'
#' @noRd
#'
deg_sel_AIC_grid_only_T <- function(cop_name, Y, X, Delta,
                                    T1, X1, C0, X0, hp,
                                    indep_assumption = FALSE) {

  #### 0) Extract hyperparameters ####

  sample_size <- hp[["sample_size"]]
  max_deg <- hp[["lag_deg_upper"]]
  init_value_number_intermed <- hp[["init_value_number_intermed"]]
  init_value_number_basis <- hp[["init_value_number_basis"]]
  lower_bound_on_phis <- hp[["lower_bound_on_phis"]]
  upper_bound_on_phis <- hp[["upper_bound_on_phis"]]
  it_NM <- hp[["it_NM"]]
  beta_full_index <- hp[["beta_full_index"]]
  gamma_index <- hp[["gamma_index"]]
  lambda_index <- hp[["lambda_index"]]
  ktau_length <- 1 - as.numeric(indep_assumption)

  #### 1) first perform optimization under zero degrees (using Nelder-Mead) ####

  init_val_list <- determine_init_basis(T1, X1, C0, X0, Delta, cop_name, hp)
  mll_basis <- perform_mll_optimisation(c(0, 0), cop_name,
                                        Y, X, Delta,
                                        init_val_list = init_val_list,
                                        opt_algo = "NM", hp,
                                        indep_assumption = indep_assumption)
  para_basis <- mll_basis[[2]]

  # Benchmark: AIC/BIC score for mll optimisation under degrees (0,0)
  basis_para_num <- length(para_basis)
  BIC_factor <- log(sample_size)
  basis_AIC <- 2*basis_para_num - 2* mll_basis[[1]]
  basis_BIC <- BIC_factor*basis_para_num - 2* mll_basis[[1]]

  T_indices <- ktau_length + c(beta_full_index, gamma_index, lambda_index)
  para_T_basis <- para_basis[T_indices]
  para_fixed_basis <- para_basis[-T_indices]

  # AIC/BIC degree selection
  all_AIC_scores <- matrix(rep(0, (max_deg + 1)^2),
                           nrow = (max_deg + 1), byrow = TRUE)
  all_AIC_scores[1,1] <- basis_AIC
  best_AIC_degrees <- c(0,0)
  best_AIC_score <- basis_AIC
  best_AIC_transf_paras <- para_basis

  all_BIC_scores <- matrix(rep(0, (max_deg + 1)^2),
                           nrow = (max_deg + 1), byrow = TRUE)
  all_BIC_scores[1,1] <- basis_BIC
  best_BIC_degrees <- c(0,0)
  best_BIC_score <- basis_BIC
  best_BIC_transf_paras <- para_basis

  #### 2) Then perform optimisation under range of degrees (grid approach) ####

  for (L_degree in (seq_len(max_deg + 1) - 1)) {
    for (R_degree in (seq_len(max_deg + 1) - 1)) {

      if (L_degree + R_degree > 0) {

        # Determine initial values
        total_degree <- L_degree + R_degree

        # fix everything except the indices for T (new phi's also not fixed)
        fixed_indices <- seq_len(
          length(para_basis) + total_degree
        )[-c(T_indices, max(T_indices) + seq_len(total_degree))]

        init_val_variable <- vector(mode = "list",
                                    length = init_value_number_intermed)
        for (j in 1:init_value_number_intermed){
          random_phis <- runif(total_degree,
                               lower_bound_on_phis, upper_bound_on_phis)

          if (ktau_length == 1) {
            init_val_variable[[j]] <- c(
              slight_perturb(para_T_basis),
              random_phis)
          } else {
            init_val_variable[[j]] <- c(0, slight_perturb(para_T_basis),
                                        random_phis)
          }
        }

        new_mll <- perform_mll_partial_optimisation(
          lag_degs = c(L_degree, R_degree),
          cop_name = cop_name,
          Y = Y, X = X, Delta = Delta,
          para_fixed = para_fixed_basis,
          fixed_indices = fixed_indices,
          init_values_variable = init_val_variable,
          opt_algo = "NMCob",
          hp = hp,
          indep_assumption = indep_assumption)

        new_AIC <- 2*(basis_para_num + total_degree) - 2* new_mll[[1]]
        new_BIC <- BIC_factor*(basis_para_num + total_degree) - 2* new_mll[[1]]

        # Storing result in grid
        all_AIC_scores[L_degree + 1, R_degree + 1] <- new_AIC
        all_BIC_scores[L_degree + 1, R_degree + 1] <- new_BIC

        # Update minimum
        if (new_AIC < best_AIC_score){
          best_AIC_degrees <- c(L_degree, R_degree)
          best_AIC_score <- new_AIC
          best_AIC_transf_paras <- new_mll[[2]]
        }
        if (new_BIC < best_BIC_score){
          best_BIC_degrees <- c(L_degree, R_degree)
          best_BIC_score <- new_BIC
          best_BIC_transf_paras <- new_mll[[2]]
        }
      }
    }
  }

  #### 3) Return the results ####

  return(list(best_AIC_degrees, best_AIC_transf_paras))
}


#' @title Validate input to the function \code{QRdepCens}.
#'
#' @description
#' This function checks a set of preconditions that are required for the
#' function \code{QRdepCens} to run successfully.
#'
#' @param data Data frame.
#' @param hp List of hyperparameters. See the documentation of \code{QRdepCens}
#' for an exhaustive summary of options that can be specified.
#'
#' @noRd
#'
QRdepCens_validateInput <- function(data, hp) {

  #### Checks related to the data ####

  # Initialize vector of column indices which have not been checked.
  cols.to.check <- 1:ncol(data)

  # The data should contain a column pertaining to the observed times, named 'Y'
  if ("Y" %notin% colnames(data)) {
    stop("The column pertaining to the observed times should be named 'Y'.")
  }
  cols.to.check <- setdiff(cols.to.check, which(colnames(data) == "Y"))

  # The data should contain a column pertaining to the censoring indicator,
  # named Delta. Values in this column can only be 0 or 1.
  if ("Delta" %notin% colnames(data)) {
    stop("The column pertaining to the censoring indicator should be named 'Delta'.")
  } else if (any(data$Delta %notin% c(0, 1))) {
    stop("The censoring indicator should be either 0 or 1.")
  }
  cols.to.check <- setdiff(cols.to.check, which(colnames(data) == "Delta"))

  # The data should contain at least one covariate. Columns pertaining to
  # covariates should be named X1, X2, etc.
  if ("X1" %notin% colnames(data)) {
    stop("The data should contain at least one covariate, which should be named 'X1'.")
  }
  cols.to.check <- setdiff(cols.to.check, which(grepl("X[1-9][[:digit:]]*$", colnames(data))))

  # The data should not already contain an intercept column
  if (any(apply(extract.covariates(data), 2, function(col) {all(col == 1)}))) {
    stop("The data should not already contain an intercept column.")
  }

  # The data should not contain any other columns
  if (length(cols.to.check) != 0) {
    stop("Unrecognized columns detected. See documentation of 'QRdepCens'.")
  }

  #### Checks related to the hyperparameters ####

  # The copula to be used should be specified, and be one of the available
  # options.
  if (is.null(hp[["test_cop_name"]])) {
    stop("The copula to be used should be specified.")
  } else if (hp[["test_cop_name"]] %notin% c("frank", "frankPos", "gumbel", "clayton", "indep")) {
    stop("Specified copula not implemented. See documentation of QRdepCens.")
  }

  # The maximal degree of the Laguerre polynomials should specified, and be a
  # positive integer.
  if (is.null(hp[["lag_deg_upper"]])) {
    stop("The maximum degree of Laguerre polynomials should be specified.")
  } else if (hp[["lag_deg_upper"]] <= 0) {
    stop("Maximum degree of Laguerre polynomial should be positive.")
  } else if (!is_integer(hp[["lag_deg_upper"]])) {
    stop("Maximum degree of Laguerre polynomial should be integer.")
  }

  # The maximum number of contraint optimization iterations should be specified
  # and positive.
  if (is.null(hp[["it_NMCob_constr"]])) {
    stop("The maximum number of contraint optimization iterations should be specified.")
  } else if (hp[["it_NMCob_constr"]] <= 0) {
    stop("The maximum number of contraint optimization iterations should be positive.")
  }

  # The maximum number of uncontraint optimization iterations should be
  # specified and positive.
  if (is.null(hp[["it_NMCob_unconstr"]])) {
    stop("The maximum number of uncontraint optimization iterations should be specified.")
  } else if (hp[["it_NMCob_unconstr"]] <= 0) {
    stop("The maximum number of uncontraint optimization iterations should be positive.")
  }

  # The value indicating if homoscedasticity can be assumed should be boolean.
  if (!is.logical(hp[["homoscedastic"]])) {
    stop("'homoscedastic' should be logical.")
  }

  # Check whether the quantile of interest/quantile index has been correctly
  # specified (depending on whether homoscedasticity is assumed)
  if (hp[["homoscedastic"]]) {
    if (!is.null(hp[["lambda_index"]])) {
      stop("'lambda_index' should not be specified for a homoscedastic model.")
    }
    if (is.null(hp[["lambda"]])) {
      stop("quantile of interest 'lambda' should be specified in homoscedastic model.")
    } else if (!is.numeric(hp[["lambda"]])) {
      stop("'lambda' should be numeric.")
    } else if (hp[["lambda"]] < 0 | hp[["lambda"]] > 1) {
      stop("'lambda' should be a value between 0 and 1.")
    }
  } else {
    if (is.null(hp[["lambda_index"]])) {
      stop("'lambda_index' should be specified for heteroscedastic model.")
    }
    if (!is.null(hp[["lambda"]])) {
      stop("'lambda' should not be specified for heteroscedastic model.")
    }
  }

  # Number of variance bootstrap iterations should be a positive integer.
  if (!is.null(hp[["variance.bootstrap.iterations"]])) {
    if (!is_integer(hp[["variance.bootstrap.iterations"]])) {
      stop("The specified number of variance bootstrap iterations should be integer.")
    } else if (hp[["variance.bootstrap.iterations"]] <= 1) {
      stop("The specified number of variance bootstrap iterations should be strictly larger than one.")
    }
  }

  # ToDo: init_value_number_basis, variance.b

}


#' @title Obtain default settings for hyperparameters.
#'
#' @description
#' This function returns the default settings for the hyperparameters used in
#' running the copula-based quantile regression method.
#'
#' @param data Data frame.
#' @param homoscedastic Boolean value indicating whether homoscedasticity can
#' be assumed for the model. Default is \code{homoscedastic = FALSE}.
#'
#' @returns A list containing the elements:
#' \describe{
#'  \item{homoscedastic}{Boolean value indicating whether homoscedasticity can
#'  be assumed for the model.}
#'  \item{lambda}{Quantile of interest.}
#'  \item{test_cop_name}{Name of copula to be used.}
#'  \item{lag_deg_upper}{Upper bound on Laguerre polynomial degree to be
#'  considered.}
#'  \item{it_NMCob_constr, it_NMCob_unconstr, it_NM:}{Maximum number of
#'  optimization cycles to run during runtime.}
#'  \item{init_value_number_basis, init_value_number_intermed,
#'  init_value_number_final:}{Number of initial values to be considered in
#'  optimization routines.}
#'  \item{lower_bound_on_phis, upper_bound_on_phis}{Lower and upper bound on
#'  the initial values for the Laguerre polynomial coefficients.}
#'  \item{variance.bootstrap.iterations:}{Number of bootstrap iterations during
#'  variance estimation.}
#'  \item{sample_size:}{The sample size.}
#'  \item{X_dimension:}{The number of covariates in the model.}
#'  \item{beta_full_index:}{Vector of indices in the full parameter vector
#'  containing the elements of beta.}
#'  \item{gamma_index:}{Vector of indices in the full parameter vector
#'  containing the element(s) (of) gamma.}
#'  \item{lambda_index:}{Index in the full parameter vector for quantile of
#'  interest.}
#'  \item{phi_til_index, phi_index:}{Index in the full parameter vector for
#'  Laguerre polynomial coefficients.}
#' }
#'
#' @noRd
#'
QRdepCens_getDefaultHyperparameters <- function(data, homoscedastic = FALSE) {

  # Set default hyperparameters
  test_cop_name <- "frank"
  lag_deg_upper <- 2
  it_NMCob_constr <- 100
  it_NMCob_unconstr <- 100
  it_NM <- 100
  init_value_number_basis <- 5
  init_value_number_intermed <- 5
  init_value_number_final <- 5
  lower_bound_on_phis <- -2
  upper_bound_on_phis <- 2
  variance.bootstrap.iterations <- 50

  # Define the positions of elements of subvectors of the parameter vector that
  # pertain to certain parts of the model for T.
  sample_size <- nrow(data)
  X_dimension <- ncol(extract.covariates(data))
  beta_full_index <- 1:(X_dimension + 1)

  # Depending on whether or not homoscedasticity is assumed, set the indices
  # of the parameters in the full parameter vector.
  if (homoscedastic) {
    gamma_index <- X_dimension + 2
    lambda <- 0.5
    lambda_index <- NULL
    phi_til_index <- function(lag_degs) {
      if (lag_degs[1] == 0){
        return(c())
      } else {
        return((X_dimension + 3):(X_dimension + 2 + lag_degs[1]))
      }
    }
    phi_index <- function(lag_degs) {
      if (lag_degs[2] == 0){
        return(c())
      } else {
        return((X_dimension + 3 + lag_degs[1]):
                 (X_dimension + 2 + lag_degs[1] + lag_degs[2]))
      }
    }
  } else {
    gamma_index <- (X_dimension + 2) : (2*X_dimension + 2)
    lambda <- NULL
    lambda_index <- c(2*X_dimension + 3)
    phi_til_index <- function(lag_degs) {
      if (lag_degs[1] == 0){
        return(c())
      } else {
        return((2*X_dimension + 4):(2*X_dimension + 3 + lag_degs[1]))
      }
    }
    phi_index <- function(lag_degs) {
      if (lag_degs[2] == 0){
        return(c())
      } else {
        return((2*X_dimension + 4 + lag_degs[1]):
                 (2*X_dimension + 3 + lag_degs[1] + lag_degs[2]))
      }
    }
  }

  # Return the result
  list(homoscedastic = homoscedastic,
       lambda = lambda,
       test_cop_name = test_cop_name,
       lag_deg_upper = lag_deg_upper,
       it_NMCob_constr = it_NMCob_constr,
       it_NMCob_unconstr = it_NMCob_unconstr,
       it_NM = it_NM,
       init_value_number_basis = init_value_number_basis,
       init_value_number_intermed = init_value_number_intermed,
       init_value_number_final = init_value_number_final,
       lower_bound_on_phis = lower_bound_on_phis,
       upper_bound_on_phis = upper_bound_on_phis,
       variance.bootstrap.iterations = variance.bootstrap.iterations,
       sample_size = sample_size,
       X_dimension = X_dimension,
       beta_full_index = beta_full_index,
       gamma_index = gamma_index,
       lambda_index = lambda_index,
       phi_til_index = phi_til_index,
       phi_index = phi_index)
}

#' @title Print the results of the QRdepCens model.
#'
#' @description
#' This function prints the results of the QRdepCens model to the console.
#'
#' @param results List containing the estimated model parameters, as returned by
#' \code{QRdepCens}.
#'
#' @noRd
#'
QRdepCens_printSummary <- function(results) {

  #### Unpack results ####

  est_beta_full <- results[["est_beta"]]
  est_gam <- results[["est_gamma"]]
  est_lambda <- results[["est_lambda"]]
  cov_mat <- results[["cov_mat"]]

  # Parse variance information
  if (!is.null(cov_mat)) {
    vars <- diag(cov_mat)
    b1 <- 1
    b2 <- length(est_beta_full)
    g1 <- b2 + 1
    g2 <- b2 + length(est_gam)
    l1 <- g2 + 1
    var_beta_full <- vars[b1:b2]
    var_gam <- vars[g1:g2]
    var_lambda <- vars[l1]
  } else {
    var_beta_full <- NA
    var_gam <- NA
    var_lambda <- NA
  }

  #### Arrange all results in a data frame ####

  # Results for beta.
  df.beta <- data.frame(
    "coef" = sprintf("beta%d", 0:(length(est_beta_full) - 1)),
    "est"  = est_beta_full,
    "s.d." = sqrt(var_beta_full),
    "conf. int." = make_ci(est_beta_full, sqrt(var_beta_full))
  )

  # Results for gamma
  df.gam <- data.frame(
    "coef" = sprintf("gamma%d", 0:(length(est_gam) - 1)),
    "est"  = est_gam,
    "s.d." = sqrt(var_gam),
    "conf. int." = make_ci(est_gam, sqrt(var_gam))
  )

  # Results for lambda.
  df.lambda <- data.frame(
    "coef" = "lambda",
    "est"  = est_lambda,
    "s.d." = sqrt(var_lambda),
    "conf. int." = make_ci(est_lambda, sqrt(var_lambda))
  )

  # Concatenate all data frames and convert all columns to string values
  df.res <- rbind(df.beta, df.gam, df.lambda)
  df.res[, 2] <- sprintf("%.3f", df.res[, 2])
  df.res[, 3] <- sprintf("%.3f", df.res[, 3])

  #### Print the results ####

  # Add header
  df.res <- rbind(c("coef", "est", "s.d.", "conf. int."), df.res)

  # For each column, determine the maximal string length
  max.str.lengths <- apply(df.res, 2, function(col) {max(nchar(col))})

  # Set each element of the results data frame to have this column-wise
  # maximal string length.
  for (col.idx in 1:ncol(df.res)) {
    max.length <- max.str.lengths[col.idx]
    sprintf.arg <- paste0("%", max.length, "s")
    df.res[, col.idx] <- sprintf(sprintf.arg, df.res[, col.idx])
  }

  # Define sequence of lines to print
  df.res.str <- apply(df.res, 1, function(row) {sprintf("%s  %s  %s  %s", row[1], row[2], row[3], row[4])})

  # Print the summary
  for (line in df.res.str) {
    message(line)
  }
}

#' @title Estimate the model of D'Haen et al. (2025).
#'
#' @description
#' This function estimates the parameters in the model of D'Haen et al. (2025).
#'
#' @param data Data on which the model should be estimated. Note that the data
#' should be structured in a specific form: The observed times should be put in
#' a column named \code{"Y"}, the censoring indicators in a column named
#' \code{"Delta"}, and the covariates in columns named \code{"X1"}, \code{"X2"},
#' ... (in increasing order). The given data set cannot contain any other
#' columns.
#' @param hp List of hyperparameters to be used, the elements of which will
#' overwrite the default settings. In particular, consider changing:
#' \describe{
#'  \item{test_cop_name:}{Copula to be used. Should be one of
#'  \code{"frank"}, \code{"gumbel"}, \code{"clayton"} or \code{"indep"}.}
#'  \item{homoscedastic:}{Boolean flag indicating whether
#'  homoscedasticity can be assumed.}
#'  \item{variance.bootstrap.iterations:}{Number of bootstrap resamples
#'  to use during variance estimation. Consider increase if more precision is
#'  needed; consider decreasing to reduce computation time.}
#' }
#' Other hyperparameters can be changed though it is not recommended. We refer
#' to the source code for the available options.
#' @param var.estimate Boolean value indicating whether the variance should be
#' estimated (via bootstrap). This can take a considerable amount of time.
#' Default is \code{var.estimate = FALSE}.
#' @param verbose Verbosity flag (boolean) indicating whether the results should
#' be printed to the console. Default is \code{verbose = TRUE}.
#'
#' @note The variance estimation procedure could easily be paralelized.
#' However, this is currently not implemented.
#'
#' @examples
#' \donttest{
#'
#'  # Load the data
#'  data(liver)
#'
#'  # Give standard column names (required!)
#'  colnames(liver) <- c("patient", "Y", "Delta", "X1", "X2", "X3", "X4")
#'  liver <- liver[, c(-1, -6, -7)]
#'
#'  # Run the model
#'  hp <- list(
#'    homoscedastic = FALSE,
#'    test_cop_name = "frank"
#'  )
#'  QRdepCens(liver, hp, var.estimate = FALSE)
#'  # Takes a while if var.estimate = TRUE...
#' }
#'
#' @references D'Haen, M., Van Keilegom, I. and Verhasselt, A. (2025). Quantile
#' regression under dependent censoring with unknown association. Lifetime Data
#' Analysis 31(2):253-299.
#'
#' @export
#'
QRdepCens <- function(data, hp, var.estimate = FALSE, verbose = TRUE) {

  #### Load dependencies ####

  lapply(QRdepCens.getDependencies(), require, character.only = TRUE)

  #### Set hyperparameters ####

  homoscedastic <- if ("homoscedastic" %in% names(hp)) hp$homoscedastic else FALSE
  user.hp <- hp
  hp <- QRdepCens_getDefaultHyperparameters(data, homoscedastic)
  hp[names(user.hp)] <- user.hp

  #### Input validation ####

  QRdepCens_validateInput(data, hp)

  #### Parse data and hyperparameters, and define settings ####

  # Extract different parts of the data
  X <- cbind(1, as.matrix(extract.covariates(data)))
  Y <- data$Y
  Delta <- data$Delta

  # Subvector of observed survival times
  T1 = Y[Delta == 1]
  X1 = X[Delta == 1, , drop = FALSE]

  # Subvector of observed censoring times
  C0 = Y[Delta == 0]
  X0 = X[Delta == 0, , drop = FALSE]

  # Hyperparameters
  cop_name <- hp[["test_cop_name"]]
  lag_deg_upper <- hp[["lag_deg_upper"]]
  it_NMCob_constr <- hp[["it_NMCob_constr"]]
  it_NMCob_unconstr <- hp[["it_NMCob_unconstr"]]
  X_dimension <- hp[["X_dimension"]]

  # Dependence parameter
  if (cop_name == "indep"){
    ktau_length_test <- 0
    indep_assumption <- TRUE
  } else {
    ktau_length_test <- 1
    indep_assumption <- FALSE
  }

  #### Estimate the model ####

  # Determine the optimal Laguerre degrees using AIC
  intermed_res <- deg_sel_AIC_grid_only_T(cop_name = cop_name,
                                          Y = Y, X = X, Delta = Delta,
                                          T1, X1, C0, X0, hp,
                                          indep_assumption = indep_assumption)
  lag_degrees <- intermed_res[[1]]
  current_paras <- intermed_res[[2]]

  # Set maximum likelihood optimization starting values
  final_init_values <- determine_init_final(current_paras, hp, indep_assumption)

  # Estimate the model
  mll_res <- perform_mll_optimisation(lag_degrees, cop_name,
                                      Y, X, Delta,
                                      init_val_list = final_init_values,
                                      opt_algo = "NMCob",
                                      hp = hp,
                                      indep_assumption = indep_assumption)

  # Extract the results
  est_mll <-  mll_res[[1]]
  est_transf_para_vec <- mll_res[[2]]
  est_para_vec <- para_inv_transform(est_transf_para_vec,
                                     cop_name = cop_name,
                                     hp = hp)

  #### Parse the results of model estimation ####

  # Define the positions of elements of subvectors of the parameter vector that
  # pertain to certain parts of the model for T.
  # Define functions that return the indices corresponding to the Laguerre
  # polynomial coefficients in the parameter vector.
  beta_full_index <- hp[["beta_full_index"]]
  gamma_index <- hp[["gamma_index"]]
  lambda_index <- hp[["lambda_index"]]
  phi_til_index <- hp[["phi_til_index"]]
  phi_index <- hp[["phi_index"]]

  # Indices in the full coefficient vector estimate corresponding to the model
  # for T. Note that ktau_length = 0 for when assuming independence between T
  # and C (in which case the dependence parameter was not estimated), and
  # ktau_length = 1 otherwise.
  paras_T_indices <- c(beta_full_index, gamma_index, lambda_index,
                       phi_til_index(lag_degrees),
                       phi_index(lag_degrees)) + ktau_length_test
  est_paras_T <- est_para_vec[paras_T_indices]

  # Extract subvectors of parameters corresponding to different 'parts' of the
  # model for T.
  est_beta_full <- est_paras_T[beta_full_index]
  est_gam_par <- est_paras_T[gamma_index]
  est_lambda <- est_paras_T[lambda_index]
  est_phi_til <- est_paras_T[phi_til_index(lag_degrees)]
  est_phi <- est_paras_T[phi_index(lag_degrees)]

  #### Estimate the variance ####

  # Initialize a NULL value for the covariance matrix.
  cov.mat <- NULL

  if (var.estimate) {

    # Extract number of bootstrap iterations
    B <- hp[["variance.bootstrap.iterations"]]

    # If this number was not specified, throw an error
    if (is.null(B)) {
      stop("Number of variance bootstrap iterations should be specified.")
    }

    # Initialize matrix to store the results
    boot.mat <- NULL

    # Obtain bootstrap estimates of the parameters
    for (b in seq_len(B)) {

      # Construct bootstrapped data set
      data.b <- data[sample(1:nrow(data), nrow(data), replace = TRUE), ]

      # Estimate the parameters
      res.b <- QRdepCens(data.b, hp, var.estimate = FALSE, verbose = FALSE)
      est.b <- c(res.b$est_beta, res.b$est_gamma, res.b$est_lambda)

      # Store the results
      boot.mat <- rbind(boot.mat, est.b)
    }

    # Compute bootstrap covariance matrix
    cov.mat <- var(boot.mat)
  }

  #### Print summary and return the results ####

  # List of results
  rtrn <- list(
    etimate_full = est_para_vec,
    estimate_T_full = est_paras_T,
    est_beta = est_beta_full,
    est_gamma = est_gam_par,
    est_lambda = est_lambda,
    est_phi_til = est_phi_til,
    est_phi = est_phi,
    cov_mat = cov.mat
  )

  # If the homoscedastic option was selected, add the selected quantile of
  # interest to this vector (in place of where its estimate would otherwise go).
  if (hp$homoscedastic) {
    rtrn[["est_lambda"]] <- hp$lambda
    if (var.estimate) {
      rtrn[["cov_mat"]]
    }
  }

  # Print a summary to the console
  if (verbose) {
    QRdepCens_printSummary(rtrn)
  }

  # Return the results (invisibly)
  return(invisible(rtrn))

}

#' @title Estimate quantile based on QRdepCens model.
#'
#' @description
#' This function estimates a specified quantile conditionally on the given
#' covariate values and estimated model parameters. We refer to equation (3.10)
#' in D'haen et al. (2024) for details.
#'
#' @param p Quantile to be estimated.
#' @param x Vector of covariate values at which to estimate the specified
#' quantiles (should not contain an intercept entry).
#' @param res List containing the model parameters, as returned by the function
#' \code{QRdepCens}.
#'
#' @note For the moment, this function is not made visible to the user.
#'
#' @noRd
#'
QRdepCens_estimateQuantile <- function(p, x, res) {

  # Extract the covariates pertaining to T from the model estimates
  beta <- res[["est_beta"]]
  gamma <- res[["est_gamma"]]
  lambda <- res[["est_lambda"]]
  phi_til <- res[["phi_til"]]
  phi <- res[["phi"]]

  # Add intercept to x
  x <- c(1, x)

  # Compute conditional variance of T
  if (length(gamma) == 1) {
    sd_T <- exp(gamma)
  } else {
    sd_T <- exp(x %*% gamma)
  }

  # Compute quantile of error term
  q.error <- qEAL(p, x = c(1), beta = c(0), gam_par = 0, lambda = lambda,
                  phi_til = phi_til, phi = phi)

  # Compute and return quantile estimate
  x %*% beta + sd_T * q.error
}

#' @title Standardize data format
#'
#' @description Checks the required preconditions of the data and possibly
#' restructures the data.
#'
#' @param data A data frame that should contain columns named \code{Y} and
#' \code{delta} (unless \code{comp.risks = TRUE}, see later).
#' @param admin Boolean value indicating whether the provided data frame contains
#' administrative (i.e. independent) censoring on top of the dependent censoring
#' (in the column named \code{delta}). The default is \code{admin = FALSE}.
#' @param conf Boolean value indicating whether the provided data frame contains
#' a confounded variable and a corresponding instrument. If \code{cond = TRUE},
#' the provided data frame should contain columns named \code{Z} and \code{W},
#' corresponding to the confounded variable and instrument, respectively.
#' Moreover, \code{Zbin} and \code{Wbin} should be specified. The default value
#' is \code{conf = FALSE}.
#' @param comp.risks Boolean value indicating whether the provided data frame
#' contains competing risks. If \code{comp.risks = TRUE}, the given data frame
#' should contain the columns \code{delta1}, \code{delta2}, etc., corresponding
#' to the indicators I(Y = T1), I(Y = T2), etc. respectively. The default is
#' \code{comp.risks = FALSE}.
#' @param Zbin Boolean or integer value (0, 1) indicating whether the confounded
#' variable is binary. \code{Zbin = TRUE} or \code{Zbin = 1} means that Z is
#' binary. \code{Zbin = FALSE} or \code{Zbin = 0} means that Z is continuous.
#' @param Wbin Boolean or integer value (0, 1) indicating whether the instrument
#' is binary. \code{Wbin = TRUE} or \code{Wbin = 1} means that W is binary.
#' \code{Wbin = FALSE} or \code{Wbin = 0} means that W is continuous.
#'
#' @return Returns the uniformized data set.
#'
#' @noRd
#'


uniformize.data <- function(data, admin = FALSE, conf = FALSE, comp.risks = FALSE,
                            Zbin = NULL, Wbin = NULL) {

  # ToDo?: Option to turn off precondition checking .

  #### Step 1: Precondition checks ####

  # Put all column names to lower case
  colnames(data) <- tolower(colnames(data))

  # Given data object should be a data frame
  if (!is.data.frame(data)) {
    stop("The data should be provided as a data frame")
  }

  ## Data cannot already contain an intercept

  if (any(apply(data == 1, MARGIN = 2, all))) {
    stop("The given data frame cannot already contain an intercept")
  }

  ## Necessary column names

  # The time variable
  if (!("y" %in% colnames(data))) {
    stop("The column corresponding to the time variable should be named 'Y'.")
  }

  # Censoring indicators
  if (comp.risks) {
    if ("delta" %in% colnames(data)) {
      stop(paste0("Column named 'delta' detected in data frame. Since comp.risks",
                  " = TRUE, columns containing censoring indicators should be ",
                  "named 'deltak', where k = 1, 2, ..."))
    }
    if ("xi" %in% colnames(data)) {
      stop(paste0("Column named 'xi' detected in data frame. Since comp.risks",
                  " = TRUE, naming conventions for censoring indicator columns ",
                  "are different."))
    }
    if (!("delta1" %in% colnames(data))) {
      stop(paste0("The column corresponding to the indicator I(Y = T1) should ",
                  "be named 'delta1' and analogous naming should be used for ",
                  "the column corresponding to I(Y = Ti). No column 'delta1' ",
                  "detected."))
    }
    if (admin & !("da" %in% colnames(data))) {
      stop(paste0("The column corresponding to the indicator I(Y = A) should be ",
                  "named 'da'. If there is no administrative censoring in the ",
                  "data (and hence this column doesn't exist), please set ",
                  "admin = FALSE."))
    }
  } else {
    if (!("delta" %in% colnames(data))) {
      stop("The column corresponding to the indicator I(Y = T) should be named 'delta'.")
    }
    if (admin & !("xi" %in% colnames(data))) {
      stop("The column corresponding to indicator I(Y = C) should be named 'xi'.")
    }
    if (any(grepl('delta[[:digit:]]+', colnames(data)))) {
      stop(paste0("Column(s) named 'deltak' detected for some integer(s) k. ",
                  "Since comp.risks = FALSE, naming conventions for censoring ",
                  "indicators are different."))
    }
    if ("da" %in% colnames(data)) {
      stop(paste0("Column named 'da' detected. ",
                  "Since comp.risks = FALSE, naming conventions for censoring ",
                  "indicators are different."))
    }
  }

  if (conf & !all(c("z", "w") %in% colnames(data))) {
    stop(paste0("The column corresponding to the confounded variable should be",
                " named 'Z' and the column corresponding to the instrument",
                " should be named 'W'."))
  }

  # Indices of columns corresponding to covariates
  if (comp.risks) {
    nbr.comp.risks <- length(which(substr(colnames(data), 1, 5) == "delta"))
    cov.col.idxs <- setdiff(1:ncol(data), which(colnames(data) %in% c("y", "delta", "xi", paste0("delta", 1:nbr.comp.risks), "da", "z", "w")))
  } else {
    cov.col.idxs <- setdiff(1:ncol(data), which(colnames(data) %in% c("y", "delta", "xi", "z", "w")))
  }


  ## Consistent arguments
  if (!is.numeric(data[, "y"])) {
    stop("Column corresponding to 'Y' should be numeric")
  }
  for (cens.col.idx in which(substr(colnames(data), 1, 5) == "delta")) {
    if (!all(data[, cens.col.idx] %in% c(0, 1))) {
      stop(paste0("Censoring indicator in column ", cens.col.idx, " can only ",
                  "take values 0 or 1."))
    }
  }
  if (!all(sapply(data, class)[cov.col.idxs] %in% c("numeric", "integer"))) {
    stop("The covariates should be numeric")
  }

  if (conf) {
    if (any(is.null(Zbin), is.null(Wbin))) {
      stop("When conf == TRUE, the arguments Zbin and Wbin should be provided.")
    }
    if (!((is.numeric(Zbin) | is.logical(Zbin)) & (is.numeric(Wbin) | is.logical(Wbin)))) {
      stop("Zbin and Wbin should be logical or binary")
    }
    if (is.numeric(Zbin) & !(Zbin %in% c(0, 1))) {
      stop("Invalid value for Zbin. Should be either 0 or 1.")
    }
    if (is.numeric(Wbin) & !(Wbin %in% c(0, 1))) {
      stop("Invalid value for Wbin Should be either 0 or 1.")
    }
    if (!all(is.numeric(data[, "z"]), is.numeric(data[, "w"]))) {
      stop("Confounded variable and/or instrument should be numeric")
    }
    if (Zbin & !all(data[, "z"] %in% c(0, 1))) {
      stop("Confounded variable is not a binary as indicated")
    }
    if (Wbin & !all(data[, "w"] %in% c(0, 1))) {
      stop("Instrumental variable is not a binary as indicated")
    }
  } else {
    if (!is.null(c(Zbin, Wbin))) {
      warning("Argument(s) Zbin and/or Wbin ignored as conf == FALSE")
      Zbin <- NULL
      Wbin <- NULL
    }
  }

  if (comp.risks) {
    cens.col.idxs <- which(colnames(data) %in% paste0("delta", 1:nbr.comp.risks))
    if (admin) {
      cens.col.idxs <- c(cens.col.idxs, which(colnames(data) == "da"))
      if (!all(data[, "da"] %in% c(0, 1))) {
        stop("Administrative censoring indicator can only take values 0 or 1.")
      }
    }
    if (!all(apply(X = data[, cens.col.idxs], MARGIN = 1, FUN = sum) == 1)) {
      stop("Precisely one censoring indicator should be 1 for each observation.")
    }
  } else {
    if (admin) {
      if (!all(data[, "xi"] %in% c(0, 1))) {
        stop("Administrative censoring indicator can only take values 0 or 1.")
      }
      if (!all((data[, "delta"] + data[, "xi"]) %in% c(0, 1))) {
        stop("Observations cannot be both administratively and dependently censored at once")
      }
    }
  }

  # Warn for possible user mistakes
  if ((!conf) & any(c("z", "w") %in% colnames(data))) {
    warning(paste0("Possible column(s) corresponding to a confounded variable",
                   " and/or instrument detected in data but conf == FALSE. ",
                   "Treated as exogenous covariates."))
  }
  if ((!conf) & any(c(!is.null(Zbin), !is.null(Wbin)))) {
    warning(paste0("Arguments Zbin and/or Wbin ignored since conf == FALSE"))
  }
  if (conf & !is.null(Zbin)) {
    if (!Zbin) {
      if (all(data[, "z"] %in% c(0, 1))) {
        warning("Z specified as numeric but appears to be binary")
      }
    }
  }
  if (conf & !is.null(Wbin)) {
    if (!Wbin) {
      if (all(data[, "w"] %in% c(0, 1))) {
        warning("W specified as numeric but appears to be binary")
      }
    }
  }
  if (comp.risks) {
    if ((!admin) & ("da" %in% colnames(data))) {
      warning(paste0("Possible administrative censoring column detected in data",
                     " but admin == FALSE so not taken into account"))
    }
  } else {
    if ((!admin) & ("xi" %in% colnames(data))) {
      warning(paste0("Possible administrative censoring column detected in data",
                     " but admin == FALSE so not taken into account"))
    }
  }

  # Missing values
  if (any(is.na(data))) {
    stop("Missing values detected")
  }

  #### Step 2: Create uniformized data frame ####

  # Add intercept column to the data

  data[, "intercept"] <- rep(1, nrow(data))

  if (comp.risks) {

    # Column index of the observed time
    unif.col.idxs <- which(colnames(data) == "y")

    # Column indices of the censoring indicators.
    ind.col.idx <- which(substr(colnames(data), 1, 5) == "delta")

    # Append column indices of censoring indicators to main vector in the order
    # delta1, delta2, delta3, etc.
    unif.col.idxs <- c(unif.col.idxs, ind.col.idx[order(colnames(data)[ind.col.idx])])

    # If administrative censoring: append the column index of indicator I(Y = A)
    # to the main vector.
    if (admin) {
      unif.col.idxs <- c(unif.col.idxs, which(colnames(data) == "da"))
    }

    # Append intercept column to the data
    unif.col.idxs <- c(unif.col.idxs, ncol(data))

    # Append exogenous covariate column indices to main vector.
    unif.col.idxs <- c(unif.col.idxs, cov.col.idxs)

    # If confounding: append the endogenous covariate Z and instrument column W
    # indices to main vector.
    if (conf) {
      unif.col.idxs <- c(unif.col.idxs,  which(colnames(data) == "z"),  which(colnames(data) == "w"))
    }

    # Reorder the data set in specified order.
    unif.data <- data[, unif.col.idxs]
    colnames(unif.data) <- c("y",
                             sort(colnames(data)[ind.col.idx]),
                             if (admin) "da" else NULL,
                             paste0("x", 0:length(cov.col.idxs)),
                             if (conf) "z" else NULL,
                             if (conf) "w" else NULL)

  } else {

    # Column indices of observed time Y and indicator I(Y = T)
    unif.col.idxs <- c(which(colnames(data) == "y"), which(colnames(data) == "delta"))

    # If administrative censoring: append the column index of indicator I(Y = A)
    # to the main vector.
    if (admin) {
      unif.col.idxs <- c(unif.col.idxs, which(colnames(data) == "xi"))
    }

    # Append intercept column to the data
    unif.col.idxs <- c(unif.col.idxs, ncol(data))

    # Append exogenous covariate column indices to main vector.
    unif.col.idxs <- c(unif.col.idxs, cov.col.idxs)

    # If confounding: append the endogenous covariate Z and instrument column W
    # indices to main vector.
    if (conf) {
      unif.col.idxs <- c(unif.col.idxs,  which(colnames(data) == "z"),  which(colnames(data) == "w"))
    }

    # Reorder the data set in specified order.
    unif.data <- data[, unif.col.idxs]
    colnames(unif.data) <- c("y",
                             "delta",
                             if (admin) "xi" else NULL,
                             paste0("x", 0:length(cov.col.idxs)),
                             if (conf) "z" else NULL,
                             if (conf) "z" else NULL)
  }

  return(unif.data)

}

#' @title Log-likelihood function for the independence copula.
#' @description This likelihood function is maximized to estimate the model parameters under the independence copula.
#' @param para Estimated parameter values/initial values.
#' @param Y Follow-up time.
#' @param Delta Censoring indicator.
#' @param Dist.T The distribution to  be used for the survival time T. This argument can take one of the values from \code{c("lnorm", "weibull", "llogis")}.
#' @param Dist.C The distribution to  be used for the censoring time C. This argument can take one of the values from \code{c("lnorm", "weibull", "llogis")}.
#' @import rvinecopulib rafalib stats
#' @return Maximized log-likelihood value.
#' @noRd
loglike.indep.unconstrained<-function(para,Y,Delta,Dist.T,Dist.C){
  par.T=para[1:2]
  par.C=para[3:4]
  n=length(Y)

  if(Dist.T=="lnorm"){
    FT<- plnorm(Y,meanlog=par.T[1],sdlog=exp(par.T[2]))
    fT<- (1/(sqrt(2*pi)*exp(par.T[2])*Y))*exp(-((log(Y)-par.T[1])^2/(2*(exp(par.T[2]))^2)))
  } else if(Dist.T=="weibull"){
    FT<-1-exp(-(Y/exp(par.T[2]))^(exp(par.T[1])))
    fT<-(exp(par.T[1])/exp(par.T[2]))*((Y/exp(par.T[2]))^(exp(par.T[1])-1))*(exp(-(Y/exp(par.T[2]))^(exp(par.T[1]))))
  } else if(Dist.T=="llogis"){
    FT<-1/(1+exp(-(log(Y)-par.T[1])/exp(par.T[2])))
    fT<-(1/(Y*exp(par.T[2])))*exp((log(Y)-par.T[1])/exp(par.T[2]))*(1+exp((log(Y)-par.T[1])/exp(par.T[2])))^(-2)
  }

  if(Dist.C=="lnorm"){
    FC<-plnorm(Y,meanlog=par.C[1],sdlog=exp(par.C[2]))
    fC<-(1/(sqrt(2*pi)*exp(par.C[2])*Y))*exp(-((log(Y)-par.C[1])^2/(2*(exp(par.C[2]))^2)))
  } else if(Dist.C=="weibull"){
    FC<-1-exp(-(Y/exp(par.C[2]))^(exp(par.C[1])))
    fC<-(exp(par.C[1])/exp(par.C[2]))*((Y/exp(par.C[2]))^(exp(par.C[1])-1))*(exp(-(Y/exp(par.C[2]))^(exp(par.C[1]))))
  } else if(Dist.C=="llogis"){
    FC<-1/(1+exp(-(log(Y)-par.C[1])/exp(par.C[2])))
    fC<-(1/(Y*exp(par.C[2])))*exp((log(Y)-par.C[1])/exp(par.C[2]))*(1+exp((log(Y)-par.C[1])/exp(par.C[2])))^(-2)
  }

  l1<-sum(log(pmax(fT[Delta==1],1e-100))+log(pmax(1-FC[Delta==1],1e-100)))
  l0<-sum(log(pmax(fC[Delta==0],1e-100))+log(pmax(1-FT[Delta==0],1e-100)))
  l1+l0
}

#' @title Log-likelihood function for the Frank copula.
#' @description This likelihood function is maximized to estimate the model parameters under the Frank copula.
#' @param para Estimated parameter values/initial values.
#' @param Y Follow-up time.
#' @param Delta Censoring indicator.
#' @param Dist.T The distribution to  be used for the survival time T. This argument can take one of the values from \code{c("lnorm", "weibull", "llogis")}.
#' @param Dist.C The distribution to  be used for the censoring time C. This argument can take one of the values from \code{c("lnorm", "weibull", "llogis")}.
#' @import rvinecopulib rafalib stats
#' @return Maximized log-likelihood value.
#' @noRd
loglike.frank.unconstrained<-function(para,Y,Delta,Dist.T,Dist.C){
  par.T=para[1:2]
  par.C=para[3:4]
  par.z=para[5]
  par.tau<-tanh(par.z)
  par.cop<-sign(par.tau) * ktau_to_par("frank", max(1e-20, abs(par.tau)))
  n=length(Y)

  if(Dist.T=="lnorm"){
    FT<-plnorm(Y,meanlog=par.T[1],sdlog=exp(par.T[2]))
    fT<- (1/(sqrt(2*pi)*exp(par.T[2])*Y))*exp(-((log(Y)-par.T[1])^2/(2*(exp(par.T[2]))^2)))
  } else if(Dist.T=="weibull"){
    FT<-1-exp(-(Y/exp(par.T[2]))^(exp(par.T[1])))
    fT<-(exp(par.T[1])/exp(par.T[2]))*((Y/exp(par.T[2]))^(exp(par.T[1])-1))*(exp(-(Y/exp(par.T[2]))^(exp(par.T[1]))))
  } else if(Dist.T=="llogis"){
    FT<-1/(1+exp(-(log(Y)-par.T[1])/exp(par.T[2])))
    fT<-(1/(Y*exp(par.T[2])))*exp((log(Y)-par.T[1])/exp(par.T[2]))*(1+exp((log(Y)-par.T[1])/exp(par.T[2])))^(-2)
  }

  if(Dist.C=="lnorm"){
    FC<-plnorm(Y,meanlog=par.C[1],sdlog=exp(par.C[2]))
    fC<-(1/(sqrt(2*pi)*exp(par.C[2])*Y))*exp(-((log(Y)-par.C[1])^2/(2*(exp(par.C[2]))^2)))
  } else if(Dist.C=="weibull"){
    FC<-1-exp(-(Y/exp(par.C[2]))^(exp(par.C[1])))
    fC<-(exp(par.C[1])/exp(par.C[2]))*((Y/exp(par.C[2]))^(exp(par.C[1])-1))*(exp(-(Y/exp(par.C[2]))^(exp(par.C[1]))))
  } else if(Dist.C=="llogis"){
    FC<-1/(1+exp(-(log(Y)-par.C[1])/exp(par.C[2])))
    fC<-(1/(Y*exp(par.C[2])))*exp((log(Y)-par.C[1])/exp(par.C[2]))*(1+exp((log(Y)-par.C[1])/exp(par.C[2])))^(-2)
  }

  frank_cop<-bicop_dist("frank",0,par.cop)
  hC.condT<-hbicop(cbind(FT,FC), 1,frank_cop)
  hT.condC<-hbicop(cbind(FT,FC), 2,frank_cop)
  l1<-sum(log(pmax(fT[Delta==1],1e-100))+log(pmax(1-hC.condT[Delta==1],1e-100)))
  l0<-sum(log(pmax(fC[Delta==0],1e-100))+log(pmax(1-hT.condC[Delta==0],1e-100)))
  l1+l0
}

#' @title Log-likelihood function for the Gumbel copula.
#' @description This likelihood function is maximized to estimate the model parameters under the Gumbel copula.
#' @param para Estimated parameter values/initial values.
#' @param Y Follow-up time.
#' @param Delta Censoring indicator.
#' @param Dist.T The distribution to  be used for the survival time T. This argument can take one of the values from \code{c("lnorm", "weibull")} and has to be the same as Dist.C.
#' @param Dist.C The distribution to  be used for the censoring time C. This argument can take one of the values from \code{c("lnorm", "weibull")} and has to be the same as Dist.T.
#' @import rvinecopulib rafalib stats
#' @return Maximized log-likelihood value.
#' @noRd
loglike.gumbel.unconstrained<-function(para,Y,Delta,Dist.T,Dist.C){
  par.T=para[1:2]
  par.C=para[3:4]
  par.z=min(c(para[5],700))
  par.tau<-exp(par.z)/(1+exp(par.z))
  par.cop<-sign(par.tau) * ktau_to_par("gumbel", max(1e-9, abs(par.tau)))
  par.cop<-min(50,par.cop)
  par.cop<-max(1,par.cop)
  n=length(Y)

  if(Dist.T=="lnorm"){
    FT<-plnorm(Y,meanlog=par.T[1],sdlog=exp(par.T[2]))
    fT<- (1/(sqrt(2*pi)*exp(par.T[2])*Y))*exp(-((log(Y)-par.T[1])^2/(2*(exp(par.T[2]))^2)))
  } else if(Dist.T=="weibull"){
    FT<-1-exp(-(Y/exp(par.T[2]))^(exp(par.T[1])))
    fT<-(exp(par.T[1])/exp(par.T[2]))*((Y/exp(par.T[2]))^(exp(par.T[1])-1))*(exp(-(Y/exp(par.T[2]))^(exp(par.T[1]))))
  }

  if(Dist.C=="lnorm"){
    FC<-plnorm(Y,meanlog=par.C[1],sdlog=exp(par.C[2]))
    fC<-(1/(sqrt(2*pi)*exp(par.C[2])*Y))*exp(-((log(Y)-par.C[1])^2/(2*(exp(par.C[2]))^2)))
  } else if(Dist.C=="weibull"){
    FC<-1-exp(-(Y/exp(par.C[2]))^(exp(par.C[1])))
    fC<-(exp(par.C[1])/exp(par.C[2]))*((Y/exp(par.C[2]))^(exp(par.C[1])-1))*(exp(-(Y/exp(par.C[2]))^(exp(par.C[1]))))
  }

  gumbel_cop<-bicop_dist("gumbel",0,par.cop)
  hC.condT<-hbicop(cbind(FT,FC), 1,gumbel_cop)
  hT.condC<-hbicop(cbind(FT,FC), 2,gumbel_cop)
  l1<-sum(log(pmax(fT[Delta==1],1e-100))+log(pmax(1-hC.condT[Delta==1],1e-100)))
  l0<-sum(log(pmax(fC[Delta==0],1e-100))+log(pmax(1-hT.condC[Delta==0],1e-100)))
  l1+l0
}

#' @title Log-likelihood function for the Clayton copula.
#' @description This likelihood function is maximized to estimate the model parameters under the Clayton copula.
#' @param para Estimated parameter values/initial values.
#' @param Y Follow-up time.
#' @param Delta Censoring indicator.
#' @param Dist.T The distribution to  be used for the survival time T. This argument can take one of the values from \code{c("lnorm", "weibull")} and has to be the same as Dist.C.
#' @param Dist.C The distribution to  be used for the censoring time C. This argument can take one of the values from \code{c("lnorm", "weibull")} and has to be the same as Dist.T.
#' @import rvinecopulib rafalib stats
#' @return Maximized log-likelihood value.
#' @noRd
loglike.clayton.unconstrained<-function(para,Y,Delta,Dist.T,Dist.C){
  par.T=para[1:2]
  par.C=para[3:4]
  par.z=min(c(para[5],700))
  par.tau<-exp(par.z)/(1+exp(par.z))
  par.cop<-sign(par.tau) * ktau_to_par("clayton", max(1e-9, abs(par.tau)))
  par.cop<-min(28,par.cop)
  par.cop<-max(1e-100,par.cop)
  n=length(Y)

  if(Dist.T=="lnorm"){
    FT<-plnorm(Y,meanlog=par.T[1],sdlog=exp(par.T[2]))
    fT<- (1/(sqrt(2*pi)*exp(par.T[2])*Y))*exp(-((log(Y)-par.T[1])^2/(2*(exp(par.T[2]))^2)))
  } else if(Dist.T=="weibull"){
    FT<-1-exp(-(Y/exp(par.T[2]))^(exp(par.T[1])))
    fT<-(exp(par.T[1])/exp(par.T[2]))*((Y/exp(par.T[2]))^(exp(par.T[1])-1))*(exp(-(Y/exp(par.T[2]))^(exp(par.T[1]))))
  }

  if(Dist.C=="lnorm"){
    FC<-plnorm(Y,meanlog=par.C[1],sdlog=exp(par.C[2]))
    fC<-(1/(sqrt(2*pi)*exp(par.C[2])*Y))*exp(-((log(Y)-par.C[1])^2/(2*(exp(par.C[2]))^2)))
  } else if(Dist.C=="weibull"){
    FC<-1-exp(-(Y/exp(par.C[2]))^(exp(par.C[1])))
    fC<-(exp(par.C[1])/exp(par.C[2]))*((Y/exp(par.C[2]))^(exp(par.C[1])-1))*(exp(-(Y/exp(par.C[2]))^(exp(par.C[1]))))
  }

  clayton_cop<-bicop_dist("clayton",0,par.cop)
  hC.condT<-hbicop(cbind(FT,FC), 1,clayton_cop)
  hT.condC<-hbicop(cbind(FT,FC), 2,clayton_cop)
  l1<-sum(log(pmax(fT[Delta==1],1e-100))+log(pmax(1-hC.condT[Delta==1],1e-100)))
  l0<-sum(log(pmax(fC[Delta==0],1e-100))+log(pmax(1-hT.condC[Delta==0],1e-100)))
  l1+l0
}

#' @title Log-likelihood function for the Gaussian copula.
#' @description This likelihood function is maximized to estimate the model parameters under the Gaussian copula.
#' @param para Estimated parameter values/initial values.
#' @param Y Follow-up time.
#' @param Delta Censoring indicator.
#' @param Dist.T The distribution to  be used for the survival time T. This argument can only the value \code{"lnorm"}.
#' @param Dist.C The distribution to  be used for the censoring time C. This argument can only the value \code{"lnorm"}.
#' @import rvinecopulib rafalib stats
#' @return Maximized log-likelihood value.
#' @noRd
loglike.gaussian.unconstrained<-function(para,Y,Delta,Dist.T,Dist.C){
  par.T=para[1:2]
  par.C=para[3:4]
  par.z=para[5]
  par.tau<-tanh(par.z)
  par.cop<-sign(par.tau) * ktau_to_par("gaussian", max(1e-20, abs(par.tau)))
  n=length(Y)

  if(Dist.T=="lnorm"){
    FT<-plnorm(Y,meanlog=par.T[1],sdlog=exp(par.T[2]))
    fT<- (1/(sqrt(2*pi)*exp(par.T[2])*Y))*exp(-((log(Y)-par.T[1])^2/(2*(exp(par.T[2]))^2)))
  } else if(Dist.T=="weibull"){
    FT<-1-exp(-(Y/exp(par.T[2]))^(exp(par.T[1])))
    fT<-(exp(par.T[1])/exp(par.T[2]))*((Y/exp(par.T[2]))^(exp(par.T[1])-1))*(exp(-(Y/exp(par.T[2]))^(exp(par.T[1]))))
  }

  if(Dist.C=="lnorm"){
    FC<-plnorm(Y,meanlog=par.C[1],sdlog=exp(par.C[2]))
    fC<-(1/(sqrt(2*pi)*exp(par.C[2])*Y))*exp(-((log(Y)-par.C[1])^2/(2*(exp(par.C[2]))^2)))
  } else if(Dist.C=="weibull"){
    FC<-1-exp(-(Y/exp(par.C[2]))^(exp(par.C[1])))
    fC<-(exp(par.C[1])/exp(par.C[2]))*((Y/exp(par.C[2]))^(exp(par.C[1])-1))*(exp(-(Y/exp(par.C[2]))^(exp(par.C[1]))))
  }

  gaussian_cop<-bicop_dist("gaussian",0,par.cop)
  hC.condT<-hbicop(cbind(FT,FC), 1,gaussian_cop)
  hT.condC<-hbicop(cbind(FT,FC), 2,gaussian_cop)
  l1<-sum(log(pmax(fT[Delta==1],1e-100))+log(pmax(1-hC.condT[Delta==1],1e-100)))
  l0<-sum(log(pmax(fC[Delta==0],1e-100))+log(pmax(1-hT.condC[Delta==0],1e-100)))
  l1+l0
}

#' @title Fit the dependent censoring models.
#' @description Estimates the model parameters by maximizing the log-likelihood.
#' @param Y Follow-up time.
#' @param Delta Censoring indicator.
#' @param Copula The copula family. This argument can take values from \code{c("frank","gumbel","clayton","gaussian","indep")}.
#' @param Dist.T The distribution to  be used for the survival time T. This argument can take one of the values from \code{c("lnorm", "weibull", "llogis")}.
#' @param Dist.C The distribution to  be used for the censoring time C. This argument can take one of the values from \code{c("lnorm", "weibull", "llogis")}.
#' @param start Starting values
#' @import rvinecopulib rafalib stats
#' @return A list containing the minimized negative log-likelihood using the independence copula model, the estimated parameter values for the model with the independence copula, the minimized negative log-likelihood using the specified copula model and the estimated parameter values for the model with the specified copula.
#' @noRd
optimlikelihood<-function(Y,Delta,Copula,Dist.T,Dist.C,start){

  para.start=optim(start,loglike.indep.unconstrained,Y=Y,Delta=Delta,Dist.T=Dist.T,Dist.C=Dist.C,control=list(fnscale=-1),method="BFGS")

  if(Copula=="frank"){
    result=optim(c(para.start$par,0),loglike.frank.unconstrained,Y=Y,Delta=Delta,Dist.T=Dist.T,Dist.C=Dist.C,control=list(fnscale=-1),method="BFGS")
    return(list(para.start$value,para.start$par,result$value,result$par))
  } else if(Copula=="clayton"){
    result=optim(c(para.start$par,0),loglike.clayton.unconstrained,Y=Y,Delta=Delta,Dist.T=Dist.T,Dist.C=Dist.C,control=list(fnscale=-1),method="BFGS")
    return(list(para.start$value,para.start$par,result$value,result$par))
  } else if(Copula=="gumbel"){
    result=optim(c(para.start$par,0),loglike.gumbel.unconstrained,Y=Y,Delta=Delta,Dist.T=Dist.T,Dist.C=Dist.C,control=list(fnscale=-1),method="BFGS")
    return(list(para.start$value,para.start$par,result$value,result$par))
  } else if(Copula=="gaussian"){
    result=optim(c(para.start$par,0),loglike.gaussian.unconstrained,Y=Y,Delta=Delta,Dist.T=Dist.T,Dist.C=Dist.C,control=list(fnscale=-1),method="BFGS")
    return(list(para.start$value,para.start$par,result$value,result$par))
  } else if(Copula == "indep"){
    return(list(para.start$value,para.start$par))
  }
}

#' @title Estimation of  a parametric dependent censoring model without covariates.
#' @description  Note that it is not assumed that the association parameter of the copula function is known,
#' unlike most other papers in the literature. The details for implementing the methodology can be found in Czado and Van Keilegom (2023).
#' @references Czado and Van Keilegom (2023). Dependent censoring based on parametric copulas. Biometrika, 110(3), 721-738.
#' @param Y Follow-up time.
#' @param Delta Censoring indicator.
#' @param Copula The copula family. This argument can take values from \code{c("frank","gumbel","clayton","gaussian","indep")}.
#' @param Dist.T The distribution to  be used for the survival time T. This argument can take one of the values from \code{c("lnorm", "weibull", "llogis")}.
#' @param Dist.C The distribution to  be used for the censoring time C. This argument can take one of the values from \code{c("lnorm", "weibull", "llogis")}.
#' @param start Starting values.
#' @import rvinecopulib rafalib stats
#' @return A table containing the minimized negative log-likelihood using the independence copula model, the estimated parameter values for the model with the independence copula, the minimized negative log-likelihood using the specified copula model and the estimated parameter values for the model with the specified copula.
#'
#' @examples
#'tau = 0.75
#'Copula = "frank"
#'Dist.T = "weibull"
#'Dist.C = "lnorm"
#'par.T = c(2,1)
#'par.C = c(1,2)
#'n=1000
#'
#'simdata<-TCsim(tau,Copula,Dist.T,Dist.C,par.T,par.C,n)
#'
#'Y = simdata[[1]]
#'Delta = simdata[[2]]
#'
#'ParamCop(Y,Delta,Copula,Dist.T,Dist.C)
#'
#' @export
ParamCop <- function(Y,Delta,Copula,Dist.T,Dist.C,start=c(1,1,1,1)){

  if(!(Copula %in% c("frank","gumbel","clayton","gaussian","indep"))){
    stop("ERROR : Copula needs to be frank, gumbel, clayton, gaussian or indep")
  }else if(!(Dist.T %in% c("lnorm","weibull","llogis"))){
    stop("ERROR : Dist.T needs to be lnorm, weibull or llogis")
  }else if(!(Dist.C %in% c("lnorm","weibull","llogis"))){
    stop("ERROR : Dist.C needs to be lnorm, weibull or llogis")
  }else if( (Copula == "gumbel" | Copula == "clayton" | Copula == "gaussian") & (Dist.T == "llogis" | Dist.C == "llogis")){
    stop("ERROR : log-logistic marginals cannot be specified with the gumbel, clayton or gaussian copula")
  }else if( (Copula == "gumbel" | Copula == "clayton" | Copula == "gaussian") & (Dist.T != Dist.C)){
    stop("ERROR : marginals need to be both lnorm or both weibull")
  } else {

    result = optimlikelihood(Y,Delta,Copula,Dist.T,Dist.C,start)

    if (Copula == "indep"){
      par.mat<-rbind(c(result[[1]],result[[2]]))
      rownames(par.mat)<-c("indep")
      colnames(par.mat)<-c("log-like","E(log(T))","sd(log(T))","E(log(C))","sd(log(C))")
      par.mat[,3]<-exp(par.mat[,3])
      par.mat[,5]<-exp(par.mat[,5])
      if (Dist.T=="weibull"){
        par.mat[,2]<-exp(par.mat[,2])
        colnames(par.mat)[2]<-c("shape(T)")
        colnames(par.mat)[3]<-c("scale(T)")
      }
      if (Dist.C=="weibull"){
        par.mat[,4]<-exp(par.mat[,4])
        colnames(par.mat)[4]<-c("shape(C)")
        colnames(par.mat)[5]<-c("scale(C)")
      }
      if (Dist.T=="llogis"){
        colnames(par.mat)[2]<-c("location(log(T))")
        colnames(par.mat)[3]<-c("scale(log(T))")
      }
      if (Dist.C=="llogis"){
        colnames(par.mat)[4]<-c("location(log(C))")
        colnames(par.mat)[5]<-c("scale(log(C))")
      }

    } else {
      par.mat<-rbind(c(result[[1]],result[[2]],0),c(result[[3]],result[[4]]))

      rownames(par.mat)<-c("indep",Copula)
      colnames(par.mat)<-c("log-like","E(log(T))","sd(log(T))","E(log(C))","sd(log(C))","tau")

      par.mat[,3]<-exp(par.mat[,3])
      par.mat[,5]<-exp(par.mat[,5])

      if (Dist.T=="weibull"){
        par.mat[,2]<-exp(par.mat[,2])
        colnames(par.mat)[2]<-c("shape(T)")
        colnames(par.mat)[3]<-c("scale(T)")
      }
      if (Dist.C=="weibull"){
        par.mat[,4]<-exp(par.mat[,4])
        colnames(par.mat)[4]<-c("shape(C)")
        colnames(par.mat)[5]<-c("scale(C)")
      }

      if (Dist.T=="llogis"){
        colnames(par.mat)[2]<-c("location(log(T))")
        colnames(par.mat)[3]<-c("scale(log(T))")
      }
      if (Dist.C=="llogis"){
        colnames(par.mat)[4]<-c("location(log(C))")
        colnames(par.mat)[5]<-c("scale(log(C))")
      }

      if (Copula=="frank" | Copula =="gaussian"){
        par.mat[2,6]<-tanh(par.mat[2,6])
      }
      if (Copula=="clayton" | Copula =="gumbel"){
        par.mat[2,6]<-exp(par.mat[2,6])/(1+exp(par.mat[2,6]))
      }
    }
    return(round(par.mat,digits=3))
  }
}


#' @title Function to simulate (Y,Delta) from the copula based model for (T,C).
#' @description Generates the follow-up time and censoring indicator according to the specified model.
#' @param tau Value of Kendall's tau for (T,C). The default value is 0.
#' @param Copula The copula family. This argument can take values from \code{c("frank","gumbel","clayton","gaussian","indep")}. The default copula model is "frank".
#' @param Dist.T Distribution of the survival time T. This argument can take one of the values from \code{c("lnorm", "weibull", "llogis")}. The default distribution is "lnorm".
#' @param Dist.C Distribution of the censoring time C. This argument can take one of the values from \code{c("lnorm", "weibull", "llogis")}. The default distribution is "lnorm".
#' @param par.T Parameter values for the distribution of T.
#' @param par.C Parameter values for the distribution of C.
#' @param n Sample size.
#' @return A list containing the generated follow-up times and censoring indicators.
#' @import rvinecopulib rafalib stats
#'
#' @examples
#' tau = 0.5
#' Copula = "gaussian"
#' Dist.T = "lnorm"
#' Dist.C = "lnorm"
#' par.T = c(1,1)
#' par.C = c(2,2)
#' n=1000
#'
#' simdata <- TCsim(tau,Copula,Dist.T,Dist.C,par.T,par.C,n)
#' Y = simdata[[1]]
#' Delta = simdata[[2]]
#' hist(Y)
#' mean(Delta)
#'
#' @export

TCsim<-function(tau=0,Copula="frank",Dist.T="lnorm",Dist.C="lnorm",
                par.T=c(0,1),par.C=c(0,1),n=10000)
{

  if(Copula=="indep"){
    u.TC = cbind(runif(n),runif(n))
  } else {
    parTC<-ktau_to_par(Copula,tau)
    u.TC<-rbicop(n,Copula,0, parTC)
  }

  f.TC<-rep(0,n)
  F.TC<-rep(0,n)
  hC.condT<-rep(0,n)
  hT.condC<-rep(0,n)

  if(Dist.T=="lnorm"){
    xT<-qlnorm(u.TC[,1],meanlog=par.T[1],sdlog=par.T[2])
  }
  else if(Dist.T=="weibull"){
    xT<-qweibull(u.TC[,1],par.T[1],par.T[2])
  }
  else if(Dist.T=="llogis"){
    xT<-exp(qlogis(u.TC[,1],par.T[1],par.T[2]))
  }

  if(Dist.C=="lnorm"){
    xC<-qlnorm(u.TC[,2],meanlog=par.C[1],sdlog=par.C[2])
  }
  else if(Dist.C=="weibull"){
    xC<-qweibull(u.TC[,2],par.C[1],par.C[2])
  }
  else if(Dist.C=="llogis"){
    xC<-exp(qlogis(u.TC[,2],par.C[1],par.C[2]))
  }
  Delta<-as.numeric(xT<=xC)
  Y<-pmin(xT,xC)
  return(list(Y,Delta))
}

test_that("Nonparametric transformation model check", {

  ## Check the fit of dependent censoring model

  #Check summary table
  #expect_message(output$parameterEstimates)

  expect_true(-1<rhohat|rhohat<1)
  expect_equal(l, 5)
  expect_true(inherits(output, "fitNonPar"))
})

test_that("Parametric copula fit", {

  #Check summary table
  #expect_message(fitPar)

  # check the scale and dependency parameter are positive

  expect_true(all(fitPar[,6]>=0))
  expect_equal(rownames(fitPar)[2], "frank")
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CopulaCoxPHmodel_BootstrapMethods.R
\name{summary.depFit}
\alias{summary.depFit}
\title{Summary of \code{depCensoringFit} object}
\usage{
\method{summary}{depFit}(object, ...)
}
\arguments{
\item{object}{Output of \code{\link{fitDepCens}} function}

\item{...}{Further arguments}
}
\value{
Summary of dependent censoring model fit in the form of table
}
\description{
Summary of \code{depCensoringFit} object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ParamTransfoCompRisks.R
\name{estimate.cmprsk}
\alias{estimate.cmprsk}
\title{Estimate the competing risks model of Willems et al. (2025).}
\usage{
estimate.cmprsk(
  data,
  admin,
  conf,
  eoi.indicator.names = NULL,
  Zbin = NULL,
  inst = "cf",
  realV = NULL,
  compute.var = TRUE,
  eps = 0.001
)
}
\arguments{
\item{data}{A data frame, adhering to the following formatting rules:
\itemize{
   \item The first column, named \code{"y"}, contains the observed times.
   \item The next columns, named \code{"delta1"}, \code{delta2}, etc. contain
   the indicators for each of the competing risks.
   \item The next column, named \code{da}, contains the censoring indicator
   (independent censoring).
   \item The next column should be a column of all ones (representing the
   intercept), names \code{x0}.
   \item The subsequent columns should contain the values of the covariates,
   named \code{x1}, \code{x2}, etc.
   \item When applicable, the next column should contain the values of the
   endogenous variable. This column should be named \code{z}.
   \item When \code{z} is provided and an instrument for \code{z} is
   available, the next column, named \code{w}, should contain the values for
   the instrument.
}}

\item{admin}{Boolean value indicating whether the data contains
administrative censoring.}

\item{conf}{Boolean value indicating whether the data contains confounding
and hence indicating the presence of \code{z} and, possibly, \code{w}.}

\item{eoi.indicator.names}{Vector of names of the censoring indicator columns
pertaining to events of interest. Events of interest will be modeled allowing
dependence between them, whereas all censoring events (corresponding to
indicator columns not listed in \code{eoi.indicator.names}) will be treated
as independent of every other event. If \code{eoi.indicator.names == NULL},
all events will be modeled dependently.}

\item{Zbin}{Indicator value indicating whether (\code{Zbin = TRUE}) or not
\code{Zbin = FALSE} the endogenous covariate is binary. Default is
\code{Zbin = NULL}, corresponding to the case when \code{conf == FALSE}.}

\item{inst}{Variable encoding which approach should be used for dealing with
the confounding. \code{inst = "cf"} indicates that the control function
approach should be used. \code{inst = "W"} indicates that the instrumental
variable should be used 'as is'. \code{inst = "None"} indicates that Z will
be treated as an exogenous covariate. Finally, when \code{inst = "oracle"},
this function will access the argument \code{realV} and use it as the values
for the control function. Default is \code{inst = "cf"}.}

\item{realV}{Vector of numerics with length equal to the number of rows in
\code{data}. Used to provide the true values of the instrumental function
to the estimation procedure.}

\item{compute.var}{Boolean value indicating whether the variance of the
parameter estimates should be computed as well (this can be very
computationally intensive, so may want to be disabled). Default is
\code{estimate.var = TRUE}.}

\item{eps}{Value that will be added to the diagonal of the covariance matrix
during estimation in order to ensure strictly positive variances.}
}
\value{
A list of parameter estimates in the second stage of the estimation
algorithm (hence omitting the estimates for the control function), as well
as an estimate of their variance and confidence intervals.
}
\description{
This function estimates the parameters in the competing risks
model described in Willems et al. (2025). Note that this model
extends the model of Crommen, Beyhum and Van Keilegom (2024) and as such, this
function also implements their methodology.
}
\examples{
\donttest{

n <- 200

# Set parameters
gamma <- c(1, 2, 1.5, -1)
theta <- c(0.5, 1.5)
eta1 <- c(1, -1, 2, -1.5, 0.5)
eta2 <- c(0.5, 1, 1, 3, 0)

# Generate exogenous covariates
x0 <- rep(1, n)
x1 <- rnorm(n)
x2 <- rbinom(n, 1, 0.5)

# Generate confounder and instrument
w <- rnorm(n)
V <- rnorm(n, 0, 2)
z <- cbind(x0, x1, x2, w) \%*\% gamma + V
realV <- z - (cbind(x0, x1, x2, w) \%*\% gamma)

# Generate event times
err <- MASS::mvrnorm(n, mu = c(0, 0), Sigma =
matrix(c(3, 1, 1, 2), nrow = 2, byrow = TRUE))
bn <- cbind(x0, x1, x2, z, realV) \%*\% cbind(eta1, eta2) + err
Lambda_T1 <- bn[,1]; Lambda_T2 <- bn[,2]
x.ind = (Lambda_T1>0)
y.ind <- (Lambda_T2>0)
T1 <- rep(0,length(Lambda_T1))
T2 <- rep(0,length(Lambda_T2))
T1[x.ind] = ((theta[1]*Lambda_T1[x.ind]+1)^(1/theta[1])-1)
T1[!x.ind] = 1-(1-(2-theta[1])*Lambda_T1[!x.ind])^(1/(2-theta[1]))
T2[y.ind] = ((theta[2]*Lambda_T2[y.ind]+1)^(1/theta[2])-1)
T2[!y.ind] = 1-(1-(2-theta[2])*Lambda_T2[!y.ind])^(1/(2-theta[2]))
# Generate adminstrative censoring time
C <- runif(n, 0, 40)

# Create observed data set
y <- pmin(T1, T2, C)
delta1 <- as.numeric(T1 == y)
delta2 <- as.numeric(T2 == y)
da <- as.numeric(C == y)
data <- data.frame(cbind(y, delta1, delta2, da, x0, x1, x2, z, w))
colnames(data) <- c("y", "delta1", "delta2", "da", "x0", "x1", "x2", "z", "w")

# Estimate the model
admin <- TRUE                # There is administrative censoring in the data.
conf <- TRUE                 # There is confounding in the data (z)
eoi.indicator.names <- NULL  # We will not impose that T1 and T2 are independent
Zbin <- FALSE                # The confounding variable z is not binary
inst <- "cf"                 # Use the control function approach
compute.var <- TRUE          # Variance of estimates should be computed.
# Since we don't use the oracle estimator, this argument is ignored anyway
realV <- NULL
estimate.cmprsk(data, admin, conf, eoi.indicator.names, Zbin, inst, realV,
                compute.var)

}

}
\references{
Willems et al. (2025). Flexible control function approach under competing risks (submitted).

Crommen, G., Beyhum, J., and Van Keilegom, I. (2024). An instrumental variable approach under dependent censoring. Test, 33(2), 473-495.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FullyParametricCopulas.R
\name{TCsim}
\alias{TCsim}
\title{Function to simulate (Y,Delta) from the copula based model for (T,C).}
\usage{
TCsim(
  tau = 0,
  Copula = "frank",
  Dist.T = "lnorm",
  Dist.C = "lnorm",
  par.T = c(0, 1),
  par.C = c(0, 1),
  n = 10000
)
}
\arguments{
\item{tau}{Value of Kendall's tau for (T,C). The default value is 0.}

\item{Copula}{The copula family. This argument can take values from \code{c("frank","gumbel","clayton","gaussian","indep")}. The default copula model is "frank".}

\item{Dist.T}{Distribution of the survival time T. This argument can take one of the values from \code{c("lnorm", "weibull", "llogis")}. The default distribution is "lnorm".}

\item{Dist.C}{Distribution of the censoring time C. This argument can take one of the values from \code{c("lnorm", "weibull", "llogis")}. The default distribution is "lnorm".}

\item{par.T}{Parameter values for the distribution of T.}

\item{par.C}{Parameter values for the distribution of C.}

\item{n}{Sample size.}
}
\value{
A list containing the generated follow-up times and censoring indicators.
}
\description{
Generates the follow-up time and censoring indicator according to the specified model.
}
\examples{
tau = 0.5
Copula = "gaussian"
Dist.T = "lnorm"
Dist.C = "lnorm"
par.T = c(1,1)
par.C = c(2,2)
n=1000

simdata <- TCsim(tau,Copula,Dist.T,Dist.C,par.T,par.C,n)
Y = simdata[[1]]
Delta = simdata[[2]]
hist(Y)
mean(Delta)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SemiParametricCopulawithoutCovariates.R
\name{SurvDC}
\alias{SurvDC}
\title{Semiparametric Estimation of the Survival Function under Dependent Censoring}
\usage{
SurvDC(
  yobs,
  delta,
  tm = NULL,
  copfam = "frank",
  margins = list(survfam = NULL, censfam = "lnorm"),
  cure = FALSE,
  Var = list(do = TRUE, nboot = 200, level = 0.05),
  control = list(maxit = 300, eps = 1e-06, trace = TRUE, ktau.inits = NULL)
)
}
\arguments{
\item{yobs}{a numeric vector that indicated the observed survival times.}

\item{delta}{a numeric vector that stores the right-censoring indicators.}

\item{tm}{a numeric vector that contains interested non-negative time points at which the survival probabilities will be evluated.
Note that if we omit the definition of this argument (the default value becomes \code{NULL}), our function will automatically output survival probabilities at all oberserved time points, that is, \code{yobs}.}

\item{copfam}{a character string that specifies the copula family.
Currently, it supports Archimedean copula families, including \code{"frank"} (the default value), \code{"clayton"}, \code{"gumbel"}, and \code{"joe"}.
The degenerated independent censoring case can be considered as well by setting \code{"indep"}.
(other options will be added in the near future!)}

\item{margins}{a list used to define the distribution structures of both the survival and censoring margins.
Specifically, it contains the following elements:
\describe{
  \item{\code{survfam}}{ a character string that defines the assumed distribution for the survival time random variable,
    including \code{"lnorm"} for log-normal distribution, \code{"weibull"} for weibull distribution (other options will be added in the near future).}
  \item{\code{censfam}}{ a character string that defines the assumed distribution for the censoring time random variable, and the details are the same as those shown in \code{survfam}.}
  \item{\code{survtrunc}}{ a positive numeric value thats denotes the value of truncation for the assumed distribution, that is, \code{survfam}.}
  \item{\code{censtrunc}}{ a positive numeric value thats denotes the value of truncation for the assumed distribution, that is, \code{censfam}.}
}
Note if one of the marginal distributions should be modeled nonparametrically, one can let the corresponding argument to be \code{NULL} directly.
For example if a semiparametric framework that defines the survival margin to be nonparametric and the censoring margin to be parametric, say log-normal, is desired,
we can let \code{survfam = NULL} and \code{censfam = "lnorm"}, which is indeed the default value.
Furthermore, if no truncation is imposed in \code{survfam} (or \code{censfam}), one can directly omit the specification of \code{survtrunc} (or \code{censtrunc}), which is the default specification.
We also remark here that when a cured fraction is included (\code{cure = TRUE}), if \code{survfam} is not \code{NULL} and \code{survtrunc = NULL}, we will automatically let \code{survtrunc} to be \code{max(yobs)}.
If we wants to model the data with a non-truncated survival distribution when there is a cured fraction, we can set \code{survtrunc = Inf}.}

\item{cure}{a logical value that indicates whether the existence of a cured fraction should be considered.}

\item{Var}{a list that controls the execution of the bootstrap for variance estimation,
and it contains two elements:
\code{do} is a logical value with default \code{FALSE} to tell the function whether the boostrap-based variances should be calculated;
\code{nboot} is a numeric integer that specifies the number of bootstrap samples.}

\item{control}{indicates more detailed control of the underlying model fitting procedures.
It is a list of the following three arguments:
\describe{
  \item{\code{maxit}}{ a positive integer that denotes the maximum iteration number in optimization. The default value is \code{300}.}
  \item{\code{eps}}{ a positive small numeric value that denotes the tolerance for convergence. The default value is \code{1e-6}.}
  \item{\code{trace}}{ a logical value that judges whereh the tracing information on the progress of the model fitting should be produced. The default value if \code{TRUE}.}
  \item{\code{ktau.inits}}{ a numeric vector that contains initial values of the Kendall's tau.
    The default value is \code{NULL}, meaning that a grids of initial values will be automatically generated within our function.}
}}
}
\value{
A list of fitted results is returned.
  Within this outputted list, the following elements can be found:
  \describe{
    \item{\code{probs}}{survival probabilities of the survial margin at \code{tm}.}
    \item{\code{ktau}}{Kendall's tau.}
    \item{\code{parapar}}{estimation of all parameters (except Kendall's tau) contained in the parametric part.}
    \item{\code{GoF}}{goodness-of-test results.}
    \item{\code{curerate}}{cure rate. If \code{cure = FALSE}, it is \code{NULL}.}
  }
}
\description{
Provide semiparametric approaches that can be used to model right-censored survival data under dependent censoring (without covariates).
  The copula-based approach is adopted and there is no need to explicitly specify the association parameter.
  One of the margins can be modeled nonparametrically. As a byproduct, both marginal distributions of survival and
  censoring times can be considered as fully parametric. The existence of a cured fraction concerning survival time
  can also be taken into consideration.

 @references Czado and Van Keilegom (2023). Dependent censoring based on parametric copulas. Biometrika, 110(3), 721-738.
 @references Delhelle and Van Keilegom (2024). Copula based dependent censoring in cure models. TEST (to appear).
 @references Ding and Van Keilegom (2024). Semiparametric estimation of the survival function under dependent censoring (in preparation).
}
\details{
This unified function provide approaches that can be used to model right-censored survival data under dependent censoring (without covariates).
  Various specifications of marginal distributions can be considered by choosing different combinations of the provided arguments.
  Generally speaking, the following two scenarios are what we mainly focused on:
  \describe{
    \item{\code{nonparametric survival margin and parametric censoring margin (without cure)}}{
      \code{survfam = NULL}, \code{censfam} is not \code{NULL} and \code{cure = FALSE}.}
    \item{\code{nonparametric survival margin and parametric censoring margin (with cure)}}{
      \code{survfam = NULL}, \code{censfam} is not \code{NULL} and \code{cure = TRUE}.}
  }
  As byproducts, several other scenarios (the distribution of the underlying survival time is not nonparametric
  but fully parametric) can also be considered by this R function:
  \describe{
    \item{\code{parametric survival and censoring margins (without cure)}}{
      both \code{survfam} and \code{censfam} are not \code{NULL} and \code{cure = FALSE}.}
    \item{\code{parametric survival and censoring margins (with cure)}}{
      both \code{survfam} and \code{censfam} are not \code{NULL} and \code{cure = TRUE}.}
    \item{\code{parametric survival margin and nonparametric censoring margin (without cure)}}{
      \code{survfam} is not \code{NULL}, \code{censfam = NULL} and \code{cure = FALSE}.}
  }
  Furthermore, one might expect that a scenario with "parametric survival margin and nonparametric censoring margin
  (with cure)" can also be included. Indeed, it can be done based on: \code{survfam} is not \code{NULL}, \code{censfam = NULL}
  and \code{cure = TRUE}. However, from a theoretical perspective of view, whether this type of modeling is reasonable or not
  still needs further investigations.

  We emphasize that the first scenario (in byproducts) has also be considered in another function of this package.
  Specifically, the scenario of "parametric survival margin and nonparametric censoring margin (without cure)" can be
  fitted based on \code{ParamCop()}. However, the default joint modeling of survival and censoring times are based on
  their joint survival function in line with the semiparametric case (instead of modeling joint distribution function
  directly as in  Czado and Van Keilegom (2023) <doi:10.1093/biomet/asac067>), but the idea of estimation methodology
  are exactly the same.
}
\examples{

\donttest{
#----------------------------------------------------------#
# Basic preparations before running subsequent examples ####
#----------------------------------------------------------#

# library necessary packages

#------------------------------------------------------------------------#
# simulated data from Frank copula log-Normal margins (without cure)
#------------------------------------------------------------------------#

# generate the simulated data

# - the sample size of the generated data
n <- 1000

# information on the used copula
copfam.true <- "frank"
ktau.true <- 0.5
coppar.true <- 5.74

# parameters of the underlying log-normal marginal distributions
survpar.true <- c(2.20,1.00)
censpar.true <- c(2.20,0.25)

# - true underlying survival and censoring times
set.seed(1)
u.TC <- copula::rCopula(
  n        = n,
  copula   = copula::archmCopula(
    family = copfam.true,
    param  = coppar.true,
    dim    = 2
  )
)
yobs.T <- qlnorm(1-u.TC[,1],survpar.true[1],survpar.true[2])
yobs.C <- qlnorm(1-u.TC[,2],censpar.true[1],censpar.true[2])

# observations
yobs  <- pmin(yobs.T,yobs.C)
delta <- as.numeric(yobs.T<=yobs.C)
cat("censoring rate is", mean(1-delta))

# model the data under different scenarios

# scenario 1: nonparametric survival margin and parametric censoring margin
set.seed(1)
sol.scenario1 <- SurvDC(
  yobs    = yobs,
  delta   = delta,
  tm      = quantile(yobs, c(0.25,0.50,0.75)),
  copfam  = copfam.true,
  margins = list(survfam = NULL, censfam = "lnorm"),
  Var     = list(do = FALSE, nboot = 50)
)
sol.scenario1$probs
sol.scenario1$ktau
sol.scenario1$parapar

# scenario 2: parametric survival and censoring margins
set.seed(1)
sol.scenario2 <- SurvDC(
  yobs    = yobs,
  delta   = delta,
  tm      = quantile(yobs, c(0.25,0.50,0.75)),
  copfam  = copfam.true,
  margins = list(survfam = "lnorm", censfam = "lnorm"),
  Var     = list(do = FALSE, nboot = 50)
)
sol.scenario2$probs
sol.scenario2$ktau
sol.scenario2$parapar

# scenario 3: parametric survival margin and nonparametric censoring margin
set.seed(1)
sol.scenario3 <- SurvDC(
  yobs    = yobs,
  delta   = delta,
  tm      = quantile(yobs, c(0.25,0.50,0.75)),
  copfam  = copfam.true,
  margins = list(survfam = "lnorm", censfam = NULL),
  Var     = list(do = FALSE, nboot = 50)
)
sol.scenario3$probs
sol.scenario3$ktau
sol.scenario3$parapar

#------------------------------------------------------------------------
# simulated data from Frank copula log-Normal margins (with cure)
#------------------------------------------------------------------------

# generate the simulated data

#  true underlying cure rate
curerate.true <- 0.2

# true underlying survival and censoring times
set.seed(1)
u.TC <- copula::rCopula(
  n        = n,
  copula   = copula::archmCopula(
    family = copfam.true,
    param  = coppar.true,
    dim    = 2
  )
)
yobs.T <- sapply(u.TC[,1],function(uT){
  if(uT<=curerate.true){ val <- Inf }else{
    val <- EnvStats::qlnormTrunc((1-uT)/(1-curerate.true),survpar.true[1],survpar.true[2],0,15)
  }
  return(val)
})
yobs.C <- qlnorm(1-u.TC[,2],censpar.true[1],censpar.true[2])
cat("cure rate is",mean(yobs.T==Inf))

#  observations
yobs  <- pmin(yobs.T,yobs.C)
delta <- as.numeric(yobs.T<=yobs.C)
cat("censoring rate is",mean(1-delta))

# model the data under different scenarios (with cure)

# scenario 4: parametric survival and censoring margins
set.seed(1)
sol.scenario4 <- SurvDC(
  yobs    = yobs,
  delta   = delta,
  tm      = quantile(yobs, c(0.25,0.50,0.75)),
  copfam  = copfam.true,
  margins = list(survfam = "lnorm", censfam = "lnorm"),
  Var     = list(do = FALSE, nboot = 50),
  cure    = TRUE
)
sol.scenario4$probs
sol.scenario4$ktau
sol.scenario4$parapar
sol.scenario4$curerate

# scenario 5: nonparametric survival margin and parametric censoring margin
set.seed(1)
sol.scenario5 <- SurvDC(
  yobs    = yobs,
  delta   = delta,
  tm      = quantile(yobs, c(0.25,0.50,0.75)),
  copfam  = copfam.true,
  margins = list(survfam = NULL, censfam = "lnorm"),
  Var     = list(do = FALSE, nboot = 50),
  cure    = TRUE
)
sol.scenario5$probs
sol.scenario5$ktau
sol.scenario5$parapar
sol.scenario5$curerate
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BoundingCovariateEffects.R
\name{Chronometer}
\alias{Chronometer}
\title{Chronometer object}
\description{
R6 object that mimics a chronometer. It can be started, paused, record legs
and stopped.
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Chronometer-show}{\code{Chronometer$show()}}
\item \href{#method-Chronometer-reset}{\code{Chronometer$reset()}}
\item \href{#method-Chronometer-start}{\code{Chronometer$start()}}
\item \href{#method-Chronometer-stop}{\code{Chronometer$stop()}}
\item \href{#method-Chronometer-record.leg}{\code{Chronometer$record.leg()}}
\item \href{#method-Chronometer-get.chronometer.data}{\code{Chronometer$get.chronometer.data()}}
\item \href{#method-Chronometer-get.total.time}{\code{Chronometer$get.total.time()}}
\item \href{#method-Chronometer-accumulate.legs}{\code{Chronometer$accumulate.legs()}}
\item \href{#method-Chronometer-clone}{\code{Chronometer$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chronometer-show"></a>}}
\if{latex}{\out{\hypertarget{method-Chronometer-show}{}}}
\subsection{Method \code{show()}}{
Display the values stored in this chronometer object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Chronometer$show()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chronometer-reset"></a>}}
\if{latex}{\out{\hypertarget{method-Chronometer-reset}{}}}
\subsection{Method \code{reset()}}{
Reset the chronometer.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Chronometer$reset()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chronometer-start"></a>}}
\if{latex}{\out{\hypertarget{method-Chronometer-start}{}}}
\subsection{Method \code{start()}}{
Start the chronometer
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Chronometer$start()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chronometer-stop"></a>}}
\if{latex}{\out{\hypertarget{method-Chronometer-stop}{}}}
\subsection{Method \code{stop()}}{
Stop the chronometer
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Chronometer$stop(leg.name = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{leg.name}}{(optional) Name for the stopped leg.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chronometer-record.leg"></a>}}
\if{latex}{\out{\hypertarget{method-Chronometer-record.leg}{}}}
\subsection{Method \code{record.leg()}}{
Record a leg time. The chronometer will continue running.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Chronometer$record.leg(leg.name = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{leg.name}}{Name for the recorded leg.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chronometer-get.chronometer.data"></a>}}
\if{latex}{\out{\hypertarget{method-Chronometer-get.chronometer.data}{}}}
\subsection{Method \code{get.chronometer.data()}}{
Like \code{show} method, but more rudimentary.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Chronometer$get.chronometer.data()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chronometer-get.total.time"></a>}}
\if{latex}{\out{\hypertarget{method-Chronometer-get.total.time}{}}}
\subsection{Method \code{get.total.time()}}{
Return the total time span between start and stop.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Chronometer$get.total.time(force = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{force}}{Boolean variable. If \code{TRUE}, avoids error when calling
this function while chronometer has not been stopped yet.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chronometer-accumulate.legs"></a>}}
\if{latex}{\out{\hypertarget{method-Chronometer-accumulate.legs}{}}}
\subsection{Method \code{accumulate.legs()}}{
Return total time spent per leg category (using leg names).
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Chronometer$accumulate.legs(force = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{force}}{force Boolean variable. If \code{TRUE}, avoids error when calling
this function while chronometer has not been stopped yet.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Chronometer-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Chronometer-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Chronometer$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{liver}
\alias{liver}
\title{Liver cirrhosis data set.}
\format{
A data frame with 281 rows and 7 variables:
\describe{
 \item{patient}{patient ID.}
 \item{time}{Survival time.}
 \item{status}{Censoring indicator.}
 \item{age}{Age of patient.}
 \item{gender}{Gender of patient (1 - male, 0 - female).}
 \item{bmi}{Body mass index.}
 \item{ukeld}{UK End-stage Liver Disease score. Higher scores indicate more
 severe disease state.}
}
}
\usage{
liver
}
\description{
End stage liver disease data set, as for example analyzed by
D'Haen et al. (2025).
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FullyParametricCopulas.R
\name{ParamCop}
\alias{ParamCop}
\title{Estimation of  a parametric dependent censoring model without covariates.}
\usage{
ParamCop(Y, Delta, Copula, Dist.T, Dist.C, start = c(1, 1, 1, 1))
}
\arguments{
\item{Y}{Follow-up time.}

\item{Delta}{Censoring indicator.}

\item{Copula}{The copula family. This argument can take values from \code{c("frank","gumbel","clayton","gaussian","indep")}.}

\item{Dist.T}{The distribution to  be used for the survival time T. This argument can take one of the values from \code{c("lnorm", "weibull", "llogis")}.}

\item{Dist.C}{The distribution to  be used for the censoring time C. This argument can take one of the values from \code{c("lnorm", "weibull", "llogis")}.}

\item{start}{Starting values.}
}
\value{
A table containing the minimized negative log-likelihood using the independence copula model, the estimated parameter values for the model with the independence copula, the minimized negative log-likelihood using the specified copula model and the estimated parameter values for the model with the specified copula.
}
\description{
Note that it is not assumed that the association parameter of the copula function is known,
unlike most other papers in the literature. The details for implementing the methodology can be found in Czado and Van Keilegom (2023).
}
\examples{
tau = 0.75
Copula = "frank"
Dist.T = "weibull"
Dist.C = "lnorm"
par.T = c(2,1)
par.C = c(1,2)
n=1000

simdata<-TCsim(tau,Copula,Dist.T,Dist.C,par.T,par.C,n)

Y = simdata[[1]]
Delta = simdata[[2]]

ParamCop(Y,Delta,Copula,Dist.T,Dist.C)

}
\references{
Czado and Van Keilegom (2023). Dependent censoring based on parametric copulas. Biometrika, 110(3), 721-738.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/NonparametricTransformation.R
\name{NonParTrans}
\alias{NonParTrans}
\title{Fit a semiparametric transformation model for dependent censoring}
\usage{
NonParTrans(
  resData,
  X,
  W,
  start = NULL,
  n.iter = 15,
  bootstrap = FALSE,
  n.boot = 50,
  eps = 0.001
)
}
\arguments{
\item{resData}{Data matrix with three columns;  Z = the observed survival time, d1 = the censoring indicator of T
and  d2 =  the censoring indicator of C.}

\item{X}{Data matrix with covariates related to T}

\item{W}{Data matrix with covariates related to C}

\item{start}{Initial values for the finite dimensional parameters. If \code{start} is NULL, the initial values will be obtained
by fitting an Accelerated failure time models.}

\item{n.iter}{Number of iterations; the default is \code{n.iter = 20}. The larger the number of iterations, the longer the computational time.}

\item{bootstrap}{A boolean indicating whether to compute bootstrap standard errors for making inferences.}

\item{n.boot}{Number of bootstrap samples to use in the estimation of bootstrap standard errors if \code{bootstrap = TRUE}. The default is n.boot = 50. But, higher
values  of \code{n.boot} are recommended for obtaining good estimates of bootstrap standard errors.}

\item{eps}{Convergence error. This is set by the user in such away that the desired convergence is met; the default is \code{eps = 1e-3}.}
}
\value{
This function returns a fit of a semiparametric transformation model; parameter estimates, estimate of the non-parametric transformation function, bootstrap standard
errors for finite-dimensional parameters, the nonparametric cumulative hazard function, etc.
}
\description{
This function allows to estimate the dependency parameter along all other model parameters. First, estimates a non-parametric transformation function, and
then at the second stage it estimates other model parameters assuming that the non-parametric function is known. The details for
implementing the dependent censoring methodology can be found in Deresa and Van Keilegom (2021).
}
\examples{
\donttest{
# Toy data example to illustrate implementation
n = 300
beta = c(0.5, 1); eta = c(1,1.5); rho = 0.70
sigma = matrix(c(1,rho,rho,1),ncol=2)
err = MASS::mvrnorm(n, mu = c(0,0) , Sigma=sigma)
err1 = err[,1]; err2 = err[,2]
x1 = rbinom(n,1,0.5); x2 = runif(n,-1,1)
X = matrix(c(x1,x2),ncol=2,nrow=n); W = X   # data matrix
T1 = X\%*\%beta+err1
C =  W\%*\%eta+err2
T1 = exp(T1); C = exp(C)
A = runif(n,0,8); Y = pmin(T1,C,A)
d1 = as.numeric(Y==T1)
d2 = as.numeric(Y==C)
resData = data.frame("Z" = Y,"d1" = d1, "d2" = d2)   # should be data frame
colnames(X) = c("X1", "X2")
colnames(W) = c("W1","W2")

#  Bootstrap is false by default
output = NonParTrans(resData = resData, X = X, W = W, n.iter = 2)
output$parameterEstimates

}
}
\references{
Deresa, N. and Van Keilegom, I. (2021). On semiparametric modelling, estimation and inference for survival data subject to dependent censoring, Biometrika, 108, 965–979.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BoundingCovariateEffects.R
\name{pi.surv}
\alias{pi.surv}
\title{Estimate the model of Willems et al. (2025).}
\usage{
pi.surv(
  data,
  idx.param.of.interest,
  idxs.c,
  t,
  par.space,
  search.method = "GS",
  add.options = list(),
  verbose = 0,
  picturose = FALSE,
  parallel = FALSE
)
}
\arguments{
\item{data}{Data frame containing the data on which to fit the model. The
columns should be named as follows: 'Y' = observed timed, 'Delta' = censoring
indicators, 'X0' = intercept column, 'X1' - 'Xp' = covariates.}

\item{idx.param.of.interest}{Index of element in the covariate vector for
which the identified interval should be estimated. It can also be specified
as \code{idx.param.of.interest = "all"}, in which case identified intervals
will be computed for all elements in the parameter vector. Note that
\code{idx.param.of.interest = 1} corresponds to the intercept parameter.}

\item{idxs.c}{Vector of indices of the continuous covariates. Suppose the
given data contains 5 covariates, of which 'X2' and 'X5' are continuous, this
argument should be specified as \code{idxs.c = c(2, 5)}.}

\item{t}{Time point for which to estimate the identified set of
\eqn{\beta(t)}.}

\item{par.space}{Matrix containing bounds on the space of the parameters. The
first column corresponds to lower bounds, the second to upper bounds. The i'th
row corresponds to the bounds on the i'th element in the parameter vector.}

\item{search.method}{The search method to be used to find the identified
interval. Default is \code{search.method = "GS"}.}

\item{add.options}{List of additional options to be specified to the method.
Notably, it can be used to select the link function \eqn{\Lambda(t))} that
should be considered. Currently, the link function leading to an accelerated
failure time model (\code{"AFT_ll"}, default) and the link function
leading to a Cox proportional hazards model (\code{"Cox_wb"}) are implemented.
Other options can range from 'standard' hyperparameters such as the
confidence level of the test and number of instrumental functions to be used,
to technical hyperparameters regarding the search method and test
implementation.

General hyperparameters:
\describe{
 \item{cov.ranges:}{known bounds on each of the covariates in the data set.}
 \item{norm.func.name:}{Name of the normalization function to be used. Can
 be either "normalize.covariates1" or "normalize.covariates2" (default).
 The former is a simple elementwise rescaling. The latter uses the PCA
 approach.}
 \item{inst.func.family:}{Family of instrumental functions to be used for
 all covariates. Options are "box", "spline" and "cd". The former two are
 only applicable for continuous covariates. The latter can also handle
 discrete covariates. Default is "cd".}
 \item{G.c:}{The class of instrumental functions used for the continuous
 covariates in the model, in case "cd" is selected as
 \code{inst.func.family:}. Options are "box" and "spline". Default is
 "spline".}
 \item{degree:}{The degree of the B-spline functions, should they be used as
 instrumental functions for the continuous covariates. Default is 3.}
 \item{link.function:}{Name of the link function to be used. Options are
 "AFT_ll" for the AFT model with log-logistic baseline, or "Cox_wb" for the
 Cox PH model (originally with Weibull baseline, but now for a general)
 baseline hazard).}
 \item{K.bar:}{Number of refinement steps when obtaining the critical value.
 See Bei (2024).}
 \item{B:}{Number of bootstrap samples to be used when obtaining the
 bootstrap distribution of the test statistic.}
 \item{ignore.empty.IF:}{Boolean value indicating whether instrumental
 functions with empty support should be ignored.
 Default is FALSE. The feature \code{ignore.empty.IF = TRUE} is experimental,
 so there might exist edge cases for which the implementation will fail to
 run.}
 \item{c:}{User-specified projection vector. If supplied, the method will
 estimate the identified interval of \eqn{c^\top \beta} as opposed to the
 identified interval of \eqn{\beta_k}, where \eqn{k} is specified as the
 argument \code{idx.param.of.interest}.}
}

Hyperparameters specific to the EAM implementation:
\describe{
 \item{min.dist/max.dist:}{The minimum/maximum distance of sampled points
 from the current best value for the coefficient of interest.}
 \item{min.eval/max.eval:}{The minimum/maximum number of points evaluated
 in the initial feasible point search.}
 \item{nbr.init.sample.points:}{The total number of drawn points required in
 the initial drawing process.}
 \item{nbr.init.unif:}{The total number of uniformly drawn points in the
 initial set of starting values.}
 \item{nbr.points.per.iter.init:}{Number of points sampled per iteration in
 the initial drawing process.}
 \item{nbr.start.vals:}{Number of starting values for which to run the
 optimization algorithm for the expected improvement.}
 \item{nbr.opt.EI:}{Number of optimal theta values found by the optimization
 algorithm to return.}
 \item{nbr.extra:}{Number of extra randomly drawn points to add to the set
 of optimal theta values (to be supplied to the next E-step).}
 \item{min.improvement:}{Minimum amount that the current best root of the
 violation curve should improve by wrt. the its previous value.}
 \item{min.possible.improvement:}{Minimum amount that the next iteration
 should be able to improve upon the current best value of the root.}
 \item{EAM.min.iter:}{Minimum amount of EAM iterations to run.}
 \item{max.iter:}{Maximum amount of EAM iterations to run.}
}

Hyperparameters specific to the gridsearch implementation:
\describe{
 \item{min.eval/max.eval:}{Minimum and maximum number of evaluations.}
 \item{next.gs.point:}{Function that determines the next point in the grid
 search sequence.}
 \item{step.size:}{Step size of the grid.}
 \item{bin.search.tol:}{Binary search tolerance.}
 \item{max.iter:}{Maximum number of iterations that the algorithm can run.}
}

Other (hidden) options can also be overwritten, though we highly discourage
this. If necessary, you can consult the source code of this functions to
find the names of the desired parameters and add their name alongside their
desired value as an entry in \code{options} (e.g.
\code{options$min.var <- 1e-4}. Again, not recommended!).}

\item{verbose}{Verbosity level. The higher the value, the more verbose the
method will be. Default is \code{verbose = 0}.}

\item{picturose}{Picturosity flag. If \code{TRUE}, a plot illustrating the
workings of the algorithm will updated during runtime. Default is
\code{picturose = FALSE}.}

\item{parallel}{Flag for whether or not parallel computing should be used.
Default is \code{parallel = FALSE}. When \code{parallel = TRUE}, this
implementation will use \code{min(detectCores() - 1, 10)} cores to construct
the parallel back-end.}
}
\value{
Matrix containing the identified intervals of the specified
coefficients, as well as corresponding convergence information of the
estimation algorithm.
}
\description{
This function estimates bounds on the coefficients the single-
index model \eqn{\Lambda(x^\top \beta(t))} for the conditional cumulative
distribution function of the event time.
}
\examples{
\donttest{

  # Clear workspace
  rm(list = ls())

  # Load the survival package
  library(survival)

  # Set random seed
  set.seed(123)

  # Load and preprocess data
  data <- survival::lung
  data[, "intercept"] <- rep(1, nrow(data))
  data[, "status"] <- data[, "status"] - 1
  data <- data[, c("time", "status", "intercept", "age", "sex")]
  colnames(data) <- c("Y", "Delta", "X0", "X1", "X2")

  # Standardize age variable
  data[, "X1"] <- scale(data[, "X1"])

  ## Example:
  ## - Link function: AFT link function (default setting)
  ## - Number of IF: 5 IF per continuous covariate (default setting)
  ## - Search method: Binary search
  ## - Type of IF: Cubic spline functions for continuous covariate, indicator
  ##   function for discrete covariate (default setting).

  # Settings for main estimation function
  idx.param.of.interest <- 2 # Interest in effect of age
  idxs.c <- 1                # X1 (age) is continuous
  t <- 200                   # Model imposed at t = 200
  search.method <- "GS"      # Use binary search
  par.space <- matrix(rep(c(-10, 10), 3), nrow = 3, byrow = TRUE)
  add.options <- list()
  picturose <- TRUE
  parallel <- FALSE

  # Estimate the identified intervals
  pi.surv(data, idx.param.of.interest, idxs.c, t, par.space,
          search.method = search.method, add.options = add.options,
          picturose = picturose, parallel = parallel)
}


}
\references{
Willems, I., Beyhum, J. and Van Keilegom, I. (2025). Partial
identification for a class of survival models under dependent censoring.
(Submitted).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/QRdepcens.R
\name{QRdepCens}
\alias{QRdepCens}
\title{Estimate the model of D'Haen et al. (2025).}
\usage{
QRdepCens(data, hp, var.estimate = FALSE, verbose = TRUE)
}
\arguments{
\item{data}{Data on which the model should be estimated. Note that the data
should be structured in a specific form: The observed times should be put in
a column named \code{"Y"}, the censoring indicators in a column named
\code{"Delta"}, and the covariates in columns named \code{"X1"}, \code{"X2"},
... (in increasing order). The given data set cannot contain any other
columns.}

\item{hp}{List of hyperparameters to be used, the elements of which will
overwrite the default settings. In particular, consider changing:
\describe{
 \item{test_cop_name:}{Copula to be used. Should be one of
 \code{"frank"}, \code{"gumbel"}, \code{"clayton"} or \code{"indep"}.}
 \item{homoscedastic:}{Boolean flag indicating whether
 homoscedasticity can be assumed.}
 \item{variance.bootstrap.iterations:}{Number of bootstrap resamples
 to use during variance estimation. Consider increase if more precision is
 needed; consider decreasing to reduce computation time.}
}
Other hyperparameters can be changed though it is not recommended. We refer
to the source code for the available options.}

\item{var.estimate}{Boolean value indicating whether the variance should be
estimated (via bootstrap). This can take a considerable amount of time.
Default is \code{var.estimate = FALSE}.}

\item{verbose}{Verbosity flag (boolean) indicating whether the results should
be printed to the console. Default is \code{verbose = TRUE}.}
}
\description{
This function estimates the parameters in the model of D'Haen et al. (2025).
}
\note{
The variance estimation procedure could easily be paralelized.
However, this is currently not implemented.
}
\examples{
\donttest{

 # Load the data
 data(liver)

 # Give standard column names (required!)
 colnames(liver) <- c("patient", "Y", "Delta", "X1", "X2", "X3", "X4")
 liver <- liver[, c(-1, -6, -7)]

 # Run the model
 hp <- list(
   homoscedastic = FALSE,
   test_cop_name = "frank"
 )
 QRdepCens(liver, hp, var.estimate = FALSE)
 # Takes a while if var.estimate = TRUE...
}

}
\references{
D'Haen, M., Van Keilegom, I. and Verhasselt, A. (2025). Quantile
regression under dependent censoring with unknown association. Lifetime Data
Analysis 31(2):253-299.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CopulaBasedCoxPH_pseudoLikelihoodFunctions.R
\name{SolveL}
\alias{SolveL}
\title{Cumulative hazard function of survival time under dependent censoring}
\usage{
SolveL(
  theta,
  resData,
  X,
  W,
  cop = c("Frank", "Gumbel", "Normal"),
  dist = c("Weibull", "lognormal")
)
}
\arguments{
\item{theta}{Estimated parameter values/initial values for finite dimensional parameters}

\item{resData}{Data matrix with three columns;  Z = the observed survival time, d1 = the censoring indicator of T
and  d2 =  the censoring indicator of C.}

\item{X}{Data matrix with covariates related to T}

\item{W}{Data matrix with covariates related to C. First column of W should be ones}

\item{cop}{Which copula should be computed to account for dependency between T and C. This argument can take
one of the values from \code{c("Gumbel", "Frank", "Normal")}. The default copula model is "Frank".}

\item{dist}{The distribution to  be used for the dependent censoring C. Only two distributions are allowed, i.e, Weibull
and lognormal distributions. With the value \code{"Weibull"} as the
  default.}
}
\value{
This function returns an estimated hazard function, cumulative hazard function and distinct observed survival times;
}
\description{
This function estimates the cumulative hazard function of survival time (T) under dependent censoring (C). The estimation
makes use of the estimating equations derived based on martingale ideas.
}
\examples{
\donttest{
n = 200
beta = c(0.5)
lambd = 0.35
eta = c(0.9,0.4)
X = cbind(rbinom(n,1,0.5))
W = cbind(rep(1,n),rbinom(n,1,0.5))
frank.cop <- copula::frankCopula(param = 5,dim = 2)
U = copula::rCopula(n,frank.cop)
T1 = (-log(1-U[,1]))/(lambd*exp(X*beta))         # Survival time'
T2 = (-log(1-U[,2]))^(1.1)*exp(W\%*\%eta)          # Censoring time
A = runif(n,0,15)                                # administrative censoring time
Z = pmin(T1,T2,A)
d1 = as.numeric(Z==T1)
d2 = as.numeric(Z==T2)
resData = data.frame("Z" = Z,"d1" = d1, "d2" = d2)
theta = c(0.3,1,0.3,1,2)

# Estimate cumulative hazard function
cumFit <- SolveL(theta, resData,X,W)
cumhaz = cumFit$cumhaz
time = cumFit$times

# plot hazard vs time

plot(time, cumhaz, type = "l",xlab = "Time",
ylab = "Estimated cumulative hazard function")

}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FitCopulaBasedCoxPHmodels.R
\name{fitIndepCens}
\alias{fitIndepCens}
\title{Fit Independent Censoring Models}
\usage{
fitIndepCens(
  resData,
  X,
  W,
  dist = c("Weibull", "lognormal"),
  start = NULL,
  n.iter = 50,
  bootstrap = TRUE,
  n.boot = 150,
  ncore = 7,
  eps = 1e-04
)
}
\arguments{
\item{resData}{Data matrix with three columns;  Z = the observed survival time, d1 = the censoring indicator of T
and  d2 =  the censoring indicator of C.}

\item{X}{Data matrix with covariates related to T.}

\item{W}{Data matrix with covariates related to C. First column of W should be ones.}

\item{dist}{The distribution to be used for the censoring time C. Only two distributions are allowed, i.e, Weibull
and lognormal distributions. With the value \code{"Weibull"} as the
  default.}

\item{start}{Initial values for the finite dimensional parameters. If \code{start} is NULL, the initial values will be obtained
by fitting a Cox model for survival time T and a Weibull model for censoring time C.}

\item{n.iter}{Number of iterations; the default is \code{n.iter = 50}. The larger the number of iterations, the longer the computational time.}

\item{bootstrap}{A boolean indicating whether to compute bootstrap standard errors for making inferences.}

\item{n.boot}{Number of bootstrap samples to use in the estimation of bootstrap standard errors if \code{bootstrap = TRUE}. The default is n.boot = 150. But, higher
values  of \code{n.boot} are recommended for obtaining good estimates of bootstrap standard errors.}

\item{ncore}{The number of cores to use for parallel computation is configurable, with the default \code{ncore = 7}.}

\item{eps}{Convergence error. This is set by the user in such away that the desired convergence is met; the default is \code{eps = 1e-4}.}
}
\value{
This function returns a fit of independent censoring model; parameter estimates, estimate of the cumulative hazard function, bootstrap standard
errors for finite-dimensional parameters, the nonparametric cumulative hazard function, etc.
}
\description{
This function allows to estimate all model parameters under the assumption of independent censoring. First, estimates the cumulative hazard function, and
then at the second stage it estimates other model parameters assuming that the cumulative hazard is known.
}
\examples{
\donttest{

# Toy data example to illustrate implementation
n = 300
beta = c(0.5)
lambd = 0.35
eta = c(0.9,0.4)
X = cbind(rbinom(n,1,0.5))
W = cbind(rep(1,n),rbinom(n,1,0.5))
# generate dependency structure from Frank
frank.cop <- copula::frankCopula(param = 5,dim = 2)
U = copula::rCopula(n,frank.cop)
T1 = (-log(1-U[,1]))/(lambd*exp(X*beta))                  # Survival time'
T2 = (-log(1-U[,2]))^(1.1)*exp(W\%*\%eta)                   # Censoring time
A = runif(n,0,15)                                         # administrative censoring time
Z = pmin(T1,T2,A)
d1 = as.numeric(Z==T1)
d2 = as.numeric(Z==T2)
resData = data.frame("Z" = Z,"d1" = d1, "d2" = d2)     # should be data frame

colnames(W) <- c("ones","cov1")
colnames(X) <- "cov.surv"

# Fit independent censoring model

fitI <- fitIndepCens(resData = resData, X = X, W = W, bootstrap = FALSE)

# parameter estimates

fitI$parameterEstimates

# summary fit results
summary(fitI)

# plot cumulative hazard vs time

 plot(fitI$observedTime, fitI$cumhazardFunction, type = "l",xlab = "Time",
 ylab = "Estimated cumulative hazard function")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/FitCopulaBasedCoxPHmodels.R
\name{fitDepCens}
\alias{fitDepCens}
\title{Fit Dependent Censoring Models}
\usage{
fitDepCens(
  resData,
  X,
  W,
  cop = c("Frank", "Gumbel", "Normal"),
  dist = c("Weibull", "lognormal"),
  start = NULL,
  n.iter = 50,
  bootstrap = TRUE,
  n.boot = 150,
  ncore = 7,
  eps = 1e-04
)
}
\arguments{
\item{resData}{Data matrix with three columns;  Z = the observed survival time, d1 = the censoring indicator of T
and  d2 =  the censoring indicator of C.}

\item{X}{Data matrix with covariates related to T.}

\item{W}{Data matrix with covariates related to C. First column of W should be a vector of ones.}

\item{cop}{Which copula should be computed to account for dependency between T and C. This argument can take
one of the values from \code{c("Gumbel", "Frank", "Normal")}.}

\item{dist}{The distribution to be used for the censoring time C. Only two distributions are allowed, i.e, Weibull
and lognormal distributions. With the value \code{"Weibull"} as the
default.}

\item{start}{Initial values for the finite dimensional parameters. If \code{start} is NULL, the initial values will be obtained
by fitting a Cox model for survival time T and a Weibull model for dependent censoring C.}

\item{n.iter}{Number of iterations; the default is \code{n.iter = 50}. The larger the number of iterations, the longer the computational time.}

\item{bootstrap}{A boolean indicating whether to compute bootstrap standard errors for making inferences.}

\item{n.boot}{Number of bootstrap samples to use in the estimation of bootstrap standard errors if \code{bootstrap = TRUE}. The default is n.boot = 150. But, higher
values  of \code{n.boot} are recommended for obtaining good estimates of bootstrap standard errors.}

\item{ncore}{The number of cores to use for parallel computation in bootstrapping, with the default \code{ncore = 7}.}

\item{eps}{Convergence error. This is set by the user in such away that the desired convergence is met; the default is \code{eps = 1e-4}.}
}
\value{
This function returns a fit of dependent censoring model; parameter estimates, estimate of the cumulative hazard function, bootstrap standard
errors for finite-dimensional parameters, the nonparametric cumulative hazard function, etc.
}
\description{
This function allows to estimate the dependency parameter along all other model parameters. First, estimates the cumulative hazard function, and
then at the second stage it estimates other model parameters assuming that the cumulative hazard function is known. The details for
implementing the dependent censoring methodology can be found in Deresa and Van Keilegom (2024).
}
\examples{
\donttest{
# Toy data example to illustrate implementation
n = 300
beta = c(0.5)
lambd = 0.35
eta = c(0.9,0.4)
X = cbind(rbinom(n,1,0.5))
W = cbind(rep(1,n),rbinom(n,1,0.5))
# generate dependency structure from Frank
frank.cop <- copula::frankCopula(param = 5,dim = 2)
U = copula::rCopula(n,frank.cop)
T1 = (-log(1-U[,1]))/(lambd*exp(X*beta))           # Survival time
T2 = (-log(1-U[,2]))^(1.1)*exp(W\%*\%eta)            # Censoring time
A = runif(n,0,15)                                  # administrative censoring time
Z = pmin(T1,T2,A)
d1 = as.numeric(Z==T1)
d2 = as.numeric(Z==T2)
resData = data.frame("Z" = Z,"d1" = d1, "d2" = d2)   # should be data frame
colnames(W) <- c("ones","cov1")
colnames(X) <- "cov.surv"

# Fit dependent censoring model

fit <- fitDepCens(resData = resData, X = X, W = W, bootstrap = FALSE)

# parameter estimates

fit$parameterEstimates

# summary fit results
summary(fit)

# plot cumulative hazard vs time

plot(fit$observedTime, fit$cumhazardFunction, type = "l",xlab = "Time",
ylab = "Estimated cumulative hazard function")
}

}
\references{
Deresa and Van Keilegom (2024). Copula based Cox proportional hazards models for dependent censoring, Journal of the American Statistical Association, 119:1044-1054.
}
